/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 *
 * $Id: hash.h,v 1.18 1998/02/07 03:44:28 libpa Exp $
 *
 */

#ifndef HASH_INCLUDED
#define HASH_INCLUDED

#include <stdio.h>
#include "common.h"
#include "hash.class.h"

/* Description.
   Space-efficient hash table for storing arbitrary <key,datum> pairs.
   Implements open addressing, double hashing, and automatic resizing (up/down). 
   lookup and insert are O(1) expected time, remove is O(n) expected time.
 */

#define KEY_NOT_FOUND                 (UINT_MAX)
#define HASH_TABLE_MAXKEYSIZE         (UCHAR_MAX)
#define HASH_TABLE_MAXDATUMSIZE       (USHRT_MAX)
#define HASH_TABLE_MAXCAPACITYINDEX   (UCHAR_MAX)

typedef struct hash_table_tag
{
  unsigned char       capacity_index;
  unsigned char       key_size;
  unsigned short      datum_size;
  unsigned int        slot_count;     /* number of allocated slots  */
  unsigned int        entry_count;    /* number of distinct entries */
  unsigned int       *slots;
  unsigned char      *keys;
  unsigned char      *datums;
  hash_class_t       *class;
  const void         *parameters;
} hash_table_t;


/************************************************************
 * CONSTRUCTOR/DESTRUCTOR
 ************************************************************/

hash_table_t *hash_table_create(unsigned key_size, unsigned datum_size, 
				hash_class_t *hash_class,
				const void *parameters);
              /* Arguments: expected number of table entries, at least 0        */
              /*            caller is responsible for hash_class and parameters */
              /* Note: hash_class_t is typically a const static global variable */
              /* Returns: empty table for <key,datum> pairs of specified sizes  */
void          hash_table_destroy(hash_table_t *table);
              /* Effect: deallocates all memory used by given hash_table_t      */

/************************************************************
 * BASIC SELECTORS
 ************************************************************/

#define hash_table_size(table_p)        ((unsigned)(table_p)->entry_count)
        /* Returns: number of entries in table                    */
#define hash_table_capacity(table_p)    ((unsigned)(table_p)->slot_count)
        /* Returns: number of slots in table                      */
#define hash_table_load(table_p)                                  \
        ((double)hash_table_size(table_p) / (double)hash_table_capacity(table_p))
        /* Returns: current hash table load factor                */

#define hash_table_key_size(table_p)     ((unsigned)(table_p)->key_size)
        /* Returns: number of bytes necessary to store each key   */
#define hash_table_datum_size(table_p)   ((unsigned)(table_p)->datum_size)
        /* Returns: number of bytes necessary to store each datum */
#define hash_table_parameters(table_p)   ((table_p)->parameters)
        /* Returns: caller parameters for the hash_table_t        */
#define hash_table_class(table_p)        ((table_p)->class)
        /* Returns: hash_class_t for the hash_table_t             */

/************************************************************
 * PREDICATES
 ************************************************************/

boolean_t  hash_table_is_valid(const hash_table_t *table);
           /* Returns: TRUE iff table is a valid hash table             */
boolean_t  hash_table_are_equivalent(const hash_table_t *left, const hash_table_t *right);
           /* Returns: TRUE iff both tables have same <key,datum> pairs */
boolean_t  hash_table_is_empty(const hash_table_t *table);
           /* Returns: TRUE iff table contains no entries               */
boolean_t  hash_table_includes(const hash_table_t *table, const void *key_p);
           /* Returns: TRUE iff table contains pair with *key_p         */
           /* Note: use this if hash_table_datum_size(table) == 0       */

/************************************************************
 * SELECTORS
 ************************************************************/

void      *hash_table_lookup(const hash_table_t *table, const void *key_p);
           /* Returns: datum_p associated with *key_p or NULL if none exists */
           /* Note: always returns NULL if hash_table_datum_size(table) == 0 */
           /*       in which case you should use hash_table_includes()       */

/************************************************************
 * MUTATORS
 ************************************************************/

void       hash_table_insert(hash_table_t *table, const void *key_p, const void *datum_p);
           /* Effect: unconditionally inserts <*key_p,*datum_p> into table   */
void      *hash_table_update(hash_table_t *table, const void *key_p, const void *initial_p);
           /* Effect: inserts <*key_p,*initial_p> if *key_p is new           */
           /* Returns: datum_p inside table after conditional insertion      */
           /* Note: always returns NULL if hash_table_datum_size(table) == 0 */

void       hash_table_remove(hash_table_t *table, const void *key_p);
           /* Effect: removes <*key_p,datum> pair from table in O(n) time */
void       hash_table_clear(hash_table_t *table);
           /* Effect: removes all entries from table in O(1) time         */
void       hash_table_prune(hash_table_t *table, 
			    boolean_t (*prune_me)(const void *key_p, const void *datum_p, 
						  void *parameters),
			    void *parameters);
           /* Effect: removes every <key,datum> entry for which prune_me() is TRUE */
           /*         in O(n) time and resizes table appropriately.                */
           /* Note: this is nearly as fast as a single hash_table_remove()!        */

void       hash_table_resize_datum(hash_table_t *table, 
				   unsigned new_datum_size,
				   void (*coerce)(const void *old_datum_p, 
						  void *new_datum_p,
						  void *parameters),
				   void *parameters);
           /* Arguments: coerce() mutates old *datum_p to new *datum_p */
           /* Effect: mutates table so that datum have new size        */
           /* Note: coercion occurs even if old and new have same size */

/************************************************************
 *  HIGHER ORDER FUNCTION
 ************************************************************/

void       hash_table_map(hash_table_t *table, 
			  boolean_t (*proc)(const void *key_p, void *datum_p, 
					    void *parameters),
			  void *parameters);
           /* Effect: calls proc() on every <key,datum> entry in table */
           /*         while proc() returns TRUE                        */
           /* Note: proc() can mutate *datum_p but not *key_p!         */

/************************************************************
 * INDEXING
 ************************************************************/

unsigned    hash_table_index(const hash_table_t *table, const void *key_p);
            /* Returns: maps *key_p to index in [0,entry_count-1] or        */
            /*          KEY_NOT_FOUND if not found                          */
            /* Warning: mapping changes when any entry is deleted.  Hence,  */
            /*          this function should only be used for static table. */
const void *hash_table_key_p(const hash_table_t *table, unsigned index);
            /* Returns: &key for table entry at given index                 */
void       *hash_table_datum_p(const hash_table_t *table, unsigned index);
            /* Returns: &datum for table entry at given index               */

void        hash_table_index_map(hash_table_t *table, 
				 boolean_t (*proc)(const void *key_p, void *datum_p, 
						   unsigned index, void *parameters),
				 void *parameters);
            /* Effect: calls proc() on every entry in table with its index  */
            /*         while proc() returns TRUE                            */
            /* Warning: indexing changes when any entry is deleted!         */
            /* Note: proc() can mutate *datum_p but not *key_p              */

/************************************************************
 * PERFORMANCE TUNING
 ************************************************************/

void        hash_table_monitor(const hash_table_t *table, FILE *stream);
            /* Effect: writes performance information to stream             */

/************************************************************
 * SEMIPRIVATE OPERATIONS
 ************************************************************/

#define hash_table_compare(table_p, left_p, right_p) \
        ((table_p)->class->compare((left_p),(right_p),(table_p)->parameters))
#define hash_table_primary_hash(table_p, key_p) \
        ((table_p)->class->primary_hash((key_p),(table_p)->parameters))
#define hash_table_secondary_hash(table_p, key_p) \
        ((table_p)->class->secondary_hash((key_p),(table_p)->parameters))

#endif

