/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 *
 * $Id: string.hash.h,v 1.5 1998/02/07 03:44:28 libpa Exp $
 *
 */

#ifndef STRING_HASH_INCLUDED
#define STRING_HASH_INCLUDED

#include <string.h>
#include "hash.h"

/* Description.
 * Hash table for storing arbitrary <string,datum> pairs
 * where strings are NULL-terminated character strings.
 */

#define STRING_NOT_FOUND   KEY_NOT_FOUND

typedef hash_table_t string_hash_t;

string_hash_t *string_hash_create(unsigned datum_size);
               /* Arguments: number of bytes used by datum                      */
               /* Returns: empty table for string keys (ie., associative array) */
               /* Note: caller is responsible for all string memory             */
void           string_hash_destroy(string_hash_t *table);
               /* Effect: deallocates all storage used by string_hash_t         */

/************************************************************
 * BASIC SELECTORS
 ************************************************************/

#define  string_hash_size(table_p)           ((unsigned)hash_table_size(table_p))
         /* Returns: number of <string,datum> entries in table      */
#define  string_hash_datum_size(table_p)     (hash_table_datum_size(table_p))
         /* Returns: number of bytes necessary to store each datum  */

/************************************************************
 * PREDICATES
 ************************************************************/

boolean_t string_hash_is_valid(const string_hash_t *table);
          /* Returns: TRUE iff table is a valid string_hash_t     */
#define   string_hash_is_empty(table_p)           (hash_table_is_empty(table_p))
          /* Returns: TRUE iff table contains no entries          */
boolean_t string_hash_includes(const string_hash_t *table, const char *string);
          /* Returns: TRUE iff table contains pair with string    */
          /* Note: use this if string_hash_datum_size(table) == 0 */
#define   string_hash_are_equivalent(left_p,right_p) \
          (hash_table_are_equivalent((left_p),(right_p)))
          /* Returns: TRUE iff both tables have same <string,datum> pairs */

/************************************************************
 * SELECTOR
 ************************************************************/

void     *string_hash_lookup(const string_hash_t *table, const char *string);
          /* Returns: datum_p associated with string or NULL if none exists    */
          /* Note: always returns NULL if string_hash_datum_size(table_p) == 0 */
          /*       in which case you should use string_hash_includes() instead */

/************************************************************
 * MUTATORS
 ************************************************************/

void      string_hash_insert(string_hash_t *table, const char *string, const void *datum_p);
          /* Effect: unconditionally inserts <string,*datum_p> into table    */
void     *string_hash_update(string_hash_t *table, const char *string, const void *inital_p);
          /* Effect: inserts <*symbol_p,*initial_p> if *key_p is new         */
          /* Returns: datum_p inside table after conditional insertion       */
          /* Note: always returns NULL if symbol_hash_datum_size(table) == 0 */

void      string_hash_remove(string_hash_t *table, const char *string);
          /* Effect: removes <string,datum> pair from table in O(n) time     */
void      string_hash_clear(string_hash_t *table);
          /* Effect: removes all entries from string_hash_t in O(n) time     */
void      string_hash_prune(string_hash_t *table,
			    boolean_t (*prune_me)(const char *string, 
						  const void *datum_p, 
						  void *parameters),
			    void *parameters);
          /* Effect: removes every <string,datum> entry for which prune_me() */
          /*         is TRUE in O(n) time and resizes table appropriately.   */
          /* Note: this is nearly as fast as a single string_hash_remove()!  */

/************************************************************
 *  HIGHER ORDER FUNCTION
 ************************************************************/

void      string_hash_map(string_hash_t *table, 
			  boolean_t (*proc)(const char *string, void *datum_p, 
					    void *parameters),
			  void *parameters);
          /* Effect: calls proc() on every <string,datum> entry in table */
          /* Note: proc() can mutate *datum_p but not string!            */

/************************************************************
 * INDEXING
 ************************************************************/

unsigned    string_hash_index(const string_hash_t *table, const char *string);
            /* Returns: maps string to index in [0,entry_count-1] or         */
            /*          STRING_NOT_FOUND if not found                        */
            /* Warning: mapping changes when any entry is deleted.  Hence,   */
            /*          this function should only be used for static table.  */
const char *string_hash_string(const string_hash_t *table, unsigned index);
            /* Returns: string associated with given index if index is valid */
void       *string_hash_datum_p(const string_hash_t *table, unsigned index);
            /* Returns: &datum associated with given index if index is valid */

void        string_hash_index_map(string_hash_t *table, 
				  boolean_t (*proc)(const char *string, void *datum_p, 
						    unsigned index, void *parameters),
				  void *parameters);
            /* Effect: calls proc() on every <string,datum> entry in table  */
            /*         along with its index while proc() returns TRUE       */
            /* Warning: indexing changes when any entry is deleted!         */
            /* Note: proc() can mutate *datum_p but not string!             */

/************************************************************
 * PERFORMANCE TUNING
 ************************************************************/

void      string_hash_monitor(const string_hash_t *table, FILE *stream);
          /* Effect: writes performance information to stream             */

#endif

