/*
 * (c) 1996 Eric Sven Ristad.  All rights reserved.
 *
 * $Id: suffix.array.h,v 2.1 1998/01/28 14:52:30 libpa Exp $
 *
 */

#ifndef SUFFIX_ARRAY_INCLUDED
#define SUFFIX_ARRAY_INCLUDED
#include <stdio.h>
#include "vector.h"
#include "suffix.h"

/* DESCRIPTION
 *  Simple implementation of traditional suffix array w/ logical pointers.
 */

typedef struct suffix_array_tag
{
  unsigned          length_limit;   /* maximal string length */
  unsigned          pruned;         /* number of pruned suffixes */
  unsigned          corpus_length;  /* for safety */
  unsigned          corpus_parity;  /* for safety */
  const vector_t   *corpus;  
  suffix_index_t   *indices;
} suffix_array_t;

/************************************************************
 *  CONSTRUCTOR/DESTRUCTOR
 ************************************************************/

suffix_array_t *suffix_array_create(int length_limit, const vector_t *corpus);
               /* Arguments: upper bound on suffix length and base corpus */
               /* Warning: suffix_array_t doesn't copy or mutate corpus   */
void           suffix_array_destroy(suffix_array_t *array);
               /* Effect: deallocates all storage used by given array     */
               /*         but not the corpus provided on creation         */
 
/************************************************************
 *  BASIC SELECTORS
 ************************************************************/

#define    suffix_array_size(s)             (vector_length((s)->indices))
           /* Returns: number of suffixes in internal memory        */
#define    suffix_array_symbol_size(s)      (vector_object_size((s)->corpus))
           /* Returns: number of bytes per symbol in suffix array   */
#define    suffix_array_corpus(s)           ((const void *)((s)->corpus))
           /* Returns: actual corpus used by the suffix_array_t     */
#define    suffix_array_corpus_length(s)    (vector_length((s)->corpus))
           /* Returns: number of symbols in suffix_array_t corpus   */
#define    suffix_array_length_limit(s)     ((s)->length_limit)

/************************************************************
 *  PREDICATES
 ************************************************************/

boolean_t  suffix_array_is_valid(const suffix_array_t *array); 
           /* Returns: TRUE iff suffix_array_t is valid             */
boolean_t  suffix_array_are_equivalent(const suffix_array_t *left, 
				       const suffix_array_t *right); 
           /* Returns: TRUE iff both suffix_array_t are equivalent  */

/************************************************************
 *  STRING SELECTORS
 ************************************************************/

unsigned   suffix_array_maximal_length(const suffix_array_t *array, 
				       unsigned string_length, const void *string);
           /* Arguments: a string containing string_length symbols        */
           /* Returns: length of maximal suffix of given string in memory */
boolean_t  suffix_array_includes(const suffix_array_t *array, 
				 unsigned string_length, const void *string);
           /* Returns: TRUE iff string is included in suffix_array_t      */
           /* Note: this is faster than suffix_array_frequency()          */
unsigned   suffix_array_frequency(const suffix_array_t *array, 
				  unsigned string_length, const void *string);
           /* Arguments: a string containing string_length symbols        */
           /* Returns: number of occurrances of string in memory          */

/************************************************************
 *  HIGHER ORDER FUNCTION
 ************************************************************/

void       suffix_array_map(const suffix_array_t *array, 
			    unsigned string_length,
			    boolean_t (*proc)(unsigned string_length,
					      const void *string,
					      unsigned frequency,
					      void *parameters),
			    void *parameters);
           /* Effect: call proc on every distinct string in corpus */
           /*         of given length along with its frequency     */
           /*         until proc() returns FALSE                   */ 

/************************************************************
 *  INPUT/OUTPUT
 ************************************************************/

suffix_array_t *suffix_array_fscanf(FILE *stream);
               /* Argument: number of bytes per symbol, ascii stream  */
void           suffix_array_fprintf(const suffix_array_t *array, FILE *stream);
               /* Argument: number of bytes per symbol, ascii stream  */

suffix_array_t *suffix_array_read(FILE *stream);
               /* Argument: number of bytes per symbol, binary stream */
void           suffix_array_write(const suffix_array_t *array, FILE *stream);
               /* Argument: number of bytes per symbol, binary stream */

#endif

