/*
 * (c) 1997 Eric Sven Ristad.  All rights reserved.
 *
 */

#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <math.h>    /* should define nextafter() */
#include <string.h>  /* for strlen() */
#include <getopt.h>  /* for getopt() */
#include "common.h"
#include "memory.h"
#include "vector.h"

static char L_make_rcsid[] = "$Id: L.make.c,v 1.5 1997/10/22 08:56:35 libpa Exp $";
static char L_make_version[] = "$Revision: 1.5 $ $Date: 1997/10/22 08:56:35 $";

#define DEBUG           FALSE
#define EXTRA_PRECISE   TRUE
#define BANNER          "/* (c) 1997 Eric Sven Ristad.  All rights reserved. */\n"

#define MIN_BASE        1.0
#define MAX_BASE        1.2
#define MIN_TABLE_SIZE  8
#define MAX_FILENAME    256
#define MAX_EXTENSION   9
#define MAX_LINE        512

#ifndef M_LN2
#define M_LN2           6.9314718055994530942E-1
#endif
#define LN(x)           (log(x))
#define LG(x)           (log(x)/M_LN2)

extern double pow(double, double);
extern double expm1(double);
extern double log1p(double);
extern double nextafter(double, double);

/* DESCRIPTION.
 *   This executable makes new L_t systems in IEEE-compliant environments.
 *   (Unfortunately, linux 2.0 is not currently IEEE-compliant.)
 *
 *   To make a new L_t system using this executable, you must either
 *      1. specify the base that you want and the name of your system to
 *         have L.make create the necessary files with the desired
 *         values, for example, "L.make -b 1.01 b01" creates the files
 *         L.b1.[ch] for an L_t system whose base is 1.01.
 *      2. specify the table size that you want and the name of your system
 *         to have L.make create the necessary files with the desired
 *         values, for example, "L.make -s 1024 1k" creates the files
 *         L.1k.[ch] for an L_t system with the smallest base possible
 *         to require a table with exactly 1024 entries.
 *
 *   Each L_t system is determined by the value of L_BASE and must
 *   also specify the corresponding values for the following symbols
 *        L_LNBASE, L_LGBASE, L_ROUNDUP, L_TABLE_SIZE, L_delta, L_*_Table[]
 *
 *   The following additional symbols are also defined in an L_t system,
 *   to specify the range of probability values that each fixed-point
 *   type (Lchar_t, Lshort_t, Lint_t) represents
 *        LCHAR_MIN, LSHRT_MIN, LINT_MIN
 *        LCHAR_MIN_EXP, LSHRT_MIN_EXP, LINT_MIN_EXP
 *   but these values are only included if they are representable by
 *   double and long, respectively.
 *
 */

/* NOTES
 *  . how does truncation affect precision? is there a better way?
 *  . support value specification
 *  . support interpolation for n-fold reduction in table size
 *
 */
#ifdef NDEBUG
#undef NDEBUG
#error __FILE__ must not be compiled with NDEBUG
#endif

int main(int argc, char *argv[])
{
  extern int	optind;
  extern char	*optarg;
  int           is_error, opt;
  double   base, lgbase, lnbase, delta, value, *table;
  double   roundup_scalar, roundup_increment;
  double   lower_base, upper_base, penultimate_delta, ultimate_delta;
  double   minimum_values[5];
  char    *name, filename[MAX_FILENAME];
  int      i, table_size, target_table_size;
  unsigned l_half;
  FILE  *stream;

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  is_error = FALSE;
  base = 0;
  target_table_size = 0;
  for(i = 0; i < 5; i++) { minimum_values[i] = 0; }
  while ((opt = getopt(argc, argv, "b:s:1:2:4:")) != -1) 
    {
      switch (opt) 
	{
	case 'b':
	  base = atof(optarg);
	  if ((base <= MIN_BASE) || (base >= MAX_BASE))
	    { 
	      fprintf(stderr, "%s: infeasible base %e\n", argv[0], base); 
	      is_error = TRUE; 
	    }
	  break;
	case 's':
	  target_table_size = atoi(optarg);
	  if (target_table_size < MIN_TABLE_SIZE)
	    { 
	      fprintf(stderr, "%s: impossible table size %d\n", argv[0], target_table_size); 
	      is_error = TRUE; 
	    }
	  break;
	default:  is_error = TRUE; break;
	case '1': minimum_values[1] = atof(optarg); break;
	case '2': minimum_values[2] = atof(optarg); break;
	case '4': minimum_values[4] = atof(optarg); break;
	}
    }

  name = NULL;
  if ((argc - optind) > 0) { name = argv[0+optind]; }

  if ((is_error) || ((argc - optind) != 1) || (optind != 3) ||
      ((target_table_size != 0) && (base != 0)) ||
      ((target_table_size == 0) && (base == 0)) ||
      (name == NULL) || (strlen(name) < 1))
    {
      /* printf("argc %d, optind %d\n", argc, optind); */
      printf("Usage:\t%s [-b base | -s size] name\n", argv[0]);
      return(-1);
    }
  assert(name); assert(strlen(name) > 0);
  assert(strlen(name)+MAX_EXTENSION < MAX_FILENAME);

  /* binary search for desired base in [MIN_BASE,MAX_BASE] */
  if (target_table_size != 0)
    {
      assert(base == 0);
      assert(target_table_size >= MIN_TABLE_SIZE);

      lower_base = MIN_BASE;
      upper_base = MAX_BASE;
      assert(lower_base < upper_base);
      while(lower_base < upper_base)
	{
	  base = lower_base + ((upper_base - lower_base) / 2);
	  lnbase = LN(base);
	  assert(lower_base <= base); assert(base <= upper_base);

	  value = pow(base, -(target_table_size-1));
	  penultimate_delta = log1p(value) / lnbase;
#if DEBUG
	  printf("\t%e\t[%e,%e]\t%e", base, lower_base, upper_base, penultimate_delta);
#endif
	  if (penultimate_delta < 0.5)
	    { 
	      /* need more precision */
	      upper_base = nextafter(base, MIN_BASE); 
	      assert(lower_base <= upper_base);
	    }
	  else 
	    { 
	      assert(penultimate_delta >= 0.5);
	      value = pow(base, -target_table_size);
	      ultimate_delta = log1p(value) / lnbase;
#if DEBUG
	      printf("\t%e", ultimate_delta);
#endif
	      if (ultimate_delta >= 0.5)
		{ 
		  /* too much precision */
		  lower_base = nextafter(base, MAX_BASE); 
		  assert(lower_base <= upper_base);
		}
	      else 
		{ 
#if EXTRA_PRECISE
		  /* try to further increase precision */
		  lower_base = nextafter(lower_base, MAX_BASE);
		  upper_base = base;
#else
		  /* don't try to furth increase precision */
		  lower_base = MAX_BASE;
		  upper_base = MIN_BASE;
		  assert(!(lower_base < upper_base));
#endif
		}
	    }
#if DEBUG
	  printf("\n");
#endif
	}
      assert(lnbase == LN(base));
      assert((log1p(pow(base, -target_table_size)) / lnbase) < 0.5);
      assert((log1p(pow(base, -(target_table_size-1))) / lnbase) >= 0.5);
    }
  assert(MIN_BASE < base); assert(base < MAX_BASE);

  lnbase = LN(base);
  lgbase = LG(base);
  roundup_scalar = (2 * base) / (1 + base);
  roundup_increment = LN(roundup_scalar) / lnbase;
  l_half = log(2*roundup_scalar) / lnbase;
  assert(l_half <= ULONG_MAX);
  assert(lnbase < lgbase);
  assert(1 < roundup_scalar); assert(roundup_scalar < base);

  /* generate table from base alone */
  assert((base >= MIN_BASE) && (base <= MAX_BASE));
  table = vector_create_surplus(sizeof(double), 0, SHRT_MAX);
  assert(vector_is_valid(table));
  assert(vector_object_size(table) == sizeof(double));
      
  value = (1+1);
  delta = log(value) / lnbase;
  delta = floor(delta+roundup_increment);
  assert(delta >= 1);
  table_size = 0;
  while(delta >= 1)
    {
      assert(table_size < INT_MAX);
      table_size++;
      table = vector_append(table, &delta);
      assert(vector_is_valid(table));
      assert(vector_length(table) == table_size);
      assert(table[table_size-1] >= 1.0);
      assert((table_size == 1) || (table[table_size-2] >= table[table_size-1]));
      value = 1+pow(base, -table_size);
      delta = log(value) / lnbase;
      delta += roundup_increment;
      assert(delta > 0);
      delta = floor(delta);
      assert(delta >= 0);
    }
  assert(l_half == (unsigned)table[0]);
  assert((target_table_size == 0) || (table_size == target_table_size));

  /**************************************************
   * write header file 
   **************************************************/
  sprintf(filename, "L.%s.h", name);
  stream = fopen(filename, "w");
  if (stream == NULL)
    {
      fprintf(stderr, "%s: cannot open %s for writing\n", argv[0], filename);
      return(-1);
    }
  fprintf(stream, BANNER);
  fprintf(stream, "/* %s(%g, %d, %s) %s */\n\n", 
	  argv[0], base, table_size, name, L_make_version);

  fprintf(stream, "#ifndef L_%s_INCLUDED\n", name);
  fprintf(stream, "#define L_%s_INCLUDED\n\n", name);
  fprintf(stream, "#define L_BASE          %.32g\n", base);
  fprintf(stream, "#define L_LNBASE        %.16e\n", lnbase);
  fprintf(stream, "#define L_LGBASE        %.16e\n", lgbase);
  fprintf(stream, "#define L_ROUNDUP       %.16e\n", roundup_increment);
  fprintf(stream, "\n");

  if (((-(double)UCHAR_MAX)*lgbase) > DBL_MIN_EXP)
    { fprintf(stream, "#define LCHAR_MIN       %.16e\n", pow(base, -(double)UCHAR_MAX)); }
  if (((-(double)USHRT_MAX)*lgbase) > DBL_MIN_EXP)
    { fprintf(stream, "#define LSHRT_MIN       %.16e\n", pow(base, -(double)USHRT_MAX)); }
  if (((-(double)UINT_MAX)*lgbase) > DBL_MIN_EXP)
    { fprintf(stream, "#define LINT_MIN        %.16e\n", pow(base, -(double)UINT_MAX)); }
  fprintf(stream, "\n");

  assert(CHAR_MIN >= SHRT_MIN); assert(SHRT_MIN >= INT_MIN); assert(INT_MIN >= -DBL_MAX);
  if ((-(double)UCHAR_MAX * lgbase) >= INT_MIN)
    { fprintf(stream, "#define LCHAR_MIN_EXP   %d\n", (int)(lgbase * -(double)UCHAR_MAX)); }
  if ((-(double)USHRT_MAX * lgbase) >= INT_MIN)
    { fprintf(stream, "#define LSHRT_MIN_EXP   %d\n", (int)(lgbase * -(double)USHRT_MAX)); }
  if ((-(double)UINT_MAX * lgbase) >= INT_MIN)
    { fprintf(stream, "#define LINT_MIN_EXP    %d\n", (int)(lgbase * -(double)UINT_MAX)); }
  fprintf(stream, "\n");

  fprintf(stream, "#define L_TABLE_SIZE    %d\n", table_size);
  fprintf(stream, "#define L_delta(n)      (((n) < L_TABLE_SIZE) ? L_%s_Table[(n)] : 0)\n\n",
	  name);
  assert(l_half <= ULONG_MAX);
  if (l_half <= UCHAR_MAX)
    { fprintf(stream, "extern unsigned char  L_%s_Table[];\n\n", name); }
  else if (l_half <= USHRT_MAX)
    { fprintf(stream, "extern unsigned short L_%s_Table[];\n\n", name); }
  else if (l_half <= UINT_MAX)
    { fprintf(stream, "extern unsigned int   L_%s_Table[];\n\n", name); }
  else { fprintf(stream, "extern unsigned long  L_%s_Table[];\n\n", name); }

  fprintf(stream, "#include \"L.h\"\n\n");
  fprintf(stream, "#endif\n\n");
  fclose(stream); stream = NULL;

  /**************************************************
   * write source code
   **************************************************/
  assert(table_size > 1);
  sprintf(filename, "L.%s.c", name);
  stream = fopen(filename, "w");
  if (stream == NULL)
    {
      fprintf(stderr, "%s: cannot open %s for writing\n", argv[0], filename);
      return(-1);
    }
  fprintf(stream, BANNER);
  fprintf(stream, "/* %s(%g, %d, %s) %s */\n\n", 
	  argv[0], base, table_size, name, L_make_version);
  if (l_half <= UCHAR_MAX) { fprintf(stream, "unsigned char  L_%s_Table[] = {", name); }
  else if (l_half <= USHRT_MAX) { fprintf(stream, "unsigned short L_%s_Table[] = {", name); }
  else if (l_half <= UINT_MAX) { fprintf(stream, "unsigned int   L_%s_Table[] = {", name); }
  else { fprintf(stream, "unsigned long  L_%s_Table[] = {", name); }
  for(i = 0; i < table_size-1; i++)
    { 
      if ((i % MAX_LINE) == 0) { fprintf(stream, "\\\n\t"); }
      fprintf(stream, "%lu,", (unsigned long)table[i]); 
    }
  fprintf(stream, "%lu};\n\n", (unsigned long)table[table_size-1]);
  fclose(stream); stream = NULL;

  /**************************************************
   * write test suite
   **************************************************/
  sprintf(filename, "L.%s.test.c", name);
  stream = fopen(filename, "w");
  if (stream == NULL)
    {
      fprintf(stderr, "%s: cannot open %s for writing\n", argv[0], filename);
      return(-1);
    }
  fprintf(stream, BANNER);
  fprintf(stream, "/* %s(%g, %d, %s) %s */\n\n", 
	  argv[0], base, table_size, name, L_make_version);
  fprintf(stream, "#include \"L.%s.h\"\n", name);
  fprintf(stream, "#include \"L.test.c\"\n\n");
  fclose(stream); stream = NULL;

  assert(vector_is_valid(table));
  vector_destroy(table); table = NULL;
  assert(memory_total_bytes() == 0); 
  assert(memory_total_blocks() == 0);

  return(0);
}






