/*
 * (c) 1997 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <errno.h>
#include <signal.h>
#include "vector.h"

static char L_test_rcsid[] = "$Id: L.test.c,v 1.16 1997/10/26 00:40:17 libpa Exp $";

#define DIVERGENCE_SAMPLES       1000
#define CONVERSION_SAMPLES       100000
#define ARITHMETIC_SAMPLES       100000
#define CODELENGTH_SAMPLES       100000
#define MAXIMUM_ALPHABET         16
#define MAX_DIVERGENCE           1e-2
#define MAX_DISCREPANCY          1e-2
#define MIN_PRECISION            8.0
#define INTEL_TOLERANCE          1e-15

#define DEBUG_INTEL              FALSE
#define DEBUG_SAMPLE             FALSE
#define DEBUG_ARITHMETIC         FALSE
#define DEBUG_CODELENGTH         FALSE
#define TRACK_PRECISION          FALSE
#define SEED                     28090069
#ifndef M_LN2
#define M_LN2                    6.9314718055994530942E-1
#endif
#ifndef M_LOG2E
#define M_LOG2E                  1.4426950408889634074E0
#endif
#define LN(x)                    (log(x))
#define LG(x)                    (log(x)/M_LN2)
#define MIN_PR                   1e-20
#define PR_EPSILON               1e-10
#ifndef ESUCCESS
#define ESUCCESS                 0
#endif

extern  double drand48(void);
extern  void   srand48(long);
extern  double expm1(double);
extern  double log1p(double);
extern  int    finite(double);

static double Minimum_Precision = DBL_MANT_DIG;
static double Total_Precision = 0;
static double Conversion_Count = 0;

static boolean_t L_Is_Raised = FALSE;
void L_test_handler(int sig) 
{ 
  L_Is_Raised = TRUE;
  printf(" L_SIGNAL");
  (void) L_signal(&L_test_handler);
}

void pr_sample(double *pr)
{
  int i, alphabet_size;
  double total_pr;

  alphabet_size = vector_length(pr);
  total_pr = 0;
  for(i = 0; i < alphabet_size; i++)
    { pr[i] = drand48() + MIN_PR; pr[i] = pr[i] * pr[i]; total_pr += pr[i]; }
  for(i = 0; i < alphabet_size; i++)
    { pr[i] /= total_pr; }
}

void reset_precision(void)
{
  Minimum_Precision = DBL_MANT_DIG;
  Total_Precision = 0;
  Conversion_Count = 0;
}

double precision(double x, double y)
{
  double relative_error, bits_precision;
  assert(errno != EDOM); assert(errno != ERANGE);

  if (!finite(x) || !finite(y)) 
    { bits_precision = ((x == y) ? DBL_MANT_DIG : 0); }
  else if ((x != 0) && (y != 0))
    {
      relative_error = fabs(x - y) / fabs(x);
      assert(relative_error >= 0); assert(relative_error <= 1);
      if (relative_error == 0) 
	{ bits_precision = DBL_MANT_DIG; }
      else 
	{ 
	  bits_precision = -LG(relative_error); 
	  if (bits_precision > DBL_MANT_DIG) { bits_precision = DBL_MANT_DIG; }
	}
    }
  else { bits_precision = DBL_MANT_DIG; }
  assert(0 <= bits_precision); assert(bits_precision <= DBL_MANT_DIG);
  if (bits_precision < Minimum_Precision)
    { 
#if TRACK_PRECISION
      printf("precision(%e,%e) => %.2f bits\n", x, y, bits_precision);
#endif
      Minimum_Precision = bits_precision; 
    }
  Total_Precision += bits_precision;
  Conversion_Count++;
  assert(errno != EDOM); assert(errno != ERANGE);
  return(bits_precision);
}

double double_sample(void)
{
  unsigned l_sample;
  double sample;

  l_sample = drand48() * -((double)(DBL_MIN_EXP+1) / L_LGBASE);
  assert(l_sample < UINT_MAX);
  sample = L2double(l_sample);
  assert(finite(sample));
#if DEBUG_SAMPLE
  printf("sample(%u) = %e\n", l_sample, sample);
#endif
  return(sample);
}

boolean_t dbl_are_eq(double x, double y) { return(x == y); }

int main(int argc, char **argv)
{
  double dw, dx, dy, dz;
  Lint_t lw, lx, ly, lz;
  int    k, t, i;
  double divergence, discrepancy, max_divergence, max_discrepancy;
  double increment, result, total_pr, truncated_pr, *pr;
#ifndef NDEBUG
  void (*handler)(int);
#endif

  srand48(SEED);
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("\n\tTesting L_t library with base %g (%d entries)...\n", L_BASE, L_TABLE_SIZE);
  if (errno != ESUCCESS) 
    { fprintf(stderr, "\tWarning: errno %d != ESUCCESS on start up\n", (int)errno); }
  errno = ESUCCESS;

  /* verify central constants */
#ifdef __i386__
  /* due to known Intel bug re extra precision for doubles */
  assert(fabs(L_LNBASE - LN(L_BASE)) < INTEL_TOLERANCE);
  assert(fabs(L_LGBASE - LG(L_BASE)) < INTEL_TOLERANCE);
#ifdef L_ROUNDUP
  result = LN((2 * L_BASE) / (1 + L_BASE)) / L_LNBASE;
  if (fabs(L_ROUNDUP - result) >= INTEL_TOLERANCE) { 
#if DEBUG_INTEL
    fprintf(stderr, "\n\t|L_ROUNDUP=%.16e - %.16e| = %.16e >= %.16e\n",
	    L_ROUNDUP, result, fabs(L_ROUNDUP - result), INTEL_TOLERANCE);
#endif
    printf("\tverified constants L_LNBASE L_LGBASE approximately.\n");
    printf("\tWarning: failed to verify L_ROUNDUP constant!\n");
  } else {
    assert(fabs(L_ROUNDUP - result) < INTEL_TOLERANCE);
    printf("\tverified constants L_LNBASE L_LGBASE L_ROUNDUP approximately.\n");
  }
#else
  printf("\tverified constants L_LNBASE L_LGBASE approximately.\n");
#endif
#else
  assert(L_LNBASE == LN(L_BASE));
  assert(L_LGBASE == LG(L_BASE));
#ifdef L_ROUNDUP
  assert(L_ROUNDUP == LN((2 * L_BASE) / (1 + L_BASE)) / L_LNBASE);
  printf("\tverified constants L_LNBASE L_LGBASE L_ROUNDUP exactly.\n");
#else
  printf("\tverified constants L_LNBASE L_LGBASE exactly.\n");
#endif
#endif

  /* verify L_Max[] table */
  assert(L_Max[0] == 0);
  assert(L_Max[sizeof(unsigned char)] == UCHAR_MAX);
  assert(L_Max[sizeof(unsigned short)] == USHRT_MAX);
  assert(L_Max[sizeof(unsigned int)] == UINT_MAX);
  assert(L_Max[sizeof(unsigned long)] == ULONG_MAX);
  printf("\tverified L_Max[] global variable.\n");

  for(i = 0; i < L_TABLE_SIZE; i++)
    { 
      assert(L_delta(i) > 0); 
      assert(L_delta(i) == (unsigned)L_truncate(L_delta_raw(i)));
    }
  assert(errno != EDOM); assert(errno != ERANGE);
  printf("\tverified L_delta() macro with L_*_Table[] variable.\n");

#ifndef NDEBUG
  /* verify that safe versions raise exceptions */
  handler = L_signal(&L_test_handler);
  assert(handler != SIG_ERR);
  printf("\traising 6 exceptions:");

  assert(!L_Is_Raised);
  lx = double2L(2.0);
  assert(L_Is_Raised);
  L_Is_Raised = FALSE;

  lx = double2L(-0.5);
  assert(L_Is_Raised);
  L_Is_Raised = FALSE;

  lx = double2L(1.0+DBL_EPSILON);
  assert(L_Is_Raised);
  L_Is_Raised = FALSE;

  lx = double2L(-DBL_EPSILON);
  assert(L_Is_Raised);
  L_Is_Raised = FALSE;

  lx = UINT_MAX; ly = UINT_MAX; lz = L_multiply(lx, ly);
  assert(L_Is_Raised);
  L_Is_Raised = FALSE;

  lx = double2L(0.99); ly = double2L(0.01); lz = L_divide(lx, ly);
  assert(L_Is_Raised);
  L_Is_Raised = FALSE;

  L_signal(SIG_DFL); /* restore default behavior */
  assert(errno != EDOM); assert(errno != ERANGE);
  printf("\n\tverified L_t exception handling and restored default handler\n");
#else
  printf("\tWARNING: cannot verify L_t exception handling in unsafe version\n");
#endif

  /* verify extreme complementation */
  assert(L_are_equal(L_complement(L_UNITY), L_ZERO));
  assert(errno != EDOM); assert(errno != ERANGE);
  assert(L_are_equal(L_complement(L_ZERO), L_UNITY));
  assert(errno != EDOM); assert(errno != ERANGE);
  printf("\tverified extremal complementation ~0 == 1, ~1 == 0\n");

  /* verify exponent representational range */
  lw = L_UNITY;
  lx = 1;
  for(i = 0; i < USHRT_MAX; i++)
    { 
      assert(lw == i); 
      lw = L_multiply(lw, lx); 
    }
  if (lw != USHRT_MAX) { abort(); }
  for(i = 0; i < USHRT_MAX; i++)
    { 
      assert(lw == USHRT_MAX - i); 
      lw = L_divide(lw, lx); 
    }
  if (lw != L_UNITY) { abort(); }
  assert(errno != EDOM); assert(errno != ERANGE);
  printf("\tverified representational range [0,%u]\n", (unsigned)USHRT_MAX);

  /* calculate divergence and discrepancy for random distributions */
  printf("\tgenerating %d distributions per alphabet size:\n", (int)DIVERGENCE_SAMPLES);
  printf("\t\tk\tmax diverge\tmax discrep\tmean precision\tmin precision\n");
  for(k = 2; k < MAXIMUM_ALPHABET; k++)
    {
      reset_precision();
      max_divergence = 0;
      max_discrepancy = 0;
      pr = vector_create(sizeof(double), k);
      assert(vector_is_valid(pr));
      for(t = 0; t < DIVERGENCE_SAMPLES; t++)
	{
	  pr_sample(pr);

	  total_pr = 0;
	  divergence = 0;
	  discrepancy = -1;
	  assert(vector_length(pr) == k);
	  for(i = 0; i < k; i++)
	    {
	      total_pr += pr[i];  
	      assert(0 <= pr[i]); assert(pr[i] <= 1);
	      assert(0 <= total_pr); assert(total_pr <= 1+PR_EPSILON);
	      if (pr[i] > 0)
		{
		  lx = double2L(pr[i]);
		  truncated_pr = L2double(lx);
		  assert(0 <= truncated_pr); assert(truncated_pr <= 1);
		  (void) precision(pr[i], truncated_pr);
		  discrepancy += truncated_pr;
		  divergence += pr[i] * log( truncated_pr / pr[i] );
		}
	    }
	  assert(fabs(total_pr - 1.0) < PR_EPSILON);
	  if (fabs(divergence) > fabs(max_divergence)) { max_divergence = divergence; }
	  if (fabs(discrepancy) > fabs(max_discrepancy)) { max_discrepancy = discrepancy; }
	}
      vector_destroy(pr); pr = NULL;
      printf("\t\t%d\t%+e\t%+e\t%e\t%e\n", k, max_divergence, max_discrepancy,
	     Total_Precision/Conversion_Count, Minimum_Precision);
      if (fabs(max_divergence) > MAX_DIVERGENCE) { abort(); }
      if (fabs(max_discrepancy) > MAX_DISCREPANCY) { abort(); }
      if (Minimum_Precision < MIN_PRECISION)
	{ fprintf(stderr, "L_t error: inadequate precision\n"); abort(); }
    }

  /* calculate precision for extreme DBL/FLT values */
  reset_precision();
  dw = FLT_MIN; dx = FLT_EPSILON; dy = DBL_MIN*2; dz = DBL_EPSILON;
  lw = double2L(dw); lx = double2L(dx); ly = double2L(dy); lz = double2L(dz);
  printf("\tprecision: FLT_MIN %.2f  FLT_EPSILON %.2f  DBL_MIN*2 %.2f  DBL_EPSILON %.2f\n",
	 precision(dw, L2double(lw)), precision(dx, L2double(lx)),
	 precision(dy, L2double(ly)), precision(dz, L2double(lz)));
  if (Minimum_Precision < MIN_PRECISION)
    { fprintf(stderr, "L_t error: inadequate precision\n"); abort(); }

  /* verify precision with respect to finer base */
  increment = 1.0/(double)CONVERSION_SAMPLES;
  reset_precision();
  dx = 1.0;
  for(i = 1; i < CONVERSION_SAMPLES; i++)
    {
      dx -= increment;
      assert(0 < dx); assert(dx < 1);
      lx = double2L(dx);
      truncated_pr = L2double(lx);
      assert(0 < truncated_pr); assert(truncated_pr <= 1);
      (void) precision(dx, truncated_pr);
    }
  printf("\tverified %d conversions in [1,0]:\n", (int)CONVERSION_SAMPLES);
  printf("\t\tmean precision %.2f bits", Total_Precision/Conversion_Count);
  printf(", minimum precision %.2f bits\n", Minimum_Precision);
  if (errno != ESUCCESS) { fprintf(stderr, "errno %d\n", (int)errno); abort(); }
  if (Minimum_Precision < MIN_PRECISION) 
    { fprintf(stderr, "L_t error: inadequate precision\n"); abort(); }

  /* verify arithmetic operations by sampling */
  reset_precision();
  for(i = 0; i < ARITHMETIC_SAMPLES; i++)
    {
      dx = double_sample(); lx = double2L(dx);
      dy = double_sample(); ly = double2L(dy);
#if DEBUG_ARITHMETIC
      printf("\t%e\t%u\t%s\t%e\t%u\n", 
	     dx, ly, ((dx < dy) ? "<" : ((dx > dy) ? ">" : "=")), dy, ly);
#endif

      if (dx < dy)
	{
	  assert(L_compare(lx, ly) < 0);
	  assert(!L_are_equal(lx, ly));
	  assert(!L_is_greater(lx, ly));
	  assert(L_is_less(lx, ly));
	  assert(!L_is_geq(lx, ly));
	  assert(L_is_leq(lx, ly));
	}
      else if (dx > dy)
	{
	  assert(L_compare(lx, ly) > 0);
	  assert(!L_are_equal(lx, ly));
	  assert(L_is_greater(lx, ly));
	  assert(!L_is_less(lx, ly));
	  assert(L_is_geq(lx, ly));
	  assert(!L_is_leq(lx, ly));
	}
      else
	{
	  assert(L_compare(lx, ly) == 0);
	  assert(L_are_equal(lx, ly));
	  assert(!L_is_greater(lx, ly));
	  assert(!L_is_less(lx, ly));
	  assert(L_is_geq(lx, ly));
	  assert(L_is_leq(lx, ly));
	}

      dz = 1 - dx;
      lz = L_complement(lx);
#if DEBUG_ARITHMETIC
      printf("~\t%e\t%u\n", dz, lz);
#endif
      (void) precision(dz, L2double(lz));

      result = dx + dy;
      if ((0 < result) && (result <= 1))
	{
	  dz = dx + dy; 
	  lz = L_add(lx, ly);
#if DEBUG_ARITHMETIC
	  printf("+\t%e\t%u\n", dz, lz);
#endif
	  (void) precision(dz, L2double(lz));
	}

      result = log10(dx) + log10(dy);
      if ((DBL_MIN_10_EXP < result) && (result < 0))
	{
	  dz = dx * dy; 
	  lz = L_multiply(lx, ly);
#if DEBUG_ARITHMETIC
	  printf("*\t%e\t%u\n", dz, lz);
#endif
	  (void) precision(dz, L2double(lz));
	}

      result = log10(dx) - log10(dy);
      if ((DBL_MIN_EXP < result) && (result < 0))
	{
#ifdef __i386__
	  /* due to known Intel bug re extra precision in doubles */
	  assert(dx <= dy); assert(lx >= ly);
#else
	  assert(dx < dy); assert(lx > ly);
#endif
	  dz = dx / dy; 
	  lz = L_divide(lx, ly);
#if DEBUG_ARITHMETIC
	  printf("/\t%e\t%u\n", dz, lz);
#endif
	  (void) precision(dz, L2double(lz));
	}
    }
  printf("\tverified arithmetic operations on %d pairs:\n", ARITHMETIC_SAMPLES);
  printf("\t\tmean precision %.2f bits", Total_Precision/Conversion_Count);
  printf(", minimum precision %.2f bits\n", Minimum_Precision);
  if (errno != ESUCCESS) { fprintf(stderr, "errno %d\n", (int)errno); abort(); }
  if (Minimum_Precision < MIN_PRECISION) 
    { fprintf(stderr, "L_t error: inadequate precision\n"); abort(); }

  reset_precision();
  for(i = 0; i < CODELENGTH_SAMPLES; i++)
    {
      dx = drand48() * (double)(DBL_MIN_EXP+1);
      assert(DBL_MIN_EXP < dx); assert(dx <= 0);
      lx = bits2L(-dx);
      dy = pow(2, dx); dz = L2double(lx);
#if DEBUG_CODELENGTH
      printf("pow(2,%e)\t%e\t%e\n", dx, dy, dz);
#endif
      precision(dy, dz);

      dx = dx / M_LOG2E;
      lx = nats2L(-dx);
      dy = exp(dx); dz = L2double(lx);
#if DEBUG_CODELENGTH
      printf("exp(%e)\t%e\t%e\n", dx, dy, dz);
#endif
      precision(dy, dz);
    }
  if (errno != ESUCCESS) { fprintf(stderr, "errno %d\n", (int)errno); abort(); }
  printf("\tverified codelength coversions on %d pairs:\n", CODELENGTH_SAMPLES);
  printf("\t\tmean precision %.2f bits", Total_Precision/Conversion_Count);
  printf(", minimum precision %.2f bits\n", Minimum_Precision);
  if (Minimum_Precision < MIN_PRECISION) 
    { fprintf(stderr, "L_t error: inadequate precision\n"); abort(); }

  printf("\tL_t library passes all tests for base %g with %d entries.\n", 
	 L_BASE, L_TABLE_SIZE);
  return(0);
}


