/*
 * (c) 1996 by Eric Sven Ristad and Peter N. Yianilos.  All rights reserved.
 *
 */

#include <assert.h>
#include <stdio.h> 
#include <stdlib.h>          /* for abort() only */
#include <math.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include "balanced.h"

static char balanced_rcsid[] = "$Id: balanced.c,v 1.10 1997/06/05 21:44:38 ristad Exp $";

#define ANNOUNCE_RCSID      FALSE
#define DEBUG_CONVERSION    FALSE

/* Note: this choice seems to have no effect on precision.  */
#define MAXIMUM_ACCURACY    FALSE
#if MAXIMUM_ACCURACY
 extern double expm1(double);
 extern double log1p(double);
 #define EXP(x)        (expm1(x)+1)
 #define LOG(x)        (log1p((x)-1))
#else
 #define EXP(x)        (exp(x))
 #define LOG(x)        (log(x))
#endif

#ifndef DBL_INFINITY
  #define DBL_INFINITY    (DBL_MAX*2)
#endif
#ifndef ESUCCESS
  #define ESUCCESS        0
#endif

extern int   kill(pid_t, int);
extern pid_t getpid(void);
extern int   finite(double);

double Balanced_Table[BALANCED_PRECISION+1];  /* can't be static */
static boolean_t Balanced_Is_Initialized = FALSE;


/************************************************************
 * Initialization
 ************************************************************/

static void balanced_initialize(void)
{
  double   x;
  int      i;

  assert(1.0 == BALANCED_RADIX * BALANCED_MANT_LOW);

  x = 1.0;
  for (i = 0; i <= BALANCED_PRECISION; i++)
    { Balanced_Table[i] = x;  x /= BALANCED_RADIX;  }
  Balanced_Is_Initialized = TRUE;

#if (FLT_RADIX != 2)
#error FLT_RADIX != 2
#endif
#if (BALANCED_RADIX != FLT_RADIX)
#error BALANCED_RADIX != FLT_RADIX
#endif
#if (BALANCED_PRECISION != FLT_MANT_DIG)
#error BALANCED_PRECISION != FLT_MANT_DIG
#endif
#if (INT_MAX < 2147483647)
#error INT_MAX < 2147483647
#endif

#if ANNOUNCE_RCSID
  fprintf(stderr, "\t%s", balanced_rcsid);
#ifdef NDEBUG
  fprintf(stderr, "\t>unsafe<");
#endif
  fprintf(stderr, "\n");
#endif
}

/************************************************************
 * Constructors
 ************************************************************/

balanced_t double2balanced(double x)
{
  balanced_t bx;
  double sig;

  if (!Balanced_Is_Initialized) { balanced_initialize(); }
  assert(finite(x)); assert(errno != EDOM); assert(errno != ERANGE); 

  if (x == 0) 
    { 
      bx.sig = x;       /* preserve sign */
      bx.exp = BALANCED_MIN_EXP; 
    } 
  else 
    { 
      sig = frexp(x, &bx.exp); 
      bx.sig = sig;
      if (fabs(bx.sig) == BALANCED_MANT_HIGH) /* conversion from double to float */
	{ 
	  assert(bx.exp < BALANCED_MAX_EXP);  /* cannot happen! */
	  bx.sig = sig / BALANCED_RADIX; 
	  bx.exp++; 
	}
    }
  assert(balanced_is_valid(bx));
#if DEBUG_CONVERSION
  printf("\tdouble2balanced(%e) = %e,%d\n", x, (double)bx.sig, bx.exp);
#endif
  assert(errno != EDOM); assert(errno != ERANGE);
  return(bx);
}

double balanced2double(balanced_t bx) 
{ 
  double x;
  assert(errno != EDOM); assert(errno != ERANGE);
  x = ldexp((double)bx.sig, bx.exp); 
  assert(errno != EDOM);
  if (errno == ERANGE) 
    { 
      assert((bx.exp < DBL_MIN_EXP) || (bx.exp > DBL_MAX_EXP));
      if (bx.exp < 0) { x = ((bx.sig < 0) ? -0 : +0); }
      else { x = ((bx.sig < 0) ? -DBL_INFINITY : DBL_INFINITY); }
      errno = ESUCCESS;  /* reset the error */
    }
  assert(errno != EDOM); assert(errno != ERANGE);
#if DEBUG_CONVERSION
  printf("\tbalanced2double(%e,%d) = %e\n", (double)bx.sig, bx.exp, x);
#endif
  return(x);
}

balanced_t balanced_zero(void) { return(double2balanced(0.0)); }
balanced_t balanced_unity(void) { return(double2balanced(1.0)); }

/************************************************************
 * Predicates
 ************************************************************/

boolean_t balanced_is_zero(balanced_t bx)   
{ return((bx.sig == 0) && (bx.exp == BALANCED_MIN_EXP)); }
boolean_t balanced_is_unity(balanced_t bx)  
{ return((bx.sig == 0.5) && (bx.exp == 1)); }
boolean_t balanced_is_pr(balanced_t bx)
{ return(((bx.sig >= 0.5) && (bx.sig < 1.0) && (bx.exp <= 0)) ||
	 ((bx.sig == 0) && (bx.exp == BALANCED_MIN_EXP)) ||
	 ((bx.sig == 0.5) && (bx.exp == 1))); }
boolean_t balanced_is_valid(balanced_t bx)
{
  return(((BALANCED_MANT_LOW <= fabs(bx.sig)) && (fabs(bx.sig) < BALANCED_MANT_HIGH)) ||
	 ((bx.sig == 0) && (bx.exp == BALANCED_MIN_EXP)));
}

boolean_t balanced_is_probability(balanced_t bx, balanced_t epsilon)
{
  balanced_t bv;
  int cmp;

  cmp = balanced_exact_compare(bx, epsilon);
  if (cmp < 0)
    {
      bv = balanced_zero();
      bx = balanced_add(bx, epsilon);
      return(balanced_exact_compare(bv, bx) <= 0);
    }
  else if (cmp > 0)
    {
      bv = balanced_unity();
      bv = balanced_add(bv, epsilon);
      return(balanced_exact_compare(bx, bv) <= 0);
    }
  else /* cmp == 0 */
    {
      return(TRUE);
    }
}


/************************************************************
 * Comparison
 ************************************************************/

int balanced_compare(balanced_t bx, balanced_t by, balanced_t tolerance)
{
  balanced_t bz;
  int sign;

  bz = balanced_subtract(bx, by);
  sign = (bz.sig < 0) ? -1 : 1;
  bz.sig = fabs(bz.sig);
  if (balanced_exact_compare(bz, tolerance) <= 0) { return(0); }
  else { assert(bz.sig != 0); return(sign); }
}

int balanced_exact_compare(balanced_t bx, balanced_t by)
{
  if (bx.exp == by.exp)
    { 
      if (bx.sig == by.sig) { return(0); }
      else { return( (bx.sig < by.sig) ? -1 : +1 ); }
    }
  else if (bx.sig < 0)
    {
      if (by.sig >= 0) { return(-1); }
      assert(bx.sig < 0); assert(by.sig < 0);
      return((bx.exp < by.exp) ? +1 : -1);
    }
  else
    {
      assert(bx.sig >= 0); 
      assert((bx.sig != 0) || (by.sig != 0));
      if (by.sig < 0) { return(+1); }
      assert(bx.sig >= 0); assert(by.sig >= 0);
      return((bx.exp < by.exp) ? -1 : +1);
    }
}

/************************************************************
 * Arithmetic
 ************************************************************/

balanced_t balanced_add(balanced_t bx, balanced_t by)
{
  balanced_t bz;
  unsigned delta;

  assert(Balanced_Is_Initialized);
  assert(balanced_is_valid(bx));
  assert(balanced_is_valid(by));

  if (bx.exp >= by.exp) 
    {
      delta = (unsigned) (bx.exp - by.exp);
      bz.exp = bx.exp;
      bz.sig = bx.sig;
      if (delta <= BALANCED_PRECISION)    /* could smaller affect larger? */
	{ bz.sig += by.sig * Balanced_Table[delta]; }
    } 
  else
    {
      delta = (unsigned) (by.exp - bx.exp);
      bz.exp = by.exp;
      bz.sig = by.sig;
      if (delta <= BALANCED_PRECISION)
	{ bz.sig += bx.sig * Balanced_Table[delta]; }
    }
  assert(fabs(bz.sig) >= 0); assert(fabs(bz.sig) < 2);

  /* now balance significand */
  if (fabs(bz.sig) >= BALANCED_MANT_HIGH)
    {
      if (bz.exp == BALANCED_MAX_EXP) { kill(getpid(), SIGFPE); }
      bz.sig /= BALANCED_RADIX;
      bz.exp++;
    }
  else if (bz.sig == 0) 
    { bz.exp = BALANCED_MIN_EXP; }
  else
    {
      while(fabs(bz.sig) < BALANCED_MANT_LOW)
	{
	  if (bz.exp == BALANCED_MIN_EXP) { kill(getpid(), SIGFPE); }
	  bz.sig *= BALANCED_RADIX;
	  bz.exp--;
	}
    }
  assert(balanced_is_valid(bz));
  return(bz);
}

balanced_t balanced_subtract(balanced_t bx, balanced_t by)
{
  balanced_t bz, bw;
  bw.sig = -by.sig;
  bw.exp =  by.exp;
  bz = balanced_add(bx, bw);
  return(bz);
}

balanced_t balanced_multiply(balanced_t bx, balanced_t by)
{
  balanced_t bz;

  assert(balanced_is_valid(bx));
  assert(balanced_is_valid(by));

  bz.sig = bx.sig * by.sig;
  if (bz.sig == 0) 
    { 
      assert((bx.sig == 0) || (by.sig == 0));
      assert((bx.exp == BALANCED_MIN_EXP) || (by.exp == BALANCED_MIN_EXP));
      bz.exp = BALANCED_MIN_EXP; 
    } 
  else
    {
      if (((bx.exp > 0) && (by.exp > 0) && (bx.exp >= BALANCED_MAX_EXP - by.exp)) ||
	  ((bx.exp < 0) && (by.exp < 0) && (bx.exp <= BALANCED_MIN_EXP - by.exp)))
	{ kill(getpid(), SIGFPE); }
      bz.exp = bx.exp + by.exp;
      assert(fabs(bz.sig) >= BALANCED_MANT_LOW * BALANCED_MANT_LOW);
      assert(fabs(bz.sig) < BALANCED_MANT_HIGH);
      if (fabs(bz.sig) < BALANCED_MANT_LOW) 
	{
	  if (bz.exp == BALANCED_MIN_EXP) { kill(getpid(), SIGFPE); }
	  bz.sig *= BALANCED_RADIX;
	  bz.exp--;
	}
    }
  assert(balanced_is_valid(bz));
  return(bz);
}

balanced_t balanced_divide(balanced_t bx, balanced_t by)
{
  balanced_t bz;

  assert(balanced_is_valid(bx));
  assert(balanced_is_valid(by));

  bz.sig = bx.sig / by.sig;
  if (bz.sig == 0) 
    {
      assert(bx.exp == BALANCED_MIN_EXP);
      bz.exp = BALANCED_MIN_EXP;
    }
  else
    {
      if (((bx.exp > 0) && (by.exp < 0) && (bx.exp >= BALANCED_MAX_EXP + by.exp)) ||
	  ((bx.exp < 0) && (by.exp > 0) && (bx.exp <= BALANCED_MIN_EXP + by.exp)))
	{ kill(getpid(), SIGFPE); }
      bz.exp = bx.exp - by.exp;
      assert(fabs(bz.sig) >= BALANCED_MANT_LOW);
      if (fabs(bz.sig) >= BALANCED_MANT_HIGH) 
	{
	  if (bz.exp == BALANCED_MAX_EXP) { kill(getpid(), SIGFPE); }
	  bz.sig /= BALANCED_RADIX;
	  bz.exp++;
	} 
    }
  assert(balanced_is_valid(bz));
  return(bz);
}

balanced_t balanced_magnitude(balanced_t bx)
{
  balanced_t bz;
  bz.sig = fabs(bx.sig);
  bz.exp = bx.exp;
  return(bz);
}

/************************************************************
 * Vector Arithmetic
 ************************************************************/

void balanced_inner_step(balanced_t *bx_p, const balanced_t *by_p, const balanced_t *bz_p)
{
  balanced_t bw, bx;
  bx = *bx_p;
  bw = balanced_multiply(*by_p, *bz_p);
  *bx_p = balanced_add(bx, bw);
}

/************************************************************
 * Transcendental
 ************************************************************/

balanced_t balanced_power(balanced_t bx, double n)
{
  double fractional, whole, base, residue;
  balanced_t bz;

  assert(balanced_is_valid(bx));
  assert(errno != EDOM); assert(errno != ERANGE);

  fractional = modf(n * (double)bx.exp, &whole);
  residue = EXP(M_LN2 * fractional);
  base = pow((double)bx.sig, n) * residue;
  bz.sig = frexp(base, &(bz.exp));
  assert(errno != EDOM); assert(errno != ERANGE);

  if (bz.sig == 0) 
    { 
      bz.exp = BALANCED_MIN_EXP; 
    }
  else 
    { 
      if (((bz.exp > 0) && (whole > 0) && (bz.exp > BALANCED_MAX_EXP - (int)whole)) ||
	  ((bz.exp < 0) && (whole < 0) && (bz.exp <= BALANCED_MIN_EXP - (int)whole)))
	{ kill(getpid(), SIGFPE); }
      bz.exp += (int)whole; 
    }
  assert(balanced_is_valid(bz));
  return(bz);
}

balanced_t balanced_exp(double x)
{
  balanced_t bx;

  if (!Balanced_Is_Initialized) { balanced_initialize(); }
  assert(finite(x));

  if (x == 0) 
    { 
      bx = balanced_unity(); 
    }
  else
    {
      x = x / M_LN2;
      if ((BALANCED_MIN_EXP > floor(x)) || (x > BALANCED_MAX_EXP-1)) 
	{ kill(getpid(), SIGFPE); }
      bx.exp = floor(x)+1; 
      assert(x+1 >= bx.exp); assert((x - bx.exp) < 0);
      bx.sig = exp((x - bx.exp) * M_LN2);
      assert(0.5 <= bx.sig); assert(bx.sig < 1);
    }
  assert(balanced_is_valid(bx));
  return(bx);
}

double balanced_log(balanced_t bx)
{
  double a, b;
  assert(balanced_is_valid(bx));
  if ((bx.exp == BALANCED_MIN_EXP) && (bx.sig == 0))
    { 
      return(-DBL_INFINITY); 
    }
  else
    {
      /* Note: b = log1p(bx.sig - 1) does not improve accuracy */
      a = bx.exp * M_LN2;
      b = LOG(bx.sig);
      return(a+b);
    }
}

double balanced_log2(balanced_t bx)
{
  double a, b;
  assert(balanced_is_valid(bx));
  if ((bx.exp == BALANCED_MIN_EXP) && (bx.sig == 0))
    { 
      return(-DBL_INFINITY); 
    }
  else
    {
      /* Note: b = log1p(bx.sig - 1) does not improve accuracy */
      a = bx.exp;
      b = LOG(bx.sig) / M_LN2;
      return(a+b);
    }
}

/************************************************************
 * Input/Output
 ************************************************************/

int balanced_fprintf(balanced_t bx, FILE *stream)
{
  int success;
  success = fprintf(stream, "%fb%d", bx.sig, bx.exp);
  return((success > 0) ? 1 : 0);
}

int balanced_printf(balanced_t bx) 
{ 
  return(balanced_fprintf(bx, stdout)); 
}


