/*
 * (c) 1996 by Eric Sven Ristad and Peter N. Yianilos.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <errno.h>
#include "balanced.h"

static char balanced_test_rcsid[] = "$Id: balanced.test.c,v 1.9 1997/06/05 21:45:02 ristad Exp $";

#define ARITHMETIC_SAMPLES       1000000
#define TRANSCENDENTAL_SAMPLES   1000000
#define COMPARISON_SAMPLES       1000000

#define DEBUG_SAMPLE             FALSE
#define DEBUG_ARITHMETIC         FALSE
#define DEBUG_TRANSCENDENTAL     FALSE
#define TRACK_PRECISION          FALSE
#define SEED                     28090069
#define M_LG10                   3.32192809488736234783
#define LG(x)            (log(x)/M_LN2)
#define NEXTAFTERF(x,y)  (((y) > (x)) ? (x)+FLT_EPSILON : (x)-FLT_EPSILON)

#define MIN_ARITHMETIC_PRECISION       12
#define MIN_TRANSCENDENTAL_PRECISION   13
#define MIN_LOGARITHMIC_PRECISION      32

extern double drand48(void);
extern void   srand48(long);
extern int    finite(double);

static double Minimum_Precision = DBL_MANT_DIG;
static double Total_Precision = 0;
static double Conversion_Count = 0;


double relative_precision(double x, double y)
{
  double relative_error, bits_precision;
  assert(errno != EDOM); assert(errno != ERANGE);

  if (!finite(x) || !finite(y)) 
    { bits_precision = ((x == y) ? DBL_MANT_DIG : 0); }
  else if ((x != 0) && (y != 0))
    {
      relative_error = fabs(x - y) / fabs(x);
      assert(relative_error >= 0); assert(relative_error <= 1);
      if (relative_error == 0) 
	{ bits_precision = DBL_MANT_DIG; }
      else 
	{ 
	  bits_precision = -LG(relative_error); 
	  if (bits_precision > DBL_MANT_DIG) { bits_precision = DBL_MANT_DIG; }
	}
    }
  else { bits_precision = DBL_MANT_DIG; }
  assert(0 <= bits_precision); assert(bits_precision <= DBL_MANT_DIG);
  assert(errno != EDOM); assert(errno != ERANGE);
  return(bits_precision);
}

double track_precision(double x, double y)
{
  double bits_precision;
  bits_precision = relative_precision(x, y);
  if (bits_precision < Minimum_Precision)
    { 
#if TRACK_PRECISION
      printf("precision(%e,%e) => %.2f bits\n", x, y, bits_precision);
#endif
      Minimum_Precision = bits_precision; 
    }
  Total_Precision += bits_precision;
  Conversion_Count++;
  return(bits_precision);
}

double double_sample(void)
{
  double mantissa, sample;
  int exponent;
  assert(errno != EDOM); assert(errno != ERANGE);

  mantissa = drand48() - 0.5;
  mantissa += (mantissa < 0) ? -0.5 : 0.5;
  assert(-1 <= mantissa); assert(mantissa < 1);
  assert((mantissa < -0.5) || (mantissa >= 0.5));
  exponent = (drand48() * (double)(DBL_MAX_EXP - DBL_MIN_EXP)) + DBL_MIN_EXP;
  assert(DBL_MIN_EXP <= exponent); assert(exponent <= DBL_MAX_EXP);
  sample = ldexp(mantissa, exponent);
#if DEBUG_SAMPLE
  printf("sample(%e,%d) = %e\n", mantissa, exponent, sample);
#endif
  assert(errno != EDOM); assert(errno != ERANGE);
  assert(finite(sample));
  return(sample);
}

#ifdef NDEBUG
#undef NDEBUG
#error __FILE__ must not be compiled with NDEBUG
#endif

int main(int argc, char **argv)
{
  double dw, dx, dy, dz;
  balanced_t bw, bx, by, bz;
  balanced_t zero, unity, minus_unity;
  int i, min_i, max_i;
  double exponent;

  srand48(SEED);
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("\n\tTesting balanced_t library...\n");

  /* verify distinguished values */
  zero = balanced_zero();
  unity = balanced_unity();
  minus_unity = double2balanced(-1);
  assert(balanced_is_valid(zero));
  assert(balanced_is_valid(unity));
  assert(balanced_is_valid(minus_unity));
  assert(balanced2double(zero) == 0);
  assert(balanced2double(unity) == 1);
  assert(balanced2double(minus_unity) == -1);
  assert(balanced_is_pr(zero));
  assert(balanced_is_pr(unity));
  assert(!balanced_is_pr(minus_unity));
  assert(balanced_are_equal(zero, zero));
  assert(balanced_are_equal(unity, unity));
  assert(balanced_are_equal(minus_unity, minus_unity));
  assert(balanced_are_equal(zero, balanced_add(unity, minus_unity)));
  assert(balanced_are_equal(unity, balanced_add(zero, unity)));
  assert(balanced_are_equal(minus_unity, balanced_add(zero, minus_unity)));
  assert(balanced_are_equal(unity, balanced_subtract(unity, zero)));
  assert(balanced_are_equal(unity, balanced_subtract(zero, minus_unity)));
  assert(balanced_are_equal(minus_unity, balanced_subtract(zero, unity)));
  assert(balanced_are_equal(zero, balanced_multiply(zero, unity)));
  assert(balanced_are_equal(zero, balanced_multiply(zero, minus_unity)));
  assert(balanced_are_equal(zero, balanced_divide(zero, unity)));
  assert(balanced_are_equal(zero, balanced_divide(zero, minus_unity)));
  assert(balanced_exact_compare(minus_unity, zero) < 0);
  assert(balanced_exact_compare(zero, unity) < 0);
  assert(balanced_are_equal(zero, balanced_magnitude(zero)));
  assert(balanced_are_equal(unity, balanced_magnitude(unity)));
  assert(balanced_are_equal(unity, balanced_magnitude(minus_unity)));
  printf("\tverified distinguished values -1,0,+1 for arithmetic\n");

  /* verify conversion for values in [.5,1) */
  dw = NEXTAFTERF(0.5, 0); dx = 0.5; dy = NEXTAFTERF(1.0, 0); dz = 1.0; 
  assert(dw < dx); assert(dx < dy); assert(dy < dz);
  bw = double2balanced(dw); assert(balanced_is_valid(bw)); assert(bw.exp == -1); 
  bx = double2balanced(dx); assert(balanced_is_valid(bx)); assert(bx.exp ==  0); 
  by = double2balanced(dy); assert(balanced_is_valid(by)); assert(by.exp ==  0); 
  bz = double2balanced(dz); assert(balanced_is_valid(bz)); assert(bz.exp ==  1);
  assert(balanced2double(bw) == dw);
  assert(balanced2double(bx) == dx);
  assert(balanced2double(by) == dy);
  assert(balanced2double(bz) == dz);
  assert(balanced_is_pr(bw)); assert(balanced_is_pr(bx));
  assert(balanced_is_pr(by)); assert(balanced_is_pr(bz));
  printf("\tverified conversions in and around [0.5,1)\n");

  /* verify accumulation of small values around thresholds */
  dw = 0.5; dx = dw - NEXTAFTERF(0.5, 0); 
  assert(dx > 0); assert(dw - dx == NEXTAFTERF(0.5, 0));
  bw = double2balanced(dw); assert(bw.sig == 0.5); assert(bw.exp == 0);
  bx = double2balanced(dx); assert(balanced_is_valid(bx));
  assert(balanced_exact_compare(bw, bw) == 0);
  assert(balanced_exact_compare(bx, bx) == 0);
  assert(balanced_exact_compare(bw, bx) > 0);
  assert(balanced_exact_compare(zero, bx) < 0);
  by = balanced_subtract(bw, bx); assert(by.sig < 1); assert(by.exp == -1);
  assert(balanced_exact_compare(by, bw) < 0);
  assert(balanced_are_equal(bw, balanced_add(by, bx)));

  dx = NEXTAFTERF(0.5, 1) - dw;
  assert(dx > 0); assert(dw + dx == NEXTAFTERF(0.5, 1));
  assert(bw.sig == 0.5); assert(bw.exp == 0);
  bx = double2balanced(dx); assert(balanced_is_valid(bx));
  by = balanced_add(bw, bx); assert(by.sig > 0.5); assert(by.exp == 0);
  assert(balanced_exact_compare(bw, by) < 0);
  assert(balanced_are_equal(bw, balanced_subtract(by, bx)));
  printf("\tverified incremental arithmetic around 0.5 threshold\n");

  dw = 1.0; dx = dw - NEXTAFTERF(1.0, 0);
  assert(dx > 0); assert(dw - dx == NEXTAFTERF(1.0, 0));
  bw = double2balanced(dw); assert(bw.sig == 0.5); assert(bw.exp == 1);
  bx = double2balanced(dx); assert(balanced_is_valid(bx));
  by = balanced_subtract(bw, bx); assert(by.exp == 0);
  assert(balanced_exact_compare(by, bw) < 0);
  assert(balanced_are_equal(bw, balanced_add(by, bx)));

  dx = NEXTAFTERF(1.0, 2) - dw;
  assert(dx > 0); assert(dw + dx == NEXTAFTERF(1.0, 2));
  assert(bw.sig == 0.5); assert(bw.exp == 1);
  bx = double2balanced(dx); assert(balanced_is_valid(bx));
  by = balanced_add(bw, bx); assert(by.sig > 0.5); assert(by.exp == 1);
  assert(balanced_exact_compare(bw, by) < 0);
  assert(balanced_are_equal(bw, balanced_subtract(by, bx)));
  printf("\tverified incremental arithmetic around 1.0 threshold\n");

  /* verify exponent representational range */
  i = 1; bw = double2balanced(2);
  assert(bw.sig == 0.5); assert(bw.exp == 2);
  while(i < BALANCED_MAX_EXP/2)
    { 
      i *= 2; bw = balanced_multiply(bw, bw); 
      assert(balanced_is_valid(bw)); 
      assert(bw.sig == 0.5); assert(bw.exp == i+1);
    }
  max_i = i;
  i = -1; bw = double2balanced(0.5);
  assert(bw.sig == 0.5); assert(bw.exp == 0);
  while(i > BALANCED_MIN_EXP/2)
    {
      i *= 2; bw = balanced_multiply(bw, bw);
      assert(balanced_is_valid(bw)); 
      assert(bw.sig == 0.5); assert(bw.exp == i+1);
    }
  min_i = i;
  printf("\tverified vast exponent range [%d,%d]\n", min_i, max_i);

  /* verify table range */
  Minimum_Precision = DBL_MANT_DIG;
  Total_Precision = 0; Conversion_Count = 0;
  dx = 0.5; bx = double2balanced(0.5);
  by.sig = NEXTAFTERF(1.0, 0); by.exp = 0;
  dy = balanced2double(by);
  assert(balanced_is_valid(bx)); assert(balanced_is_valid(by));
  for(i = 0; i <= BALANCED_PRECISION; i++)
    {
      dz = dx + dy; bz = balanced_add(bx, by);
      assert(dx < dz); assert(balanced_exact_compare(bx, bz) < 0);
      track_precision(dz, balanced2double(bz));
      dy /= 2.0; by.exp--; 
    }
  dz = dx + dy; bz = balanced_add(bx, by);
  assert(dx < dz); assert(balanced_exact_compare(bx, bz) == 0);
  printf("\tverified addition throughout table range [0,%d]:\n", (int)BALANCED_PRECISION);
  printf("\t\tmean precision %.2f bits (%.1f digits)\n",
	 Total_Precision/Conversion_Count, (Total_Precision/Conversion_Count)/M_LG10);
  printf("\t\tminimum precision %.2f bits (%.1f digits)\n",
	 Minimum_Precision, Minimum_Precision/M_LG10);

  /* calculate precision for extreme DBL/FLT values */
  Minimum_Precision = DBL_MANT_DIG;
  Total_Precision = 0; Conversion_Count = 0;
  dw = FLT_MIN; dx = FLT_EPSILON; dy = FLT_MAX;
  bw = double2balanced(dw); bx = double2balanced(dx); by = double2balanced(dy);
  assert(balanced_is_valid(bw)); assert(balanced_is_valid(bx)); assert(balanced_is_valid(by));
  printf("\tprecision: FLT_MIN %.1f  FLT_EPSILON %.1f  FLT_MAX   %.1f\n",
	 track_precision(dw, balanced2double(bw)),
	 track_precision(dx, balanced2double(bx)),
	 track_precision(dy, balanced2double(by)));
  dw = DBL_MIN; dx = DBL_EPSILON; dy = DBL_MAX/2;
  bw = double2balanced(dw); bx = double2balanced(dx); by = double2balanced(dy);
  assert(balanced_is_valid(bw)); assert(balanced_is_valid(bx)); assert(balanced_is_valid(by));
  printf("\tprecision: DBL_MIN %.1f  DBL_EPSILON %.1f  DBL_MAX/2 %.1f\n",
	 track_precision(dw, balanced2double(bw)),
	 track_precision(dx, balanced2double(bx)),
	 track_precision(dy, balanced2double(by)));
  assert(errno != EDOM); assert(errno != ERANGE);
  if (Minimum_Precision < DBL_MANT_DIG)
    { printf("\tWARNING: Minimum_Precision %.1f < DBL_MANT_DIG %.1f\n\a",
	     Minimum_Precision, (double)DBL_MANT_DIG); }
#ifndef __hpux /* because hpsux */
  assert(Minimum_Precision == DBL_MANT_DIG);
#endif

  /* verify arithmetic operations by sampling */
  Minimum_Precision = DBL_MANT_DIG;
  Total_Precision = 0; Conversion_Count = 0;
  for(i = 0; i < ARITHMETIC_SAMPLES; i++)
    {
      dx = double_sample(); bx = double2balanced(dx); 
      track_precision(dx, balanced2double(bx));
      assert((dx < 0) || (dx > 1) || balanced_is_pr(bx));

      dy = double_sample(); by = double2balanced(dy);
      track_precision(dy, balanced2double(by));
      assert((dy < 0) || (dy > 1) || balanced_is_pr(by));

      dz = dx + dy; 
      bz = balanced_add(bx, by);
#if DEBUG_ARITHMETIC
      printf("+\t%e\t%e,%d\n", dz, (double)bz.sig, bz.exp);
#endif
      track_precision(dz, balanced2double(bz));

      dz = dx - dy; 
      bz = balanced_subtract(bx, by);
#if DEBUG_ARITHMETIC
      printf("-\t%e\t%e,%d\n", dz, (double)bz.sig, bz.exp);
#endif
      track_precision(dz, balanced2double(bz));

      exponent = (double)bx.exp + (double)by.exp;
      if ((DBL_MIN_EXP < exponent) && (exponent < DBL_MAX_EXP))
	{
	  dz = dx * dy; 
	  bz = balanced_multiply(bx, by);
#if DEBUG_ARITHMETIC
	  printf("*\t%e\t%e,%d\n", dz, (double)bz.sig, bz.exp);
#endif
	  track_precision(dz, balanced2double(bz));
	}

      exponent = (double)bx.exp - (double)by.exp;
      if ((DBL_MIN_EXP < exponent) && (exponent < DBL_MAX_EXP))
	{
	  dz = dx / dy; 
	  bz = balanced_divide(bx, by);
#if DEBUG_ARITHMETIC
	  printf("/\t%e\t%e,%d\n", dz, (double)bz.sig, bz.exp);
#endif
	  track_precision(dz, balanced2double(bz));
	}
    }
  printf("\tverified arithmetic operations on %d pairs:\n", ARITHMETIC_SAMPLES);
  printf("\t\tmean precision %.2f bits (%.1f digits)\n",
	 Total_Precision/Conversion_Count, (Total_Precision/Conversion_Count)/M_LG10);
  printf("\t\tminimum precision %.2f bits (%.1f digits)\n",
	 Minimum_Precision, Minimum_Precision/M_LG10);
  assert(Minimum_Precision > MIN_ARITHMETIC_PRECISION);

  /* verify transcendental operations by sampling */
  Minimum_Precision = DBL_MANT_DIG;
  Total_Precision = 0; Conversion_Count = 0;
  for(i = 0; i < TRANSCENDENTAL_SAMPLES; i++)
    {
      dx = (drand48() * ((DBL_MAX_10_EXP-1) - (DBL_MIN_10_EXP+1))) + (DBL_MIN_10_EXP+1);
      assert(DBL_MIN_10_EXP < dx); assert(dx < DBL_MAX_10_EXP);

      dy = exp(dx); by = balanced_exp(dx);
#if DEBUG_TRANSCENDENTAL
      printf("exp(%e)\t%e\t%e,%d\n", dx, dy, (double)by.sig, by.exp);
#endif
      track_precision(dy, balanced2double(by));

      dz = log(dy); dw = balanced_log(by);
#if DEBUG_TRANSCENDENTAL
      printf("log(%e)\t%e\t%e,%d\n", dy, dz, dw);
#endif
      track_precision(dz, dw);
      assert(relative_precision(balanced_log2(by), dw/M_LN2) > MIN_LOGARITHMIC_PRECISION);
      assert(relative_precision(balanced_log10(by), dw/M_LN10) > MIN_LOGARITHMIC_PRECISION);

      dy = drand48()/2 + 0.5; by = double2balanced(dy);
      dz = pow(dy, dx);
      bz = balanced_power(by, dx);
#if DEBUG_TRANSCENDENTAL
      printf("pow(%e,%e)\t%e\t%e,%d\n", dy, dx, dz, (double)bz.sig, bz.exp);
#endif
      track_precision(dz, balanced2double(bz));
    }
  printf("\tverified transcendental operations on %d pairs:\n", TRANSCENDENTAL_SAMPLES);
  printf("\t\tmean precision %.2f bits (%.1f digits)\n",
	 Total_Precision/Conversion_Count, (Total_Precision/Conversion_Count)/M_LG10);
  printf("\t\tminimum precision %.2f bits (%.1f digits)\n",
	 Minimum_Precision, Minimum_Precision/M_LG10);
  assert(Minimum_Precision > MIN_TRANSCENDENTAL_PRECISION);

  /* verify comparison operations by sampling */
  assert(balanced_exact_compare(zero, double2balanced(FLT_EPSILON)) < 0);
  assert(balanced_exact_compare(minus_unity, double2balanced(-1)) == 0);
  for(i = 0; i < COMPARISON_SAMPLES; i++)
    {
      dx = double_sample(); bx = double2balanced(dx);
      assert(balanced_exact_compare(bx, bx) == 0);

      bw = balanced_multiply(bx, minus_unity);
      assert(balanced_exact_compare(bx, bw) != 0);
      assert((balanced_exact_compare(bx, zero) > 0) || (balanced_exact_compare(bx, bw) < 0));
      assert((balanced_exact_compare(bx, zero) < 0) || (balanced_exact_compare(bx, bw) > 0));

      by.sig = NEXTAFTERF(bx.sig,  2); by.exp = bx.exp;
      assert(!balanced_is_valid(by) || (balanced_exact_compare(bx, by) < 0));

      bz.sig = NEXTAFTERF(bx.sig, -2); bz.exp = bx.exp;
      assert(!balanced_is_valid(bz) || (balanced_exact_compare(bx, bz) > 0));
    }
  printf("\tverified comparison operations on %d samples\n", COMPARISON_SAMPLES);

  printf("\tsizeof(balanced_t) == %d bytes\n", (int)sizeof(balanced_t));
  printf("\tbalanced_t library passes all tests.\n");
  return(0);
}


