/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "memory.h"
#include "memory.spartan.h"
#include "hash.h"

static char  hash_rcsid[] = "$Id: hash.c,v 1.22 1998/02/07 03:44:28 libpa Exp $";

#define ANNOUNCE_RCSID        FALSE
#define ANNOUNCE_WARNINGS     TRUE
#define ANNOUNCE_REHASH       FALSE
#define TRACK_PROBE           FALSE
#define TRACK_PRUNE           FALSE
#define TRACK_PERFORMANCE     FALSE
#define EXTRA_SAFE            FALSE    /* verify after every mutation */

#define PROBE_WARNING_THRESHOLD         1000
#define MEAN_COLLISION_THRESHOLD        3.0
#define MEAN_REDUNDANCY_THRESHOLD       0.01

/***********************************************************
 * FUTURE IMPROVEMENTS
 * . variable slot index size for space efficiency
 *
 ************************************************************/

/************************************************************/

#define HASH_TABLE_MINSLOTS             (8-1)
#define HASH_TABLE_NOT_FOUND            (UINT_MAX)
#define hash_table_null_index(table_p)  (UINT_MAX)

#define hash_table_maxload(table_p)      ((table_p)->class->maxload)
#define hash_table_minload(table_p)      ((table_p)->class->minload)
#define hash_table_is_stuffed(table_p)        \
        ((double)hash_table_size(table_p) >   \
	 hash_table_maxload(table_p)*(double)hash_table_capacity(table_p))
#define hash_table_will_be_stuffed(table_p)   \
        ((double)hash_table_size(table_p)+1 > \
	 hash_table_maxload(table_p)*(double)hash_table_capacity(table_p))
#define hash_table_is_starved(table_p)        \
        ((double)hash_table_size(table_p) <   \
	 hash_table_minload(table_p)*(double)hash_table_capacity(table_p))

#define hash_table_detect_collision(table_p)  ((table_p)->class->collisions++)
#define hash_table_detect_redundancy(table_p) ((table_p)->class->redundancies++)
#define hash_table_detect_probe(table_p)      ((table_p)->class->probes++)

#define hash_table_capacity_index(table_p)   ((table_p)->capacity_index)
#define hash_table_capacity_count(table_p)   ((table_p)->class->capacity_count)
#define hash_table_maximum_capacity(table_p) (hash_class_maximum_capacity((table_p)->class))
#define hash_table_capacity_index_is_valid(table_p)                         \
        (hash_table_capacity_index(table_p) < hash_table_capacity_count(table_p))
#define hash_table_target_capacity(table_p)                                 \
        ((table_p)->class->capacity[hash_table_capacity_index(table_p)])
        /* Returns: desired capacity for current capacity index        */

static boolean_t  hash_table_are_equal(const hash_table_t *left, const hash_table_t *right);
                  /* Returns: TRUE iff each table is a copy of the other       */
static hash_table_t *hash_table_copy(const hash_table_t *original);
                     /* Returns: exact copy of original hash table that shares */
                     /*          original hash_class_t and parameters as well  */
                     /* Warning: don't use this for symbol table!              */

static void hash_table_initialize(hash_table_t *blank_table, 
				  unsigned key_size, unsigned datum_size, 
				  hash_class_t *hash_class,
				  const void *parameters);
            /* Arguments: blank hash_table_t *, freshly allocated             */
            /* Effect: makes table into a valid hash_table_t                  */
            /* Note: use this to create an efficient array of hash_table_t    */
            /*       caller is responsible for deallocating blank_table       */
            /*       after calling hash_table_finalize() to clean up          */
static void hash_table_finalize(hash_table_t *table);
            /* Effect: deallocates all internal memory structures             */
            /* Note: call this before you deallocate a hash_table_t that you  */
            /*       allocated yourself and then hash_table_initialized.      */


static boolean_t hash_class_is_valid(const hash_class_t *class);
                 /* Returns: TRUE iff hash_class_t is valid                             */
static void      hash_table_minimize(hash_table_t *table);
                 /* Effect: rehashes all entries into smallest possible table           */
static void      hash_table_maximize(hash_table_t *table);
                 /* Effect: rehashes all entries into largest possible table            */
static unsigned  hash_table_probe(const hash_table_t *table, const void *key_p);
                 /* Returns: slot_index hwere *key_p is or should be                    */
static void      hash_table_rehash(hash_table_t *table, unsigned new_slot_count);
                 /* Effect: rehashes all entries into table of specified new_slot_count */


/************************************************************/


hash_table_t *hash_table_create(unsigned key_size, unsigned datum_size, 
				hash_class_t *class,
				const void *parameters)
{
  hash_table_t *table;
  table = memory_spartan_create(sizeof(hash_table_t));
  assert(table);
  hash_table_initialize(table, 
			key_size, datum_size, 
			class, parameters);
  return(table);
}


static void hash_table_initialize(hash_table_t *blank_table, 
				  unsigned key_size, unsigned datum_size,
				  hash_class_t *class,
				  const void *parameters)
{
  unsigned null_index;
  int i;
#if ANNOUNCE_RCSID
  static boolean_t rcsid_is_announced = FALSE;
  if (!rcsid_is_announced)
    {
#ifndef NDEBUG
      fprintf(stderr, "%s\n", hash_rcsid);
#else
      fprintf(stderr, "%s\t>unsafe<\n", hash_rcsid);
#endif
      rcsid_is_announced = TRUE;
    }
  assert(rcsid_is_announced);
#endif

  assert(blank_table);
  assert(class);
  assert(hash_class_is_valid(class));

  if ((1 > key_size) || (key_size > HASH_TABLE_MAXKEYSIZE) ||
      (0 > datum_size) || (datum_size > HASH_TABLE_MAXDATUMSIZE))
    {
      fprintf(stderr, "hash_table_t error: infeasible key,datum sizes %u,%u\n",
	      key_size, datum_size);
      abort();
    }

  blank_table->capacity_index = 0;
  blank_table->key_size = key_size;
  blank_table->datum_size = datum_size;
  blank_table->class = class;
  blank_table->parameters = parameters;

  /* initialize all slots with null indices */
  null_index = hash_table_null_index(table);
  blank_table->slot_count = HASH_TABLE_MINSLOTS;
  blank_table->slots = memory_spartan_create(sizeof(unsigned) * HASH_TABLE_MINSLOTS);
  assert(blank_table->slots);
  for(i = 0; i < blank_table->slot_count; i++)
    { blank_table->slots[i] = null_index; }

  /* initialize keys,datums arrays to contain no entries */
  blank_table->entry_count = 0;
  blank_table->keys = NULL;
  blank_table->datums = NULL;

  assert(memory_spartan_create(0) == NULL);
  assert(hash_table_is_starved(blank_table));
  assert(hash_table_minload(blank_table) < hash_table_maxload(blank_table));
  assert(hash_table_is_valid(blank_table));
}

hash_table_t *hash_table_copy(const hash_table_t *original)
{
  unsigned slot_count, entry_count, key_size, datum_size;
  hash_table_t *copy;
  assert(original);
  assert(hash_table_is_valid(original));

  slot_count = original->slot_count;
  entry_count = original->entry_count;
  key_size = original->key_size;
  datum_size = original->datum_size;

  copy = memory_spartan_create(sizeof(hash_table_t));
  copy->capacity_index = original->capacity_index;
  copy->key_size = key_size;
  copy->datum_size = datum_size;
  copy->slot_count = slot_count;
  copy->entry_count = entry_count;
  copy->slots = memory_spartan_copy(original->slots, slot_count*sizeof(unsigned));
  copy->keys = memory_spartan_copy(original->keys, entry_count * key_size);
  copy->datums = memory_spartan_copy(original->datums, entry_count * datum_size);
  copy->class = original->class;
  copy->parameters = original->parameters;

  assert(copy->slots);
  assert(copy->keys);
  assert((datum_size == 0) || (copy->datums != NULL));
  assert(hash_table_are_equal(copy, original));

  return(copy);
}

static void hash_table_finalize(hash_table_t *table)
{
  assert(table);
  assert(hash_table_is_valid(table));

#if ANNOUNCE_WARNINGS
  { 
    static boolean_t is_warned = FALSE;
    double mean_collisions = hash_class_mean_collisions(table->class);
    double mean_redundancies = hash_class_mean_redundancies(table->class);
    if (!is_warned && 
	(hash_class_probes(table->class) > PROBE_WARNING_THRESHOLD) &&
	((mean_collisions > MEAN_COLLISION_THRESHOLD) ||
	 (mean_redundancies > MEAN_REDUNDANCY_THRESHOLD)))
      { 
	fprintf(stderr, "[ hash_table_t warning: serious performance problem!\n");
	/* fprintf(stderr, "\t"); hash_table_monitor(table, stderr); */
	if (mean_collisions > MEAN_COLLISION_THRESHOLD)
	  { fprintf(stderr, "\tmean collisions/probe %f are unacceptably high (> %f)\n", 
		    mean_collisions, (double)MEAN_COLLISION_THRESHOLD); }
	if (mean_redundancies > MEAN_REDUNDANCY_THRESHOLD)
	  { fprintf(stderr, "\tmean redundancies/probe %f are unacceptably high (> %f)\n", 
		    mean_redundancies, (double)MEAN_REDUNDANCY_THRESHOLD); }
	fprintf(stderr, "\t=> please revise primary and/or secondary hash functions ]\n");
	is_warned = TRUE;
      }
  }
#endif

#ifndef NDEBUG
  /* clear all storage before deallocation for extra safety */
  (void) memset(table->slots, 0, sizeof(unsigned) * table->slot_count);
  (void) memset(table->keys, 0, hash_table_key_size(table) * table->entry_count);
  (void) memset(table->datums, 0, hash_table_datum_size(table) * table->entry_count);
#endif

  memory_spartan_destroy(table->slots, table->slot_count * sizeof(unsigned));
  memory_spartan_destroy(table->keys, table->entry_count * table->key_size);
  memory_spartan_destroy(table->datums, table->entry_count * table->datum_size);

#ifndef NDEBUG
  /* clear all storage before deallocation for extra safety */
  table->capacity_index = 0;
  table->key_size = 0;
  table->datum_size = 0;
  table->slot_count = 0;
  table->entry_count = 0;
  table->slots = NULL;
  table->keys = NULL;
  table->datums = NULL;
  table->class = NULL;
  table->parameters = NULL;
#endif
}

void hash_table_destroy(hash_table_t *table)
{
  assert(table);
  hash_table_finalize(table);
  memory_spartan_destroy(table, sizeof(hash_table_t));
}

static boolean_t hash_class_is_valid(const hash_class_t *class)
{
  int i;

  if ((class == NULL) ||
      (class->maxload <= class->minload) ||
      (class->collisions < 0) ||
      (class->redundancies < 0) ||
      (class->probes < 0) ||
      (class->capacity_count < 1) ||
      (class->capacity == NULL) ||
      (class->compare == NULL) ||
      (class->primary_hash == NULL) ||
      (class->secondary_hash == NULL))
    { fprintf(stderr, "hash_class_t error: invalid struct\n"); return(FALSE); }

  for(i = 1; i < class->capacity_count; i++)
    { 
      if (class->capacity[i-1] >= class->capacity[i]) 
	{ 
	  fprintf(stderr, "hash_class_t error: capacity %d is nonmonotonic %u >= %u\n",
		  i, class->capacity[i-1], class->capacity[i]);
	  return(FALSE); 
	}
    }
  
  return(TRUE);
}

boolean_t hash_table_is_valid(const hash_table_t *table)
{
  if (table == NULL) 
    { fprintf(stderr, "hash_table_t error: NULL table\n"); return(FALSE); }

  /* verify all fields are properly initialized */
  if ((table->key_size == 0) ||
      (table->slots == NULL) ||
      (table->class == NULL))
    { fprintf(stderr, "hash_table_t error: uninitialized field\n"); return(FALSE); }

  assert(table->class);
  if (!hash_class_is_valid(table->class))
    { fprintf(stderr, "hash_table_t error: invalid class\n"); return(FALSE); }

  if ((table->entry_count > 0) ? 
      ((table->keys == NULL) || ((table->datum_size > 0) && (table->datums == NULL))) :
      ((table->keys != NULL) || (table->datums != NULL)))
    { fprintf(stderr, "hash_table_t error: key/datum misallocation\n"); return(FALSE); }

  if ((table->slot_count < HASH_TABLE_MINSLOTS) ||
      (table->slot_count != hash_table_target_capacity(table)))
    {
      fprintf(stderr, "hash_table_t error: slot_count %u != capacity_index %u\n",
	      (unsigned)table->slot_count, (unsigned)table->capacity_index);
      return(FALSE);
    }
  
  /* verify that table has been resized appropriately */
  if (hash_table_is_stuffed(table))
    {
      fprintf(stderr, "hash_table_t error: failure to resize\n");
      return(FALSE);
    }
#if TRACK_PERFORMANCE
  {
    static boolean_t is_warned = FALSE;
    if (!is_warned && hash_table_is_starved(table))
      { fprintf(stderr, "[ hash_table_t warning: starved table ]\n"); is_warned = TRUE; }
  }
#endif

  /* verify that all slots are valid */
  if (table->entry_count == 0)
    {
      unsigned null_index; int i;
      null_index = hash_table_null_index(table);
      if (table->slot_count != HASH_TABLE_MINSLOTS) 
	{ fprintf(stderr, "hash_table_t error: invalid minslots\n"); return(FALSE); }
      for(i = 0; i < table->slot_count; i++)
	{ 
	  if (table->slots[i] != null_index) 
	    { fprintf(stderr, "hash_table_t error: \n"); return(FALSE); }
	}
    }
  else
    {
      boolean_t is_valid;
      char *entry_is_indexed;
      int i, entry_count; 
      unsigned entry_index, null_index;

      assert(hash_table_size(table) > 0);

      /* allocate and initialize set-type data structure */
      entry_is_indexed = memory_create(sizeof(char) * table->entry_count); 
      assert(memory_is_valid(entry_is_indexed));
      (void) memset(entry_is_indexed, 0, sizeof(char) * table->entry_count);

      is_valid = TRUE;
      entry_count = 0;
      null_index = hash_table_null_index(table);
      for(i = 0; i < table->slot_count; i++)
	{
	  if (table->slots[i] != null_index)
	    {
	      entry_index = table->slots[i];
	      if (entry_index >= table->entry_count) 
		{ is_valid = FALSE; } 
	      else 
		{
		  assert(memory_address_is_valid(entry_is_indexed, 
						 entry_is_indexed+entry_index));
		  if (entry_is_indexed[entry_index]) { is_valid = FALSE; }
		  entry_is_indexed[entry_index] = TRUE;
		}
	      entry_count++;
	    }
	}

      memory_destroy(entry_is_indexed);
      entry_is_indexed = NULL;

      if (!is_valid || (entry_count != table->entry_count))
	{ fprintf(stderr, "hash_table_t error: invalid slot\n"); return(FALSE); }
    }

  /* verify that every key is equal to itself and probes to itself */
  {
    unsigned char *key_p, *datum_p;
    unsigned key_size, datum_size;
    int i;

    key_size = hash_table_key_size(table);
    datum_size = hash_table_datum_size(table);

    key_p = table->keys;
    datum_p = table->datums;
    for(i = 0; i < table->entry_count; i++)
      {
	if (hash_table_compare(table, key_p, key_p) != 0) 
	  { fprintf(stderr, "hash_table_t error: invalid compare\n"); return(FALSE); }
	if ((datum_size > 0) && (hash_table_lookup(table, key_p) != datum_p))
	  { fprintf(stderr, "hash_table_t error: invalid probe\n"); return(FALSE); }
	if (!hash_table_includes(table, key_p))
	  { fprintf(stderr, "hash_table_t error: invalid probe\n"); return(FALSE); }
	key_p += key_size;
	datum_p += datum_size;
      }
  }

  return(TRUE);
}

static boolean_t hash_table_are_equal(const hash_table_t *left, const hash_table_t *right)
{
  unsigned key_size, datum_size, slot_count, entry_count;

  key_size = left->key_size;
  datum_size = left->datum_size;
  slot_count = left->slot_count;
  entry_count = left->entry_count;

  return ((key_size == right->key_size) &&
	  (datum_size == right->datum_size) &&
	  (slot_count == right->slot_count) &&
	  (entry_count == right->entry_count) &&
	  (left->class == right->class) &&
	  (left->parameters == right->parameters) &&
	  (memcmp(left->slots, right->slots, slot_count * sizeof(unsigned)) == 0) &&
	  (memcmp(left->keys, right->keys, key_size * entry_count) == 0) &&
	  (memcmp(left->datums, right->datums, datum_size * entry_count) == 0)
	  );
}

typedef struct equivalence_frame_tag
{
  boolean_t           are_equivalent;
  const hash_table_t *reference;
} equivalence_frame_t;

static boolean_t equivalence_verifier(const void *key_p, void *datum_p, void *parameters)
{
  equivalence_frame_t *frame;
  void *reference_datum_p;
  unsigned datum_size;

  frame = (equivalence_frame_t *)parameters;
  assert(frame->are_equivalent);

  /* verify that both compare functions are equivalent in finite domain  */
  /* and that the keys are associated with the same datum in both tables */
  if (!hash_table_includes(frame->reference, key_p))
    { frame->are_equivalent = FALSE; return(FALSE); }

  datum_size = hash_table_datum_size(frame->reference);
  assert((datum_p != NULL) || (datum_size == 0));
  if (datum_size > 0)
    {
      reference_datum_p = hash_table_lookup(frame->reference, key_p);
      if ((reference_datum_p == NULL) ||
	  (memcmp(datum_p, reference_datum_p, datum_size) != 0) ||
	  (hash_table_compare(frame->reference, key_p, key_p) != 0))
	{ frame->are_equivalent = FALSE; }
    }
  return(frame->are_equivalent);
}

boolean_t hash_table_are_equivalent(const hash_table_t *left, const hash_table_t *right)
{
  equivalence_frame_t frame;

  if ((hash_table_key_size(left) != hash_table_key_size(right)) ||
      (hash_table_datum_size(left) != hash_table_datum_size(right)) ||
      (hash_table_size(left) != hash_table_size(right)))
    { return(FALSE); }
  
  frame.are_equivalent = TRUE;
  frame.reference = right;
  hash_table_map((hash_table_t *)left, &equivalence_verifier, &frame);

  return(frame.are_equivalent);
}

boolean_t hash_table_is_empty(const hash_table_t *table)
{
  return (hash_table_size(table) == 0);
}


static unsigned hash_table_probe(const hash_table_t *table, const void *key_p)
{
  unsigned long slot_index, initial_slot_index;            /* to prevent overflow */
  unsigned int  primary_hash_value, secondary_hash_value;
  unsigned int  null_index, entry_index, slot_count;
  unsigned char *entry_key_p;
  unsigned key_size;
#ifndef NDEBUG
  int bounce_count;
#endif

  assert(hash_table_size(table) < hash_table_capacity(table));
  assert(hash_table_capacity(table) >= HASH_TABLE_MINSLOTS);

  hash_table_detect_probe(table);    /* maintain statistics */

  key_size = hash_table_key_size(table);
  null_index = hash_table_null_index(table);
  slot_count = table->slot_count;

  primary_hash_value = hash_table_primary_hash(table, key_p);
  initial_slot_index = primary_hash_value % slot_count;
  assert(initial_slot_index < table->slot_count);
  entry_index = table->slots[initial_slot_index];

#if TRACK_PROBE 
  printf("\n[probing: %d", (int)initial_slot_index); 
#endif
  if (entry_index == null_index) 
    {
#if TRACK_PROBE 
      printf("-]"); 
#endif
      return(initial_slot_index);
    }

  assert(table->keys);
  assert(table->entry_count > 0);
  assert(entry_index < table->entry_count);
  assert(key_size == hash_table_key_size(table));

  entry_key_p = table->keys + (entry_index * key_size);
  if (hash_table_compare(table, entry_key_p, key_p) == 0) 
    {
#if TRACK_PROBE 
      printf("+]");
#endif
      return(initial_slot_index);
    }

  /* calculate secondary hash value */
  assert(slot_count == hash_table_capacity(table));
  secondary_hash_value = hash_table_secondary_hash(table, key_p);
  if (primary_hash_value == secondary_hash_value) 
    { hash_table_detect_redundancy(table); }                  /* maintain statistics */
  secondary_hash_value = secondary_hash_value % slot_count;   /* prevent overflow */
  
#ifndef NDEBUG
  bounce_count = 0;
#endif
  slot_index = initial_slot_index;
  while(1)
    {
      hash_table_detect_collision(table);    /* maintain statistics */
      slot_index = (slot_index + secondary_hash_value) % slot_count;
      if (slot_index == initial_slot_index)  
	{
	  /* secondary_hash_value not relatively prime to slot_count */
	  /* => switch to linear probing from initial_hash_value     */
	  assert((secondary_hash_value%slot_count) != 1);
	  assert(hash_table_size(table) < hash_table_capacity(table));
	  secondary_hash_value = 1;                      /* make it so! */
	  slot_index = (initial_slot_index + secondary_hash_value) % slot_count;
#if TRACK_PROBE 
	  printf(" ^");
#endif
#ifndef NDEBUG
	  bounce_count = 1;
#endif
	}
#if TRACK_PROBE 
      printf(" %d", (int)slot_index);
#endif
      assert(slot_index != initial_slot_index);
      assert(slot_index < table->slot_count);
      entry_index = table->slots[slot_index];
      if (entry_index == null_index) 
	{
#if TRACK_PROBE 
	  printf("-]");
#endif
	  return(slot_index);
	}
      assert(entry_index < table->entry_count);
      entry_key_p = table->keys + (entry_index * key_size);
      if (hash_table_compare(table, entry_key_p, key_p) == 0) 
	{
#if TRACK_PROBE 
	  printf("+]");
#endif
	  return(slot_index);
	}
#ifndef NDEBUG
      bounce_count++;
      assert(bounce_count < table->slot_count);
#endif
    }
}

boolean_t hash_table_includes(const hash_table_t *table, const void *key_p)
{
  unsigned slot_index, entry_index, null_index;

  null_index = hash_table_null_index(table);
  slot_index = hash_table_probe(table, key_p);
  assert(slot_index < table->slot_count);

  entry_index = table->slots[slot_index];
  if (entry_index == null_index) { return(FALSE); }
  assert(entry_index < table->entry_count);
  return(TRUE);
}

void *hash_table_lookup(const hash_table_t *table, const void *key_p)
{
  unsigned slot_index, entry_index, null_index, datum_size;
  unsigned char *datum_p;
#if ANNOUNCE_WARNINGS
  static boolean_t is_warned = FALSE;
  if (!is_warned && (hash_table_datum_size(table) == 0))
    { 
      fprintf(stderr, "[ hash_table_t warning: ");
      fprintf(stderr, "_lookup() always returns NULL when datum_size == 0 ]\n"); 
      is_warned = TRUE;
    }
  assert(is_warned || (hash_table_datum_size(table) > 0));
#endif

  null_index = hash_table_null_index(table);
  slot_index = hash_table_probe(table, key_p);
  assert(slot_index < table->slot_count);
  entry_index = table->slots[slot_index];

  if (entry_index == null_index) { return(NULL); }
  assert(entry_index < table->entry_count);

  datum_size = hash_table_datum_size(table);
  datum_p = table->datums + (entry_index * datum_size);
  assert((datum_size == 0) ? (datum_p == NULL) : (datum_p != NULL));
  return(datum_p);
}

unsigned hash_table_index(const hash_table_t *table, const void *key_p)
{
  unsigned slot_index, entry_index;

  slot_index = hash_table_probe(table, key_p);
  assert(slot_index < table->slot_count);
  entry_index = table->slots[slot_index];
  assert((entry_index < table->entry_count) ||
	 (entry_index == hash_table_null_index(table)));
  return(entry_index);
}

const void *hash_table_key_p(const hash_table_t *table, unsigned index)
{
  unsigned key_size;
  unsigned char *key_p;
  assert(table); assert(index < hash_table_size(table));
  key_size = hash_table_key_size(table);
  key_p = table->keys + (key_size * index);
  return((const void *)key_p);
}

void *hash_table_datum_p(const hash_table_t *table, unsigned index)
{
  unsigned datum_size;
  unsigned char *datum_p;
  assert(table); assert(index < hash_table_size(table));
  datum_size = hash_table_datum_size(table);
  datum_p = table->datums + (datum_size * index);
  return((void *)datum_p);
}


static void *hash_table_mutate(hash_table_t *table, const void *key_p, 
			       const void *initial_p, const void *replacement_p)
{
  unsigned slot_index, entry_index, null_index;
  unsigned key_size, datum_size;

  assert(table != NULL);
  assert(key_p != NULL);

  /* first make sure that there's room for another <key,datum> entry    */
  /* note: slightly inefficient because table always has one extra slot */
  if (hash_table_will_be_stuffed(table))
    {
      unsigned target_capacity;
      if (table->capacity_index == HASH_TABLE_MAXCAPACITYINDEX) {
	fprintf(stderr, "hash_table_t error: capacity index overflow\n");
	abort();
      } else if (table->capacity_index == hash_table_capacity_count(table)) { 
	fprintf(stderr, "hash_class_t error: insufficient capacities\n"); 
	abort();
      }
      assert(table->capacity_index < HASH_TABLE_MAXCAPACITYINDEX);
      assert(table->capacity_index < hash_table_capacity_count(table));
      table->capacity_index++;
      target_capacity = hash_table_target_capacity(table);
      assert(target_capacity > hash_table_capacity(table));
      hash_table_rehash(table, target_capacity);
    }

  key_size = hash_table_key_size(table);
  datum_size = hash_table_datum_size(table);

  null_index = hash_table_null_index(table);
  slot_index = hash_table_probe(table, key_p);
  assert(slot_index < table->slot_count);
  entry_index = table->slots[slot_index];
  assert((entry_index == null_index) || (entry_index < table->entry_count));

  if (entry_index == null_index)
    {
      unsigned char *new_key_p, *new_datum_p;

      /* insert new <key,datum> pair at end of table->keys,datums */
      entry_index = table->entry_count;
      table->slots[slot_index] = entry_index;
      table->entry_count++;
      table->keys = memory_spartan_recreate(table->keys, 
					    key_size * entry_index,
					    key_size * table->entry_count);
      table->datums = memory_spartan_recreate(table->datums, 
					      datum_size * entry_index,
					      datum_size * table->entry_count);
      assert(table->keys);
      assert((datum_size == 0) ? (table->datums == NULL) : (table->datums != NULL));
      assert(entry_index+1 == hash_table_size(table));
      assert((initial_p != NULL) || (datum_size == 0));

      new_key_p = table->keys + (key_size * entry_index);
      new_datum_p = table->datums + (datum_size * entry_index);
      memcpy(new_key_p, key_p, key_size);
      memcpy(new_datum_p, initial_p, datum_size);
      assert(hash_table_compare(table, new_key_p, key_p) == 0);
      assert(hash_table_probe(table, key_p) == slot_index);
      assert(table->slots[slot_index] == entry_index);
      assert(entry_index < table->entry_count);
#if EXTRA_SAFE
      assert(hash_table_is_valid(table));
#endif
      return(new_datum_p);
    }
  else
    {
      unsigned char *old_datum_p;
  
      /* must mutate existing <key,datum> pair */
      assert(entry_index < hash_table_size(table));
      assert(hash_table_compare(table, key_p, table->keys+(entry_index*key_size)) == 0);
      assert(hash_table_probe(table, key_p) == slot_index);
      assert(table->slots[slot_index] == entry_index);

      old_datum_p = table->datums + (datum_size * entry_index);
      if ((replacement_p != NULL) && (old_datum_p != replacement_p))
	{ memcpy(old_datum_p, replacement_p, datum_size); }
#if EXTRA_SAFE
      assert(hash_table_is_valid(table));
#endif
      return(old_datum_p);
    }
}

void hash_table_insert(hash_table_t *table, const void *key_p, const void *datum_p)
{ 
  (void) hash_table_mutate(table, key_p, datum_p, datum_p); 
}

void *hash_table_update(hash_table_t *table, const void *key_p, const void *initial_p)
{
#if ANNOUNCE_WARNINGS
  static boolean_t is_warned = FALSE;
  if (!is_warned && (hash_table_datum_size(table) == 0))
    { 
      fprintf(stderr, "[ hash_table_t warning: ");
      fprintf(stderr, "_update() always returns NULL when datum_size == 0 ]\n"); 
      is_warned = TRUE;
    }
  assert(is_warned || (hash_table_datum_size(table) > 0));
#endif
  return( hash_table_mutate(table, key_p, initial_p, NULL) );
}


void hash_table_remove(hash_table_t *table, const void *key_p)
{
  unsigned deleted_slot_index, deleted_entry_index;
  unsigned last_entry_index, null_index;
  unsigned char *deleted_key_p, *last_key_p;
  unsigned char *deleted_datum_p, *last_datum_p;
  unsigned key_size, datum_size;

  null_index = hash_table_null_index(table);
  deleted_slot_index = hash_table_probe(table, key_p);
  assert(deleted_slot_index < table->slot_count);
  deleted_entry_index = table->slots[deleted_slot_index];
  assert((deleted_entry_index == null_index) || (deleted_entry_index < table->entry_count));

  /* simple case: *key_p is not in the hash table */
  if (deleted_entry_index == null_index) { return; }

  /* complex case: *key_p is actually in the table */
  /* first delete the actual entry by overwriting it with the last entry  */
  /* note that this changes the index of the last entry in table->slots */
  key_size = hash_table_key_size(table);
  datum_size = hash_table_datum_size(table);
  last_entry_index = table->entry_count - 1;
  assert(deleted_entry_index <= last_entry_index);
  assert(table->entry_count > 0);
  if (deleted_entry_index < last_entry_index)
    {
      last_key_p = table->keys + (last_entry_index * key_size);
      deleted_key_p = table->keys + (deleted_entry_index * key_size);
      memcpy(deleted_key_p, last_key_p, key_size);

      last_datum_p = table->datums + (last_entry_index * datum_size);
      deleted_datum_p = table->datums + (deleted_entry_index * datum_size);
      memcpy(deleted_datum_p, last_datum_p, datum_size);
    }
  table->keys = memory_spartan_recreate(table->keys, 
					key_size * table->entry_count,
					key_size * last_entry_index);
  table->datums = memory_spartan_recreate(table->datums, 
					  datum_size * table->entry_count,
					  datum_size * last_entry_index);
  table->entry_count--;
  assert((table->entry_count == 0) || (table->keys != NULL));
  assert((datum_size == 0) ? (table->datums == NULL) :
	 ((table->entry_count == 0) || (table->datums != NULL)));

  /* now rehash because deleted entry may have bounced another entry */
  if (hash_table_is_starved(table))
    {
      /* rehash entries into smaller table */
      assert(hash_table_capacity_index_is_valid(table));
      if (table->capacity_index > 0) { table->capacity_index--; }
      hash_table_rehash(table, hash_table_target_capacity(table));
    }
  else
    {
      /* rehash entries into same table */
      hash_table_rehash(table, table->slot_count);
    }
}

void hash_table_clear(hash_table_t *table)
{
  unsigned null_index;
  unsigned key_size, datum_size, entry_count, slot_count;
  int i;

  null_index = hash_table_null_index(table);
  key_size = hash_table_key_size(table);
  datum_size = hash_table_datum_size(table);
  slot_count = table->slot_count;
  entry_count = table->entry_count;

  table->keys = memory_spartan_recreate(table->keys, key_size * entry_count, 0);
  table->datums = memory_spartan_recreate(table->datums, datum_size * entry_count, 0);
  table->entry_count = 0;

  table->capacity_index = 0;
  table->slots = memory_spartan_recreate(table->slots, 
					 sizeof(unsigned) * slot_count,
					 sizeof(unsigned) * HASH_TABLE_MINSLOTS);
  table->slot_count = HASH_TABLE_MINSLOTS;
  assert(table->slots);
  for(i = 0; i < table->slot_count; i++)
    { table->slots[i] = null_index; }

  assert(hash_table_size(table) == 0);
  assert(hash_table_is_starved(table));
  assert(hash_table_is_valid(table));
}

/* Effect: prune unwanted <key,datum> entries in reverse order */
/* Note: this is simpler but less efficient than forward order */
static void hash_table_reverse_prune(hash_table_t *table, 
				     boolean_t (*prune_me)(const void *key_p, 
							   const void *datum_p, 
							   void *parameters),
				     void *parameters)
{
  unsigned char *this_key_p, *this_datum_p, *last_key_p, *last_datum_p;
  unsigned initial_entry_count;
  unsigned key_size, datum_size;

  key_size = hash_table_key_size(table);
  datum_size = hash_table_datum_size(table);
  initial_entry_count = table->entry_count;

  last_key_p = table->keys + (key_size * (initial_entry_count-1));
  last_datum_p = table->datums + (datum_size * (initial_entry_count-1));
  this_key_p = last_key_p;
  this_datum_p = last_datum_p;
  while(this_key_p >= table->keys)
    {
      assert(this_key_p >= table->keys); assert(this_key_p <= last_key_p);
      assert(this_datum_p >= table->datums); assert(this_datum_p <= last_datum_p);
      assert(last_key_p == table->keys + (key_size * (table->entry_count-1)));
      assert(last_datum_p == table->datums + (datum_size * (table->entry_count-1)));
      assert((datum_size * (last_key_p - this_key_p)) ==
	     (key_size * (last_datum_p - this_datum_p)));

      if (prune_me((const void *)this_key_p, (const void *)this_datum_p, parameters))
	{
	  if (this_key_p < last_key_p)
	    {
	      assert(this_datum_p+datum_size <= last_datum_p);
	      memcpy(this_key_p, last_key_p, key_size);
	      memcpy(this_datum_p, last_datum_p, datum_size);
	    }
	  last_key_p -= key_size;
	  last_datum_p -= datum_size;
	  table->entry_count--;
	}
      this_key_p -= key_size;
      this_datum_p -= datum_size;
    }
  assert(this_key_p+key_size == table->keys); 
  assert(this_datum_p+datum_size == table->datums);
  table->keys = memory_spartan_recreate(table->keys, 
					key_size * initial_entry_count,
					key_size * table->entry_count);
  table->datums = memory_spartan_recreate(table->datums, 
					  datum_size * initial_entry_count,
					  datum_size * table->entry_count);
  assert((table->entry_count == 0) || (table->keys != NULL));
  assert((datum_size == 0) || (table->entry_count == 0) || (table->datums != NULL));

#if EXTRA_SAFE
  assert(!hash_table_is_valid(table));
#endif
  hash_table_minimize(table);
  assert(hash_table_is_valid(table));
}

void hash_table_prune(hash_table_t *table, 
		      boolean_t (*prune_me)(const void *key_p, const void *datum_p, 
					    void *parameters),
		      void *parameters)
{
  unsigned char *this_key_p, *this_datum_p, *last_key_p, *last_datum_p;
  unsigned char *open_key_p, *open_datum_p;
  unsigned initial_entry_count;
  unsigned key_size, datum_size;

  key_size = hash_table_key_size(table);
  datum_size = hash_table_datum_size(table);
  initial_entry_count = table->entry_count;

  if (initial_entry_count == 0) { return; }

  assert(initial_entry_count > 0);
  last_key_p = table->keys + (key_size * (initial_entry_count-1));
  last_datum_p = table->datums + (datum_size * (initial_entry_count-1));
  this_key_p = table->keys;
  this_datum_p = table->datums;
  open_key_p = this_key_p;
  open_datum_p = this_datum_p;
  while(this_key_p <= last_key_p)
    {
      assert(this_key_p >= table->keys); assert(this_key_p <= last_key_p);
      assert(this_datum_p >= table->datums); assert(this_datum_p <= last_datum_p);
      assert(open_key_p >= table->keys); assert(open_key_p <= this_key_p);
      assert(open_datum_p >= table->datums); assert(open_datum_p <= this_datum_p);

      if (prune_me((const void *)this_key_p, (const void *)this_datum_p, parameters))
	{ 
	  assert(table->entry_count > 0);
	  table->entry_count--; 
	}
      else 
	{
	  if (this_key_p > open_key_p)
	    {
	      memcpy(open_key_p, this_key_p, key_size);
	      memcpy(open_datum_p, this_datum_p, datum_size);
	    }
	  open_key_p += key_size;
	  open_datum_p += datum_size;
	}
      this_key_p += key_size;
      this_datum_p += datum_size;
    }
  assert(this_key_p == last_key_p + key_size); 
  assert(this_datum_p == last_datum_p + datum_size);
  table->keys = memory_spartan_recreate(table->keys, 
					key_size * initial_entry_count,
					key_size * table->entry_count);
  table->datums = memory_spartan_recreate(table->datums, 
					  datum_size * initial_entry_count,
					  datum_size * table->entry_count);
  assert((table->entry_count == 0) || (table->keys != NULL));
  assert((datum_size == 0) || (table->entry_count == 0) || (table->datums != NULL));

#if EXTRA_SAFE
  assert(!hash_table_is_valid(table));
#endif
  hash_table_minimize(table);
  assert(hash_table_is_valid(table));
}

static void hash_table_minimize(hash_table_t *table)
{
  unsigned target_capacity, minimum_capacity;

  assert(hash_table_maxload(table) > hash_table_minload(table));

  /* search for lowest capacity index that won't trigger rehash up */
  target_capacity = hash_table_capacity(table);
  minimum_capacity = (double)(table->entry_count) / hash_table_maxload(table);
  assert(target_capacity == table->slot_count);
  assert(table->entry_count == hash_table_size(table));
  while((table->capacity_index > 0) && 
	(target_capacity > minimum_capacity))
    {
      assert(table->capacity_index > 0);
      table->capacity_index--;
      assert(hash_table_target_capacity(table) < target_capacity);
      target_capacity = hash_table_target_capacity(table);
    }
  if (target_capacity <= minimum_capacity) /* did we go too far? */
    {
      assert(table->capacity_index < HASH_TABLE_MAXCAPACITYINDEX);
      assert(table->capacity_index < hash_table_capacity_count(table));
      table->capacity_index++;
      assert(hash_table_capacity_index_is_valid(table));
      assert(hash_table_target_capacity(table) > target_capacity);
      target_capacity = hash_table_target_capacity(table);
    }
  assert(target_capacity == hash_table_target_capacity(table));
  assert(target_capacity > hash_table_size(table));

  /* rehash the table using new capacity */
  hash_table_rehash(table, target_capacity);
  assert(!hash_table_is_stuffed(table));  /* note: could be starved! */
}

static void hash_table_maximize(hash_table_t *table)
{
  unsigned target_capacity, maximum_capacity;

  { fprintf(stderr, "hash_table_maximize() not tested\n"); abort(); }

  /* search for largest capacity index that won't trigger rehash down */
  target_capacity = hash_table_capacity(table);
  maximum_capacity = (double)(table->entry_count) / hash_table_minload(table);
  assert(target_capacity == table->slot_count);
  assert(table->entry_count == hash_table_size(table));
  while((target_capacity < maximum_capacity) &&
	(table->capacity_index < HASH_TABLE_MAXCAPACITYINDEX) && 
	(table->capacity_index < hash_table_capacity_count(table)))
    {
      assert(table->capacity_index < HASH_TABLE_MAXCAPACITYINDEX);
      assert(table->capacity_index < hash_table_capacity_count(table));
      table->capacity_index++;
      assert(hash_table_target_capacity(table) > target_capacity);
      target_capacity = hash_table_target_capacity(table);
    }
  if (target_capacity >= maximum_capacity)  /* did we go too far? */
    {
      assert(table->capacity_index > 0);
      table->capacity_index--;
      assert(hash_table_target_capacity(table) < target_capacity);
      target_capacity = hash_table_target_capacity(table);
    }
  assert(target_capacity == hash_table_target_capacity(table));
  assert(target_capacity > hash_table_size(table));

  /* rehash the table using new capacity */
  hash_table_rehash(table, target_capacity);
  assert(!hash_table_is_starved(table));
  assert(!hash_table_is_stuffed(table));
}

static void hash_table_rehash(hash_table_t *table, unsigned new_slot_count)
{
  unsigned char *key_p;
  unsigned null_index, slot_index;
  unsigned key_size, entry_count;
  int i;

  assert(table);
  assert(new_slot_count >= HASH_TABLE_MINSLOTS);
  assert(new_slot_count >= table->entry_count);
  assert(new_slot_count == hash_table_target_capacity(table));
  assert(hash_table_minload(table) < hash_table_maxload(table));

#if ANNOUNCE_REHASH
  fprintf(stderr, "\n%u\thash_table_rehash(%u entries): %u -> %u\t",
	  memory_total_bytes(), hash_table_size(table), 
	  table->slot_count, new_slot_count);
#endif

  key_size = hash_table_key_size(table);
  null_index = hash_table_null_index(table);

  /* realloc index array to requested size and clear it */
  table->slots = memory_spartan_recreate(table->slots, 
					 sizeof(unsigned) * table->slot_count,
					 sizeof(unsigned) * new_slot_count);
  table->slot_count = new_slot_count;
  assert(table->slots);
  for(i = 0; i < new_slot_count; i++)
    { table->slots[i] = null_index; }

  /* rehash all entries into newly cleared index array */
  key_p = table->keys;                  /* first entry */
  entry_count = table->entry_count;
  for(i = 0; i < entry_count; i++)
    {
      slot_index = hash_table_probe(table, key_p);
      assert(table->slots[slot_index] == null_index);
      table->slots[slot_index] = i;
      key_p += key_size;                /* next entry */
    }
#if EXTRA_SAFE
  assert(hash_table_is_valid(table));
#endif
}


void hash_table_map(hash_table_t *table, 
		    boolean_t (*proc)(const void *key_p, void *datum_p, void *parameters),
		    void *parameters)
{
  unsigned char *key_p, *datum_p;
  unsigned key_size, datum_size, entry_count;
  boolean_t is_alive;
  int i;

  key_size = hash_table_key_size(table);
  datum_size = hash_table_datum_size(table);

  key_p = table->keys;
  datum_p = table->datums;
  entry_count = table->entry_count;
  for(i = 0; i < entry_count; i++)
    {
      is_alive = proc((const void *)key_p, (void *)datum_p, parameters);
      key_p += key_size;
      datum_p += datum_size;
      if (!is_alive) { return; }
    }
}


void hash_table_index_map(hash_table_t *table, 
			  boolean_t (*proc)(const void *key_p, void *datum_p, 
					    unsigned index, void *parameters),
			  void *parameters)
{
  unsigned char *key_p, *datum_p;
  unsigned key_size, datum_size;
  unsigned index, entry_count;
  boolean_t is_alive;

  key_size = hash_table_key_size(table);
  datum_size = hash_table_datum_size(table);

  key_p = table->keys;
  datum_p = table->datums;
  entry_count = table->entry_count;
  for(index = 0; index < entry_count; index++)
    {
      assert(hash_table_key_p(table, index) == key_p);
      assert(hash_table_datum_p(table, index) == datum_p);
      is_alive = proc((const void *)key_p, (void *)datum_p, index, parameters);
      key_p += key_size;
      datum_p += datum_size;
      if (!is_alive) { return; }
    }
}


void hash_table_resize_datum(hash_table_t *table, unsigned new_datum_size,
			     void (*coerce)(const void *old_datum_p, 
					    void *new_datum_p,
					    void *parameters),
			     void *parameters)
{
  unsigned char *old_datums, *new_datums, *old_datum_p, *new_datum_p;
  unsigned entry_count, old_datum_size;
  int i;

  assert(new_datum_size <= USHRT_MAX);

  entry_count = hash_table_size(table);
  old_datum_size = hash_table_datum_size(table);
  old_datums = table->datums;

  new_datums = memory_spartan_create(new_datum_size * entry_count);
  assert((new_datum_size * entry_count == 0) || (new_datums != NULL));
  table->datums = new_datums;
  table->datum_size = new_datum_size;

  new_datum_p = new_datums;
  old_datum_p = old_datums;
  for(i = 0; i < entry_count; i++)
    {
      coerce(old_datum_p, new_datum_p, parameters);
      old_datum_p += old_datum_size;
      new_datum_p += new_datum_size;
    }

  memory_spartan_destroy(old_datums, old_datum_size * entry_count);
  old_datums = NULL;

  assert(hash_table_size(table) == entry_count);
  assert(hash_table_datum_size(table) == new_datum_size);
  assert(hash_table_is_valid(table));
}


/************************************************************
 * PERFORMANCE TUNING
 ************************************************************/

void hash_table_monitor(const hash_table_t *table, FILE *stream)
{
  const hash_class_t *class;
  class = hash_table_class(table);
  fprintf(stream, "%u entries in %u slots (%2.2f load)",
	  hash_table_size(table), hash_table_capacity(table), hash_table_load(table));
  fprintf(stream, ": %d probes, %d collisions (%.2f/probe), %d redundancies\n",
	  (int)hash_class_probes(class), (int)hash_class_collisions(class), 
	  hash_class_mean_collisions(class), (int)hash_class_redundancies(class));
}

