/*
 * (c) 1995 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "memory.h"
#include "vector.h"
#include "hash.class.h"

static char hash_meter_rcsid[] = "$Id: hash.meter.c,v 1.4 1996/11/08 04:25:22 ristad Exp $";

/* DESCRIPTION 
   Compares joint performance of different pairs of hash functions.  
   Estimates number of collisions as well as hash value redundancy.  
*/

#define TRACK_PROBE       FALSE
#define MONITOR_PROGRESS  FALSE

#define MAX_SYMBOL          UCHAR_MAX
#define MAX_STRING_LENGTH   10
#define MAX_SAMPLE_SIZE     (((unsigned long)1) << 20)
#define SAMPLE_SIZE(i)      (((unsigned long)1) << ((i)*4))  /* density = 16^{-i} */

extern double drand48(void);
extern void   srand48();


/*
  for each string length
    sample many strings of that length
     accumulate redundancy statistics
     accumulate number of samples with same primary/secondary hash value
       note: overestimates number of collisions because sample w/ replacement
     (accumulate mean probes/lookup ?)
 */


/************** PROTOTYPES FOR HASH OPERATIONS ******************/

int       string_compare(const void *lkey_p, const void *rkey_p, const void *);
unsigned  string_xor_hash(const void *key_p, const void *);
unsigned  string_reverse_xor_hash(const void *key_p, const void *);
unsigned  string_nonlocal_xor_hash(const void *key_p, const void *);
unsigned  string_nonlocal_reverse_xor_hash(const void *key_p, const void *);
unsigned  string_accumulate_hash(const void *key_p, const void *);
unsigned  string_shift_accumulate_hash(const void *key_p, const void *);
unsigned  string_linear_hash(const void *key_p, const void *);

/************** IMPLEMENTATIONS ******************/

int string_compare(const void *left_p, const void *right_p, const void *parameters)
{
  const char *left_string, *right_string;
  assert(parameters == NULL);

  left_string = *(char **) left_p;
  right_string = *(char **) right_p;
  return strcmp(left_string, right_string);
}

unsigned string_linear_hash(const void *key_p, const void *parameters) { return(1); }

unsigned string_accumulate_hash(const void *key_p, const void *parameters)
{
  unsigned char *string;
  unsigned hash_value;
  int i;

  string = *(unsigned char **) key_p;

  i = 0;
  hash_value = 0;
  while(string[i] != '\0')
    {
      hash_value += string[i];
      i++;
    }
  return(hash_value);
}

unsigned string_shift_accumulate_hash(const void *key_p, const void *parameters)
{
  unsigned char *string;
  unsigned hash_value;
  int i;

  string = *(unsigned char **)key_p;

  i = 0;
  hash_value = 0;
  while(string[i] != '\0')
    {
      assert(string[0] != '\0');
      hash_value = hash_value << 4;   /* shift by 4 of 8 bits */
      hash_value += string[0];
      i++;
    }
  return(hash_value);
}

unsigned string_xor_hash(const void *key_p, const void *parameters)
{
  unsigned int  hash_value;
  unsigned char *hash_bytes, *string;
  int hash_byte_index, i;

  string = *((unsigned char **)key_p);
  hash_bytes = (unsigned char *) &hash_value;

  i = 0;
  hash_value = 0;
  hash_byte_index = 0;
  while(string[i] != '\0')
    {
      assert(hash_byte_index >= 0);
      assert(hash_byte_index < sizeof(unsigned));

      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ string[i];
      hash_byte_index = (hash_byte_index + 1) % sizeof(unsigned);
      i++;
    }
  return(hash_value);
}

unsigned string_nonlocal_xor_hash(const void *key_p, const void *parameters)
{
  unsigned char *hash_bytes, *string;
  unsigned int   hash_value;
  int      hash_byte_index, i;

  hash_bytes = (unsigned char *) &hash_value;
  string = *(unsigned char **)key_p;

  i = 0;
  hash_value = 0;
  hash_byte_index = 0;
  while(string[i] != '\0')
    {
      assert(hash_byte_index >= 0);
      assert(hash_byte_index < sizeof(unsigned));

      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ string[i];
      hash_byte_index = (hash_byte_index + string[i]) % sizeof(unsigned);
      i++;
    }
  return(hash_value);
}

unsigned string_reverse_xor_hash(const void *key_p, const void *parameters)
{
  unsigned char *hash_bytes, *string;
  unsigned int   hash_value;
  int hash_byte_index, i;

  hash_bytes = (unsigned char *) &hash_value;
  string = *(unsigned char **) key_p;

  i = 0;
  hash_value = UINT_MAX;
  hash_byte_index = sizeof(unsigned)-1;
  while(string[i] != '\0')
    {
      assert(hash_byte_index >= 0);
      assert(hash_byte_index < sizeof(unsigned));

      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ string[i];
      hash_byte_index = (hash_byte_index - 1) % sizeof(unsigned);
      i++;
    }
  return(hash_value);
}

unsigned string_nonlocal_reverse_xor_hash(const void *key_p, const void *parameters)
{
  unsigned char *hash_bytes;
  unsigned int   hash_value;
  int hash_byte_index, i;
  const unsigned char *string;

  hash_bytes = (unsigned char *)&hash_value;
  string = *(unsigned char **)key_p;

  i = 0;
  hash_value = UINT_MAX;
  hash_byte_index = sizeof(unsigned)-1;
  while(string[i] != '\0')
    {
      assert(string[i] != '\0');
      assert(hash_byte_index >= 0);
      assert(hash_byte_index < sizeof(unsigned));

      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ string[i];
      hash_byte_index = (hash_byte_index - string[i]) % sizeof(unsigned);
      i++;
    }
  return(hash_value);
}


static unsigned (*String_Hash_Functions[])(const void *, const void *)
     = { &string_xor_hash,
	   &string_reverse_xor_hash,
	   &string_nonlocal_xor_hash,
	   &string_nonlocal_reverse_xor_hash,
	   &string_accumulate_hash,
	   &string_shift_accumulate_hash,
	   &string_linear_hash 
       };

int main(int argc, char **argv)
{
  int            primary_index, secondary_index;
  unsigned int   primary_value, secondary_value;
  unsigned long  index, initial_index;  /* to prevent overflow */
  hash_class_t class;
  char      *string;
  UCHAR      symbol;
  int        string_length, sample, sample_size, i, j;
  int       *hit_table; 
  double     redundant_count, collision_count;
  double     mean_redundancies, mean_collisions;
  int        bounce_count;

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  if (argc == 2) 
    {
      fprintf(stdout, "#len\t1st\t2nd\tsamples\tcol/ins\tredund\n");;
      return(1);
    }

  if ((argc < 3) || (argc > 5))
    {
      printf("Usage:\t%s string_length [0-5] ([0-6])\n", argv[0]);
      printf("\t\t0\txor\n");
      printf("\t\t1\treverse_xor\n");
      printf("\t\t2\tnonlocal_xor\n");
      printf("\t\t3\tnonlocal_reverse_xor\n");
      printf("\t\t4\taccumulate\n");
      printf("\t\t5\tshift_accumulate\n");
      printf("\t\t6\tconstant (default secondary hash)\n");
      return(1);
    }

  string_length = atoi(argv[1]);
  primary_index = atoi(argv[2]);
  secondary_index = ((argc == 3) ? 6 : atoi(argv[3]));

  if ((string_length < 1) || (string_length > MAX_STRING_LENGTH))
    {
      printf("%s: string length %d must be in [1,%d]\n", 
	     argv[0], string_length, MAX_STRING_LENGTH);
      return(1);
    }
  if ((primary_index < 0) || (primary_index > 5) ||
      (secondary_index < 0) || (secondary_index > 6))
    {
      printf("%s: invalid hash function indices %d,%d\n", 
	     argv[0], primary_index, secondary_index);
      return(1);
    }

  class.compare = &string_compare;
  class.primary_hash = String_Hash_Functions[primary_index];
  class.secondary_hash = String_Hash_Functions[secondary_index];
  
  string = memory_create(string_length+1);
  assert(string);

  sample_size = SAMPLE_SIZE(string_length);
  sample_size = MIN(MAX_SAMPLE_SIZE, sample_size);
  sample_size--;		/* Mersenne Prime */

  /* create hit_table[] to count collisions */
  hit_table = vector_create(sizeof(int), sample_size);
  assert(hit_table);
  assert(vector_length(hit_table) == sample_size);
  for(j = 0; j < sample_size; j++) hit_table[j] = 0;

  /* generate samples and hash them */
  redundant_count = 0;
  collision_count = 0;
  string[string_length] = '\0';	/* terminate string */
  for(sample = 0; sample < sample_size; sample++)
    {
      /* generate the string uniformly */
      for(i = 0; i < string_length; i++)
	{
	  symbol = 1 + (drand48() * (double)UCHAR_MAX);
	  assert((symbol > 0) && (symbol <= UCHAR_MAX));
	  string[i] = symbol;
	}
      assert(strlen(string) == string_length);

#if MONITOR_PROGRESS
      printf(".");
#endif
#if FALSE
      /* debugging */
      string_xor_hash(&string, NULL);
      string_accumulate_hash(&string, NULL);
      string_nonlocal_reverse_xor_hash(&string, NULL);
#endif

      /* hash random string */
      assert(class.compare(&string, &string, NULL) == 0);
      primary_value = class.primary_hash(&string, NULL);
      secondary_value = class.secondary_hash(&string, NULL);
      if (primary_value == secondary_value) { redundant_count++; }
      /* truncate secondary hash value to prevent overflow in probing */
      secondary_value = secondary_value % sample_size;
      
      /* simulate hash table probe */
      initial_index = primary_value % sample_size;
      assert((initial_index >= 0) && (initial_index < sample_size));

      index = initial_index;
      bounce_count = 0;
#if TRACK_PROBE
      printf("\n[%d", initial_index);
#endif
      while(hit_table[index] != 0)
	{
	  bounce_count++;
	  collision_count++;
	  hit_table[index]++;
	  index = (index + secondary_value) % sample_size;
	  if (index == initial_index) /* not relatively prime! */
	    {
	      assert((secondary_value%sample_size) != 1);
	      assert(sample < sample_size);
	      secondary_value = 1;         /* switch to linear probing */
	      index = (initial_index + secondary_value) % sample_size;
	      bounce_count = 1;
#if TRACK_PROBE
	      printf(" ^");
#endif
	    }
	  assert(bounce_count <= sample_size);
	  assert((index >= 0) && (index < sample_size));
#if TRACK_PROBE
	  printf(" %d", index);
#endif
	}
#if TRACK_PROBE
      printf("]");
#endif
      assert(hit_table[index] == 0);
      hit_table[index]++;
    }
#ifndef NDEBUG
  {
    double hit_count_total;
    assert(vector_length(hit_table) == sample_size);
    hit_count_total = 0;
    for(j = 0; j < sample_size; j++)
      {
	assert(hit_table[j] > 0);
	hit_count_total += (hit_table[j] - 1);
      }
    assert(hit_count_total == collision_count);
  }
#endif
  vector_destroy(hit_table);
  hit_table = NULL;
  memory_destroy(string);
  string = NULL;

  mean_collisions = (double)collision_count / (double)sample_size;
  mean_redundancies = (double)redundant_count / (double)sample_size;

  if (stdout)
    {
      fprintf(stdout, "%d\t%d\t%d\t%d\t%.2f\t%d\n", string_length,
	      primary_index, secondary_index, sample_size,
	      mean_collisions, (int)redundant_count);
    }

  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  return(0);
}


