/*
 * (c) 1995 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "common.h"

static char hash_primes_rcsid[] = "$Id: hash.primes.c,v 1.3 1998/02/02 21:17:28 libpa Exp $";

/* DESCRIPTION 
   Enumerates primes closest to target capacities.
   Used to generate capacity series for hash.primes.h
*/

#define INITIAL_TARGET        8
#define INITIAL_PRIOR_PRIME   3
#define TRACK_BOUNDS          FALSE

extern double drand48(void);
extern void   srand48();
extern double ceil(double);
extern double sqrt(double);

#define number_has_factor(number,factor)  (((number)%(factor))==0)
#define number_has_subfactor(number,factor) \
        (((factor)*(factor)<=(number)) && number_has_factor((number),(factor)))


#define number_bit_mask(i)      (((unsigned)1)<<(i))
#define number_bit(n,i)         ((n) & number_bit_mask(i))
#define number_is_composite(n)  (!number_is_prime(n))

boolean_t number_is_prime(unsigned n)
{
  unsigned factor, factor_limit;
  assert(n > 1);

  if (n == 2) { return(TRUE); }
  else if ((n % 2) == 0) { return(FALSE); }

  factor_limit = 1+ceil(sqrt((double)n));
  if (factor_limit >= n) { factor_limit = n - 1; }
  assert((double)factor_limit * (double)factor_limit > (double)n);
  for(factor = 3; factor < factor_limit; factor += 2)
    { if ((n % factor) == 0) { return(FALSE); } }
  return(TRUE);
}


boolean_t primality_test_is_valid(void)
{
  int i;
  for(i = 0; i < sizeof(unsigned)*8; i++)
    {
      assert(number_bit(UINT_MAX, i));
      assert(!number_bit(0, i));
      assert(number_bit((((unsigned)1)<<i), i));
      assert(!number_bit((((unsigned)2)<<i), i));
      assert(number_bit((((unsigned)3)<<i), i));
    }

  return(number_is_prime(2) &&
	 number_is_prime(3) &&
	 number_is_composite(4) &&
	 number_is_prime(5) &&
	 number_is_composite(6) &&
	 number_is_prime(7) &&
	 number_is_composite(8) &&
	 number_is_composite(9) &&
	 number_is_composite(10) &&
	 number_is_prime(11) &&
	 number_is_composite(12) &&
	 number_is_prime(13) &&
	 number_is_composite(14) &&
	 number_is_composite(15) &&
	 number_is_composite(16) &&
	 number_is_prime(17) &&
	 number_is_composite(18) &&
	 number_is_prime(19) &&
	 number_is_composite(20) &&
	 number_is_composite(21) &&
	 number_is_composite(22) &&
	 number_is_prime(23) &&
	 number_is_composite(24) &&
	 number_is_composite(25) &&
	 number_is_composite(26) &&
	 number_is_composite(27) &&
	 number_is_composite(28) &&
	 number_is_prime(29) &&
	 number_is_composite(30) &&
	 number_is_composite(UINT_MAX)
	 );
}

/* TRUE iff number a proves that number n is composite */
boolean_t number_is_witness(unsigned n, unsigned a)
{
  unsigned d, x, n_pred;
  int i, k;

  d = 1;
  n_pred = n-1;
  k = (sizeof(unsigned)*8)-1;
  for(i = k; i >= 0; i--)
    {
      x = d;
      d = (d * d) % n;
      if ((d == 1) && (x != 1) && (x != (n-1))) { return(TRUE); }
      if (number_bit(n_pred,i)) { d = (d * a) % n; }
    }
  if (d != 1) { return(TRUE); }
  return(FALSE);
}

/* TRUE iff number is not composite after s trials of the test */
boolean_t number_miller_rabin_primality_test(unsigned n, int s)
{
  unsigned n_pred, a;
  int j;

  assert(n > 1);

  if ((n == 2) || (n == 3) || (n == 5) || (n == 7)) 
    { return(TRUE); }

  if (number_has_factor(n,2) || 
      number_has_factor(n,3) ||
      number_has_factor(n,5) ||
      number_has_factor(n,7))
    { return(FALSE); }

  n_pred = n - 1;
  for(j = 0; j < s; j++)
    {
      a = 1 + (drand48() * (double)n_pred);
      assert((a > 0) && (a < n));
      if (number_is_witness(n, a)) { return(FALSE); }
    }
  return(TRUE);
}


int main(int argc, char **argv)
{
  unsigned target_number, barrier_number, prior_prime_number;
  unsigned lower_number, upper_number;
  double   expansion_factor;
  int      series_length, i;

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  if ((argc < 3) || (argc > 3))
    {
      printf("Usage:\t%s length factor\n", argv[0]);
      return(1);
    }
  series_length = atoi(argv[1]);
  expansion_factor = atof(argv[2]);
  assert(series_length > 0);
  assert(expansion_factor > 1.0);

  if (!primality_test_is_valid()) 
    { fprintf(stderr, "%s error: invalid primality test\n", argv[0]); abort(); }

  target_number = INITIAL_TARGET;
  prior_prime_number = INITIAL_PRIOR_PRIME;
  barrier_number = prior_prime_number+1;
  for(i = 0; i < series_length; i++)
    {
#if TRACK_BOUNDS
      printf("\n%d: %u < %u\n", i, barrier_number, target_number);
#endif
      if ((target_number > prior_prime_number) &&
	  (number_is_prime(target_number)))
	{
	  prior_prime_number = target_number;
	  printf(", %uu", target_number);
	} 
      else 
	{
	  lower_number = target_number-1;
	  upper_number = target_number+1;
	  assert(lower_number < upper_number);
	  while((lower_number != prior_prime_number) &&
		(upper_number != prior_prime_number))
	    {
	      assert(lower_number < target_number);
	      assert(target_number < upper_number);
	      assert(barrier_number > prior_prime_number);
	      assert(lower_number >= barrier_number);
	      assert((upper_number < UINT_MAX) || (lower_number > barrier_number));
#if TRACK_BOUNDS
	      printf("\t%u,%u\n", lower_number, upper_number);
#endif
	      if ((lower_number > barrier_number) && (number_is_prime(lower_number)))
		{
		  printf(", %uu", lower_number);
		  prior_prime_number = lower_number;
		}
	      else if ((upper_number < UINT_MAX) && (number_is_prime(upper_number)))
		{
		  printf(", %uu", upper_number);
		  prior_prime_number = upper_number;
		}
	      else
		{
		  if ((upper_number == UINT_MAX) && (lower_number == barrier_number))
		    { return(1); }
		  if (upper_number < UINT_MAX) { upper_number++; }
		  if (lower_number > barrier_number) { lower_number--; }
		}
	    }
	  assert((lower_number == prior_prime_number) ||
		 (upper_number == prior_prime_number));
	}
      
      barrier_number = prior_prime_number+1;
      assert(prior_prime_number < barrier_number);
      assert(barrier_number <= UINT_MAX); 
      assert(!number_is_prime(barrier_number));
      if ((double)target_number * expansion_factor >= (double)UINT_MAX)
	{ target_number = UINT_MAX-3; }
      else
	{ target_number = (double)target_number * expansion_factor; }
      assert(target_number <= UINT_MAX);
      if (target_number <= barrier_number) { 
	assert(barrier_number >= UINT_MAX-3);
	break; 
      }
    }

  printf("\n");
  return(0);
}



