/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>    /* for drand48() */
#include "memory.h"
#include "vector.h"
#include "hash.h"
#include "hash.primes.h"

static char hash_test_rcsid[] = "$Id: hash.test.c,v 1.15 1998/02/07 03:44:28 libpa Exp $";

#define DEBUG                 FALSE
#define VERBOSE               FALSE
#define MONITOR_REDUNDANCY    TRUE

typedef unsigned short hkey_t;
typedef short          datum_t;
typedef int            big_datum_t;


#define MAX_KEY                USHRT_MAX
#define KEY_SIZE              (sizeof(hkey_t))
#define DATUM_SIZE            (sizeof(datum_t))
#define BIG_DATUM_SIZE        (sizeof(big_datum_t))

#define SEED                   8907489
#define INSERT_TEST_SIZE      (SHRT_MAX)
#define REMOVE_TEST_SIZE      (UCHAR_MAX)
#define BAD_SYMBOL            (64)

extern double drand48(void);
extern void   srand48(long);


/************** PROTOTYPES FOR HASH OPERATIONS ******************/

int       test_compare(const void *lkey_p, const void *rkey_p, const void *);
unsigned  test_xor_hash(const void *key_p, const void *);
unsigned  test_reverse_xor_hash(const void *key_p, const void *);

/************** GLOBAL VARIABLE ******************/

/* Note: minload = optimal_load/expansion_factor */
/* warning: maxload / expansion_factor > minload */
/* warning: minload * expansion_factor < maxload */
static hash_class_t Test_Hash_Class = { 0.7, 0.45,
					    0, 0, 0,
					    PRIME_13_COUNT,
					    Primes_13,
					    &test_compare,
					    &test_xor_hash,
					    &test_reverse_xor_hash
					  };

/************** IMPLEMENTATIONS ******************/


int test_compare(const void *left_p, const void *right_p, const void *parameters)
{
  unsigned key_size;
  key_size = *(unsigned *)parameters;
  assert(key_size == KEY_SIZE);
  return(memcmp(left_p, right_p, key_size));
}

unsigned test_xor_hash(const void *key_p, const void *parameters)
{
  unsigned int   hash_value, key_size;
  unsigned char *hash_bytes, *key_bytes;
  int            hash_byte_index;

  assert(key_p); assert(parameters);

  key_size = *(unsigned *) parameters;
  assert(key_size == KEY_SIZE);
  key_bytes = (unsigned char *) key_p;
  hash_bytes = (unsigned char *) &hash_value;

  hash_value = 0;
  hash_byte_index = 0;
  while(key_size > 0)
    {
      assert(hash_byte_index >= 0);
      assert(hash_byte_index < sizeof(unsigned));
      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ *key_bytes;
      hash_byte_index = (hash_byte_index + 1) % sizeof(unsigned);
      key_bytes++;
      key_size -= sizeof(char);
    }
  return(hash_value);
}

unsigned test_reverse_xor_hash(const void *key_p, const void *parameters)
{
  unsigned int   hash_value, key_size;
  unsigned char *hash_bytes, *key_bytes;
  int            hash_byte_index;

  assert(key_p); assert(parameters);

  key_size = *(unsigned *) parameters;
  assert(key_size == KEY_SIZE);
  key_bytes = (unsigned char *) key_p;
  hash_bytes = (unsigned char *) &hash_value;

  hash_value = UINT_MAX;
  hash_byte_index = sizeof(unsigned)-1;
  while(key_size > 0)
    {
      assert(hash_byte_index >= 0);
      assert(hash_byte_index < sizeof(unsigned));
      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ *key_bytes;
      hash_byte_index = (hash_byte_index - 1) % sizeof(unsigned);
      key_bytes++;
      key_size -= sizeof(char);
    }
  return(hash_value);
}

/************************************************************/

void little2nothing(const void *old_datum_p, void *new_datum_p, void *parameters)
{
  const datum_t *old_p;

  assert(new_datum_p == NULL);
  old_p = old_datum_p;
  assert(*old_p == 0);
  return;
}

void nothing2little(const void *old_datum_p, void *new_datum_p, void *parameters)
{
  datum_t *new_p;
  assert(old_datum_p == NULL);
  new_p = new_datum_p;
  *new_p = 0;
  return;
}

void little2big(const void *old_datum_p, void *new_datum_p, void *parameters)
{
  const datum_t *old_p;
  big_datum_t *new_p;

  old_p = old_datum_p;
  new_p = new_datum_p;
  *new_p = *old_p;
}

void big2little(const void *old_datum_p, void *new_datum_p, void *parameters)
{
  const big_datum_t *old_p;
  datum_t *new_p;

  old_p = old_datum_p;
  new_p = new_datum_p;
  *new_p = *old_p;
}

boolean_t everything_is_bad(const void *key_p, const void *datum_p, void *parameters)
{
  return(TRUE);
}

boolean_t nothing_is_bad(const void *key_p, const void *datum_p, void *parameters)
{
  return(FALSE);
}

boolean_t symbol_is_bad(const void *key_p, const void *datum_p, void *parameters)
{
  hkey_t symbol;

  symbol = *(hkey_t *)key_p;
  return(symbol <= BAD_SYMBOL);
}

boolean_t bad_symbol_counter(const void *key_p, void *datum_p, void *parameters)
{
  hkey_t symbol;
  int *bad_symbol_count_p;

  symbol = *(hkey_t *)key_p;
  bad_symbol_count_p = parameters;
  assert(*bad_symbol_count_p >= 0);
  if (symbol <= BAD_SYMBOL) (*bad_symbol_count_p)++;
  return(TRUE);
}

boolean_t verify_indexing(const void *key_p, void *datum_p, unsigned index, void *parameters)
{
  char *index_is_visited;
  hkey_t key;
  datum_t datum;

  key = *(hkey_t *)key_p;
  datum = *(datum_t *)datum_p;
  index_is_visited = (char *)parameters;

  assert(vector_is_valid(index_is_visited));
  assert(vector_index_is_valid(index_is_visited, index));
  assert(!index_is_visited[index]);
  index_is_visited[index] = TRUE;
  return(TRUE);
}


boolean_t symbol_negate_vector(const void *key_p, void *datum_p, void *parameters)
{
  datum_t *vector, datum;
  hkey_t key;

  key = *(hkey_t *)key_p;
  datum = *(datum_t *)datum_p;
  vector = (datum_t *)parameters;

  assert(datum > 0);
  assert(datum == vector[key]);

  vector[key] *= -1;
  return(TRUE);
}


int main(int argc, char **argv)
{
  datum_t       *vector;
  char          *index_is_visited;
  hash_table_t  *table, *predicate;
  hash_class_t  *class;
  hkey_t         key, *key_p;
  datum_t        datum, *datum_p, bad_datum;
  big_datum_t    big_datum, *big_datum_p;
  int            i, size, bad_symbol_count, bad_symbol_recount;
  unsigned       key_size;

  printf("\n\tTesting hash_table_t library.\n");

  srand48(SEED);
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  key_size = KEY_SIZE;
  table = hash_table_create(KEY_SIZE, DATUM_SIZE, &Test_Hash_Class, &key_size);
  assert(table);
  assert(hash_table_size(table) == 0);
  assert(hash_table_load(table) == 0.0);
  assert(hash_table_key_size(table) == KEY_SIZE);
  assert(hash_table_datum_size(table) == DATUM_SIZE);
  assert(hash_table_parameters(table) == &key_size);
  assert(hash_class_minload(hash_table_class(table)) < 
	 hash_class_maxload(hash_table_class(table)));
  assert(hash_class_collisions(hash_table_class(table)) == 0);
  assert(hash_class_redundancies(hash_table_class(table)) == 0);
  assert(hash_class_probes(hash_table_class(table)) == 0);
  printf("%u\tcreated empty table(%d,%d)\n", memory_total_bytes(), 
	 hash_table_key_size(table), hash_table_datum_size(table));

  predicate = hash_table_create(KEY_SIZE, 0, &Test_Hash_Class, &key_size);
  assert(predicate);
  assert(hash_table_size(predicate) == 0);
  assert(hash_table_load(predicate) == 0.0);
  assert(hash_table_key_size(predicate) == KEY_SIZE);
  assert(hash_table_datum_size(predicate) == 0);
  assert(hash_table_parameters(predicate) == &key_size);
  assert(hash_class_minload(hash_table_class(predicate)) < 
	 hash_class_maxload(hash_table_class(predicate)));
  assert(hash_class_collisions(hash_table_class(predicate)) == 0);
  assert(hash_class_redundancies(hash_table_class(predicate)) == 0);
  assert(hash_class_probes(hash_table_class(predicate)) == 0);
  printf("%u\tcreated empty predicate(%d,%d)\n", memory_total_bytes(), 
	 hash_table_key_size(predicate), hash_table_datum_size(predicate));

  class = hash_table_class(table);
  assert(class == hash_table_class(predicate));
  assert(class == &Test_Hash_Class);

  vector = vector_create(sizeof(datum_t), MAX_KEY);
  assert(vector);
  assert(vector_length(vector) == MAX_KEY);
  for(i = 0; i < MAX_KEY; i++) { vector[i] = 0; }
  printf("%u\tallocated test vector[%d]\n", memory_total_bytes(), MAX_KEY);

  /* insert keys into table */
  size = 0;
  bad_symbol_count = 0;
  bad_datum = SHRT_MIN;
#if VERBOSE
  printf("\t");
#endif
  key = 0;
  key_p = &key;
  assert(hash_table_index(table, key_p) == KEY_NOT_FOUND);
  for(i = 0 ; i < INSERT_TEST_SIZE; i++)
    {
      key = (hkey_t)(drand48() * (double)MAX_KEY);
      assert(key < MAX_KEY); 
      assert(key_p == &key);
      
#if MONITOR_REDUNDANCY
      if (hash_table_primary_hash(table, key_p) 
	  == hash_table_secondary_hash(table, key_p))
	{ printf(" *%d", (int)key); }
#endif

      datum_p = hash_table_lookup(table, key_p);
      if (datum_p == NULL)
	{
#if VERBOSE
	  printf(" %d", i);
#endif
	  assert(!hash_table_includes(table, key_p));
	  assert(hash_table_lookup(predicate, key_p) == NULL);
	  assert(!hash_table_includes(predicate, key_p));
	  assert(vector[key] == 0);
	  datum = 0;
	  size++;
	  datum_p = hash_table_update(table, key_p, &datum);
	  assert(&datum != datum_p);
	  assert(*datum_p == datum);
	  if (key <= BAD_SYMBOL) { bad_symbol_count++; }
	}
      else
	{
#if VERBOSE
	  printf(" _");
#endif
	  datum = *datum_p;
	  assert(datum > 0);
	}
      assert(datum_p != NULL);
      assert(datum_p == hash_table_lookup(table, key_p));
      assert(datum == (datum_t)vector[key]);
      assert(datum_p == hash_table_update(table, key_p, &bad_datum));
      assert(*datum_p != bad_datum);
      assert(*datum_p == (datum_t)vector[key]);

      vector[key]++;
      datum++;
      hash_table_insert(table, key_p, &datum);
      datum_p = hash_table_lookup(table, key_p);
      assert(datum == *datum_p);
      assert(datum_p == (datum_t *)hash_table_update(table, key_p, &bad_datum));
      assert(datum_p == (datum_t *)hash_table_lookup(table, key_p));
      assert(datum == *datum_p);
      assert(datum == (datum_t)vector[key]);
      assert(hash_table_includes(table, key_p));
      assert(hash_table_size(table) == size);

      datum_p = hash_table_update(predicate, key_p, &datum);
      hash_table_insert(predicate, key_p, &datum);
      assert(datum_p == NULL);
      assert(hash_table_lookup(predicate, key_p) == NULL);
      assert(hash_table_includes(predicate, key_p));
      assert(hash_table_size(predicate) == size);
    }
#if VERBOSE
  printf("\n");
#endif
  assert(hash_table_is_valid(table));
  assert(!hash_table_is_empty(table));
  assert(hash_table_is_valid(predicate));
  assert(!hash_table_is_empty(predicate));
  assert(bad_symbol_count <= size);
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      datum_p = hash_table_lookup(table, key_p);
      datum = ((datum_p == NULL) ? 0 : *datum_p);
      assert(datum == (datum_t)vector[key]);
      assert(hash_table_includes(predicate, key_p) ? 
	     (datum_p != NULL) : (datum_p == NULL));
    }
  assert(hash_table_size(table) == size);
  assert(hash_table_size(predicate) == hash_table_size(table));
  assert(hash_table_load(predicate) == hash_table_load(table));
  assert(hash_table_capacity(predicate) == hash_table_capacity(table));
  printf("%u\tfilled %d slots with %d symbols (%2.2f load)\n", 
	 memory_total_bytes(), hash_table_capacity(table), 
	 hash_table_size(table), hash_table_load(table));
  printf("\t"); hash_table_monitor(table, stdout);

  /* verify hash_table_prune() with no pruning */
  hash_table_prune(table, &nothing_is_bad, NULL);
  assert(hash_table_is_valid(table));
  assert(!hash_table_is_empty(table));
  hash_table_prune(predicate, &nothing_is_bad, NULL);
  assert(hash_table_is_valid(predicate));
  assert(!hash_table_is_empty(predicate));
  assert(bad_symbol_count <= size);
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      datum_p = hash_table_lookup(table, key_p);
      datum = ((datum_p == NULL) ? 0 : *datum_p);
      assert(datum == (datum_t)vector[key]);
      assert(hash_table_includes(predicate, key_p) ? 
	     (datum_p != NULL) : (datum_p == NULL));
    }
  assert(hash_table_size(table) == size);
  assert(hash_table_size(predicate) == hash_table_size(table));
  assert(hash_table_load(predicate) == hash_table_load(table));
  assert(hash_table_capacity(predicate) == hash_table_capacity(table));
  printf("%u\tverified _prune() identity operation (%d slots, %d symbols)\n", 
	 memory_total_bytes(), hash_table_capacity(table), 
	 hash_table_size(table));

  /* verify map and bad_symbol_counter() */
  bad_symbol_recount = 0;
  hash_table_map(table, &bad_symbol_counter, &bad_symbol_recount);
  assert(bad_symbol_recount == bad_symbol_count);
  assert(hash_table_is_valid(table));
  bad_symbol_recount = 0;
  hash_table_map(predicate, &bad_symbol_counter, &bad_symbol_recount);
  assert(bad_symbol_recount == bad_symbol_count);
  assert(hash_table_is_valid(predicate));
  printf("\tverified %d bad symbols in table/predicate\n", bad_symbol_count);

  /* remove a few keys from the table (if they're in the table) */
  for(i = 0 ; i < REMOVE_TEST_SIZE; i++)
    {
      key = (hkey_t)(drand48() * (double)MAX_KEY);
      key_p = &key;
      assert(key < MAX_KEY); 
      assert((vector[key] == 0) || 
	     ((hash_table_lookup(table, key_p) != NULL) &&
	      hash_table_includes(predicate, key_p)));
      assert((vector[key] > 0) || 
	     ((hash_table_lookup(table, key_p) == NULL) &&
	      !hash_table_includes(predicate, key_p)));
      
      if (vector[key] > 0) 
	{
	  size--;
	  if (key <= BAD_SYMBOL) { bad_symbol_count--; }
	}
      hash_table_remove(table, key_p);
      hash_table_remove(predicate, key_p);
      assert(hash_table_lookup(table, key_p) == NULL);
      assert(!hash_table_includes(table, key_p));
      assert(!hash_table_includes(predicate, key_p));

      vector[key] = 0;
    }
  assert(hash_table_size(table) == size);
  assert(hash_table_size(predicate) == size);
  assert(bad_symbol_count <= size);
  assert(hash_table_is_valid(table));
  assert(hash_table_is_valid(predicate));
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      datum_p = hash_table_lookup(table, key_p);
      datum = ((datum_p == NULL) ? 0 : *datum_p);
      assert(datum == (datum_t)vector[key]);
      assert(hash_table_includes(predicate, key_p) ?
	     (vector[key] > 0) : (vector[key] == 0));
    }
  printf("%u\tshrunk table/predicate to %d entries in %d slots (%2.2f load)\n", 
	 memory_total_bytes(), hash_table_size(table),
	 hash_table_capacity(table), hash_table_load(table));
  printf("\t: %d probes, %d collisions (%.1f/probe), %d redundancies\n",
	 (int)hash_class_probes(class), (int)hash_class_collisions(class), 
	 hash_class_mean_collisions(class), (int)hash_class_redundancies(class));

  /* verify hash_table_resize_datum() by sizing up and then down */
  hash_table_resize_datum(table, BIG_DATUM_SIZE, &little2big, NULL);
  assert(hash_table_datum_size(table) == BIG_DATUM_SIZE);
  assert(hash_table_is_valid(table));
  hash_table_resize_datum(predicate, DATUM_SIZE, &nothing2little, NULL);
  assert(hash_table_datum_size(predicate) == DATUM_SIZE);
  assert(hash_table_is_valid(predicate));
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      big_datum_p = hash_table_lookup(table, key_p);
      big_datum = ((big_datum_p == NULL) ? 0 : *big_datum_p);
      assert(big_datum == (datum_t)vector[key]);
      assert(hash_table_includes(predicate, key_p) ?
	     (vector[key] > 0) : (vector[key] == 0));
    }
  printf("%u\tupsized table/predicate datum_p by 2 bytes\n", memory_total_bytes());

  hash_table_resize_datum(table, DATUM_SIZE, &big2little, NULL);
  assert(hash_table_datum_size(table) == DATUM_SIZE);
  assert(hash_table_is_valid(table));
  hash_table_resize_datum(predicate, 0, &little2nothing, NULL);
  assert(hash_table_datum_size(predicate) == 0);
  assert(hash_table_is_valid(predicate));
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      datum_p = hash_table_lookup(table, key_p);
      datum = ((datum_p == NULL) ? 0 : *datum_p);
      assert(datum == (datum_t)vector[key]);
      assert(hash_table_includes(predicate, key_p) ?
	     (vector[key] > 0) : (vector[key] == 0));
    }
  printf("%u\tdownsized table/predicate datum_p by 2 bytes\n", memory_total_bytes());

  /* verify hash_table_index_map() */
  index_is_visited = vector_create(sizeof(char), hash_table_size(table));
  assert(vector_is_valid(index_is_visited));
  assert(vector_length(index_is_visited) == hash_table_size(table));
  (void) memset(index_is_visited, 0, hash_table_size(table));
  hash_table_index_map(table, &verify_indexing, (void *)index_is_visited);
  for(i = 0; i < hash_table_size(table); i++) { assert(index_is_visited[i]); }
  vector_destroy(index_is_visited); index_is_visited = NULL;
  printf("%u\tverified hash_table_index_map() for table\n", memory_total_bytes());

  /* verify hash_table_map() */
  assert(hash_table_datum_size(table) == DATUM_SIZE);
  assert(hash_table_datum_size(predicate) == 0);
  hash_table_map(table, &symbol_negate_vector, (void *)vector);
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      datum_p = hash_table_lookup(table, key_p);
      datum = ((datum_p == NULL) ? 0 : *datum_p);
      assert(vector[key] <= 0);
      assert(datum == (datum_t)(-vector[key]));
    }
  vector_destroy(vector); vector = NULL;
  printf("%u\tverified hash_table_map() for table\n", memory_total_bytes());

  /* verify bad symbol count and then prune all bad symbols */
  bad_symbol_recount = 0;
  hash_table_map(table, &bad_symbol_counter, &bad_symbol_recount);
  assert(bad_symbol_recount == bad_symbol_count);
  assert(hash_table_size(table) == size);
  hash_table_prune(table, &symbol_is_bad, NULL);
  size -= bad_symbol_count;
  assert(hash_table_is_valid(table));
  assert(hash_table_size(table) == size);
  bad_symbol_recount = 0;
  hash_table_map(table, &bad_symbol_counter, &bad_symbol_recount);
  assert(bad_symbol_recount == 0);
  printf("%u\tcorrectly pruned %d bad symbols from table\n", 
	 memory_total_bytes(), bad_symbol_count);

  bad_symbol_recount = 0;
  hash_table_map(predicate, &bad_symbol_counter, &bad_symbol_recount);
  assert(bad_symbol_recount == bad_symbol_count);
  hash_table_prune(predicate, &symbol_is_bad, NULL);
  assert(hash_table_size(predicate) == size);
  hash_table_prune(predicate, &symbol_is_bad, NULL);
  assert(hash_table_is_valid(predicate));
  assert(hash_table_size(predicate) == size);
  assert(hash_table_is_valid(predicate));
  bad_symbol_recount = 0;
  hash_table_map(predicate, &bad_symbol_counter, &bad_symbol_recount);
  assert(bad_symbol_recount == 0);
  printf("%u\tcorrectly pruned %d bad symbols from predicate\n", 
	 memory_total_bytes(), bad_symbol_count);

  /* prune remaining symbols from table and predicate */
  hash_table_prune(table, &everything_is_bad, NULL);
  assert(hash_table_size(table) == 0);
  assert(hash_table_is_empty(table));
  assert(hash_table_is_valid(table));
  hash_table_prune(predicate, &everything_is_bad, NULL);
  assert(hash_table_size(predicate) == 0);
  assert(hash_table_is_empty(predicate));
  assert(hash_table_is_valid(predicate));
  printf("%u\tcorrectly pruned all remaining symbols from table/predicate\n", 
	 memory_total_bytes());

  /* verify hash_table_clear() */
  hash_table_clear(table);
  assert(hash_table_size(table) == 0);
  assert(hash_table_is_empty(table));
  assert(hash_table_is_valid(table));
  for(key = 0; key < MAX_KEY; key++) 
    { 
      assert(!hash_table_includes(table, &key));
      assert(hash_table_lookup(table, &key) == NULL); 
      assert(hash_table_index(table, &key) == KEY_NOT_FOUND); 
    }
  hash_table_clear(predicate);
  assert(hash_table_size(predicate) == 0);
  assert(hash_table_is_empty(predicate));
  assert(hash_table_is_valid(predicate));
  for(key = 0; key < MAX_KEY; key++) 
    { 
      assert(!hash_table_includes(predicate, &key)); 
      assert(hash_table_index(predicate, &key) == KEY_NOT_FOUND); 
      assert(hash_table_lookup(predicate, &key) == NULL); 
    }
  printf("%u\tverified hash_table_clear() for table/predicate\n", 
	 memory_total_bytes());

  /* prune empty hash_table_t */
  hash_table_prune(table, &everything_is_bad, NULL);
  hash_table_is_valid(table);
  hash_table_prune(predicate, &everything_is_bad, NULL);
  hash_table_is_valid(predicate);
  printf("%u\tcorrectly pruned empty table/predicate\n", memory_total_bytes());

  hash_table_destroy(table); table = NULL;
  hash_table_destroy(predicate); predicate = NULL;
  printf("%u\tdestroyed symbol hash_table_t\n", memory_total_bytes());
  printf("\tmemory leak == %u bytes in %u blocks\n", 
	 memory_total_bytes(), memory_total_blocks());
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);

  printf("\tsizeof(hash_table_t) == %d\n", sizeof(hash_table_t));
  printf("\thash_table_t library passes all tests.\n");
  return(0);
}



