/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "memory.h"
#include "vector.h"
#include "string.hash.h"
#include "hash.primes.h"

static char string_hash_rcsid[] = "$Id: string.hash.c,v 1.6 1998/02/07 03:44:28 libpa Exp $";

/************** PROTOTYPES FOR HASH OPERATIONS ******************/

int       string_compare(const void *lkey_p, const void *rkey_p, const void *);
unsigned  string_xor_hash(const void *key_p, const void *);
unsigned  string_reverse_xor_hash(const void *key_p, const void *);
unsigned  string_nonlocal_xor_hash(const void *key_p, const void *);
unsigned  string_nonlocal_reverse_xor_hash(const void *key_p, const void *);
unsigned  string_accumulate_hash(const void *key_p, const void *);
unsigned  string_shift_accumulate_hash(const void *key_p, const void *);
unsigned  string_linear_hash(const void *key_p, const void *);

/************** GLOBAL VARIABLE ******************/

/* warning: maxload / expansion_factor > minload */
/* warning: minload * expansion_factor < maxload */
static hash_class_t String_Hash_Class = { 0.8, 0.3,
					    0, 0, 0,
					    PRIME_18_COUNT,
					    Primes_18,
					    &string_compare,
					    &string_nonlocal_xor_hash,
					    &string_reverse_xor_hash
					  };


/************** IMPLEMENTATIONS ******************/

int string_compare(const void *left_p, const void *right_p, const void *parameters)
{
  const char *left_string, *right_string;
  assert(left_p); assert(right_p); assert(parameters == NULL);
  left_string = *(char **) left_p;
  right_string = *(char **) right_p;
  return strcmp(left_string, right_string);
}

unsigned string_linear_hash(const void *key_p, const void *parameters) { return(1); }

unsigned string_accumulate_hash(const void *key_p, const void *parameters)
{
  unsigned char *string;
  unsigned hash_value;
  int i;

  assert(key_p); assert(parameters == NULL);
  string = *(unsigned char **) key_p;

  i = 0;
  hash_value = 0;
  while(string[i] != '\0')
    {
      hash_value += string[i];
      i++;
    }
  return(hash_value);
}

unsigned string_shift_accumulate_hash(const void *key_p, const void *parameters)
{
  unsigned char *string;
  unsigned hash_value;
  int i;

  assert(key_p); assert(parameters == NULL);
  string = *(unsigned char **)key_p;

  i = 0;
  hash_value = 0;
  while(string[i] != '\0')
    {
      assert(string[0] != '\0');
      hash_value = hash_value << 4;   /* shift by 4 of 8 bits */
      hash_value += string[0];
      i++;
    }
  return(hash_value);
}

unsigned string_xor_hash(const void *key_p, const void *parameters)
{
  unsigned int  hash_value, hash_byte_index, i;
  unsigned char *hash_bytes, *string;

  assert(key_p); assert(parameters == NULL);
  string = *(unsigned char **)key_p;
  hash_bytes = (unsigned char *) &hash_value;

  i = 0;
  hash_value = 0;
  hash_byte_index = 0;
  while(string[i] != '\0')
    {
      assert(hash_byte_index < sizeof(unsigned));

      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ string[i];
      hash_byte_index = (hash_byte_index + 1) % sizeof(unsigned);
      i++;
    }
  return(hash_value);
}

unsigned string_nonlocal_xor_hash(const void *key_p, const void *parameters)
{
  unsigned char *hash_bytes, *string;
  unsigned int   hash_value, hash_byte_index, i;

  assert(key_p); assert(parameters == NULL);
  string = *(unsigned char **)key_p;
  hash_bytes = (unsigned char *) &hash_value;

  i = 0;
  hash_value = 0;
  hash_byte_index = 0;
  while(string[i] != '\0')
    {
      assert(hash_byte_index < sizeof(unsigned));

      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ string[i];
      hash_byte_index = (hash_byte_index + string[i]) % sizeof(unsigned);
      i++;
    }
  return(hash_value);
}

unsigned string_reverse_xor_hash(const void *key_p, const void *parameters)
{
  unsigned char *hash_bytes, *string;
  unsigned int   hash_value, hash_byte_index, i;

  assert(key_p); assert(parameters == NULL);
  string = *(unsigned char **) key_p;
  hash_bytes = (unsigned char *) &hash_value;

  i = 0;
  hash_value = UINT_MAX;
  hash_byte_index = sizeof(unsigned)-1;
  while(string[i] != '\0')
    {
      assert(hash_byte_index < sizeof(unsigned));

      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ string[i];
      hash_byte_index = (hash_byte_index - 1) % sizeof(unsigned);
      i++;
    }
  return(hash_value);
}

unsigned string_nonlocal_reverse_xor_hash(const void *key_p, const void *parameters)
{
  unsigned char *hash_bytes;
  unsigned int   hash_value, hash_byte_index, i;
  const unsigned char *string;

  assert(key_p); assert(parameters == NULL);
  string = *(unsigned char **)key_p;
  hash_bytes = (unsigned char *)&hash_value;

  i = 0;
  hash_value = UINT_MAX;
  hash_byte_index = sizeof(unsigned)-1;
  while(string[i] != '\0')
    {
      assert(string[i] != '\0');
      assert(hash_byte_index < sizeof(unsigned));

      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ string[i];
      hash_byte_index = (hash_byte_index - string[i]) % sizeof(unsigned);
      i++;
    }
  return(hash_value);
}

/************************************************************/

string_hash_t *string_hash_create(unsigned datum_size)
{
  string_hash_t *table;
  assert(datum_size >= 0);
  table = hash_table_create(sizeof(char *),
			    datum_size, 
			    &String_Hash_Class,
			    NULL);
  assert(string_hash_is_valid(table));
  return(table);
}

void string_hash_destroy(string_hash_t *table)
{
  string_hash_clear(table);
  hash_table_destroy(table);
}

boolean_t string_hash_is_valid(const string_hash_t *table)
{
  if (!hash_table_is_valid(table)) { return(FALSE); }

  return ((hash_table_key_size(table) == sizeof(char *)) &&
	  (hash_table_parameters(table) == NULL) &&
	  (hash_table_class(table) == &String_Hash_Class));
}

boolean_t string_hash_includes(const string_hash_t *table, const char *string)
{
  return(hash_table_includes(table, &string));
}

void *string_hash_lookup(const string_hash_t *table, const char *string)
{
  return(hash_table_lookup(table, &string));
}

/************************************************************
 * MUTATORS
 ************************************************************/

void string_hash_insert(string_hash_t *table, const char *string, const void *datum_p)
{
  unsigned string_length, index, datum_size;
  char *copied_string;
  void *table_datum_p;

  assert(table); assert(string);

  datum_size = hash_table_datum_size(table);
  index = hash_table_index(table, &string);
  if (index != KEY_NOT_FOUND)
    { 
      table_datum_p = hash_table_datum_p(table, index);
      (void) memcpy(table_datum_p, datum_p, datum_size); 
      return;
    }
  assert(index == KEY_NOT_FOUND);

  string_length = strlen(string);
  copied_string = memory_create(string_length+1);
  assert(memory_is_valid(copied_string));
  assert(memory_bytes(copied_string) == string_length+1);
  (void) strcpy(copied_string, string);
  assert(string_compare(&copied_string, &string, NULL) == 0);

  hash_table_insert(table, &copied_string, datum_p);
  assert(string_hash_includes(table, string));
}

void *string_hash_update(string_hash_t *table, const char *string, const void *initial_p)
{
  unsigned string_length, index, datum_size;
  char *copied_string;
  void *table_datum_p;

  assert(table); assert(string);

  datum_size = hash_table_datum_size(table);
  index = hash_table_index(table, &string);
  if (index != KEY_NOT_FOUND)
    { 
      table_datum_p = hash_table_datum_p(table, index);
      return(table_datum_p); 
    }
  assert(index == KEY_NOT_FOUND);

  string_length = strlen(string);
  copied_string = memory_create(string_length+1);
  assert(memory_is_valid(copied_string));
  assert(memory_bytes(copied_string) == string_length+1);
  (void) strcpy(copied_string, string);
  assert(string_compare(&copied_string, &string, NULL) == 0);

  table_datum_p = hash_table_update(table, &copied_string, initial_p);
  assert(memcmp(table_datum_p, initial_p, datum_size) == 0);
  assert(string_hash_includes(table, string));
  return(table_datum_p);
}

void string_hash_remove(string_hash_t *table, const char *string)
{
  void *block;
  unsigned index;

  index = hash_table_index(table, &string);
  if (index != KEY_NOT_FOUND)
    {
      block = *(void **)hash_table_key_p(table, index);
      assert(memory_is_valid(block));
      hash_table_remove(table, &string);
      memory_destroy(block);
    }
}

boolean_t prune_all(const char *string, const void *datum_p, void *parameters) 
{ 
  return(TRUE); 
}
void string_hash_clear(string_hash_t *table) 
{ 
  string_hash_prune(table, &prune_all, NULL);
  assert(string_hash_size(table) == 0);
  assert(string_hash_is_valid(table));
}

typedef struct prune_enclosure_tag
{
  boolean_t (*predicate)(const char *, const void *, void *);
  void      *parameters;
} prune_enclosure_t;

boolean_t string_pruner(const void *string_p, const void *datum_p, void *parameters)
{
  char *string;
  prune_enclosure_t *enclosure;
  boolean_t prune_me;

  string = *(char **)string_p;
  enclosure = parameters;
  prune_me = enclosure->predicate(string, datum_p, enclosure->parameters);
  if (prune_me)
    {
      assert(memory_is_valid(string));
      memory_destroy(string); 
#ifndef NDEBUG
      *(char **)string_p = NULL;
#endif
    }
  return(prune_me);
}

void string_hash_prune(string_hash_t *table, 
		       boolean_t (*prune_me)(const char *string, 
					     const void *datum_p, 
					     void *parameters),
		       void *parameters)
{
  prune_enclosure_t enclosure;
  enclosure.predicate = prune_me;
  enclosure.parameters = parameters;
  hash_table_prune(table, &string_pruner, &enclosure);
}


/************************************************************
 *  HIGHER ORDER FUNCTION
 ************************************************************/

typedef struct map_enclosure_tag
{
  boolean_t (*proc)(const char *, void *, void *);
  void       *parameters;
} map_enclosure_t;

boolean_t string_mapper(const void *string_p, void *datum_p, void *parameters)
{
  const char *string = *(char **)string_p;
  map_enclosure_t *enclosure = parameters;
  boolean_t is_alive;
  is_alive = enclosure->proc(string, datum_p, enclosure->parameters);
  return(is_alive);
}

void string_hash_map(string_hash_t *table, 
		     boolean_t (*proc)(const char *string, void *datum_p, void *parameters),
		     void *parameters)
{
  map_enclosure_t enclosure;
  enclosure.proc = proc;
  enclosure.parameters = parameters;
  hash_table_map(table, &string_mapper, &enclosure);
}

/************************************************************
 * INDEXING
 ************************************************************/

unsigned string_hash_index(const string_hash_t *table, const char *string)
{
  return(hash_table_index(table, &string));
}

const char *string_hash_string(const string_hash_t *table, unsigned index)
{
  const char **string_p;
  string_p = (const char **)hash_table_key_p(table, index);
  return(*string_p);
}

void *string_hash_datum_p(const string_hash_t *table, unsigned index)
{
  return(hash_table_datum_p(table, index));
}


typedef struct index_enclosure_tag
{
  boolean_t (*proc)(const char *, void *, unsigned, void *);
  void       *parameters;
} index_enclosure_t;

boolean_t string_indexer(const void *string_p, void *datum_p, unsigned index, 
			 void *parameters)
{
  const char *string = *(char **)string_p;
  index_enclosure_t *enclosure = parameters;
  boolean_t is_alive;
  is_alive = enclosure->proc(string, datum_p, index, enclosure->parameters);
  return(is_alive);
}

void string_hash_index_map(string_hash_t *table, 
			   boolean_t (*proc)(const char *string, void *datum_p, 
					     unsigned index, void *parameters),
			   void *parameters)
{
  index_enclosure_t enclosure;
  enclosure.proc = proc;
  enclosure.parameters = parameters;
  hash_table_index_map(table, &string_indexer, &enclosure);
}

/************************************************************
 * PERFORMANCE TUNING
 ************************************************************/

void string_hash_monitor(const string_hash_t *table, FILE *stream)
{
  hash_table_monitor(table, stream);
}

