/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>    /* for drand48() */
#include "memory.h"
#include "vector.h"
#include "hash.class.h"
#include "string.hash.h"

static char string_hash_test_rcsid[] = "$Id: string.hash.test.c,v 1.4 1998/02/07 03:44:28 libpa Exp $";

#define DEBUG                 TRUE
#define VERBOSE               TRUE

typedef short    datum_t;
typedef int      big_datum_t;
typedef unsigned index_t;

#define DATUM_SIZE            (sizeof(datum_t))
#define BIG_DATUM_SIZE        (sizeof(big_datum_t))

#define SEED                   19613549
#define SUPERSTRING_SIZE       USHRT_MAX
#define STRING_PRUNE_SIZE      64
#define BOUNDARY_PR            0.1
#define BAD_SYMBOL     (32)


extern double drand48(void);
extern void   srand48(long);

boolean_t string_is_bad(const char *string, const void *datum_p, void *parameters)
{
  assert(string[0] != '\0');
  while(string[0] != '\0')
    {
      assert(string[0] > '\0'); assert(string[0] < CHAR_MAX);
      if (string[0] <= BAD_SYMBOL) { return(TRUE); }
      string++;
    }
  return(FALSE);
}

boolean_t bad_string_counter(const char *string, void *datum_p, void *parameters)
{
  int *bad_string_count_p;

  bad_string_count_p = parameters;
  assert(*bad_string_count_p >= 0);
  assert(string[0] != '\0');

  while(string[0] != '\0')
    {
      assert(string[0] > 0); assert(string[0] < CHAR_MAX);
      if (string[0] <= BAD_SYMBOL) 
	{ (*bad_string_count_p)++; return(TRUE); }
      string++;
    }
  return(TRUE);
}

boolean_t verify_string2index(const char *string, void *datum_p, void *parameters)
{
  char    **index2string = (char **)parameters;
  index_t  *index_p = (index_t *)datum_p;
  if (strcmp(string, index2string[*index_p]) != 0) { abort(); }
  return(TRUE);
}


int main(int argc, char **argv)
{
  char   symbol;  
  char **index2string, *superstring, *substring;
  string_hash_t *table;
  index_t *index_p, index, bad_index;
  int i, j, string_count, distinct_string_count, substring_length;
  int bad_string_count, bad_string_recount;

  srand48(SEED);
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("\n\tTesting string_hash_t library...\n");

  /* allocate the string hash table (inverse table) */
  table = string_hash_create(sizeof(index_t));
  assert(table);
  assert(string_hash_size(table) == 0);
  assert(string_hash_datum_size(table) == sizeof(index_t));
  assert(string_hash_is_valid(table));
  printf("%u\tcreated empty table(%d) for %d-byte indices\n",
	  memory_total_bytes(), string_hash_datum_size(table), sizeof(index_t));

  /* allocate superstring and initialize its contents */
  superstring = memory_create(SUPERSTRING_SIZE+1);
  assert(superstring);
  superstring[SUPERSTRING_SIZE] = '\0'; /* terminate it */
  for(i = 0; i < SUPERSTRING_SIZE; i++)
    {
      symbol = 1 + (drand48() * (double)(CHAR_MAX-1));
      assert((symbol > 0) && (symbol < CHAR_MAX));
      superstring[i] = symbol;
    }
  printf("%u\tallocated %d-character superstring\n",
	  memory_total_bytes(), SUPERSTRING_SIZE);

  /* allocate index2string[] forward table and insert inital substring */
  index2string = vector_create(sizeof(char *), 1);
  assert(index2string);
  assert(vector_length(index2string) == 1);
  index2string[0] = superstring;
  string_count = 1;

  /* randomly partition superstring into nonempty substrings */
  i = 1;
  substring_length = 1;
  substring = superstring;
  while(i < SUPERSTRING_SIZE-1)
    {
      if (drand48() < BOUNDARY_PR) 
	{
	  /* terminate active substring */
	  assert(i < SUPERSTRING_SIZE);
	  superstring[i] = '\0'; 
	  i++;
	  assert(substring[0] != '\0');
	  assert(strlen((char *)substring) == substring_length);
	  assert(substring_length > 0);

	  /* activate next substring and put it in the table */
	  substring = superstring+i;
	  i++;
	  substring_length = 1;
	  index2string = vector_append(index2string, &substring);
	  assert(index2string);
	  assert(index2string[string_count] == substring);
	  string_count++;
	}
      else
	{
	  i++;
	  substring_length++;
	}
    }
  assert(string_count > 1);
  assert(superstring[0] != '\0');
  assert(superstring[SUPERSTRING_SIZE] == '\0');
  assert(superstring[SUPERSTRING_SIZE-1] != '\0');
  assert(vector_length(index2string) == string_count);
  printf("%u\tpartitioned %d-character superstring into %d substrings\n",
	  memory_total_bytes(), SUPERSTRING_SIZE, string_count);

  /* now insert each distinct substring and its index into the table */
  assert(table);
  assert(string_hash_is_empty(table));
  bad_index = UINT_MAX;
  bad_string_count = 0;
  distinct_string_count = 0;
  for(i = 0; i < string_count; i++)
    {
      substring = index2string[i];
      assert(strlen(substring) > 0);

      index_p = string_hash_lookup(table, substring);
      assert((index_p == NULL) || (strcmp(index2string[*index_p], substring) == 0));
      if (index_p == NULL)  /* is this substring a virgin? */
	{
	  assert(!string_hash_includes(table, substring));
	  assert(string_hash_index(table, substring) == STRING_NOT_FOUND);
	  index = i;
	  distinct_string_count++;
	  index_p = string_hash_update(table, substring, &index);
	  assert(index_p != &index); assert(*index_p == index);
	  assert(index_p == (index_t *)string_hash_lookup(table, substring));
	  assert(index_p == (index_t *)string_hash_update(table, substring, &bad_index));
	  assert(index_p != &bad_index); assert(*index_p == index);
	  string_hash_insert(table, substring, &bad_index);
	  assert(index_p == (index_t *)string_hash_lookup(table, substring));
	  assert(index_p != &bad_index); assert(*index_p == bad_index);
	  string_hash_insert(table, substring, &index);
	  assert(index_p == (index_t *)string_hash_update(table, substring, &bad_index));
	  assert(index_p != &bad_index); assert(*index_p == index);
	  assert(string_hash_includes(table, substring));
	  assert(string_hash_size(table) == distinct_string_count);

	  /* count first time a bad substring occurs */
	  assert(substring == index2string[i]);
	  while(substring[0] != '\0')
	    {
	      if (substring[0] <= BAD_SYMBOL) { bad_string_count++; break; }
	      assert(substring[0] > BAD_SYMBOL);
	      substring++;
	    }
	}
      else
	{
	  assert(string_hash_includes(table, substring));
	  index = string_hash_index(table, substring);
	  assert(strcmp(string_hash_string(table, index), substring) == 0);
	  assert(index_p == (index_t *)string_hash_datum_p(table, index));
	  /* verify that _update() has no effect for existing string */
	  assert(index_p == (index_t *)string_hash_update(table, substring, &bad_index));
	  assert(*index_p < i); assert(*index_p != bad_index);
	  assert(index_p == string_hash_lookup(table, substring));
	}
    }
  assert(string_hash_is_valid(table));
  assert(distinct_string_count > 0);
  assert(distinct_string_count <= string_count);
  assert(bad_string_count >= 0);
  assert(bad_string_count <= distinct_string_count);
  assert(string_hash_size(table) == distinct_string_count);
  printf("\tgenerated %d substrings, %d are distinct\n\t",
	  string_count, distinct_string_count);
  string_hash_monitor(table, stdout);

  /* verify bad string count */
  bad_string_recount = 0;
  string_hash_map(table, &bad_string_counter, &bad_string_recount);
  assert(bad_string_recount == bad_string_count);
  printf("\tverified bad string count %d\n", bad_string_count);

  /* verify hash table against index2string[] table */
  string_hash_map(table, &verify_string2index, (void *)index2string);
  printf("\tverified string2index() using index2string[]\n");

  /* remove some strings chosen at random */
  for(i = 0; i < STRING_PRUNE_SIZE; i++)
    {
      j = drand48() * (double)string_count;
      substring = index2string[j];
      string_hash_remove(table, substring);
      index2string[j] = NULL;
      assert(string_hash_lookup(table, substring) == NULL);
      assert(!string_hash_includes(table, substring));
      assert(string_hash_index(table, substring) == STRING_NOT_FOUND);
  }
  assert(string_hash_is_valid(table));
  string_hash_map(table, &verify_string2index, (void *)index2string);
  printf("%u\tshrunk table to %d entries\n\t", 
	 memory_total_bytes(), string_hash_size(table));
  string_hash_monitor(table, stdout);

  /* recalculate bad string count and prune all bad strings */
  bad_string_recount = 0;
  string_hash_map(table, &bad_string_counter, &bad_string_recount);
  assert(bad_string_recount <= bad_string_count);
  bad_string_count = bad_string_recount;
  distinct_string_count = string_hash_size(table);
  string_hash_prune(table, &string_is_bad, NULL);
  distinct_string_count -= bad_string_count;
  assert(string_hash_is_valid(table));
  assert(string_hash_size(table) == distinct_string_count);
  bad_string_recount = 0;
  string_hash_map(table, &bad_string_counter, &bad_string_recount);
  assert(bad_string_recount == 0);
  printf("%u\tcorrectly pruned %d bad strings from table\n", 
	  memory_total_bytes(), bad_string_count);

  /* verify string_hash_clear() */
  string_hash_clear(table);
  assert(string_hash_size(table) == 0);
  assert(string_hash_is_empty(table));
  assert(string_hash_is_valid(table));
  printf("%u\tverified string_hash_clear()\n", memory_total_bytes());

  /* clean up */
  vector_destroy(index2string); index2string = NULL;
  memory_destroy(superstring); superstring = NULL;
  printf("%d\tdestroyed superstring and index2string vector\n", memory_total_bytes());

  string_hash_destroy(table); table = NULL;
  printf("%u\tdestroyed string_hash_t\n", memory_total_bytes());
  printf("\tmemory leak == %u bytes in %u blocks\n", 
	  memory_total_bytes(), memory_total_blocks());
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  printf("\tsizeof(string_hash_t) == %d\n", sizeof(string_hash_t));
  printf("\tstring_hash_t library passes all tests.\n");
  return(0);
}



