/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <string.h>
#include <limits.h>
#include "memory.h"
#include "symbol.hash.h"
#include "hash.primes.h"

static char symbol_hash_rcsid[] = "$Id: symbol.hash.c,v 1.6 1998/02/07 03:44:28 libpa Exp $";


/************** HASH_CLASS_T PROTOTYPES and GLOBAL VARIABLE ******************/

int       symbol_compare(const void *lkey_p, const void *rkey_p, const void *);
unsigned  symbol_xor_hash(const void *key_p, const void *);
unsigned  symbol_reverse_xor_hash(const void *key_p, const void *);
unsigned  symbol_accumulate_hash(const void *key_p, const void *);
unsigned  symbol_shift_accumulate_xor_hash(const void *key_p, const void *);

/* warning: maxload / expansion_factor > minload */
/* warning: minload * expansion_factor < maxload */
static hash_class_t Symbol_Hash_Class = { 0.8, 0.4,
					    0, 0, 0,
					    PRIME_15_COUNT,
					    Primes_15,
					    &symbol_compare,
					    &symbol_xor_hash,
					    &symbol_reverse_xor_hash
					  };

/************** IMPLEMENTATIONS ******************/


int symbol_compare(const void *left_p, const void *right_p, const void *parameters)
{
  unsigned key_size;
  key_size = *(unsigned *)parameters;
  return(memcmp(left_p, right_p, key_size));
}

unsigned symbol_accumulate_hash(const void *key_p, const void *parameters)
{
  unsigned int   hash_value, key_size;
  unsigned char *key_bytes;

  assert(key_p); assert(parameters);

  hash_value = 0;
  key_bytes = (unsigned char *)key_p;
  key_size = *(unsigned *)parameters;
  assert(key_size > 0);
  while(key_size > 0)
    {
      hash_value += *key_bytes;
      key_bytes++;
      key_size--;
    }
  return(hash_value);
}

unsigned symbol_shift_accumulate_hash(const void *key_p, const void *parameters)
{
  unsigned int   hash_value, key_size;
  unsigned char *key_bytes;

  assert(key_p); assert(parameters);

  hash_value = 0;
  key_bytes = (unsigned char *)key_p;
  key_size = *(unsigned *)parameters;
  assert(key_size > 0);
  while(key_size > 0)
    {
      hash_value *= 64;
      hash_value += *key_bytes;
      key_bytes++;
      key_size--;
    }

  return(hash_value);
}

unsigned symbol_xor_hash(const void *key_p, const void *parameters)
{
  unsigned int   hash_value, key_size, hash_byte_index;
  unsigned char *hash_bytes, *key_bytes;

  assert(key_p); assert(parameters);

  key_size = *(unsigned *) parameters;
  key_bytes = (unsigned char *) key_p;
  hash_bytes = (unsigned char *) &hash_value;

  hash_value = 0;
  hash_byte_index = 0;
  while(key_size > 0)
    {
      assert(hash_byte_index >= 0);
      assert(hash_byte_index < sizeof(unsigned));
      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ *key_bytes;
      hash_byte_index = (hash_byte_index + 1) % sizeof(unsigned);
      key_bytes++;
      key_size -= sizeof(char);
    }
  return(hash_value);
}

unsigned symbol_reverse_xor_hash(const void *key_p, const void *parameters)
{
  unsigned int   hash_value, key_size, hash_byte_index;
  unsigned char *hash_bytes, *key_bytes;

  assert(key_p); assert(parameters);

  key_size = *(unsigned *) parameters;
  key_bytes = (unsigned char *) key_p;
  hash_bytes = (unsigned char *) &hash_value;

  hash_value = UINT_MAX;
  hash_byte_index = sizeof(unsigned)-1;
  while(key_size > 0)
    {
      assert(hash_byte_index >= 0);
      assert(hash_byte_index < sizeof(unsigned));
      hash_bytes[hash_byte_index] = hash_bytes[hash_byte_index] ^ *key_bytes;
      hash_byte_index = (hash_byte_index - 1) % sizeof(unsigned);
      key_bytes++;
      key_size -= sizeof(char);
    }
  return(hash_value);
}

unsigned linear_secondary_hash(const void *key_p, const void *parameters)
{
  assert(key_p);
  return(1);
}

/************************************************************/
/************************************************************/

symbol_hash_t *symbol_hash_create(unsigned symbol_size, unsigned datum_size)
{
  unsigned *key_size_p;
  symbol_hash_t *table;

  key_size_p = memory_create(sizeof(unsigned int));
  *key_size_p = symbol_size;
  table = hash_table_create(symbol_size, 
			    datum_size, 
			    &Symbol_Hash_Class,
			    key_size_p);
  if (table == NULL) { memory_destroy(key_size_p); }
  assert(symbol_hash_is_valid(table));
  return(table);
}

void symbol_hash_destroy(symbol_hash_t *table)
{
  unsigned *key_size_p;

  assert(symbol_hash_is_valid(table));
  key_size_p = (unsigned *)table->parameters;
  hash_table_destroy(table);
  memory_destroy(key_size_p);
}

boolean_t symbol_hash_is_valid(const symbol_hash_t *table)
{
  unsigned *key_size_p;

  if (!hash_table_is_valid(table)) { return(FALSE); }
  key_size_p = (unsigned *) table->parameters;
  return((table->class == &Symbol_Hash_Class) &&
	 (key_size_p != NULL) &&
	 memory_is_valid(key_size_p) &&
	 (memory_bytes(key_size_p) == sizeof(unsigned int)) &&
	 (*key_size_p == hash_table_key_size(table)));
}

void symbol_hash_prune(symbol_hash_t *table, 
		       boolean_t (*prune_me)(const void *symbol_p, 
					     const void *datum_p, 
					     void *parameters),
		       void *parameters)
{
  hash_table_prune(table, prune_me, parameters);
}

void symbol_hash_resize_datum(symbol_hash_t *table, 
			      unsigned new_datum_size,
			      void (*coerce)(const void *old_datum_p, 
					     void *new_datum_p,
					     void *parameters),
			      void *parameters)
{
  hash_table_resize_datum(table, new_datum_size, coerce, parameters);
}

void symbol_hash_map(symbol_hash_t *table, 
		     boolean_t (*proc)(const void *symbol_p, void *datum_p, void *parameters),
		     void *parameters)
{
  hash_table_map(table, proc, parameters);
}

void symbol_hash_index_map(symbol_hash_t *table, 
			   boolean_t (*proc)(const void *symbol_p, void *datum_p, 
					unsigned index, void *parameters),
			   void *parameters)
{
  hash_table_index_map(table, proc, parameters);
}

/************************************************************
 * PERFORMANCE TUNING
 ************************************************************/

void symbol_hash_monitor(const symbol_hash_t *table, FILE *stream)
{
  hash_table_monitor(table, stream);
}

