/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 *
 * $Id: symbol.hash.h,v 1.5 1998/02/07 03:44:28 libpa Exp $
 *
 */

#ifndef SYMBOL_HASH_INCLUDED
#define SYMBOL_HASH_INCLUDED

#include <stdio.h>
#include "hash.h"

/* Description.
 * Hash table for storing arbitrary <symbol,datum> pairs.
 * Note that symbols must contain hashable data, ie., no pointers,
 * and symbol equality must be defined by memcmp().
 */

#define SYMBOL_NOT_FOUND   KEY_NOT_FOUND

typedef hash_table_t symbol_hash_t;

symbol_hash_t *symbol_hash_create(unsigned symbol_size, unsigned datum_size);
               /* Arguments: number of bytes used by symbols and datum, resp. */
               /* Returns: empty table for specified <key,datum> pairs        */
void           symbol_hash_destroy(symbol_hash_t *table);
               /* Effect: deallocates all storage used by symbol_hash_t       */

/************************************************************
 * BASIC SELECTORS
 ************************************************************/

#define   symbol_hash_size(table_p)           ((unsigned)hash_table_size(table_p))
          /* Returns: number of <symbol,datum> entries in table      */
#define   symbol_hash_symbol_size(table_p)    (hash_table_key_size(table_p))
          /* Returns: number of bytes necessary to store each symbol */
#define   symbol_hash_datum_size(table_p)     (hash_table_datum_size(table_p))
          /* Returns: number of bytes necessary to store each datum  */

/************************************************************
 * PREDICATES
 ************************************************************/

boolean_t symbol_hash_is_valid(const symbol_hash_t *table);
          /* Returns: TRUE iff table is a valid symbol_hash_t     */
#define   symbol_hash_is_empty(table_p)           (hash_table_is_empty(table_p))
          /* Returns: TRUE iff table contains no entries          */
#define   symbol_hash_includes(table_p,symbol_p)  (hash_table_includes(table_p,symbol_p))
          /* Returns: TRUE iff table contains pair with *symbol_p */
          /* Note: use this if symbol_hash_datum_size(table) == 0 */
#define   symbol_hash_are_equivalent(left_p,right_p) \
          (hash_table_are_equivalent(left_p,right_p))
          /* Returns: TRUE iff both tables have same <symbol,datum> pairs */

/************************************************************
 * SELECTOR
 ************************************************************/

#define   symbol_hash_lookup(table_p,symbol_p)   (hash_table_lookup(table_p,symbol_p))
          /* Returns: datum_p associated with *symbol_p or NULL if none exists */
          /* Note: always returns NULL if symbol_hash_datum_size(table_p) == 0 */
          /*       in which case you should use symbol_hash_includes() instead */

/************************************************************
 * MUTATORS
 ************************************************************/

#define   symbol_hash_insert(table_p,symbol_p,datum_p) \
          (hash_table_insert(table_p,symbol_p,datum_p))
          /* Effect: unconditionally inserts <*symbol_p,*datum_p> into table */
#define   symbol_hash_update(table_p,symbol_p,initial_p) \
          (hash_table_update(table_p,symbol_p,initial_p))
          /* Effect: inserts <*symbol_p,*initial_p> if *key_p is new         */
          /* Returns: datum_p inside table after conditional insertion       */
          /* Note: always returns NULL if symbol_hash_datum_size(table) == 0 */

#define   symbol_hash_remove(table_p,symbol_p) (hash_table_remove(table_p,symbol_p))
          /* Effect: removes <*symbol_p,datum> pair from table in O(n) time  */
#define   symbol_hash_clear(table_p)           (hash_table_clear(table_p))
          /* Effect: removes all entries from symbol_hash_t in O(1) time     */
void      symbol_hash_prune(symbol_hash_t *table, 
			    boolean_t (*prune_me)(const void *symbol_p, 
						  const void *datum_p, 
						  void *parameters),
			    void *parameters);
          /* Effect: removes every <symbol,datum> entry for which prune_me() */
          /*         is TRUE in O(n) time and resizes table appropriately.   */
          /* Note: this is nearly as fast as a single symbol_hash_remove()!  */

void      symbol_hash_resize_datum(symbol_hash_t *table, 
				   unsigned new_datum_size,
				   void (*coerce)(const void *old_datum_p, 
						  void *new_datum_p,
						  void *parameters),
				   void *parameters);
          /* Arguments: coerce() mutates old *datum_p to new *datum_p   */
          /* Effect: mutates symbol_hash_t so that datum have new size  */
          /* Note: coercion occurs even if old and new have same size   */

/************************************************************
 * HIGHER ORDER FUNCTION
 ************************************************************/

void      symbol_hash_map(symbol_hash_t *table, 
			  boolean_t (*proc)(const void *symbol_p, void *datum_p, 
					    void *parameters),
			  void *parameters);
          /* Effect: calls proc() on every <symbol,datum> entry in table */
          /*         while proc() returns TRUE                           */
          /* Note: proc() can mutate *datum_p but not *symbol_p!         */

/************************************************************
 * INDEXING
 ************************************************************/

#define   symbol_hash_index(table_p,symbol_p)   (hash_table_index((table_p),(symbol_p)))
          /* Returns: maps *symbol_p to index in [0,entry_count-1] or     */
          /*          SYMBOL_NOT_FOUND if not found                       */
          /* Warning: mapping changes when any entry is deleted.  Hence,  */
          /*          this function should only be used for static table. */
void      symbol_hash_index_map(symbol_hash_t *table, 
				boolean_t (*proc)(const void *symbol_p, void *datum_p, 
						  unsigned index, void *parameters),
				void *parameters);
          /* Effect: calls proc() on every <symbol,datum> entry in table  */
          /*         along with its index while proc() returns TRUE       */
          /* Warning: indexing changes when any entry is deleted!         */
          /* Note: proc() can mutate *datum_p but not *symbol_p!          */

/************************************************************
 * PERFORMANCE TUNING
 ************************************************************/

void      symbol_hash_monitor(const symbol_hash_t *table, FILE *stream);
          /* Effect: writes performance information to stream             */

#endif

