/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>    /* for drand48() */
#include "memory.h"
#include "vector.h"
#include "symbol.hash.h"

static char symbol_hash_test_rcsid[] = "$Id: symbol.hash.test.c,v 1.4 1998/02/07 03:44:28 libpa Exp $";

#define DEBUG                 TRUE
#define VERBOSE               TRUE
#define SEED                  198743549
#define ALPHABET_SIZE         (8192)

extern double drand48(void);
extern void   srand48(long);

typedef unsigned symbol_t;
typedef unsigned datum_t;
typedef struct new_datum_tag { datum_t old, new; } new_datum_t;

static void *Default_Parameters;

boolean_t map_verifier(const void *symbol_p, void *datum_p, void *parameters)
{
  symbol_t *alphabet, symbol;
  datum_t   datum;

  symbol = *(symbol_t *)symbol_p;
  datum = *(datum_t *)datum_p;
  alphabet = parameters;

  assert(symbol < ALPHABET_SIZE);
  assert(datum < ALPHABET_SIZE);
  assert(vector_is_valid(alphabet));
  assert(vector_length(alphabet) == ALPHABET_SIZE);
  assert(alphabet[datum] == symbol);
  alphabet[datum] += ALPHABET_SIZE;
  return(TRUE);
}

boolean_t index_verifier(const void *symbol_p, void *datum_p, unsigned index, 
			 void *parameters)
{
  symbol_t *alphabet, symbol;
  datum_t   datum;

  symbol = *(symbol_t *)symbol_p;
  datum = *(datum_t *)datum_p;
  alphabet = parameters;

  assert(symbol < ALPHABET_SIZE);
  assert(datum < ALPHABET_SIZE);
  assert(datum == index);
  assert(vector_is_valid(alphabet));
  assert(vector_length(alphabet) == ALPHABET_SIZE);
  assert(alphabet[datum] == symbol+ALPHABET_SIZE);
  alphabet[datum] -= ALPHABET_SIZE;
  return(TRUE);
}

boolean_t no_pruner(const void *symbol_p, const void *datum_p, void *parameters)
{
  return(FALSE);
}

boolean_t odd_pruner(const void *symbol_p, const void *datum_p, void *parameters)
{
  symbol_t  symbol;
  symbol = *(symbol_t *)symbol_p;
  return(symbol%2);
}

void old2new(const void *in_p, void *out_p, void *parameters)
{
  datum_t *old_p;
  new_datum_t *new_p;

  assert(in_p); assert(out_p);  
  assert(parameters == Default_Parameters);
  old_p = (datum_t *)in_p;
  new_p = (new_datum_t *)out_p;

  new_p->old = *old_p;
  new_p->new = *old_p + ALPHABET_SIZE;
}

void new2old(const void *in_p, void *out_p, void *parameters)
{
  datum_t *old_p;
  new_datum_t *new_p;

  assert(in_p); assert(out_p);  
  assert(parameters == Default_Parameters);
  new_p = (new_datum_t *)in_p;
  old_p = (datum_t *)out_p;

  assert(new_p->new == new_p->old + ALPHABET_SIZE);
  *old_p = new_p->old;
}


int main(int argc, char **argv)
{
  symbol_hash_t *table;
  symbol_t      *alphabet, symbol;
  datum_t       *datum_p, datum, bad_datum;
  unsigned       k;

  srand48(SEED);
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("\n\tTesting symbol_hash_t library...\n");

  alphabet = vector_create(sizeof(symbol_t), ALPHABET_SIZE);
  assert(vector_is_valid(alphabet));
  assert(vector_length(alphabet) == ALPHABET_SIZE);
  for(k = 0; k < ALPHABET_SIZE; k++) { alphabet[k] = k; }
  vector_shuffle(alphabet);
  printf("%u\tallocated and permuted %d-symbol alphabet\n", 
	 memory_total_bytes(), (int)ALPHABET_SIZE);

  table = symbol_hash_create(sizeof(symbol_t), sizeof(datum_t));
  assert(table);
  assert(symbol_hash_symbol_size(table) == sizeof(symbol_t));
  assert(symbol_hash_datum_size(table) == sizeof(datum_t));
  assert(symbol_hash_size(table) == 0);
  assert(symbol_hash_is_empty(table));
  assert(symbol_hash_is_valid(table));
  printf("%u\tcreated symbol_hash_t for %d-byte symbols and %d-byte data\n", 
	 memory_total_bytes(), (int)sizeof(symbol_t), (int)sizeof(datum_t));

  /* insert the symbols in random order paired with their order */
  bad_datum = UINT_MAX;
  for(k = 0; k < ALPHABET_SIZE; k++)
    {
      datum = k;
      symbol = alphabet[k];
      assert(symbol_hash_size(table) == k);
      assert(!symbol_hash_includes(table, &symbol));
      assert(symbol_hash_lookup(table, &symbol) == NULL);
      assert(symbol_hash_index(table, &symbol) == SYMBOL_NOT_FOUND);

      /* add new symbol with bad value */
      datum_p = symbol_hash_update(table, &symbol, &bad_datum);
      assert(datum_p != &bad_datum);
      assert(*datum_p == bad_datum);

      /* fail to overwrite bad value */
      datum_p = symbol_hash_update(table, &symbol, &datum);
      assert(datum_p != &datum);
      assert(*datum_p != datum);
      assert(*datum_p == bad_datum);

      /* successfully overwrite bad value with good value */
      symbol_hash_insert(table, &symbol, &datum);
      datum_p = symbol_hash_lookup(table, &symbol);
      assert(datum_p != &datum); 
      assert(*datum_p == datum);
      assert(symbol_hash_includes(table, &symbol));
      assert(symbol_hash_index(table, &symbol) == k);
    }
  assert(symbol_hash_size(table) == ALPHABET_SIZE);
  printf("%u\tcorrectly inserted %d symbols\n\t", memory_total_bytes(), (int)ALPHABET_SIZE);
  symbol_hash_monitor(table, stdout);

  /* remove no symbols */
  symbol_hash_prune(table, &no_pruner, NULL);

  /* verify map and index_map */
  symbol_hash_map(table, &map_verifier, alphabet);
  symbol_hash_index_map(table, &index_verifier, alphabet);
  for(k = 0; k < ALPHABET_SIZE; k++)
    {
      datum = k;
      symbol = alphabet[k];
      assert(symbol_hash_includes(table, &symbol));
      datum_p = symbol_hash_lookup(table, &symbol);
      assert(datum_p); assert(*datum_p == datum);
    }
  printf("%u\tverified symbol_hash_map() and symbol_hash_index_map()\n", 
	 memory_total_bytes());

  /* resize the datum up and down */
  assert(ALPHABET_SIZE <= USHRT_MAX);
  symbol_hash_resize_datum(table, sizeof(new_datum_t), &old2new, Default_Parameters);
  symbol_hash_resize_datum(table, sizeof(datum_t), &new2old, Default_Parameters);
  assert(symbol_hash_is_valid(table));
  for(k = 0; k < ALPHABET_SIZE; k++)
    {
      datum = k;
      symbol = alphabet[k];
      assert(symbol_hash_includes(table, &symbol));
      datum_p = symbol_hash_lookup(table, &symbol);
      assert(datum_p); assert(*datum_p == datum);
    }
  printf("%u\tcorrectly resized datum up to %d and back down to %d\n\t", 
	 memory_total_bytes(), (int)sizeof(new_datum_t), (int)sizeof(datum_t));
  symbol_hash_monitor(table, stdout);

  /* remove all the big symbols manually */
  for(k = ALPHABET_SIZE/2; k < ALPHABET_SIZE; k++)
    {
      symbol = k;
      symbol_hash_remove(table, &symbol);
      assert(!symbol_hash_includes(table, &symbol));
      assert(symbol_hash_lookup(table, &symbol) == NULL);
    }

  /* remove all the odd symbols automatically */
  symbol_hash_prune(table, &odd_pruner, NULL);

  /* verify that all the right symbols were removed properly */
  for(k = 0; k < ALPHABET_SIZE; k++)
    {
      symbol = k;
      if ((k%2 == 1) || (k >= ALPHABET_SIZE/2))
	{
	  assert(!symbol_hash_includes(table, &symbol));
	  assert(symbol_hash_lookup(table, &symbol) == NULL);
	  assert(symbol_hash_index(table, &symbol) == SYMBOL_NOT_FOUND);
	}
      else
	{
	  assert(symbol_hash_includes(table, &symbol));
	  assert(symbol_hash_index(table, &symbol) < symbol_hash_size(table));
	  datum_p = symbol_hash_lookup(table, &symbol);
	  assert(datum_p); 
	  assert(*datum_p < ALPHABET_SIZE);
	  assert(alphabet[*datum_p] == symbol);
	}
    }
  printf("%u\tcorrectly removed big symbols and pruned odd symbols\n\t", 
	 memory_total_bytes());
  symbol_hash_monitor(table, stdout);

  vector_destroy(alphabet); alphabet = NULL;
  symbol_hash_destroy(table); table = NULL;
  printf("%u\tdestroyed all symbol_hash_t\n", memory_total_bytes());
  printf("\tmemory leak == %u bytes in %u blocks\n", 
	  memory_total_bytes(), memory_total_blocks());
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);

  printf("\tsizeof(symbol_hash_t) == %d\n", sizeof(symbol_hash_t));
  printf("\tsymbol_hash_t library passes all tests.\n");
  return(0);
}



