/*
 *  (c) 1996,1997 Eric Sven Ristad and Peter N. Yianilos. All Rights Reserved. 
 *
 */

#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <stdio.h>
#include "common.h"  /* for boolean_t */
#include "memory.h"
#include "memory.spartan.h"

static char memory_rcsid[] = "$Id";

#define ANNOUNCE_FAILURE    TRUE
#define EXIT_ON_FAILURE     TRUE

/******************** GLOBAL VARIABLES ********************/
unsigned int  Memory_Total_Bytes;
unsigned int  Memory_Total_Blocks;
unsigned int  Memory_Maximum_Bytes;
/**********************************************************/

memory_t *memory_create(unsigned bytes)
{
  memory_header_t *header;
  memory_t        *block;

  assert(sizeof(memory_header_t) == sizeof(double)); /* verify alignment */

  if (bytes == 0) { return(NULL); }
  header = malloc(bytes + sizeof(memory_header_t));
  if (header == NULL)
    { 
#if ANNOUNCE_FAILURE
      fprintf(stderr, "memory_t error: unable to allocate %u+%u bytes\n", 
	      Memory_Total_Bytes, bytes);
#endif
#if EXIT_ON_FAILURE
      fprintf(stderr, "memory_t fatal error: exiting process %d....\n", (int)getpid());
      exit(-1);
#else
      return(NULL); 
#endif
    }

  block = header+1;
  header->bytes = bytes;
  header->workspace = UINT_MAX;   /* for safety */
  assert(memory_bytes(block) == bytes);

  Memory_Total_Bytes  += bytes + sizeof(memory_header_t);
  Memory_Total_Blocks += 1;

  if (Memory_Total_Bytes > Memory_Maximum_Bytes)
    { Memory_Maximum_Bytes = Memory_Total_Bytes; }

  assert(memory_is_valid(block));
  assert(memory_bytes(block) == bytes);
  return(block);
}

memory_t *memory_recreate(memory_t *block, unsigned new_bytes)
{
  memory_header_t *new_header, *old_header;
  memory_t        *new_block;
  unsigned         old_bytes;
#ifndef NDEBUG
  unsigned         old_workspace;
#endif

  if (block == NULL) { return(memory_create(new_bytes)); }
  else if (new_bytes == 0) { memory_destroy(block); return(NULL); }
      
  assert(memory_is_valid(block)); assert(new_bytes > 0);
  assert(memory_total_bytes() >= (sizeof(memory_header_t) + memory_bytes(block)));
  assert(memory_total_blocks() > 0);

  old_bytes = memory_bytes(block);
#ifndef NDEBUG
  old_workspace = *memory_workspace_p(block);
#endif

  old_header = memory_header(block);
  new_header = realloc(old_header, new_bytes + sizeof(memory_header_t));
  if (new_header == NULL) 
    { 
#if ANNOUNCE_FAILURE
      fprintf(stderr, "memory_t error: unable to reallocate %u+%u bytes\n", 
	      Memory_Total_Bytes, new_bytes);
#endif
#if EXIT_ON_FAILURE
      fprintf(stderr, "memory_t fatal error: exiting process %d....\n", (int)getpid());
      exit(-1);
#else
      /* unexpected deallocation! */
      Memory_Total_Bytes  -= (old_bytes + sizeof(memory_header_t));
      Memory_Total_Blocks -= 1;
      return(NULL); 
#endif
    }
  assert(new_header->bytes == old_bytes);
  assert(new_header->workspace == old_workspace);

  new_block = new_header+1;
  new_header->bytes = new_bytes;
  assert(memory_bytes(new_block) == new_bytes);

  Memory_Total_Bytes -= old_bytes;  /* + sizeof(memory_header_t) */
  Memory_Total_Bytes += new_bytes;  /* + sizeof(memory_header_t) */

  if (Memory_Total_Bytes > Memory_Maximum_Bytes)
    { Memory_Maximum_Bytes = Memory_Total_Bytes; }

  assert(memory_is_valid(new_block));
  assert(memory_bytes(new_block) == new_bytes);
  return(new_block);
}

memory_t *memory_copy(const memory_t *original_block)
{
  memory_t *copy_block;
  unsigned  bytes;

  if (original_block == NULL) { return(NULL); }

  assert(memory_is_valid(original_block));
  assert(memory_total_blocks() > 0);
  assert(memory_bytes(original_block) > 0);
  assert(memory_total_bytes() >= memory_bytes(original_block)+sizeof(memory_header_t));

  bytes = memory_bytes(original_block);
  copy_block = memory_create(bytes);
  if (copy_block == NULL) { return(NULL); }
  
  copy_block = memcpy(copy_block, original_block, bytes);
  memory_header(copy_block)->workspace = memory_header(original_block)->workspace;

  assert(memory_is_valid(copy_block));
  assert(copy_block != original_block);
  assert(memory_bytes(copy_block) == memory_bytes(original_block));
  assert(memory_workspace_p(copy_block) != memory_workspace_p(original_block));
  assert(*memory_workspace_p(copy_block) == *memory_workspace_p(original_block));
  assert(memory_compare(copy_block, original_block) == 0);

  return(copy_block);
}

void memory_destroy(memory_t *block)
{
  assert(memory_is_valid(block));
  if (block != NULL)
    {
      memory_header_t *header;

      assert(memory_bytes(block) > 0);
      assert(memory_total_blocks() > 0);
      assert(memory_total_bytes() >= (memory_bytes(block) + sizeof(memory_header_t)));

      Memory_Total_Bytes  -= (memory_bytes(block) + sizeof(memory_header_t));
      Memory_Total_Blocks -= 1;

      header = memory_header(block);
      header->bytes = 0;                /* invalidate block     */
      header->workspace = UINT_MAX;     /* invalidate workspace */
      free(header);
    }
}

void *memory_move(void *destination, const void *source, unsigned bytes)
{
  unsigned char	*destination_p;
  const unsigned char *source_p;

  assert(source != NULL);
  assert(destination != NULL);

  source_p = source;
  destination_p = destination;
  if (destination > source)            /* must go high to low */
    {
      source_p += bytes;
      destination_p += bytes;
      while(bytes > 0)
	{
	  bytes--;
	  source_p--;
	  destination_p--;
	  *destination_p = *source_p;
	}
    }
  else if (destination < source)       /* must go low to high */
    {
      while(bytes > 0) 
	{ 
	  *destination_p = *source_p;
	  destination_p++;
	  source_p++;
	  bytes--;
	}
    }
  return(destination);
  /* p.s. beautiful, n'est ce pas? */
}

boolean_t memory_is_valid(const memory_t *block)
{
  if (block == NULL) { return(TRUE); }
  return((memory_total_blocks() > 0) &&
	 (memory_bytes(block) > 0) &&
	 (memory_total_bytes() >= memory_bytes(block) + sizeof(memory_header_t)));
}

boolean_t memory_address_is_valid(const memory_t *block, const void *address)
{
  unsigned bytes;
  assert(memory_is_valid(block));
  if (block == NULL) { return(FALSE); }
  bytes = memory_bytes(block);
  return((bytes > 0) && 
	 (address >= block) && 
	 ((unsigned char *)address < (unsigned char *)block + bytes));
}

int memory_compare(const memory_t *left, const memory_t *right)
{
  unsigned left_bytes, right_bytes, bytes;
  int result;

  if ((left == NULL) || (right == NULL)) 
    { return((long)left - (long)right); }

  assert(memory_is_valid(left));
  assert(memory_is_valid(right));

  left_bytes = memory_bytes(left);
  right_bytes = memory_bytes(right);
  bytes = MIN(left_bytes, right_bytes);
  result = memcmp(left, right, bytes);
  if (result == 0) { result = (long)left_bytes - right_bytes; }
  return(result);
}


/**********************************************************************
 *                         SPARTAN INTERFACE                          *
 **********************************************************************/

void *memory_spartan_create(unsigned bytes)
{
  memory_t *block;

  if (bytes == 0) { return(NULL); }
  block = malloc(bytes);
  if (block == NULL) 
    { 
#if ANNOUNCE_FAILURE
      fprintf(stderr, "memory_t error: unable to allocate %u+%u bytes\n", 
	      Memory_Total_Bytes, bytes);
#endif
#if EXIT_ON_FAILURE
      fprintf(stderr, "memory_t fatal error: exiting process %d....\n", (int)getpid());
      exit(-1);
#else
      return(NULL); 
#endif
    }
  
  Memory_Total_Bytes  += bytes;
  Memory_Total_Blocks += 1;

  if (Memory_Total_Bytes > Memory_Maximum_Bytes)
    { Memory_Maximum_Bytes = Memory_Total_Bytes; }

  return(block);
}

void *memory_spartan_recreate(void *block, unsigned old_bytes, unsigned new_bytes)
{
  memory_t *new_block;

  if (block == NULL) { assert(old_bytes == 0); return(memory_spartan_create(new_bytes)); }
  else if (new_bytes == 0) { memory_spartan_destroy(block, old_bytes); return(NULL); }

  assert(old_bytes > 0);
  assert(memory_total_blocks() > 0);
  assert(memory_total_bytes() >= old_bytes);

  Memory_Total_Bytes  -= old_bytes;
  Memory_Total_Blocks -= 1;

  new_block = realloc(block, new_bytes);
  if (new_block == NULL) 
    {  
#if ANNOUNCE_FAILURE
      fprintf(stderr, "memory_t error: unable to reallocate %u+%u bytes\n", 
	      Memory_Total_Bytes, new_bytes);
#endif
#if EXIT_ON_FAILURE
      fprintf(stderr, "memory_t fatal error: exiting process %d....\n", (int)getpid());
      exit(-1);
#else
      return(NULL); 
#endif
    }

  Memory_Total_Bytes  += new_bytes;
  Memory_Total_Blocks += 1;

  if (Memory_Total_Bytes > Memory_Maximum_Bytes)
    { Memory_Maximum_Bytes = Memory_Total_Bytes; }

  return(new_block);
}

void *memory_spartan_copy(const void *original_block, unsigned bytes)
{
  memory_t *copy_block;

  if (original_block == NULL) { assert(bytes == 0); return(NULL); }
  assert(bytes > 0);
  assert(memory_total_bytes() >= bytes);
  assert(memory_total_blocks() > 0);

  copy_block = memory_spartan_create(bytes);
  if (copy_block == NULL) { return(NULL); }
  copy_block = memcpy(copy_block, original_block, bytes);
  return(copy_block);
}

void memory_spartan_destroy(void *block, unsigned bytes)
{
  assert((bytes == 0) ? (block == NULL) : (block != NULL));
  if (block != NULL)
    {
      assert(bytes > 0);
      assert(memory_total_blocks() > 0);
      assert(memory_total_bytes() >= bytes);

      Memory_Total_Bytes  -= bytes;
      Memory_Total_Blocks -= 1;
      free(block);
    }
}

boolean_t memory_spartan_address_is_valid(const void *block, unsigned bytes, 
					const void *address)
{
  return((block != NULL) &&
	 (bytes > 0) && 
	 (address >= block) && 
	 ((unsigned char *)address < (unsigned char *)block + bytes));
}

int memory_spartan_compare(const memory_t *left, unsigned left_bytes,
			   const memory_t *right, unsigned right_bytes)
{
  unsigned bytes;
  int      result;

  if ((left == NULL) || (right == NULL)) { return((long)left - (long)right); }
  bytes = MIN(left_bytes, right_bytes);
  result = memcmp(left, right, bytes);
  if (result == 0) { result = (long)left_bytes - right_bytes; }
  return(result);
}



