/*
 * (c) 1996 Eric Sven Ristad and Peter N. Yianilos.  All rights reserved.
 *
 * $Id: memory.h,v 1.8 1997/02/11 20:33:51 ristad Exp $
 *
 */

#ifndef MEMORY_INCLUDED
#define MEMORY_INCLUDED
#include "common.h"   /* for boolean_t */

/* Description.
 *   Simple and consistent interface to machine memory resource.
 *   Extremely useful for eliminating memory leaks.
 *
 */

#define MEMORY_WORKSPACE_SIZE        sizeof(unsigned)

typedef void  memory_t;
typedef struct memory_header_tag 
{
  unsigned  bytes;       /* number of currently allocated bytes       */
  unsigned  workspace;   /* caller workspace, necessary for alignment */
} memory_header_t;

/************************************************************
 *  CONSTRUCTORS/DESTRUCTORS
 ************************************************************/

memory_t *memory_create(unsigned bytes);
          /* Returns: address of newly allocated memory block   */
void      memory_destroy(memory_t *block);
          /* Requires: block was returned by memory_t operation */
          /* Effect: deallocates specified block of memory      */
memory_t *memory_recreate(memory_t *block, unsigned bytes);
          /* Requires: block was returned by a memory_t operation      */
          /* Effect: destroys original block if new block is allocated */
          /* Returns: exact copy of block but of requested size        */
memory_t *memory_copy(const memory_t *block);
          /* Requires: block was returned by memory_t operation        */
          /* Returns: new block of storage that is exact copy of block */
          /* Warning: this is NOT equivalent to memcpy() in <string.h> */

/************************************************************
 *  MUTATOR
 ************************************************************/

void     *memory_move(void *destination, const void *source, unsigned bytes);
          /* Effect: moves specified number of bytes from source to destination */
          /*         overlapping moves are handled correctly.                   */
          /* Note: this replaces memmove() in <string.h> and does not require   */
          /*       either destination or source to be be valid memory_t block   */
          /* Returns: destination pointer                                       */

/************************************************************
 *  SELECTORS
 ************************************************************/

#define   memory_total_bytes()       ((unsigned)Memory_Total_Bytes)
          /* Returns: total number of allocated bytes, including overhead */
#define   memory_total_blocks()      ((unsigned)Memory_Total_Blocks)
          /* Returns: total number of blocks allocated by memory_t        */
#define   memory_maximum_bytes()     ((unsigned)Memory_Maximum_Bytes)
          /* Returns: maximum number of bytes allocated by memory_t       */

#define   memory_bytes(mem_p)        ((unsigned)(((mem_p) == NULL) ? 0 : memory_header(mem_p)->bytes))
          /* Requires: mem_p was returned by a memory_t operation */
          /* Returns: number of bytes in block mem_p w/o overhead */
#define   memory_workspace_p(mem_p)  (((mem_p) == NULL) ? NULL : &(memory_header(mem_p)->workspace))
          /* Requires: mem_p was returned from memory_t operation */
          /* Returns: pointer to unsigned int for caller to use   */
          /* Warning: use of this macro is strongly discouraged   */

/************************************************************
 *  PREDICATES
 ************************************************************/

boolean_t memory_is_valid(const memory_t *block);
          /* Returns: TRUE iff block is valid memory_t reference           */
boolean_t memory_address_is_valid(const memory_t *block, const void *address);
          /* Returns: TRUE iff address falls within given memory_t block   */

int       memory_compare(const memory_t *left, const memory_t *right);
          /* Returns: total order on all memory_t blocks, 0 is equivalence */


/************************* PRIVATE *************************/
/* Warning: use of these symbols violates the abstraction! */
extern unsigned int  Memory_Total_Bytes;
extern unsigned int  Memory_Total_Blocks;
extern unsigned int  Memory_Maximum_Bytes;
#define memory_header(mem_p) (((memory_header_t *)(mem_p))-1)
/************************************************************/

#endif	/* MEMORY_INCLUDED */


