#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include "common.h"
#include "memory.h"
#include "memory.spartan.h"

static char memory_test_rcsid[] = "$Id: memory.test.c,v 1.11 1997/02/11 20:33:51 ristad Exp $";


#define BE_VERBOSE         FALSE
#define BLOCK_MINSIZE      1024
#define BLOCK_COUNT        64
#define BLOCK_OFFSET       16
#define block_bytes(i)    (BLOCK_MINSIZE+(i))

int main(int argc, char **argv)
{
  unsigned char *block[BLOCK_COUNT], *block_copy, *byte_p;
  unsigned char *spartan[BLOCK_COUNT], *spartan_copy;
  unsigned bytes, *workspace_p, total_bytes, total_blocks;
  unsigned maximum_bytes;
  int i, j;

  printf("\n\tTesting memory_t library...\n");
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  assert(memory_maximum_bytes() == 0);

  if ((malloc(0) == NULL) &&
      (realloc(NULL, 1) != NULL) &&     /* secret memory leak! */
      (realloc(malloc(1), 0) == NULL))  /* secret memory leak! */
    { 
      printf("\tstdlib malloc(0)/realloc(*,0) satisfies null specification\n"); 
    }
  else 
    { 
      printf("\tWarning: malloc(0) != NULL on your operating system\n");
      printf("\t\tmemory_create(0) differs from your malloc(0)\n"); 
      printf("\t\tmemory_recreate(*,0) differs from your realloc(*,0)\n"); 
    }
  
  assert(memory_create(0) == NULL);
  assert(memory_copy(NULL) == NULL);
  assert(memory_is_valid(NULL));
  assert(!memory_address_is_valid(NULL, NULL));
  assert(memory_bytes(NULL) == 0);
  assert(memory_workspace_p(NULL) == NULL);

  block[0] = memory_recreate(NULL, 1);
  assert(block[0] != NULL);
  assert(memory_is_valid(block[0]));
  assert(memory_address_is_valid(block[0], block[0]));  
  assert(!memory_address_is_valid(block[0], NULL));
  assert(memory_bytes(block[0]) == 1);
  assert(memory_total_blocks() == 1);
  assert(memory_total_bytes() == 1+sizeof(memory_header_t));
  memory_destroy(block[0]);
  block[0] = NULL;
  assert(memory_total_blocks() == 0);
  assert(memory_total_bytes() == 0);
  assert(memory_maximum_bytes() == 1+sizeof(memory_header_t));
  printf("\tregular memory_t implementation satisfies null specification\n");

  assert(memory_spartan_create(0) == NULL);
  assert(memory_spartan_copy(NULL, 0) == NULL);
  spartan[0] = memory_spartan_recreate(NULL, 0, 1);
  assert(spartan[0] != NULL);
  assert(memory_total_blocks() == 1);
  assert(memory_total_bytes() == 1);
  memory_spartan_destroy(spartan[0], 1);
  spartan[0] = NULL;
  assert(memory_total_blocks() == 0);
  assert(memory_total_bytes() == 0);
  assert(memory_maximum_bytes() == 1+sizeof(memory_header_t));
  printf("\tspartan memory_t implementation satisfies null specification\n");
#if BE_VERBOSE
  printf("\n");
#endif

  /****************************** REGULAR MEMORY_T *******************************/

  total_bytes = 0;
  total_blocks = 0;
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  maximum_bytes = 1+sizeof(memory_header_t);
  assert(memory_maximum_bytes() == 1+sizeof(memory_header_t));

  for(i = 0; i < BLOCK_COUNT; i++)
    {
      bytes = block_bytes(i);
      block[i] = memory_create(bytes);
      assert(memory_is_valid(block[i]));
      assert(memory_bytes(block[i]) == bytes);

      total_blocks++;
      total_bytes += bytes + sizeof(memory_header_t);
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);

      if (total_bytes > maximum_bytes) { maximum_bytes = total_bytes; }
      assert(memory_maximum_bytes() == maximum_bytes);

      workspace_p = memory_workspace_p(block[i]);
      *workspace_p = UINT_MAX - bytes;
      assert(*memory_workspace_p(block[i]) == UINT_MAX - bytes);
      
      assert(!memory_address_is_valid(block[i], block[i]-1));
      assert(!memory_address_is_valid(block[i], block[i]+bytes));
      assert(!memory_address_is_valid(NULL, block[i]));
      assert(!memory_address_is_valid(block[i], NULL));

      /* initialize in ascending order */
      for(j = 0; j < bytes; j++)
	{
	  block[i][j] = j;
	  assert(block[i][j] == (unsigned char)j);
	  byte_p = block[i] + j;
	  assert(*byte_p == (unsigned char)j);
	  assert(memory_address_is_valid(block[i], byte_p));
	}
      assert(memory_compare(block[i], block[i]) == 0);
      assert((i == 0) || (memory_compare(block[i-1], block[i]) != 0));
    }
  assert(memory_total_blocks() == BLOCK_COUNT);
  printf("%u\tallocated %d valid regular blocks\n", 
	 memory_total_bytes(), BLOCK_COUNT);

  for(i = 0; i < BLOCK_COUNT; i++)
    {
      bytes = block_bytes(i);

      block_copy = memory_copy(block[i]);
      assert(memory_is_valid(block_copy));
      assert(memory_bytes(block_copy) == memory_bytes(block[i]));
      assert(*memory_workspace_p(block_copy) == *memory_workspace_p(block[i]));
      assert(memory_compare(block_copy, block[i]) == 0);
      assert(memcmp(block_copy, block[i], bytes) == 0);
      assert(memory_total_bytes() == total_bytes + bytes + sizeof(memory_header_t));
      assert(memory_total_blocks() == total_blocks + 1);

      if (memory_total_bytes() > maximum_bytes) { maximum_bytes = memory_total_bytes(); }
      assert(memory_maximum_bytes() == maximum_bytes);

      memory_destroy(block_copy);
      block_copy = NULL;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  printf("%u\tcopied %d regular blocks correctly\n", 
	 memory_total_bytes(), BLOCK_COUNT);

  for(i = 0; i < BLOCK_COUNT; i++)
    {
      bytes = block_bytes(i);
      block[i] = memory_recreate(block[i], 4*bytes);
      assert(memory_is_valid(block[i]));
      assert(memory_bytes(block[i]) == 4*bytes);

      assert(memory_total_bytes() == total_bytes + ((4-1)*bytes));
      assert(memory_total_blocks() == total_blocks);
      if (memory_total_bytes() > maximum_bytes) { maximum_bytes = memory_total_bytes(); }
      assert(memory_maximum_bytes() == maximum_bytes);

      block[i] = memory_recreate(block[i], bytes);
      assert(memory_is_valid(block[i]));
      assert(memory_bytes(block[i]) == bytes);
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
      assert(memory_maximum_bytes() == maximum_bytes);
    }
  printf("%u\treallocated %d valid regular blocks up and then down\n", 
	 memory_total_bytes(), BLOCK_COUNT);

  for(i = 0; i < BLOCK_COUNT; i++)
    {
      /* shift big block down and then shift it back to where it came from */
      memory_move(block[i], block[i]+BLOCK_OFFSET, 2 * BLOCK_OFFSET);
      for(j = 0; j < 2 * BLOCK_OFFSET; j++)
	{ assert(block[i][j] == (unsigned char)(j+BLOCK_OFFSET)); }
      memory_move(block[i]+BLOCK_OFFSET, block[i], 2 * BLOCK_OFFSET);
      for(j = BLOCK_OFFSET; j < 3 * BLOCK_OFFSET; j++)
	{ assert(block[i][j] == (unsigned char)j); }
    }
  assert(memory_total_bytes() == total_bytes);
  assert(memory_total_blocks() == total_blocks);
  assert(memory_maximum_bytes() == maximum_bytes);
  printf("%u\tmoved %d overlapping regular blocks down and then up\n", 
	 memory_total_bytes(), BLOCK_COUNT);
#if BE_VERBOSE
  printf("\n");
#endif

  /****************************** SPARTAN MEMORY_T *******************************/

  for(i = 0; i < BLOCK_COUNT; i++)
    {
      bytes = block_bytes(i);
      spartan[i] = memory_spartan_create(bytes);
      assert(spartan[i]);

      total_blocks++;
      total_bytes += bytes;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);

      assert(!memory_spartan_address_is_valid(spartan[i], bytes, spartan[i]-1));
      assert(!memory_spartan_address_is_valid(spartan[i], bytes, spartan[i]+bytes));
      assert(!memory_spartan_address_is_valid(NULL, 0, spartan[i]));
      assert(!memory_spartan_address_is_valid(spartan[i], bytes, NULL));

      /* initialize in ascending order */
      for(j = 0; j < bytes; j++)
	{
	  spartan[i][j] = j;
	  byte_p = spartan[i] + j;
	  assert(*byte_p == (unsigned char)j);
	  assert(spartan[i][j] == (unsigned char)j);
	  assert(memory_spartan_address_is_valid(spartan[i], bytes, byte_p));
	}
      assert(memory_spartan_compare(spartan[i], bytes, spartan[i], bytes) == 0);
      assert((i == 0) || (memory_spartan_compare(spartan[i-1], bytes-1, 
						 spartan[i], bytes) != 0));
    }
  maximum_bytes = total_bytes;
  assert(memory_total_blocks() == BLOCK_COUNT + BLOCK_COUNT);
  assert(memory_maximum_bytes() == maximum_bytes);
  printf("%u\tallocated additional %d valid spartan blocks\n", 
	 memory_total_bytes(), BLOCK_COUNT);

  for(i = 0; i < BLOCK_COUNT; i++)
    {
      bytes = block_bytes(i);
      spartan_copy = memory_spartan_copy(spartan[i], bytes);
      assert(spartan_copy);
      assert(memcmp(spartan_copy, spartan[i], bytes) == 0);
      assert(memory_spartan_compare(spartan_copy, bytes, spartan[i], bytes) == 0);
      assert(memory_total_bytes() == total_bytes + bytes);
      assert(memory_total_blocks() == total_blocks + 1);
      memory_spartan_destroy(spartan_copy, bytes);
      spartan_copy = NULL;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  maximum_bytes = total_bytes + block_bytes(BLOCK_COUNT-1);
  assert(memory_maximum_bytes() == maximum_bytes);
  printf("%u\tcopied %d spartan blocks correctly\n", 
	 memory_total_bytes(), BLOCK_COUNT);

  for(i = 0; i < BLOCK_COUNT; i++)
    {
      bytes = block_bytes(i);
      spartan[i] = memory_spartan_recreate(spartan[i], bytes, 4*bytes);
      assert(spartan[i]);
      assert(memory_total_bytes() == total_bytes + ((4-1)*bytes));
      assert(memory_total_blocks() == total_blocks);
      spartan[i] = memory_spartan_recreate(spartan[i], 4*bytes, bytes);
      assert(spartan[i]);
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  maximum_bytes = total_bytes + (4-1)*block_bytes(BLOCK_COUNT-1);
  assert(memory_maximum_bytes() == maximum_bytes);
  printf("%u\treallocated %d spartan blocks correctly\n", 
	 memory_total_bytes(), BLOCK_COUNT);
#if BE_VERBOSE
  printf("\n");
#endif

  /****************************** EPILOGUE *******************************/

  /* deallocate regular memory_t */
  for(i = 0; i < BLOCK_COUNT; i++)
    {
      bytes = block_bytes(i);
      assert(memory_is_valid(block[i]));
      assert(memory_bytes(block[i]) == bytes);
      assert(*memory_workspace_p(block[i]) == UINT_MAX - bytes);

      total_blocks--;
      total_bytes -= (bytes + sizeof(memory_header_t));
      memory_destroy(block[i]);
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  assert(memory_maximum_bytes() == maximum_bytes);
  printf("%u\tdeallocated %d regular blocks correctly\n", 
	 memory_total_bytes(), BLOCK_COUNT);

  /* deallocate spartan memory_t */
  for(i = 0; i < BLOCK_COUNT; i++)
    {
      bytes = block_bytes(i);
      total_blocks--;
      total_bytes -= bytes;
      memory_spartan_destroy(spartan[i], bytes);
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  assert(memory_maximum_bytes() == maximum_bytes);
  printf("%u\tdeallocated %d spartan blocks correctly\n", 
	 memory_total_bytes(), BLOCK_COUNT);

  printf("\tmaximal memory usage is %u bytes\n", memory_maximum_bytes());
  printf("\tmemory leak == %u bytes in %u blocks\n", 
	 memory_total_bytes(), memory_total_blocks());
  assert(total_bytes == 0);
  assert(total_blocks == 0);
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  printf("\tmemory_t library passed all tests.\n");
  return(0);
}

#ifdef NDEBUG
#error invalid compiler option -DNDEBUG
#endif

