/*
 * (c) 1997 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <math.h>
#include "L.1k.h"
#include "balanced.h"
#include "logpr.h"

#define BE_INFORMATIVE     TRUE
#define ITERATIONS         1
#define ALPHABET_SIZE      2
#define STATE_COUNT        64
#define CORPUS_SIZE        1048576
#define MIN_PR             1e-10
#define SYMBOL_PR          (1.0/ALPHABET_SIZE)
#ifndef M_LN2
#define M_LN2              6.9314718055994530942E-1
#endif

static char pr_bench_rcsid[] = "$Id: pr.bench.c,v 1.7 1997/08/27 12:40:00 ristad Exp $";

extern double drand48(void);
extern double log1p(double);
extern double exp(double);
extern pid_t  getpid(void);
extern int    kill(pid_t, int);
extern int    atoi(const char *);


int main(int argc, char *argv[])
{
  double     dA[STATE_COUNT][STATE_COUNT], dx;
  double     dAlpha[2][STATE_COUNT], d_scale, total_scale, d_symbol;
  logpr_t    nA[STATE_COUNT][STATE_COUNT], nx;
  logpr_t    nAlpha[2][STATE_COUNT], n_symbol;
  Lint_t     lA[STATE_COUNT][STATE_COUNT], lx;
  Lint_t     lAlpha[2][STATE_COUNT], l_symbol;
  balanced_t bA[STATE_COUNT][STATE_COUNT], bx;
  balanced_t bAlpha[2][STATE_COUNT], b_symbol;
  int	     i, j, k, t, t1, t2, tf;
  int        system;
  double     result;

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  system = 0;
  if (argc >= 2) { system = atoi(argv[1]); }

  if ((argc != 2) || (system < 1) || (system > 4))
    {
      printf("Usage: pr.bench 1|2|3|4 \n");
      printf("\t1\tdouble precision arithmetic with scaling\n");
      printf("\t2\tlogarithmic arithmetic\n");
      printf("\t3\tbalanced_t arithmetic\n");
      printf("\t4\tL_t arithmetic (base %g, %d entries)\n", L_BASE, L_TABLE_SIZE);
      return(-1);
    }
  printf("\tpr.bench: %d iterations, %d alphabet, %d states, %d corpus\n",
	 ITERATIONS, ALPHABET_SIZE, STATE_COUNT, CORPUS_SIZE);

  /* generate transition probabilities A[i][j] */
  for(i = 0; i < STATE_COUNT; i++) 
    {
      dx = 0;
      for(j = 0; j < STATE_COUNT; j++)
	{ 
	  dA[i][j] = drand48() + MIN_PR; 
	  dA[i][j] *= dA[i][j];
	  dA[i][j] *= dA[i][j];
	  dx += dA[i][j]; 
	}
      for(j = 0; j < STATE_COUNT; j++)
	{ dA[i][j] /= dx; }
    }

  d_symbol = SYMBOL_PR;
  assert(d_symbol > 0); assert(d_symbol <= 1);
  switch (system) 
    {
    case 1:
#if BE_INFORMATIVE
      printf("\tdouble precision arithmetic with scaling");
#endif
      for(k = 0; k < ITERATIONS; k++)
	{
	  dx = 1.0/(double)STATE_COUNT;
	  for(i = 0; i < STATE_COUNT; i++)
	    { dAlpha[0][i] = dx; }

	  d_scale = 1;
	  total_scale = 0;
	  for(t = 0; t < CORPUS_SIZE; t++)
	    {
	      t1 = t % 2;
	      t2 = (t+1) % 2;
	      dx = 0;
	      for(j = 0; j < STATE_COUNT; j++)
		{
		  dAlpha[t2][j] = dAlpha[t1][0] * dA[0][j];
		  for(i = 1; i < STATE_COUNT; i++)
		    { dAlpha[t2][j] += dAlpha[t1][i] * dA[i][j]; }
		  dAlpha[t2][j] *= d_scale;
		  dAlpha[t2][j] *= d_symbol;
		  dx += dAlpha[t2][j];
		}
	      total_scale += log(d_scale);
	      d_scale = 1 / dx;
	      assert(d_scale > .9);
	    }

	  tf = CORPUS_SIZE % 2;
	  dx = dAlpha[tf][0];
	  for(i = 1; i < STATE_COUNT; i++)
	    { dx += dAlpha[tf][i]; }
	  result = total_scale - log(dx);
	  printf(".");
	}
      break;

    case 2:
#if BE_INFORMATIVE
      printf("\tlogarithmic arithmetic");
#endif
      for(k = 0; k < ITERATIONS; k++)
	{
	  dx = 1.0/(double)STATE_COUNT;
	  nx = double2logpr(dx);
	  for(i = 0; i < STATE_COUNT; i++) 
	    {
	      nAlpha[0][i] = nx;
	      for(j = 0; j < STATE_COUNT; j++)
		{ nA[i][j] = double2logpr(dA[i][j]); }
	    }
	  n_symbol = double2logpr( d_symbol );

	  for(t = 0; t < CORPUS_SIZE; t++)
	    {
	      t1 = t % 2;
	      t2 = (t+1) % 2;
	      for(j = 0; j < STATE_COUNT; j++)
		{
		  nAlpha[t2][j] = logpr_multiply(nAlpha[t1][0], nA[0][j]); 
		  for(i = 1; i < STATE_COUNT; i++)
		    { 
		      nx = logpr_multiply(nAlpha[t1][i], nA[i][j]);
		      nAlpha[t2][j] = logpr_add(nAlpha[t2][j], nx);
		      assert(nAlpha[t2][j] >= logpr_multiply(nAlpha[t1][i], nA[i][j])); 
		    }
		  nAlpha[t2][j] = logpr_multiply(nAlpha[t2][j], n_symbol);
		}
	    }

	  tf = CORPUS_SIZE % 2;
	  nx = nAlpha[tf][0];
	  for(i = 1; i < STATE_COUNT; i++)
	    { nx = logpr_add(nAlpha[tf][i], nx); }
	  result = logpr2nats(nx);
	  printf(".");
	}
      break;

    case 3:
#if BE_INFORMATIVE
      printf("\tbalanced_t arithmetic");
#endif
      for(k = 0; k < ITERATIONS; k++)
	{
	  dx = 1.0/(double)STATE_COUNT;
	  bx = double2balanced(dx);
	  for(i = 0; i < STATE_COUNT; i++) 
	    {
	      bAlpha[0][i] = bx; 
	      for(j = 0; j < STATE_COUNT; j++)
		{ bA[i][j] = double2balanced(dA[i][j]); }
	    }
	  b_symbol = double2balanced(d_symbol);

	  for(t = 0; t < CORPUS_SIZE; t++)
	    {
	      t1 = t % 2;
	      t2 = (t+1) % 2;
	      for(j = 0; j < STATE_COUNT; j++)
		{
		  bAlpha[t2][j] = balanced_multiply(bAlpha[t1][0], bA[0][j]);
		  for(i = 1; i < STATE_COUNT; i++)
		    { 
		      bx = balanced_multiply(bAlpha[t1][i], bA[i][j]);
		      bAlpha[t2][j] = balanced_add(bAlpha[t2][j], bx);
		    }
		  bAlpha[t2][j] = balanced_multiply(bAlpha[t2][j], b_symbol);
		}
	    }

	  tf = CORPUS_SIZE % 2;
	  bx = bAlpha[tf][0];
	  for(i = 1; i < STATE_COUNT; i++)
	    { bx = balanced_add(bx, bAlpha[tf][i]); }
	  result = -balanced_log(bx);
	  printf(".");
	}
      break;

    case 4:
#if BE_INFORMATIVE
      printf("\tL_t arithmetic (base %g, %d entries)", L_BASE, L_TABLE_SIZE);
#endif
      for(k = 0; k < ITERATIONS; k++)
	{
	  dx = 1.0/(double)STATE_COUNT;
	  lx = double2L(dx);
	  for(i = 0; i < STATE_COUNT; i++) 
	    {
	      lAlpha[0][i] = lx;
	      for(j = 0; j < STATE_COUNT; j++)
		{ lA[i][j] = double2L(dA[i][j]); }
	    }
	  l_symbol = double2L(d_symbol);

	  for(t = 0; t < CORPUS_SIZE; t++)
	    {
	      t1 = t % 2;
	      t2 = (t+1) % 2;
	      for(j = 0; j < STATE_COUNT; j++)
		{
		  lAlpha[t2][j] = L_multiply(lAlpha[t1][0], lA[0][j]);
		  for(i = 1; i < STATE_COUNT; i++)
		    { 
		      lx = L_multiply(lAlpha[t1][i], lA[i][j]);
		      lAlpha[t2][j] = L_add(lAlpha[t2][j], lx);
		    }
		  lAlpha[t2][j] = L_multiply(lAlpha[t2][j], l_symbol);
		}
	    }

	  tf = CORPUS_SIZE % 2;
	  lx = lAlpha[tf][0];
	  for(i = 1; i < STATE_COUNT; i++)
	    { lx = L_add(lx, lAlpha[tf][i]); }
	  result = L2nats(lx);
	  printf(".");
	}
      break;

    default:
      printf("Please specify a number between 1 and 4.\n");
      return(-1);
    }

  printf("\n\tresult %e (bits)\n", result/M_LN2);
  return 0;
}

