/*
 * (c) 1997 Eric Sven Ristad and Peter N. Yianilos.  All rights reserved.
 *
 * $Id: pr.h,v 1.3 1997/10/26 00:37:09 libpa Exp $
 *
 */

#ifndef PR_INCLUDED
#define PR_INCLUDED

/* Description.
 *  Uniform interface to all probability value modules.
 *    BALANCED  - single-precision balanced_t floating-point representation
 *    LOGPR     - double-precision floating-point logarithmic representation
 *    L         - fixed-point negative logarithmic representation (max table)
 *    L1K       - fixed-point with low precision (1k table)
 *    L16K      - fixed-point with medium precision (16k table)
 *    L32K      - fixed-point with high precision (32k table)
 *  To choose your preferred module, compile your code with -D<MODULE>,
 *  eg., -DBALANCED to use the balanced_t representation (see Makefile).
 *
 */

#ifdef BALANCED
#include "balanced.h"
#define PR_TYPE "balanced_t"
typedef balanced_t pr_t;

 #define double2pr(x)          (double2balanced(x))
 #define pr2double(px)         (balanced2double(px))
 #define pr2nats(px)           (-balanced_log(px))
 #define nats2pr(nx)           (balanced_exp(-(nx)))
 #define pr2bits(px)           (-balanced_log2(px))
 #define bits2pr(bx)           (nats2pr((bx)*M_LN2))
 #define pr_zero()             (balanced_zero())
 #define pr_unity()            (balanced_unity())

 #define pr_is_valid(px)       (balanced_is_pr(px))
 #define pr_compare(px,py)     (balanced_exact_compare(px,py))
 #define pr_are_equal(px,py)   (balanced_are_equal(px,py))
 #define pr_is_greater(px,py)  (balanced_exact_compare(px,py) > 0)
 #define pr_is_less(px,py)     (balanced_exact_compare(px,py) < 0)
 #define pr_is_geq(px,py)      (balanced_exact_compare(px,py) >= 0)
 #define pr_is_leq(px,py)      (balanced_exact_compare(px,py) <= 0)

 #define pr_multiply(px,py)    (balanced_multiply(px,py))
 #define pr_divide(px,py)      (balanced_divide(px,py))
 #define pr_add(px,py)         (balanced_add(px,py))
 #define pr_complement(px)     (balanced_subtract(balanced_unity(), (px)))

#elif defined LOGPR
#include "logpr.h"
#define PR_TYPE "logpr_t"
typedef logpr_t pr_t;

 #define double2pr(x)          double2logpr(x)
 #define pr2double(px)         logpr2double(px)
 #define pr2nats(px)           logpr2nats(px)
 #define nats2pr(nx)           nats2logpr(nx)
 #define pr2bits(px)           logpr2bits(px)
 #define bits2pr(bx)           bits2logpr(bx)
 #define pr_zero()             LOGPR_ZERO
 #define pr_unity()            LOGPR_UNITY

 #define pr_is_valid(px)       logpr_is_valid(px)
 #define pr_compare(px,py)     logpr_compare(px,py)
 #define pr_are_equal(px,py)   logpr_are_equal(px,py)
 #define pr_is_greater(px,py)  logpr_is_greater(px,py)
 #define pr_is_less(px,py)     logpr_is_less(px,py)
 #define pr_is_geq(px,py)      logpr_is_geq(px,py)
 #define pr_is_leq(px,py)      logpr_is_leq(px,py)

 #define pr_multiply(px,py)    logpr_multiply(px,py)
 #define pr_divide(px,py)      logpr_divide(px,py)
 #define pr_add(px,py)         logpr_add(px,py)
 #define pr_complement(px)     logpr_complement(px)

#elif defined(L) || defined(L1K) || defined(L16K) || defined(L32K)
#if defined L1K
#include "L.1k.h"
#elif defined L16K
#include "L.16k.h"
#else
#include "L.32k.h"
#endif
#define PR_TYPE "L_t"
typedef Lint_t pr_t;

 #define double2pr(x)          double2L(x)
 #define pr2double(px)         L2double(px)
 #define pr2nats(px)           L2nats(px)
 #define nats2pr(nx)           nats2L(nx)
 #define pr2bits(px)           L2bits(px)
 #define bits2pr(bx)           bits2L(bx)
 #define pr_zero()             L_ZERO
 #define pr_unity()            L_UNITY

 #define pr_is_valid(px)       L_is_valid(px)
 #define pr_compare(px,py)     L_compare(px,py)
 #define pr_are_equal(px,py)   L_are_equal(px,py)
 #define pr_is_greater(px,py)  L_is_greater(px,py)
 #define pr_is_less(px,py)     L_is_less(px,py)
 #define pr_is_geq(px,py)      L_is_geq(px,py)
 #define pr_is_leq(px,py)      L_is_leq(px,py)

 #define pr_multiply(px,py)    L_multiply(px,py)
 #define pr_divide(px,py)      L_divide(px,py)
 #define pr_add(px,py)         L_add(px,py)
 #define pr_complement(px)     L_complement(px)

#else
#error __FILE__: must define BALANCED|LOGPR|L|L1K|L16K|L32K when including pr.h
#endif
#endif

