/*
 * (c) 1997 Eric Sven Ristad and Peter N. Yianilos.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <errno.h>
#include <signal.h>
#include "pr.h"        /* Note: logpr.h must preceed common.h for SunOS */
#include "vector.h"

static char pr_test_rcsid[] = "$Id: pr.test.c,v 1.6 1998/02/07 02:09:25 libpa Exp $";

#define CORPUS_SIZE              (USHRT_MAX/8)
#define ALPHABET_SIZE            256
#define ALPHABET_BITS            8
#define CORPUS_BITS              (CORPUS_SIZE*ALPHABET_BITS)     
#define CORPUS_TOLERANCE         1e-3

#define DIVERGENCE_SAMPLES       1000
#define ARITHMETIC_SAMPLES       100000
#define CODELENGTH_SAMPLES       100000
#define MAXIMUM_ALPHABET         32
#define MAX_DIVERGENCE           1e-2
#define MAX_DISCREPANCY          1e-2
#define MIN_PRECISION            0.0

#define DEBUG_ARITHMETIC         FALSE
#define DEBUG_CODELENGTH         FALSE
#define TRACK_PRECISION          FALSE
#define SEED                     28090069
#ifndef M_LN2
#define M_LN2                    6.9314718055994530942E-1
#endif
#ifndef M_LOG2E
#define M_LOG2E                  1.4426950408889634074E0
#endif
#define LG(x)                    (log(x)/M_LN2)
#define MIN_PR                   1e-20
#define PR_EPSILON               1e-10
#ifndef ESUCCESS
#define ESUCCESS                 0
#endif

extern  double drand48(void);
extern  void   srand48(long);
extern  double expm1(double);
extern  double log1p(double);
extern  double fabs(double);
extern  int    finite(double);

static double Minimum_Precision = DBL_MANT_DIG;
static double Total_Precision = 0;
static double Conversion_Count = 0;

void libpa_fail(const char *file, unsigned line, const char *expr)
{
  /* (void)fprintf(stderr, "%s:%d: Assertion `#EXPR' failed.\n", __FILE__, __LINE__)) */
  fprintf(stderr, "%s:%d: NDEBUG Assertion `%s' failed.\n", file, line, expr);
  abort();
}
#define assert_u(EXPR) (((int)(EXPR)) ? (void)0 : libpa_fail(__FILE__, __LINE__, #EXPR))

double pr_sample(void)
{
  double pr;
  pr = drand48() + MIN_PR;
  return(pr);
}

void pr_sample_vector(double *pr)
{
  int i, alphabet_size;
  double total_pr;

  alphabet_size = vector_length(pr);
  total_pr = 0;
  for(i = 0; i < alphabet_size; i++)
    { pr[i] = drand48() + MIN_PR; pr[i] = pr[i] * pr[i]; total_pr += pr[i]; }
  for(i = 0; i < alphabet_size; i++)
    { pr[i] /= total_pr; }
}

void reset_precision(void)
{
  Minimum_Precision = DBL_MANT_DIG;
  Total_Precision = 0;
  Conversion_Count = 0;
}

double precision(double x, double y)
{
  double relative_error, bits_precision;
  assert_u(errno != EDOM); assert_u(errno != ERANGE);

  if (!finite(x) || !finite(y)) 
    { bits_precision = ((x == y) ? DBL_MANT_DIG : 0); }
  else if ((x != 0) && (y != 0))
    {
      relative_error = fabs(x - y) / fabs(x);
      assert(relative_error >= 0); assert(relative_error <= 1);
      if (relative_error == 0) 
	{ bits_precision = DBL_MANT_DIG; }
      else 
	{ 
	  bits_precision = -LG(relative_error); 
	  if (bits_precision > DBL_MANT_DIG) { bits_precision = DBL_MANT_DIG; }
	}
    }
  else { bits_precision = DBL_MANT_DIG; }
  assert_u(0 <= bits_precision); assert_u(bits_precision <= DBL_MANT_DIG);
  if (bits_precision < Minimum_Precision)
    { 
#if TRACK_PRECISION
      printf("precision(%e,%e) => %.2f bits\n", x, y, bits_precision);
#endif
      Minimum_Precision = bits_precision; 
    }
  Total_Precision += bits_precision;
  Conversion_Count++;
  assert_u(errno != EDOM); assert_u(errno != ERANGE);
  return(bits_precision);
}


int main(int argc, char **argv)
{
  double dw, dx, dy, dz;
  pr_t   pw, px, py, pz;
  pr_t   unity, zero, not_unity, not_zero;
  int    k, t, i;
  double divergence, discrepancy, max_divergence, max_discrepancy;
  double result, bits, total_pr, truncated_pr, *pr;
  unsigned divergence_samples;

  srand48(SEED);
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("\n\tTesting pr_t library with %d-byte %s...\n", (int)sizeof(pr_t), PR_TYPE);
  if (errno != ESUCCESS) 
    { fprintf(stderr, "\tWarning: errno %d != ESUCCESS on start up\n", (int)errno); }
  errno = ESUCCESS;

  /* verify extreme complementation */
  zero = pr_zero();
  unity = pr_unity();
  assert_u(pr2double(zero) == 0);
  assert_u(pr2double(unity) == 1);
  not_zero = pr_complement(zero);
  assert_u(pr_are_equal(not_zero, unity));
  not_unity = pr_complement(unity);
  assert_u(pr_are_equal(not_unity, zero));
  assert_u(pr_are_equal(pr_complement(unity), zero)); assert_u(errno == ESUCCESS);
  assert_u(pr_are_equal(pr_complement(zero), unity)); assert_u(errno == ESUCCESS);
  assert_u(pr2double(zero) == 0);
  assert_u(pr2double(unity) == 1);
  assert_u(errno == ESUCCESS);
  printf("\tverified extremal complementation ~0 == 1, ~1 == 0\n");

#if FALSE
  /* verify arithmetic identities */
  pw = pr_multiply(zero, unity);
  assert_u(pr_are_equal(pw, zero));
  pw = pr_multiply(unity, zero);
  assert_u(pr_are_equal(pw, zero));
  pw = pr_multiply(zero, zero);
  assert_u(pr_are_equal(pw, zero));
  pw = pr_multiply(unity, unity);
  assert_u(pr_are_equal(pw, unity));
  pw = pr_add(zero, zero);
  assert_u(pr_are_equal(pw, zero));
  pw = pr_add(zero, unity);
  assert_u(pr_are_equal(pw, unity));
  pw = pr_add(unity, zero);
  assert_u(pr_are_equal(pw, unity));
  pw = pr_divide(zero, unity);
  assert_u(pr_are_equal(pw, zero));
  printf("\tverified arithmetic identities for 0, 1\n");
#endif

  /* verify exponent representational range */
  reset_precision();
  px = double2pr(1.0/ALPHABET_SIZE);
  assert_u(pr_are_equal(px, bits2pr(ALPHABET_BITS)));
  pw = unity;
  for(i = 0; i < CORPUS_SIZE; i++)
    { pw = pr_multiply(pw, px); }
  dw = pr2bits(pw);
  /* note: this precision test is weak but sufficient to verify range */
  bits = fabs(CORPUS_BITS - dw) / CORPUS_BITS;
  if (bits >= CORPUS_TOLERANCE)
    { fprintf(stderr, "\nerror: |%.16e - dw=%.16e| = %.16e / %.16e = %.16e >= %.1e\n", 
	      (double)CORPUS_BITS, dw, fabs(CORPUS_BITS - dw), 
	      (double)CORPUS_BITS, bits, CORPUS_TOLERANCE); abort(); }
  assert_u(bits < CORPUS_TOLERANCE);
  assert_u(errno == ESUCCESS);
  /* note: i = 0 can cause pw to exceed unity */
  for(i = 1; i < CORPUS_SIZE; i++)
    { pw = pr_divide(pw, px); }
  dw = pr2bits(pw);
  bits = fabs(ALPHABET_BITS - dw) / ALPHABET_BITS;
  if (bits >= CORPUS_TOLERANCE)
    { fprintf(stderr, "\nerror: |%.16e - dw=%.16e| = %.16e / %.16e = %.16e >= %.1e\n", 
	      (double)ALPHABET_BITS, dw, fabs(ALPHABET_BITS - dw),
	      (double)ALPHABET_BITS, bits, CORPUS_TOLERANCE); abort(); }
  assert_u(bits < CORPUS_TOLERANCE);
  assert_u(errno == ESUCCESS);
  if (bits == 0) { dw = DBL_MANT_DIG; }
  printf("\tverified representational range 2^[0,-%u] to %.1e bits\n",
	 (unsigned)CORPUS_SIZE * ALPHABET_BITS, fabs(ALPHABET_BITS - dw));

  /* calculate divergence and discrepancy for random distributions */
  printf("\tgenerating %d distributions per alphabet size:\n", (int)DIVERGENCE_SAMPLES);
  printf("\t\tk\tmax diverge\tmax discrep\tmean precision\tmin precision\n");
  for(k = 2; k < MAXIMUM_ALPHABET; k *= 2)
    {
      reset_precision();
      max_divergence = 0;
      max_discrepancy = 0;
      pr = vector_create(sizeof(double), k);
      assert(vector_is_valid(pr));
      divergence_samples = DIVERGENCE_SAMPLES * k;
      for(t = 0; t < divergence_samples; t++)
	{
	  pr_sample_vector(pr);

	  total_pr = 0;
	  divergence = 0;
	  discrepancy = -1;
	  assert(vector_length(pr) == k);
	  for(i = 0; i < k; i++)
	    {
	      total_pr += pr[i];  
	      assert(0 <= pr[i]); assert(pr[i] <= 1);
	      assert(0 <= total_pr); assert(total_pr <= 1+PR_EPSILON);
	      if (pr[i] > 0)
		{
		  px = double2pr(pr[i]);
		  truncated_pr = pr2double(px);
		  assert(0 <= truncated_pr); assert(truncated_pr <= 1);
		  (void) precision(pr[i], truncated_pr);
		  discrepancy += truncated_pr;
		  divergence += pr[i] * log( truncated_pr / pr[i] );
		}
	    }
	  assert_u(fabs(total_pr - 1.0) < PR_EPSILON);
	  if (fabs(divergence) > fabs(max_divergence)) { max_divergence = divergence; }
	  if (fabs(discrepancy) > fabs(max_discrepancy)) { max_discrepancy = discrepancy; }
	}
      vector_destroy(pr); pr = NULL;
      printf("\t\t%d\t%+e\t%+e\t%e\t%e\n", k, max_divergence, max_discrepancy,
	     Total_Precision/Conversion_Count, Minimum_Precision);
      if (fabs(max_divergence) > MAX_DIVERGENCE) { abort(); }
      if (fabs(max_discrepancy) > MAX_DISCREPANCY) { abort(); }
      if (Minimum_Precision < MIN_PRECISION) { abort(); }
    }

  /* calculate precision for extreme DBL/FLT values */
  reset_precision();
  dw = FLT_MIN; dx = FLT_EPSILON; dy = DBL_MIN*2; dz = DBL_EPSILON;
  pw = double2pr(dw); px = double2pr(dx); py = double2pr(dy); pz = double2pr(dz);
  printf("\tprecision: FLT_MIN %4.1f FLT_EPSILON %4.1f DBL_MIN*2 %4.1f DBL_EPSILON %4.1f\n",
	 precision(dw, pr2double(pw)), precision(dx, pr2double(px)),
	 precision(dy, pr2double(py)), precision(dz, pr2double(pz)));
  if (Minimum_Precision < MIN_PRECISION) { abort(); }

  /* verify arithmetic operations by sampling */
  reset_precision();
  for(i = 0; i < ARITHMETIC_SAMPLES; i++)
    {
      dx = pr_sample(); px = double2pr(dx);
      dy = pr_sample(); py = double2pr(dy);

      if (pr_is_less(px, py))
	{
	  assert_u(dx < dy);
	  assert_u(pr_compare(px, py) < 0);
	  assert_u(!pr_are_equal(px, py));
	  assert_u(!pr_is_greater(px, py));
	  assert_u(pr_is_less(px, py));
	  assert_u(!pr_is_geq(px, py));
	  assert_u(pr_is_leq(px, py));
	}
      else if (pr_is_greater(px, py))
	{
	  assert_u(dx > dy);
	  assert_u(pr_compare(px, py) > 0);
	  assert_u(!pr_are_equal(px, py));
	  assert_u(pr_is_greater(px, py));
	  assert_u(!pr_is_less(px, py));
	  assert_u(pr_is_geq(px, py));
	  assert_u(!pr_is_leq(px, py));
	}
      else
	{
	  assert_u(pr_compare(px, py) == 0);
	  assert_u(pr_are_equal(px, py));
	  assert_u(!pr_is_greater(px, py));
	  assert_u(!pr_is_less(px, py));
	  assert_u(pr_is_geq(px, py));
	  assert_u(pr_is_leq(px, py));
	}

      /*
      dz = 1 - dx;
      pz = pr_complement(px);
#if DEBUG_ARITHMETIC
      printf("~\t%e\t%u\n", dz, pz);
#endif
      (void) precision(dz, pr2double(pz));
      */

      result = dx + dy;
      if ((0 < result) && (result <= 1))
	{
	  dz = dx + dy; 
	  pz = pr_add(px, py);
#if DEBUG_ARITHMETIC
	  printf("+\t%e\t%u\n", dz, pz);
#endif
	  (void) precision(dz, pr2double(pz));
	}

      result = log10(dx) + log10(dy);
      if ((DBL_MIN_10_EXP < result) && (result < 0))
	{
	  dz = dx * dy; 
	  pz = pr_multiply(px, py);
#if DEBUG_ARITHMETIC
	  printf("*\t%e\t%u\n", dz, pz);
#endif
	  (void) precision(dz, pr2double(pz));
	}

      result = log10(dx) - log10(dy);
      if ((DBL_MIN_EXP < result) && (result <= 0))
	{
	  assert_u(dx <= dy); assert_u(pr_is_leq(px, py));
	  dz = dx / dy; 
	  pz = pr_divide(px, py);
#if DEBUG_ARITHMETIC
	  printf("/\t%e\t%u\n", dz, pz);
#endif
	  (void) precision(dz, pr2double(pz));
	}
    }
  printf("\tverified arithmetic operations on %d pairs:\n", ARITHMETIC_SAMPLES);
  printf("\t\tprecision: mean %5.2f bits", Total_Precision/Conversion_Count);
  printf(" minimum %5.2f bits\n", Minimum_Precision);
  if (errno != ESUCCESS) { fprintf(stderr, "errno %d\n", (int)errno); abort(); }
  if (Minimum_Precision < MIN_PRECISION) { abort(); }

  reset_precision();
  for(i = 0; i < CODELENGTH_SAMPLES; i++)
    {
      dx = drand48() * (double)(DBL_MIN_EXP+1);
      assert(DBL_MIN_EXP < dx); assert(dx <= 0);
      px = bits2pr(-dx);
      dy = pow(2, dx); dz = pr2double(px);
#if DEBUG_CODELENGTH
      printf("pow(2,%e)\t%e\t%e\n", dx, dy, dz);
#endif
      precision(dy, dz);

      dx = dx / M_LOG2E;
      px = nats2pr(-dx);
      dy = exp(dx); dz = pr2double(px);
#if DEBUG_CODELENGTH
      printf("exp(%e)\t%e\t%e\n", dx, dy, dz);
#endif
      precision(dy, dz);
    }
  if (errno != ESUCCESS) { fprintf(stderr, "errno %d\n", (int)errno); abort(); }
  printf("\tverified codelength coversions on %d pairs:\n", CODELENGTH_SAMPLES);
  printf("\t\tprecision: mean %5.2f bits", Total_Precision/Conversion_Count);
  printf(" minimum %5.2f bits\n", Minimum_Precision);
  if (Minimum_Precision < MIN_PRECISION) { abort(); }

  printf("\tsizeof(pr_t) == %d\n", (int)sizeof(pr_t));
  printf("\tpr_t library passes all tests with %s\n", PR_TYPE);
  return(0);
}



