/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 *
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include "common.h"
#include "memory.h"
#include "vector.h"
#include "suffix.internal.h"
#include "suffix.array.h"

static char suffix_array_rcsid[] = "$Id: suffix.array.c,v 2.6 1998/02/07 00:40:44 libpa Exp $";

#define ANNOUNCE_RCSID           FALSE
#define EXTRA_SAFE               FALSE /* validate after every mutation */
#define DEBUG                    FALSE
#define SPAN_LIMIT               1048576 /* 65536 */

/**********************************************************************/
#define suffix_array_pruned(a)    ((a)->pruned)
/**********************************************************************/

suffix_array_t *suffix_array_create(int length_limit, const vector_t *corpus)
{
  suffix_array_t *array;
  unsigned corpus_length, i;

  assert(length_limit > 0);
  assert(vector_is_valid(corpus));

  /* allocate the suffix_array_t struct */
  corpus_length = vector_length(corpus);
  if (length_limit > corpus_length) { length_limit = corpus_length; }
  array = memory_create(sizeof(suffix_array_t));
  assert(memory_is_valid(array));
  array->length_limit = length_limit;
  array->pruned = 0;
  array->corpus_length = vector_length(corpus);
  array->corpus_parity = vector_parity(corpus);
  array->corpus = corpus;
  array->indices = vector_create(sizeof(suffix_index_t), corpus_length);

  /* set the indices and then sort them */
  for(i = 0; i < corpus_length; i++)
    { array->indices[i] = (suffix_index_t)i; }
  suffix_indices_bounded_sort(array->indices, array->corpus, length_limit-1);

  assert(suffix_array_length_limit(array) == length_limit);
  assert(suffix_array_corpus_length(array) == corpus_length);
  assert(suffix_array_is_valid(array));
  return(array);
}

void suffix_array_destroy(suffix_array_t *array)
{
  assert(suffix_array_is_valid(array));
  array->corpus = NULL;
  vector_destroy(array->indices); array->indices = NULL;
  memory_destroy(array);
}

/************************************************************
 ************************************************************/

boolean_t suffix_array_is_valid(const suffix_array_t *array)
{
  suffix_index_t *indices_copy;
  unsigned i, length;

  if ((array->length_limit == 0) ||
      (array->length_limit > array->corpus_length))
    { fprintf(stderr, "suffix_array_t error: invalid length limit\n"); return(FALSE); }

  if (!vector_is_valid(array->corpus) ||
      (vector_length(array->corpus) != array->corpus_length) ||
      (vector_parity(array->corpus) != array->corpus_parity))
    { fprintf(stderr, "suffix_array_t error: invalid corpus\n"); return(FALSE); }

  if (!vector_is_valid(array->indices) ||
      (vector_object_size(array->indices) != sizeof(suffix_index_t)) ||
      (vector_length(array->corpus) != (vector_length(array->indices)+array->pruned)))
    { fprintf(stderr, "suffix_array_t error: invalid indices\n"); return(FALSE); }

  if (suffix_array_pruned(array) == 0)
    {
      /* make sure all the indices are really there */
      length = vector_length(array->indices);
      indices_copy = vector_copy(array->indices);
      vector_quicksort(indices_copy, &uint_compare, NULL);
      for(i = 0; i < length; i++) 
	{ 
	  if (indices_copy[i] != i) 
	    { 
	      fprintf(stderr, "suffix_array_t error: invalid index %u\n", i); 
	      return(FALSE); 
	    }
 	}
      vector_destroy(indices_copy); indices_copy = NULL;
    }
  else
    {
      /* make sure that all indices are valid */
      length = vector_length(array->indices);
      for(i = 0; i < length; i++) 
	{ 
	  if (array->indices[i] >= array->corpus_length) 
	    { 
	      fprintf(stderr, "suffix_array_t error: pruned index %u out of bounds\n", i); 
	      return(FALSE); 
	    } 
	}
    }

  /* verify that indices array is totally ordered */
  if (!suffix_indices_are_bounded_sorted(array->indices, 
					 array->corpus, array->length_limit-1))
    { 
      fprintf(stderr, "suffix_array_t error: indices not sorted to length %u\n",
	      (unsigned)array->length_limit);
      return(FALSE); 
    }

  return(TRUE);
}
   
boolean_t suffix_array_are_equivalent(const suffix_array_t *left, 
				      const suffix_array_t *right)
{
  return((left->length_limit == right->length_limit) &&
	 (left->pruned == right->pruned) &&
	 (left->corpus_length == right->corpus_length) &&
	 (left->corpus_parity == right->corpus_parity) &&
	 vector_are_equivalent(left->corpus, right->corpus) &&
	 vector_are_equivalent(left->indices, right->indices));
}



/************************************************************
 ************************************************************/

/* Returns: index in array->indices where affix_index is or should be */
/* Requires: indices are totally ordered, with or without duplicates  */
/* Note: if duplicates exist, then return lowest index in their midst  */
static unsigned suffix_array_lower_bound(const suffix_array_t *array,
					 unsigned affix_length, const suffix_t *affix)
{
  unsigned suffix_length, lower, upper, middle;
  suffix_index_t suffix_index;
  suffix_t *suffix;
  int comparison;
  
  assert(array); assert(affix);
  assert(affix_length > 0);
  assert(affix_length >= suffix_array_symbol_size(array));
  assert(affix_length <= (suffix_array_symbol_size(array)
			  * suffix_array_length_limit(array)));

  /* find lowest index where affix could be inserted into array */
  lower = 0;
  upper = vector_length(array->indices);
  while(lower < upper)
    {
      middle = lower + ((upper - lower) / 2);
      assert(lower <= middle); assert(middle < upper);
      assert(vector_index_is_valid(array->indices, middle));

      suffix_index = array->indices[middle];
      suffix_length = suffix_index2length(suffix_index, array->corpus);
      suffix = suffix_index2suffix(suffix_index, array->corpus);

      comparison = suffix_compare(affix_length, affix, suffix_length, suffix);
      if (comparison <= 0) { upper = middle; }
      else { lower = middle + 1; }
    } 
  assert(lower == upper); assert(lower <= vector_length(array->indices));
  return(lower);
}

/* Returns: minimal upper bound on affix in array in [lower_bound,suffix_count] */
/*          ie., lowest index that doesn't include affix whose predecessor does */
/* Requires: lower_bound is valid index */
static unsigned suffix_array_upper_bound(const suffix_array_t *array,
					 unsigned lower_bound,
					 unsigned affix_length, const suffix_t *affix)
{
  unsigned suffix_count, suffix_length, lower, upper, middle;
  suffix_index_t suffix_index;
  suffix_t *suffix;

  assert(array); assert(affix);
  assert(affix_length > 0);
  assert(affix_length >= suffix_array_symbol_size(array));
  assert(affix_length <= (suffix_array_symbol_size(array)
			  * suffix_array_length_limit(array)));
  assert(suffix_includes(suffix_index2length(array->indices[lower_bound], array->corpus),
			 suffix_index2suffix(array->indices[lower_bound], array->corpus),
			 affix_length, 
			 affix));

  abort();  /* this isn't tested */

  suffix_count = vector_length(array->indices);
  assert(lower_bound < suffix_count);
  assert(lower_bound == suffix_array_lower_bound(array, affix_length, affix));

  lower = lower_bound+1;
  upper = suffix_count;
  while(lower < upper)
    {
      /* first try at the lower bound */
      suffix_index = array->indices[lower];
      suffix = suffix_index2suffix(suffix_index, array->corpus);
      suffix_length = suffix_index2length(suffix_index, array->corpus);
      if (!suffix_includes(suffix_length, suffix, affix_length, affix))	{ return(lower); }
      else { lower++; }
      if (lower == upper) { return(upper); }

      /* then try bisecting range */
      middle = lower + ((upper - lower) / 2);
      assert(lower <= middle); assert(middle < upper);
      assert(lower_bound < lower); assert(upper <= suffix_count);

      suffix_index = array->indices[middle];
      suffix = suffix_index2suffix(suffix_index, array->corpus);
      suffix_length = suffix_index2length(suffix_index, array->corpus);
      if (suffix_includes(suffix_length, suffix, affix_length, affix)) { lower = middle+1; }
      else { upper = middle; }
    }
  assert(lower_bound <= lower); assert(lower == upper); assert(upper <= suffix_count);
  return(upper);
}


unsigned suffix_array_maximal_length(const suffix_array_t *array, 
				     unsigned string_length, const void *string)
{
  suffix_t *affix, *suffix;
  unsigned symbol_size, affix_length, suffix_index, suffix_length;
  unsigned lower_length, upper_length, index, suffix_length_limit;

  assert(array);
  assert(vector_is_valid(array->indices));
  assert(vector_is_valid(array->corpus));
  assert(array->corpus_length == vector_length(array->corpus));
  assert(array->corpus_parity == vector_parity(array->corpus));

  if (string_length == 0) { return(0); }
  else if (suffix_array_corpus_length(array) == 0) { return(0); }
  assert(string); assert(string_length > 0);

  /* convert string into bounded suffix_t representation */
  symbol_size = suffix_array_symbol_size(array);
  suffix_length_limit = suffix_array_length_limit(array) * symbol_size;
  affix_length = string_length * symbol_size;
  affix = (suffix_t *)string + affix_length;
  if (affix_length > suffix_length_limit) { affix_length = suffix_length_limit; }

  /* find lowest index where affix could be inserted into array */
  index = suffix_array_lower_bound(array, affix_length, affix);
  assert(index <= vector_length(array->indices));

  /* now see how close the match really is */
  if (index < vector_length(array->indices)) 
    { 
      suffix_index = array->indices[index];
      assert(suffix_index_is_valid(suffix_index, array->corpus));
      suffix = suffix_index2suffix(suffix_index, array->corpus);
      suffix_length = suffix_index2length(suffix_index, array->corpus);
      upper_length = suffix_common_length(suffix_length, suffix, affix_length, affix);
      assert(upper_length <= suffix_length);
      assert(upper_length <= affix_length);
      if (upper_length == affix_length) { return(string_length); }
      else if (index == 0) { return(upper_length/symbol_size); }
    }
  assert(index > 0);
  assert(vector_index_is_valid(array->indices, index-1));

  suffix_index = array->indices[index-1];
  assert(suffix_index_is_valid(suffix_index, array->corpus));
  suffix = suffix_index2suffix(suffix_index, array->corpus);
  suffix_length = suffix_index2length(suffix_index, array->corpus);
  lower_length = suffix_common_length(suffix_length, suffix, affix_length, affix);
  if ((index < vector_length(array->indices)) && (lower_length < upper_length))
    { lower_length = upper_length; }
  assert(lower_length/symbol_size < string_length);
  return(lower_length/symbol_size);
}

boolean_t suffix_array_includes(const suffix_array_t *array, 
				unsigned string_length, const void *string)
{
  unsigned length_limit, maximal_length;
  length_limit = suffix_array_length_limit(array);
  if (string_length > length_limit) { return(FALSE); }
  maximal_length = suffix_array_maximal_length(array, string_length, string);
  return(maximal_length == string_length);
}

unsigned suffix_array_frequency(const suffix_array_t *array, 
				unsigned string_length, const void *string)
{
  suffix_t *affix, *suffix;
  suffix_index_t suffix_index;
  unsigned symbol_size, length_limit, suffix_count, corpus_length;
  unsigned affix_length, suffix_length, suffix_length_limit;
  unsigned index, lower, upper, middle;

  assert(array);
  assert(vector_is_valid(array->indices));
  assert(vector_is_valid(array->corpus));
  assert(array->corpus_length == vector_length(array->corpus));
  assert(array->corpus_parity == vector_parity(array->corpus));

  suffix_count = vector_length(array->indices);
  corpus_length = vector_length(array->corpus);
  length_limit = suffix_array_length_limit(array);
  assert(suffix_count+suffix_array_pruned(array) == corpus_length);
  if ((string_length > corpus_length) || (string_length > length_limit)) { return(0); }
  else if (string_length == 0) { return(corpus_length+1); }
  assert(0 < string_length); assert(string_length <= suffix_array_length_limit(array));

  /* convert string into suffix_t representation */
  symbol_size = vector_object_size(array->corpus);
  suffix_length_limit = length_limit * symbol_size;
  affix_length = string_length * symbol_size;
  affix = (suffix_t *)string + affix_length;
  if (affix_length > suffix_length_limit) { affix_length = suffix_length_limit; }

  /* find lowest index where affix could be inserted into array */
  index = suffix_array_lower_bound(array, affix_length, affix);
  assert(index <= suffix_count);
  if (index == suffix_count) { return(0); } /* affix not found */
  assert(vector_index_is_valid(array->indices, index));
  suffix_index = array->indices[index];
  suffix = suffix_index2suffix(suffix_index, array->corpus);
  suffix_length = suffix_index2length(suffix_index, array->corpus);
  assert(suffix_compare(affix_length, affix, suffix_length, suffix) <= 0);

  if (suffix_length < affix_length) { return(0); } /* can't be there! */
  if (!suffix_includes(suffix_length, suffix, affix_length, affix)) { return(0); }
  assert(suffix_length >= affix_length);

  /* at this point the string must exist in the suffix array */
  assert(string_length == suffix_array_maximal_length(array, string_length, string));
  assert(suffix_includes(suffix_length, suffix, affix_length, affix));
  assert(suffix_count == vector_length(array->indices));

  /* so go find the string's minimal upper bound in the suffix array */
  lower = index+1;
  upper = suffix_count;
  while(lower < upper)
    {
      middle = lower + ((upper - lower) / 2);
      assert(lower <= middle); assert(middle < upper);
      assert(index < lower); assert(upper <= suffix_count);

      suffix_index = array->indices[middle];
      suffix = suffix_index2suffix(suffix_index, array->corpus);
      suffix_length = suffix_index2length(suffix_index, array->corpus);
      if (suffix_includes(suffix_length, suffix, affix_length, affix)) { lower = middle+1; }
      else { upper = middle; }
    }
  assert(index < lower); assert(lower == upper); assert(upper <= suffix_count);
  return(lower - index);
}

/************************************************************
 * INPUT/OUTPUT
 ************************************************************/

static void uchar_fprintf(const void *symbol_p, FILE *stream)
{ fprintf(stream, "%u", (unsigned)(*((unsigned char *)symbol_p))); }
static void ushrt_fprintf(const void *symbol_p, FILE *stream)
{ fprintf(stream, "%u", (unsigned)(*((unsigned short *)symbol_p))); }
static void uint_fprintf(const void *symbol_p, FILE *stream)
{ fprintf(stream, "%u", (unsigned)(*((unsigned int *)symbol_p))); }

static void uchar_fscanf(void *symbol_p, FILE *stream)
{ unsigned symbol; fscanf(stream, "%u", &symbol); *(UCHAR *)symbol_p = (UCHAR)symbol; }
static void ushrt_fscanf(void *symbol_p, FILE *stream)
{ unsigned symbol; fscanf(stream, "%u", &symbol); *(USHORT *)symbol_p = (USHORT)symbol; }
static void uint_fscanf(void *symbol_p, FILE *stream)
{ fscanf(stream, "%u", (unsigned *)symbol_p); }

static void (*Printers[])(const void *, FILE *) 
     = { NULL, &uchar_fprintf, &ushrt_fprintf, NULL, &uint_fprintf };
static void (*Scanners[])(void *, FILE *) 
     = { NULL, &uchar_fscanf, &ushrt_fscanf, NULL, &uint_fscanf };

suffix_array_t *suffix_array_fscanf(FILE *stream)
{
  unsigned symbol_size, limit, pruned, length, parity;
  suffix_array_t *array;

  fscanf(stream, "<%u,%u,%u,%u,%u>", &symbol_size, &limit, &pruned, &length, &parity);
  assert((symbol_size == 1) || (symbol_size == 2) || (symbol_size == 4));

  array = memory_create(sizeof(suffix_array_t));
  array->length_limit = limit;
  array->pruned = pruned;
  array->corpus_length = length;
  array->corpus_parity = parity;
  array->corpus = vector_fscanf(stream, Scanners[symbol_size]);
  array->indices = vector_fscanf(stream, Scanners[sizeof(suffix_index_t)]);
  assert(vector_is_valid(array->corpus));
  assert(vector_is_valid(array->indices));

  assert(vector_object_size(array->corpus) == symbol_size);
  assert(vector_object_size(array->indices) == sizeof(suffix_index_t));
  assert(vector_length(array->corpus) == length);
  assert(vector_parity(array->corpus) == parity);
  assert(vector_length(array->corpus) == vector_length(array->indices)+pruned);
  assert(suffix_array_is_valid(array));
  return(array);
}

void suffix_array_fprintf(const suffix_array_t *array, FILE *stream)
{
  unsigned symbol_size, limit, pruned, length, parity;
  symbol_size = suffix_array_symbol_size(array);
  limit = suffix_array_length_limit(array);
  pruned = suffix_array_pruned(array);
  length = array->corpus_length;
  parity = array->corpus_parity;
  fprintf(stream, "<%u,%u,%u,%u,%u>", symbol_size, limit, pruned, length, parity);
  (void) vector_fprintf(array->corpus, stream, Printers[symbol_size]);
  (void) vector_fprintf(array->indices, stream, Printers[sizeof(suffix_index_t)]);
}


suffix_array_t *suffix_array_read(FILE *stream)
{
  unsigned symbol_size, limit, pruned, length, parity;
  suffix_array_t *array;
  assert(!ferror(stream));

  symbol_size = 0;
  (void) fread(&symbol_size, sizeof(unsigned), 1, stream); assert(!ferror(stream));
  (void) fread(&limit, sizeof(unsigned), 1, stream); assert(!ferror(stream));
  (void) fread(&pruned, sizeof(unsigned), 1, stream); assert(!ferror(stream));
  (void) fread(&length, sizeof(unsigned), 1, stream); assert(!ferror(stream));
  (void) fread(&parity, sizeof(unsigned), 1, stream); assert(!ferror(stream));
  assert((symbol_size == 1) || (symbol_size == 2) || (symbol_size == 4));

  array = memory_create(sizeof(suffix_array_t));
  assert(memory_is_valid(array));
  array->length_limit = limit;
  array->pruned = pruned;
  array->corpus_length = length;
  array->corpus_parity = parity;
  array->corpus = vector_read(stream, NULL);
  array->indices = vector_read(stream, NULL);

  assert(vector_object_size(array->corpus) == symbol_size);
  assert(vector_object_size(array->indices) == sizeof(suffix_index_t));
  assert(vector_length(array->corpus) == length);
  assert(vector_parity(array->corpus) == parity);
  assert(vector_length(array->corpus) == vector_length(array->indices)+pruned);
  assert(suffix_array_is_valid(array));
  return(array);
}

void suffix_array_write(const suffix_array_t *array, FILE *stream)
{
  unsigned symbol_size, limit, pruned, length, parity;
  assert(!ferror(stream));

  symbol_size = suffix_array_symbol_size(array);
  limit = suffix_array_length_limit(array);
  pruned = suffix_array_pruned(array);
  length = array->corpus_length;
  parity = array->corpus_parity;

  (void) fwrite(&symbol_size, sizeof(unsigned), 1, stream); assert(!ferror(stream));
  (void) fwrite(&limit, sizeof(unsigned), 1, stream); assert(!ferror(stream));
  (void) fwrite(&pruned, sizeof(unsigned), 1, stream); assert(!ferror(stream));
  (void) fwrite(&length, sizeof(unsigned), 1, stream); assert(!ferror(stream));
  (void) fwrite(&parity, sizeof(unsigned), 1, stream); assert(!ferror(stream));

  (void) vector_write(array->corpus, stream, NULL);
  (void) vector_write(array->indices, stream, NULL);
}

/************************************************************
 * HIGHER ORDER FUNCTIONS
 ************************************************************/

void suffix_array_map(const suffix_array_t *array, 
		      unsigned string_length,
		      boolean_t (*proc)(unsigned string_length,
					const void *string,
					unsigned frequency,
					void *parameters),
		      void *parameters)
{
  suffix_t *suffix, *upper_suffix;
  suffix_index_t suffix_index, upper_suffix_index, minimum_suffix_index;
  unsigned minimum_suffix_length, upper_suffix_length;
  unsigned symbol_size, length_limit, corpus_length, suffix_count, frequency;
  unsigned index, lower, upper, middle;
  boolean_t is_alive;
  void *string;

  assert(array);
  assert(vector_is_valid(array->corpus));
  assert(vector_is_valid(array->indices));
  assert(vector_length(array->corpus) == 
	 vector_length(array->indices)+suffix_array_pruned(array));

  length_limit = suffix_array_length_limit(array);
  symbol_size = vector_object_size(array->corpus);
  corpus_length = vector_length(array->corpus);

  if (string_length == 0) { (void) proc(0, NULL, corpus_length+1, parameters); return; }
  else if ((string_length > length_limit) || (string_length > corpus_length)) { return; }
  assert(string_length > 0); assert(string_length <= length_limit);

  suffix_count = vector_length(array->indices); assert(suffix_count > 0);
  minimum_suffix_length = string_length * symbol_size;
  minimum_suffix_index = string_length - 1;

  index = 0;
  while(index < suffix_count)
    {
      /* find next suffix to satisfy length constraint */
      suffix_index = array->indices[index];
      while(suffix_index < minimum_suffix_index)  /* suffix_index == suffix_length-1 */
      	{
	  index++;
	  if (index >= suffix_count) { return; }
	  suffix_index = array->indices[index];
	} 
      assert(vector_index_is_valid(array->indices, index));
      assert(suffix_index == array->indices[index]);
      assert(suffix_index2length(suffix_index, array->corpus) >= minimum_suffix_length);
      assert(vector_index_is_valid(array->corpus, 1+suffix_index-string_length));

      /* found a valid string of appropriate length */
      suffix = suffix_index2suffix(suffix_index, array->corpus);
      string = vector_element_p(array->corpus, 1+suffix_index-string_length);
      assert(string_length == suffix_array_maximal_length(array, string_length, string));

      /* so go find the string's minimal upper bound in the suffix array */
      lower = index+1;
      upper = suffix_count;
      while(lower < upper)
	{
	  middle = lower + ((upper - lower) / 2);
	  assert(lower <= middle); assert(middle < upper);
	  assert(index < lower); assert(upper <= suffix_count);

	  upper_suffix_index = array->indices[middle];
	  upper_suffix = suffix_index2suffix(upper_suffix_index, array->corpus);
	  upper_suffix_length = suffix_index2length(upper_suffix_index, array->corpus);
	  if (suffix_includes(upper_suffix_length, upper_suffix, 
			      minimum_suffix_length, suffix)) 
	    { lower = middle+1; }
	  else { upper = middle; }
	}
      assert(index < lower); 
      assert(lower == upper); 
      assert(upper <= suffix_count);

      /* call the caller's procedure with the appropriate information */
      frequency = upper - index;
      assert(frequency == suffix_array_frequency(array, string_length, string));
      is_alive = proc(string_length, string, frequency, parameters);
      if (!is_alive) { return; }

      /* start search for the next string */
      index = upper;
    }
}


void suffix_array_dump(const suffix_array_t *array)
{
  unsigned suffix_count, corpus_length, suffix_length;
  int i, j;
  suffix_index_t suffix_index;
  suffix_t *suffix;

  suffix_count = suffix_array_size(array);
  corpus_length = suffix_array_corpus_length(array);
  printf("suffix_array_t for %d-byte symbols, %u-symbol corpus, %u suffixes:\n",
	 (int)suffix_array_symbol_size(array), corpus_length, suffix_count);

  for(i = 0; i < suffix_count; i++)
    {
      suffix_index = array->indices[i];
      assert(suffix_index_is_valid(suffix_index, array->corpus));
      suffix = suffix_index2suffix(suffix_index, array->corpus);
      suffix_length = suffix_index2length(suffix_index, array->corpus);

      printf("  %4.4u:", suffix_length);
      for(j = 0; j < suffix_length; j++)
	{ suffix--; printf(" %d", (int)(*suffix)); }
      printf("\n");

#ifndef NDEBUG
      if (i > 0)
	{
	  suffix_index_t previous_index;
	  unsigned  previous_length;
	  suffix_t *previous_suffix;

	  assert(suffix_index == array->indices[i]);
	  assert(suffix_length == suffix_index2length(suffix_index, array->corpus));
	  suffix = suffix_index2suffix(suffix_index, array->corpus);

	  previous_index = array->indices[i-1];
	  assert(suffix_index_is_valid(previous_index, array->corpus));
	  previous_length = suffix_index2length(previous_index, array->corpus);
	  previous_suffix = suffix_index2suffix(previous_index, array->corpus);
	  assert(suffix_compare(previous_length, previous_suffix, 
				suffix_length, suffix) < 0);
	}
#endif
    }
}


