/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>  /* for memset() */
#include <math.h>    /* for drand48() */
#include "common.h"
#include "memory.h"
#include "vector.h"
#include "suffix.internal.h"
#include "suffix.array.h"


static char suffix_array_test_rcsid[] = "$Id: suffix.array.test.c,v 2.4 1998/01/28 14:52:30 libpa Exp $";


#define BE_VERBOSE            FALSE
#define DEBUG_MONOTONOUS      FALSE
#define DEBUG_CYCLIC          FALSE
#define DEBUG_FREQUENCY       FALSE

#define BORING_STRING_LENGTH  512 /* 512 */

#define CYCLE                 16   /* 16 */
#define CYCLE_COUNT           32   /* 32 */
#define MAX_CYCLIC_STRING     (CYCLE*CYCLE_COUNT)

#define MAX_CHAR              16   /* 16 */
#define RANDOM_STRING_LENGTH  1024 /* 1024 */

#define TERMINATE_TRIALS      10
#define TERMINATE_THRESHOLD   0.95

static unsigned trigram[MAX_CHAR][MAX_CHAR][MAX_CHAR];
static unsigned bigram[MAX_CHAR][MAX_CHAR];
static unsigned unigram[MAX_CHAR];

extern double drand48(void);
extern void suffix_array_dump(const suffix_array_t *array);

void set_symbol(void *symbol_p, int symbol_size, unsigned value)
{
  switch(symbol_size)
    {
    case 1: *((unsigned char  *)symbol_p) = value; break;
    case 2: *((unsigned short *)symbol_p) = value; break;
    case 4: *((unsigned int   *)symbol_p) = value; break;
    default: abort();
    }
}

unsigned get_symbol(const void *symbol_p, int symbol_size)
{
  switch(symbol_size)
    {
    case 1: return(*((unsigned char  *)symbol_p));
    case 2: return(*((unsigned short *)symbol_p));
    case 4: return(*((unsigned int   *)symbol_p));
    default: abort();
    }
}

boolean_t premature_terminator(unsigned string_length, const void *string, 
			       unsigned frequency, void *parameter)
{
  boolean_t *is_alive_p = parameter;
  double throw;
  assert(*is_alive_p); assert(frequency > 0);
  throw = drand48();
  if (throw > TERMINATE_THRESHOLD) { *is_alive_p = FALSE; }
  return(*is_alive_p);
}

boolean_t monotone_verifier(unsigned string_length, const void *string,
			    unsigned frequency, void *symbol_size_p)
{
  int symbol_size;
  const unsigned char *symbol_p;

  symbol_p = string;
  symbol_size = *(int *)symbol_size_p;
  assert((symbol_size == 1) || (symbol_size == 2) || (symbol_size == 4));
  assert((string_length == 0) || (symbol_p != NULL));
  assert(frequency == 1+BORING_STRING_LENGTH-string_length);
  return(TRUE);
}

boolean_t ngram_verifier(unsigned string_length, const void *string,
			 unsigned frequency, void *symbol_size_p)
{
  int symbol_size, i, j, k;
  const unsigned char *symbol_p;

  symbol_p = string;
  symbol_size = *(int *)symbol_size_p;
  assert((symbol_size == 1) || (symbol_size == 2) || (symbol_size == 4));
  assert((string_length == 0) || (symbol_p != NULL));

  if (string_length > 0) { i = get_symbol(symbol_p, symbol_size); }
  if (string_length > 1) { j = get_symbol(symbol_p+symbol_size, symbol_size); }
  if (string_length > 2) { k = get_symbol(symbol_p+(2*symbol_size), symbol_size); }

  switch(string_length)
    {
    case 0: assert(frequency == RANDOM_STRING_LENGTH+1); break;
    case 1: assert(frequency == unigram[i]); unigram[i] = 0; break;
    case 2: assert(frequency == bigram[i][j]); bigram[i][j] = 0; break;
    case 3: assert(frequency == trigram[i][j][k]); trigram[i][j][k] = 0; break;
    default: abort();
    }
  return(TRUE);
}

/* uniform string */
boolean_t monotonous_test(int symbol_size)
{
  vector_t        *corpus, *string, *corpus_copy;
  suffix_array_t  *array, *array_copy;
  unsigned         i, length, frequency;
  FILE            *stream;

  corpus = vector_create(symbol_size, BORING_STRING_LENGTH);
  assert(vector_is_valid(corpus));
  assert(vector_length(corpus) == BORING_STRING_LENGTH);
  memset(corpus, 0, BORING_STRING_LENGTH*symbol_size);
  string = vector_copy(corpus);
  assert(vector_is_valid(string));

  array = suffix_array_create(BORING_STRING_LENGTH, corpus);
  assert(array);
  assert(suffix_array_size(array) == BORING_STRING_LENGTH);
  assert(suffix_array_is_valid(array));
  assert(suffix_array_are_equivalent(array, array));
#if BE_VERBOSE
  printf("%u\tcreated %d-byte suffix_array_t for monotonous pattern of length %d\n",
	 memory_total_bytes(), symbol_size, (int)BORING_STRING_LENGTH);
#endif

  assert(vector_length(string) == BORING_STRING_LENGTH);
  for(i = 0; i < BORING_STRING_LENGTH; i++)
    {
      assert(suffix_array_includes(array, i, string));
      length = suffix_array_maximal_length(array, i, string);
      assert(length == i);
      frequency = suffix_array_frequency(array, i, string);
      assert(frequency == 1+BORING_STRING_LENGTH-i);
      suffix_array_map(array, i, &monotone_verifier, &symbol_size);
    }
#if BE_VERBOSE  
  printf("%u\tverified length/freqency of all %d-byte monotonous substrings\n", 
	 memory_total_bytes(), symbol_size);
#endif

  /* test ascii input/output */
  stream = tmpfile();
  assert(stream);
  suffix_array_fprintf(array, stream);
  rewind(stream);
  array_copy = suffix_array_fscanf(stream);
  assert(!ferror(stream));
  assert(suffix_array_are_equivalent(array, array_copy));
  fclose(stream); stream = NULL;
  corpus_copy = (vector_t *)suffix_array_corpus(array_copy);
  suffix_array_destroy(array_copy); array_copy = NULL;
  vector_destroy(corpus_copy); corpus_copy = NULL;
#if BE_VERBOSE
  printf("%u\tverified %d-byte _fprintf()/_fscanf() on monotonous pattern\n", 
	 memory_total_bytes(), symbol_size);
#endif

  /* test binary input/output */
  stream = tmpfile();
  assert(stream);
  suffix_array_write(array, stream);
  rewind(stream);
  array_copy = suffix_array_read(stream);
  assert(!ferror(stream));
  assert(suffix_array_are_equivalent(array, array_copy));
  fclose(stream); stream = NULL;
  corpus_copy = (vector_t *)suffix_array_corpus(array_copy);
  suffix_array_destroy(array_copy); array_copy = NULL;
  vector_destroy(corpus_copy); corpus_copy = NULL;
#if BE_VERBOSE
  printf("%u\tverified %d-byte _write()/_read() on monotonous pattern\n", 
	 memory_total_bytes(), symbol_size);
#endif

  vector_destroy(string); string = NULL;
  suffix_array_destroy(array); array = NULL;
  vector_destroy(corpus); corpus = NULL;
#if BE_VERBOSE
  printf("%u\tdestroyed %d-byte monotonous string and its suffix_array_t\n", 
	 memory_total_bytes(), symbol_size);
#else
  printf("%u\tverified suffix_array_t for %d-byte monotonous string of length %d\n", 
	 memory_total_bytes(), symbol_size, (int)BORING_STRING_LENGTH);
#endif

  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  return(TRUE);
}


boolean_t cyclic_test(int symbol_size)
{
  suffix_array_t *array, *array_copy;
  unsigned corpus_length, length, frequency;
  void *symbol_p, *corpus, *corpus_copy, *corpus_substring;
  FILE *stream;
  boolean_t is_alive;
  int i, j, k, t;

  symbol_p = memory_create(symbol_size);
  assert(symbol_p);
  assert(memory_bytes(symbol_p) == symbol_size);

  /* create corpus with cyclic pattern */
  corpus = vector_create_surplus(symbol_size, 0, MAX_CYCLIC_STRING);
  assert(vector_is_valid(corpus));
  assert(vector_length(corpus) == 0);
#if BE_VERBOSE
  printf("%u\tinserting %d-byte %dx%d cyclic patterns:", 
	 memory_total_bytes(), symbol_size, (int)CYCLE_COUNT, (int)CYCLE);
#endif
  corpus_length = 0;
  for(i = 0; i < CYCLE_COUNT; i++)
    { 
#if BE_VERBOSE
      printf(" %d", i);
#endif
      for(j = 0; j < CYCLE; j++)
	{
	  assert(corpus_length >= 0);
	  assert(corpus_length < MAX_CYCLIC_STRING);
	  set_symbol(symbol_p, symbol_size, j);
	  corpus = vector_append(corpus, symbol_p);
	  assert(vector_is_valid(corpus));
	  corpus_length++;
	  assert(vector_length(corpus) == corpus_length);
	}
    }
#if BE_VERBOSE
  printf("\n");
#endif

  /* create suffix_array_t for cyclic pattern */
  array = suffix_array_create(corpus_length, corpus);
  assert(array);
  assert(suffix_array_size(array) == corpus_length);
  assert(suffix_array_symbol_size(array) == symbol_size);
  assert(suffix_array_corpus(array) == corpus);
  assert(suffix_array_corpus_length(array) == corpus_length);
  assert(suffix_array_length_limit(array) == corpus_length);
  assert(suffix_array_is_valid(array));
  
  /* check that all suffixes are all there correctly */
  for(k = 0; k < corpus_length; k++)
    {
      corpus_substring = vector_element_p(corpus, k);
      length = suffix_array_maximal_length(array, corpus_length-k, corpus_substring);
      assert(length == corpus_length-k);
      assert(suffix_array_includes(array, corpus_length-k, corpus_substring));
      frequency = suffix_array_frequency(array, corpus_length-k, corpus_substring);
      assert(frequency == 1 + k/CYCLE);
    }
#if BE_VERBOSE
  printf("%u\tverified %d-byte suffix_array_t on cyclic pattern\n", 
	 memory_total_bytes(), symbol_size);
#endif

  /* verify premature termination */
  for(t = 0; t < TERMINATE_TRIALS; t++) 
    {
      is_alive = TRUE;
      k = drand48() * (double)corpus_length;
      assert(0 <= k); assert(k < corpus_length);
      suffix_array_map(array, k, &premature_terminator, &is_alive);
    }
#if BE_VERBOSE
  printf("%u\tverified %d-byte _map() with premature termination\n", 
	 memory_total_bytes(), symbol_size);
#endif

  /* test ascii input/output */
  stream = tmpfile();
  assert(stream);
  suffix_array_fprintf(array, stream);
  rewind(stream);
  array_copy = suffix_array_fscanf(stream);
  assert(!ferror(stream));
  assert(suffix_array_are_equivalent(array, array_copy));
  fclose(stream); stream = NULL;
  corpus_copy = (vector_t *)suffix_array_corpus(array_copy);
  suffix_array_destroy(array_copy); array_copy = NULL;
  vector_destroy(corpus_copy); corpus_copy = NULL;
#if BE_VERBOSE
  printf("%u\tverified %d-byte _fprintf()/_fscanf() on cyclic pattern\n", 
	 memory_total_bytes(), symbol_size);
#endif

  /* test binary input/output */
  stream = tmpfile();
  assert(stream);
  suffix_array_write(array, stream);
  rewind(stream);
  array_copy = suffix_array_read(stream);
  assert(!ferror(stream));
  assert(suffix_array_are_equivalent(array, array_copy));
  fclose(stream); stream = NULL;
  corpus_copy = (vector_t *)suffix_array_corpus(array_copy);
  suffix_array_destroy(array_copy); array_copy = NULL;
  vector_destroy(corpus_copy); corpus_copy = NULL;
#if BE_VERBOSE
  printf("%u\tverified %d-byte _write()/_read() on cyclic pattern\n", 
	 memory_total_bytes(), symbol_size);
#endif

  /* clean up */
  memory_destroy(symbol_p); symbol_p = NULL;
  suffix_array_destroy(array); array = NULL;
  vector_destroy(corpus); corpus = NULL;
#if BE_VERBOSE
  printf("%u\tpassed all %d-byte cyclic pattern tests\n", 
	 memory_total_bytes(), symbol_size);
#else
  printf("%u\tverified suffix_array_t for %d-byte %dx%d cyclic pattern\n", 
	 memory_total_bytes(), symbol_size, (int)CYCLE_COUNT, (int)CYCLE);
#endif

  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  return(TRUE);
}

boolean_t ngram_test(int symbol_size)
{
  suffix_array_t *array;
  unsigned string_length, frequency, maximal_length;
  unsigned char byte, string[RANDOM_STRING_LENGTH];
  void *symbol_p, *corpus, *gram;
  boolean_t is_alive;
  int i, j, k, t;

  /* zero out ngrams */
  for(i = 0; i < MAX_CHAR; i++)
    {
      unigram[i] = 0;
      for(j = 0; j < MAX_CHAR; j++)
	{
	  bigram[i][j] = 0;
	  for(k = 0; k < MAX_CHAR; k++)
	    { trigram[i][j][k] = 0; }
	}
    }

  symbol_p = memory_create(symbol_size);
  assert(memory_is_valid(symbol_p));
  assert(memory_bytes(symbol_p) == symbol_size);

  corpus = vector_create_surplus(symbol_size, 0, RANDOM_STRING_LENGTH);
  assert(vector_is_valid(corpus));
  assert(vector_length(corpus) == 0);

  /* generate many random strings and insert them */
#if BE_VERBOSE
  printf("\tinserting random %d-byte string of length %d...", 
	 symbol_size, (int)RANDOM_STRING_LENGTH);
#endif
  string_length = 0;
  for(i = 0; i < RANDOM_STRING_LENGTH; i++)
    {
      assert(string_length == i);
      byte = drand48() * (double)MAX_CHAR;
      assert((byte >= 0) && (byte < MAX_CHAR));
      string[i] = byte;
      string_length++;
#if DEBUG_FREQUENCY
      printf(" %d", (int)byte);
#endif
      set_symbol(symbol_p, symbol_size, byte);
      assert(get_symbol(symbol_p, symbol_size) == byte);
      corpus = vector_append(corpus, symbol_p);
      assert(vector_length(corpus) == string_length);
      assert(get_symbol(vector_get(corpus, i), symbol_size) == byte);

      unigram[byte]++;
      if (i > 0) { bigram[string[i-1]][byte]++; }
      if (i > 1) { trigram[string[i-2]][string[i-1]][byte]++; }
    }
  assert(string_length == RANDOM_STRING_LENGTH);
  assert(vector_length(corpus) == RANDOM_STRING_LENGTH);
#if BE_VERBOSE
  printf(" done.\n");
#endif

  array = suffix_array_create(3, corpus);
  assert(array);
  assert(suffix_array_size(array) == string_length);
  assert(suffix_array_symbol_size(array) == symbol_size);
  assert(suffix_array_corpus(array) == corpus);
  assert(suffix_array_corpus_length(array) == string_length);
  assert(suffix_array_is_valid(array));
#if BE_VERBOSE
  printf("%u\tcreated %d-byte suffix_array_t for 3gram corpus\n", 
	 memory_total_bytes(), symbol_size);
#endif
#if DEBUG_FREQUENCY
  suffix_array_dump(array);
#endif

  gram = vector_create(symbol_size, 3);
  assert(vector_is_valid(gram));
  assert(vector_length(gram) == 3);

  /* manually verify that the frequencies are correct */
  assert(vector_length(gram) == 3);
  assert(RANDOM_STRING_LENGTH+1 == suffix_array_frequency(array, 0, NULL));
  for(i = 0; i < MAX_CHAR; i++)
    {
      set_symbol(symbol_p, symbol_size, i);
      vector_set(gram, 0, symbol_p);
      assert(get_symbol(vector_element_p(gram, 0), symbol_size) == i);
      frequency = suffix_array_frequency(array, 1, gram);
      assert(unigram[i] == frequency);
      for(j = 0; j < MAX_CHAR; j++)
	{
	  set_symbol(symbol_p, symbol_size, j);
	  vector_set(gram, 1, symbol_p);
	  assert(get_symbol(vector_element_p(gram, 1), symbol_size) == j);
	  frequency = suffix_array_frequency(array, 2, gram);
	  assert(bigram[i][j] == frequency);
	  for(k = 0; k < MAX_CHAR; k++)
	    {
	      set_symbol(symbol_p, symbol_size, k);
	      vector_set(gram, 2, symbol_p);
	      assert(get_symbol(vector_element_p(gram, 2), symbol_size) == k);
	      frequency = suffix_array_frequency(array, 3, gram);
	      assert(trigram[i][j][k] == frequency);
	      assert(suffix_array_frequency(array, 4, gram) == 0);
	      maximal_length = suffix_array_maximal_length(array, 3, gram);
	      switch(maximal_length)
		{
		case 3: assert(trigram[i][j][k] > 0);
		case 2: assert(bigram[j][k] > 0);
		case 1: assert(unigram[k] > 0);
		}
	      switch(maximal_length)
		{
		case 0: assert(unigram[k] == 0);
		case 1: assert(bigram[j][k] == 0);
		case 2: assert(trigram[i][j][k] == 0);
		}
	    }
	}
    }
#if BE_VERBOSE
  printf("%u\tmanually verified %d-byte suffix_array_t frequencies against ngrams\n",
	 memory_total_bytes(), symbol_size);
#endif

  /* verify premature termination */
  for(k = 0; k <= 5; k++) 
    for(t = 0; t < TERMINATE_TRIALS; t++) 
      {
	is_alive = TRUE;
	suffix_array_map(array, k, &premature_terminator, &is_alive);
      }
#if BE_VERBOSE
  printf("%u\tverified %d-byte _map() with premature termination\n", 
	 memory_total_bytes(), symbol_size);
#endif

  /* now verify frequencies using suffix_array_map() */
  suffix_array_map(array, 0, &ngram_verifier, &symbol_size);
  suffix_array_map(array, 1, &ngram_verifier, &symbol_size);
  suffix_array_map(array, 2, &ngram_verifier, &symbol_size);
  suffix_array_map(array, 3, &ngram_verifier, &symbol_size);
  suffix_array_map(array, 4, &ngram_verifier, &symbol_size);
  suffix_array_map(array, 5, &ngram_verifier, &symbol_size);
  /* confirm that ngram_verifier() zeroed out all ngrams */
  for(i = 0; i < MAX_CHAR; i++)
    {
      assert(unigram[i] == 0);
      for(j = 0; j < MAX_CHAR; j++)
	{
	  assert(bigram[i][j] == 0);
	  for(k = 0; k < MAX_CHAR; k++) 
	    { assert(trigram[i][j][k] == 0); }
	}
    }
#if BE_VERBOSE
  printf("%u\tverified %d-byte suffix_array_map() against ngrams\n", 
	 memory_total_bytes(), symbol_size);
#endif

  assert(vector_is_valid(corpus));
  assert(suffix_array_is_valid(array));
  memory_destroy(symbol_p); symbol_p = NULL;
  vector_destroy(gram); gram = NULL;
  suffix_array_destroy(array); array = NULL;
  vector_destroy(corpus); corpus = NULL;

#if BE_VERBOSE
  printf("%u\tdestroyed valid %d-byte suffix array and all scratch\n", 
	 memory_total_bytes(), symbol_size);
#else
  printf("%u\tverified suffix_array_t for random %d-byte string of length %d\n", 
	 memory_total_bytes(), symbol_size, (int)RANDOM_STRING_LENGTH);
#endif
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);

  return(TRUE);
}


int main(int argc, char **argv)
{
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("\n\tTesting suffix_array_t library...\n");

  assert(monotonous_test(1));
  assert(monotonous_test(2));
  assert(monotonous_test(4));
#if BE_VERBOSE
  printf("\n");
#endif
  assert(cyclic_test(1));
  assert(cyclic_test(2));
  assert(cyclic_test(4));
#if BE_VERBOSE
  printf("\n");
#endif
  assert(ngram_test(1));
  assert(ngram_test(2));
  assert(ngram_test(4));
#if BE_VERBOSE
  printf("\n");
#endif
  printf("\tmemory leak == %u bytes in %u blocks\n", 
	 memory_total_bytes(), memory_total_blocks());
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  printf("\tsuffix_array_t library passes all tests.\n");
  return(0);
}


