/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 *
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>  /* for memcpy() */
#include <limits.h>  /* for USHRT_MAX */
#include "common.h"
#include "suffix.internal.h"

static char suffix_internal_rcsid[] = "$Id: suffix.internal.c,v 1.9 1998/01/24 15:54:32 libpa Exp $";

#define DEBUG                    FALSE
#define DEBUG_BYTESORT           FALSE
#define INDICES_SPAN_LIMIT       USHRT_MAX

/**********************************************************************/

/* Returns: total order on suffixes, shorter < longer for inclusion */
int suffix_compare(unsigned left_length, const suffix_t *left_suffix, 
		   unsigned right_length, const suffix_t *right_suffix)
{
  unsigned common_length;
  int comparison;

  common_length = MIN(left_length, right_length);
  while(common_length > 0)
    {
      left_suffix--;
      right_suffix--;
      comparison = (int)(*left_suffix) - *right_suffix;
      if (comparison != 0) { return(comparison); }
      common_length--;
    }
  comparison = (long)left_length - right_length;
  return(comparison);
}

/* Does left_suffix include right_suffix? */
boolean_t suffix_includes(unsigned left_length, const suffix_t *left_suffix, 
			  unsigned right_length, const suffix_t *right_suffix)
{
  if (left_length < right_length) { return(FALSE); }
  assert(left_length >= right_length);
  assert(right_length > 0);
  while(right_length > 0)
    {
      left_suffix--;
      right_suffix--;
      if (*left_suffix != *right_suffix) { return(FALSE); }
      right_length--;
    }
  return(TRUE);
}

unsigned suffix_common_length(unsigned left_length, const suffix_t *left_suffix, 
			      unsigned right_length, const suffix_t *right_suffix)
{
  unsigned common_length, maximum_common_length;

  maximum_common_length = MIN(left_length, right_length);
  common_length = 0;
  while(common_length < maximum_common_length)
    {
      left_suffix--;
      right_suffix--;
      if (*left_suffix != *right_suffix) { return(common_length); }
      common_length++;
    }
  assert(common_length == maximum_common_length);
  return(maximum_common_length);
}

/**********************************************************************/

int suffix_index_compare(const void *left_index_p, const void *right_index_p, 
			 const vector_t *corpus)
{
  suffix_index_t left_index, right_index;
  suffix_t *left_suffix, *right_suffix;
  unsigned  left_length, right_length, symbol_size;
  int  comparison;

  assert(vector_is_valid(corpus));

  left_index = *(suffix_index_t *)left_index_p;
  right_index = *(suffix_index_t *)right_index_p;
  assert(suffix_index_is_valid(left_index, corpus));
  assert(suffix_index_is_valid(right_index, corpus));

  symbol_size = vector_object_size(corpus);
  left_length = (left_index+1) * symbol_size;
  left_suffix = (suffix_t *)corpus + left_length;
  assert(left_length == suffix_index2length(left_index, corpus));
  assert(left_suffix == suffix_index2suffix(left_index, corpus));

  right_length = (right_index+1) * symbol_size;
  right_suffix = (suffix_t *)corpus + right_length;
  assert(right_length == suffix_index2length(right_index, corpus));
  assert(right_suffix == suffix_index2suffix(right_index, corpus));

  comparison = suffix_compare(left_length, left_suffix, right_length, right_suffix);
  return(comparison);
}

void suffix_indices_sort(suffix_index_t *indices, const vector_t *corpus)
{
  vector_quicksort(indices, &suffix_index_compare, corpus);
}

boolean_t suffix_indices_are_sorted(suffix_index_t *indices, const vector_t *corpus)
{
  return(vector_is_sorted(indices, &suffix_index_compare, corpus));
}


/**********************************************************************/

/* Returns: literal comparison between indices as trivial unsigned integers */
int uint_compare(const void *left_p, const void *right_p, const void *parameters)
{
  unsigned left, right;
  assert(parameters == NULL);

  left = *(unsigned *)left_p;
  right = *(unsigned *)right_p;
  return((long)left - right);
}

/************************** BOUNDED SORTING **********************************/

int suffix_index_bounded_compare(const void *left_index_p, 
				 const void *right_index_p, 
				 const void *parameters)
{
  const bounded_corpus_t *enclosure;
  suffix_index_t left_index, right_index;
  suffix_t *left_suffix, *right_suffix;
  const vector_t *corpus;
  unsigned left_length, right_length, suffix_length_limit;
  unsigned depth, symbol_size;
  int comparison;

  left_index = *(suffix_index_t *)left_index_p;
  right_index = *(suffix_index_t *)right_index_p;

  enclosure = parameters;
  depth = enclosure->bound;
  corpus = enclosure->corpus;
  symbol_size = vector_object_size(corpus);
  suffix_length_limit = (depth+1) * symbol_size;

  left_suffix = suffix_index2suffix(left_index, corpus);
  if (left_index < depth) { left_length = suffix_index2length(left_index, corpus); }
  else { left_length = suffix_length_limit; }
  assert(left_length == MIN(suffix_index2length(left_index, corpus), suffix_length_limit));

  right_suffix = suffix_index2suffix(right_index, corpus);
  if (right_index < depth) { right_length = suffix_index2length(right_index, corpus); }
  else { right_length = suffix_length_limit; }
  assert(right_length == MIN(suffix_index2length(right_index, corpus), suffix_length_limit));

  comparison = suffix_compare(left_length, left_suffix, right_length, right_suffix);
  return(comparison);
}

void suffix_indices_bounded_sort(suffix_index_t *indices, 
				 const vector_t *corpus, 
				 unsigned depth)
{
  bounded_corpus_t enclosure;
  enclosure.corpus = corpus;
  enclosure.bound = depth;
  vector_quicksort(indices, &suffix_index_bounded_compare, &enclosure);
}

boolean_t suffix_indices_are_bounded_sorted(const suffix_index_t *indices, 
					    const vector_t *corpus, 
					    unsigned depth)
{
  bounded_corpus_t enclosure;
  enclosure.corpus = corpus;
  enclosure.bound = depth;
  return(vector_is_sorted(indices, &suffix_index_bounded_compare, &enclosure));
}

/************************************************************/

/* first byte is suffix_byte_index == 0 */
#define suffix_index2byte(index,byte,corpus) \
        (*(suffix_t *)(suffix_index2suffix(index,corpus)-((byte)+1)))

int suffix_index_byte_compare(const void *left_index_p, 
			      const void *right_index_p, 
			      const void *parameters)
{
  const bounded_corpus_t *enclosure;
  suffix_index_t left_index, right_index;
  unsigned left_length, right_length, byte_index;
  unsigned char left_byte, right_byte;
  const vector_t *corpus;
  int comparison;

  enclosure = parameters;
  byte_index = enclosure->bound;
  corpus = enclosure->corpus;

  left_index = *(suffix_index_t *)left_index_p;
  right_index = *(suffix_index_t *)right_index_p;
  assert(suffix_index_is_valid(left_index, corpus));
  assert(suffix_index_is_valid(right_index, corpus));

  left_length = suffix_index2length(left_index, corpus);
  right_length = suffix_index2length(right_index, corpus);
  assert(left_length > 0); assert(right_length > 0);
  if ((left_length <= byte_index) || (right_length <= byte_index))
    { 
      comparison = (long)left_length - right_length; 
      return(comparison);
    }
  left_byte = suffix_index2byte(left_index, byte_index, corpus);
  right_byte = suffix_index2byte(right_index, byte_index, corpus);
  comparison = (int)left_byte - right_byte;
  return(comparison);
}


void suffix_indices_byte_subsort(suffix_index_t *indices,
				 unsigned start, unsigned stop, unsigned byte_index,
				 const vector_t *corpus,
				 unsigned depth, unsigned span_limit)
{
  bounded_corpus_t enclosure;
  unsigned symbol_size;
  assert(vector_is_valid(indices));
  assert(vector_index_is_valid(indices, start));
  assert(vector_index_is_valid(indices, stop));
#if DEBUG_BYTESORT
  printf("[%u,%u,%u ", start, stop, byte_index);
#endif
  assert(byte_index / vector_object_size(corpus) <= depth);
  assert(start <= stop);
#ifndef NDEBUG
  /* make sure that indices[start..stop] are sorted at previous level */
  if (byte_index > 0)
    {
      unsigned i, index; unsigned char byte;
#if DEBUG_BYTESORT
      printf("^ ");
#endif
      assert(vector_index_is_valid(indices, start));
      index = indices[start];
      assert(suffix_index_is_valid(index, corpus));
      assert(suffix_index2length(index, corpus) >= byte_index);
      byte = suffix_index2byte(index, byte_index-1, corpus);
      for(i = start+1; i <= stop; i++)
	{
	  assert(vector_index_is_valid(indices, i));
	  index = indices[i];
	  assert(suffix_index_is_valid(index, corpus));
	  assert(suffix_index2length(index, corpus) >= byte_index);
	  assert(byte == suffix_index2byte(index, byte_index-1, corpus));
	}
    }
#endif

  symbol_size = vector_object_size(corpus);
  assert((byte_index / symbol_size) <= depth);
  if ((start + span_limit >= stop) && 
      ((byte_index % symbol_size) == 0))
    {
      unsigned i, index_depth;
#if DEBUG_BYTESORT
      printf("<%d", (int)INDICES_SPAN_LIMIT);
#endif
      assert(vector_index_is_valid(indices, start));
      assert(vector_index_is_valid(indices, stop));
      enclosure.bound = depth;
      enclosure.corpus = corpus;
      if (byte_index > symbol_size) 
	{ 
	  assert(byte_index > 0);
	  index_depth = (byte_index-1) / symbol_size; 
	  assert(enclosure.bound >= index_depth);
	  enclosure.bound -= index_depth;
	  for(i = start; i <= stop; i++) 
	    { assert(indices[i] >= index_depth); indices[i] -= index_depth; }
	  vector_quicksubsort(indices, start, stop, 
			      &suffix_index_bounded_compare, &enclosure);
	  for(i = start; i <= stop; i++) 
	    { indices[i] += index_depth; }
	}
      else 
	{ 
	  vector_quicksubsort(indices, start, stop, 
			      &suffix_index_bounded_compare, &enclosure); 
	}
    }
  else
    {
      unsigned middle, start_length;
      suffix_index_t start_index, middle_index, stop_index;
      unsigned char start_byte, middle_byte, stop_byte;

      assert(symbol_size == vector_object_size(corpus));
      assert(start < stop);
      enclosure.corpus = corpus;
      enclosure.bound = byte_index;
      vector_quicksubsort(indices, start, stop, &suffix_index_byte_compare, &enclosure);

      if (((byte_index+1) / symbol_size) > depth) 
	{ 
#if DEBUG_BYTESORT
	  printf(" !]");
#endif
	  return; 
	}

      start_index = indices[start];
      start_length = suffix_index2length(start_index, corpus);
      if (start_length <= byte_index)     /* only first index could be too short */
	{
	  start++; 
	  start_index = indices[start];
	}
      assert(suffix_index2length(start_index, corpus) > byte_index);
      start_byte = suffix_index2byte(start_index, byte_index, corpus);

      stop_index = indices[stop];
      assert(suffix_index2length(stop_index, corpus) > byte_index);
      stop_byte = suffix_index2byte(stop_index, byte_index, corpus);
      if (start_byte == stop_byte) /* nothing to partition at this level */
	{
	  suffix_indices_byte_subsort(indices, start, stop, byte_index+1, 
				      corpus, depth, span_limit);
	}
      else
	{
	  middle = start+1;
	  while(middle <= stop)
	    {
	      assert(start < middle); assert(middle <= stop);
	      assert(start_index == indices[start]);
	      assert(suffix_index2length(start_index, corpus) > byte_index);
	      assert(start_byte == suffix_index2byte(start_index, byte_index, corpus));

	      middle_index = indices[middle];
	      assert(suffix_index2length(middle_index, corpus) > byte_index);
	      middle_byte = suffix_index2byte(middle_index, byte_index, corpus);
	      if (middle_byte != start_byte)
		{ 
		  suffix_indices_byte_subsort(indices, start, middle-1, byte_index+1, 
					      corpus, depth, span_limit);
		  start = middle;
		  start_index = middle_index;
		  assert(suffix_index2length(start_index, corpus) > byte_index);
		  start_byte = suffix_index2byte(start_index, byte_index, corpus);
		}
	      middle++;
	    }
	  assert(start < stop);
	  suffix_indices_byte_subsort(indices, start, stop, byte_index+1, 
				      corpus, depth, span_limit);
	}
    }
#if DEBUG_BYTESORT
  printf("]");
#endif
}

/****** ABSTRACTION ******
  vector_bytesort(vector_t *vector, 
                  unsigned char *byte_p(const void *object_p, unsigned byte_index,
		                        void *parameters),
		  void *parameters);
 *************************/

void suffix_indices_byte_sort(suffix_index_t *indices, const vector_t *corpus, 
			      unsigned span_limit)
{
  unsigned corpus_length;
  corpus_length = vector_length(corpus);
  if (corpus_length > 0) 
    { suffix_indices_bounded_byte_sort(indices, corpus, corpus_length-1, span_limit); }
  assert(suffix_indices_are_sorted(indices, corpus));
}

void suffix_indices_bounded_byte_sort(suffix_index_t *indices, 
				      const vector_t *corpus,
				      unsigned depth, 
				      unsigned span_limit)
{
  unsigned suffix_count;
  assert(depth < vector_length(corpus));
  suffix_count = vector_length(indices);
  if (suffix_count > 1) 
    { suffix_indices_byte_subsort(indices, 0, suffix_count-1, 0, 
				  corpus, depth, span_limit); }
  assert(suffix_indices_are_bounded_sorted(indices, corpus, depth));
}

/************************************************************/

#if FALSE
#define KEY_COUNT  (UCHAR_MAX+1)
/* Notes: counting sort on suffix_t bytes */
void suffix_indices_byte_countsort(suffix_index_t *indices,
				   unsigned start, unsigned stop, unsigned depth,
				   const vector_t *corpus)
{
  static unsigned counter[KEY_COUNT];
  suffix_index_t *shadow_indices;                        /* this is real */
  suffix_index_t *indices, *next_indices, *swap_indices; /* these are fake */
  suffix_index_t  suffix_index;
  suffix_t *suffix, *key_p;
  unsigned suffix_length, maximum_suffix_length, depth, corpus_length, symbol_size;
  register int index_count, index_start;
  register int i, j, key;

  assert(vector_is_valid(corpus));
  assert(vector_is_valid(real_indices));
  assert(vector_length(corpus) == vector_length(real_indices));
  assert(depth_limit < vector_length(corpus));

  index_count = vector_length(real_indices);
  if (index_count < 2) { return; }

  /* Note: speedup is to quicksort all suffix indices and   */
  /*       then limit comparisons to indices of given depth */
  /*       This also keeps the keys in [0,255]              */
  if (!vector_is_sorted(real_indices, &uint_compare, NULL))
    { vector_quicksort(real_indices, &uint_compare, NULL); }
  assert(KEY_COUNT == 256);
  assert(index_count >= 2);
  assert(vector_index_is_valid(real_indices, index_count-2));
  assert(real_indices[index_count-2] < real_indices[index_count-1]);

  corpus_length = vector_length(corpus);
  symbol_size = vector_object_size(corpus);
  shadow_indices = vector_copy(real_indices);
  assert(vector_is_valid(shadow_indices));
  assert(vector_are_equivalent(real_indices, shadow_indices));

  indices = real_indices;
  next_indices = shadow_indices;
  index_start = index_count - 2;
  if (depth_limit > indices[index_start]) { depth_limit = indices[index_start]; }
  maximum_suffix_length = suffix_index2length(depth_limit, corpus);
  assert(vector_are_equivalent(indices, next_indices));
  for(suffix_length = maximum_suffix_length; suffix_length > 0; suffix_length--)
    {
      assert(vector_index_is_valid(indices, index_start));
      assert(((indices == real_indices) && (next_indices == shadow_indices)) ||
             ((indices == shadow_indices) && (next_indices == real_indices)));

      depth = (suffix_length-1) / symbol_size;
      assert(depth <= depth_limit);

      while((index_start > 0) && (indices[index_start-1] >= depth)) { index_start--; }
      assert(vector_index_is_valid(indices, index_start));
      assert(indices[index_start] >= depth);
      assert((index_start == 0) || (indices[index_start-1] < depth));
    
#if DEBUG
      printf("\n [%d,%d) : depth %u, bytes %u <= %u, ", 
             index_start, index_count, depth, suffix_length, maximum_suffix_length);
#endif
      /* NOTE:
         If the indices[] are dense, ie., of same length as corpus,
         then we can eliminate first loop over the suffix_index_t
         to calculate the counter[] array as follows. 
         Each new suffix_length introduces only one new symbol.
         So if we keep a counter[] array for every byte in the symbol,
         then we just need to increment a single entry every time,
         and then copy the counter[i][] array before decrementing it.
       */

      for(i = 0; i < KEY_COUNT; i++) { counter[i] = 0; }
      for(j = index_start; j < index_count; j++)
        {
          assert(vector_index_is_valid(indices, j));
          suffix_index = indices[j];
          assert(suffix_index >= depth);
          suffix = suffix_index2suffix(suffix_index, corpus);
          key_p = suffix - suffix_length;
          key = *key_p;
          assert(key == suffix_index2byte(suffix_index, suffix_length-1, corpus));
          assert(key >= 0); assert(key < KEY_COUNT);
          counter[key]++;
        }
      for(i = 1; i < KEY_COUNT; i++) { counter[i] += counter[i-1]; }
      for(j = index_count-1; j >= index_start; j--)
        {
          assert(j >= 0); assert(j < vector_length(indices));
          assert(vector_index_is_valid(indices, j));
          suffix_index = indices[j];
          assert(suffix_index >= depth);
          key = suffix_index2byte(suffix_index, suffix_length-1, corpus);
          assert(key >= 0); assert(key < KEY_COUNT);
          assert(counter[key] > 0);  
          assert(counter[key] <= index_count-index_start);
          assert(vector_index_is_valid(next_indices, index_start+(counter[key]-1)));
          counter[key]--;
          next_indices[index_start + counter[key]] = suffix_index;
        }

      swap_indices = indices;
      indices = next_indices;
      next_indices = swap_indices;
    }

  if (indices != real_indices)
    {
      assert(indices == shadow_indices);
      memcpy(real_indices, shadow_indices, index_count * sizeof(suffix_index_t));
      assert(vector_are_equivalent(real_indices, shadow_indices));
    }
  vector_destroy(shadow_indices); shadow_indices = NULL;
  assert(vector_is_valid(real_indices));
}

#endif
