/*
 * (c) 1996 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "memory.h"
#include "vector.h"
#include "suffix.internal.h"

static char suffix_internal_test_rcsid[] = "$Id: suffix.internal.test.c,v 1.7 1996/12/14 19:19:48 ristad Exp $";


#define DEBUG                 FALSE
#define BE_VERBOSE            FALSE
#define TEST_MONOTONOUS       TRUE
#define TEST_CYCLIC           TRUE

#define SPAN_LIMIT            4
#define CYCLE                 8       /* 8 */
#define CYCLE_COUNT           16      /* 16 */
#define CYCLIC_STRING_LENGTH  (CYCLE*CYCLE_COUNT)
#define BORING_STRING_LENGTH  512     /* 512 */
#define CYCLIC_DEPTH_LIMIT    CYCLIC_STRING_LENGTH /* 8 */

extern double drand48(void);

int suffix_corpus_compare(const void *left, const void *right, const vector_t *corpus)
{
  unsigned left_length, right_length;
  const suffix_t *left_suffix, *right_suffix;

  assert(left); assert(right);
  assert(vector_is_valid(corpus));

  left_suffix = *(suffix_t **)left;
  right_suffix = *(suffix_t **)right;
  assert((void *)left_suffix > corpus);
  assert((void *)right_suffix > corpus);
  left_length = suffix_length(left_suffix, corpus);
  right_length = suffix_length(right_suffix, corpus);
  assert(suffix_is_valid(left_suffix, corpus));
  assert(suffix_is_valid(right_suffix, corpus));
  return(suffix_compare(left_length, left_suffix, right_length, right_suffix));
}

boolean_t montonous_test(int symbol_size)
{
  vector_t         *corpus;
  suffix_t       **suffixes;
  suffix_index_t  *indices;
  int i;

  corpus = vector_create(symbol_size, BORING_STRING_LENGTH);
  assert(vector_is_valid(corpus));
  memset(corpus, 0, BORING_STRING_LENGTH*symbol_size);
#if BE_VERBOSE
  printf("%u\tcreated %d-byte monotonous %d symbol pattern\n",
	 memory_total_bytes(), symbol_size, (int)BORING_STRING_LENGTH);
#endif

  indices = vector_create(sizeof(suffix_index_t), BORING_STRING_LENGTH);
  suffixes = vector_create(sizeof(suffix_t *), BORING_STRING_LENGTH);
  assert(vector_length(indices) == BORING_STRING_LENGTH);
  assert(vector_length(suffixes) == BORING_STRING_LENGTH);
  for(i = 0; i < BORING_STRING_LENGTH; i++)
    {
      indices[i] = i;
      suffixes[i] = vector_element_p(corpus, i+1);
    }
#if BE_VERBOSE
  printf("%u\t created %d-byte monotonous suffix_t and suffix_index_t arrays\n", 
	 memory_total_bytes(), symbol_size);
#endif

  for(i = 0; i < BORING_STRING_LENGTH; i++)
    {
      assert(suffix_is_valid(suffixes[i], corpus));
      assert(suffix_length(suffixes[i], corpus) == (i+1)*symbol_size);
      assert(suffix2index(suffixes[i], corpus) == indices[i]);
      assert(suffix_corpus_compare(suffixes+i, suffixes+i, corpus) == 0);
      assert((i == 0) || 
	     suffix_includes(suffix_length(suffixes[i], corpus), suffixes[i], 
			     suffix_length(suffixes[i-1], corpus), suffixes[i-1]));
      assert((i == 0) || (suffix_corpus_compare(suffixes+i-1, suffixes+i, corpus) < 0));

      assert(suffix_index_is_valid(indices[i], corpus));
      assert(suffix_index2length(indices[i], corpus) == (i+1)*symbol_size);
      assert(suffix_index2suffix(indices[i], corpus) == suffixes[i]);
      assert(suffix_index_compare(indices+i, indices+i, corpus) == 0);
      assert((i == 0) || (suffix_index_compare(indices+i-1, indices+i, corpus) < 0));

    }
  assert(vector_is_sorted(indices, &suffix_index_compare, corpus));  
  assert(suffix_indices_are_sorted(indices, corpus));
  assert(vector_is_sorted(suffixes, &suffix_corpus_compare, corpus));
#if BE_VERBOSE
  printf("%u\t verified %d-byte monotonous suffix_t and suffix_index_t arrays\n", 
	 memory_total_bytes(), symbol_size);
#endif
     
  assert(vector_length(indices) == BORING_STRING_LENGTH);
  while(suffix_indices_are_sorted(indices, corpus)) { vector_shuffle(indices); }
  assert(!vector_is_sorted(indices, &suffix_index_compare, corpus));
  suffix_indices_sort(indices, corpus);
  assert(vector_is_sorted(indices, &suffix_index_compare, corpus));
  assert(suffix_indices_are_sorted(indices, corpus));
#if BE_VERBOSE
  printf("%u\t verified %d-byte monotonous _sort()\n", 
	 memory_total_bytes(), symbol_size);
#endif

  assert(vector_length(indices) == BORING_STRING_LENGTH);
  while(suffix_indices_are_sorted(indices, corpus)) { vector_shuffle(indices); }
  assert(!vector_is_sorted(indices, &suffix_index_compare, corpus));
  suffix_indices_byte_sort(indices, corpus, SPAN_LIMIT);
  assert(vector_is_sorted(indices, &suffix_index_compare, corpus));
  assert(suffix_indices_are_sorted(indices, corpus));
#if BE_VERBOSE
  printf("%u\t verified %d-byte monotonous _byte_sort()\n", 
	 memory_total_bytes(), symbol_size);
#endif

  vector_destroy(corpus); corpus = NULL;
  vector_destroy(suffixes); suffixes = NULL;
  vector_destroy(indices); indices = NULL;
#if BE_VERBOSE
  printf("%u\t destroyed all %d-byte monotonous arrays\n",
	 memory_total_bytes(), symbol_size);
#endif
#if !BE_VERBOSE
  printf("%u\tverified all %d-byte _compare() and _sort() on %d symbol pattern\n",
	 memory_total_bytes(), symbol_size, (int)BORING_STRING_LENGTH);
#endif

  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  return(TRUE);
}

boolean_t cyclic_test(int symbol_size)
{
  bounded_corpus_t  enclosure;
  vector_t         *corpus;
  suffix_t       **suffixes;
  suffix_index_t  *indices;
  void            *symbol_p;
  unsigned         string_length, index, depth;
  int              i, j, k;

  symbol_p = memory_create(symbol_size); 
  assert(memory_bytes(symbol_p) == symbol_size);
  corpus = vector_create(symbol_size, CYCLIC_STRING_LENGTH); 
  assert(vector_is_valid(corpus));
  corpus = vector_contract(corpus, CYCLIC_STRING_LENGTH);
  suffixes = vector_create(sizeof(suffix_t *), CYCLIC_STRING_LENGTH);
  indices = vector_create(sizeof(suffix_index_t), CYCLIC_STRING_LENGTH);
  assert(vector_is_valid(indices)); 
  assert(vector_is_valid(suffixes));
  assert(vector_length(indices) == CYCLIC_STRING_LENGTH);
  assert(vector_length(suffixes) == CYCLIC_STRING_LENGTH);
#if BE_VERBOSE
  printf("%u\tallocated %d-byte cyclic corpus plus suffix_t/suffix_index_t arrays\n",
	 memory_total_bytes(), symbol_size);
#endif

#if BE_VERBOSE
  printf("%u\t generating %d-byte cyclic corpus:", memory_total_bytes(), symbol_size);
#endif
  string_length = 0;
  assert(vector_length(corpus) == string_length);
  for(i = 0; i < CYCLE_COUNT; i++)
    { 
#if BE_VERBOSE
      printf(" %d", i);
#endif
      for(j = 0; j < CYCLE; j++)
	{
	  assert(string_length >= 0);
	  assert(string_length < CYCLIC_STRING_LENGTH);
	  assert(vector_length(corpus) == string_length);
	  index = string_length;

	  /* add symbol to corpus */
	  switch(symbol_size)
	    {
	    case 1: *((unsigned char  *)symbol_p) = j; break;
	    case 2: *((unsigned short *)symbol_p) = j; break;
	    case 4: *((unsigned int   *)symbol_p) = j; break;
	    default: abort();
	    }
	  corpus = vector_append(corpus, symbol_p);
	  string_length++;

	  /* add suffix_index to indices and suffix to suffixes */
	  indices[index] = index;
	  suffixes[index] = vector_element_p(corpus, index+1);

	  /* verify this suffix_t and suffix_index_t */
	  assert(index+1 == string_length);
	  assert(suffix_is_valid(suffixes[index], corpus));
	  assert(suffix2index(suffixes[index], corpus) == index);
	  assert(suffix_length(suffixes[index], corpus) == string_length*symbol_size);
	  assert(suffix_index_is_valid(index, corpus));
	  assert(suffix_index2length(index, corpus) == string_length*symbol_size);
	  assert(suffix_index2suffix(index, corpus) == vector_element_p(corpus, index+1));

	  /* compare this suffix_t and suffix_index_t to previous one */
	  if (index == 0) { continue; }
	  if (j > 0) 
	    {
	      assert(suffix_corpus_compare(suffixes+index-1, suffixes+index, corpus) < 0);
	      assert(suffix_index_compare(indices+index-1, indices+index, corpus) < 0);
	    } 
	  else 
	    {
	      assert(suffix_corpus_compare(suffixes+index-1, suffixes+index, corpus) > 0);
	      assert(suffix_index_compare(indices+index-1, indices+index, corpus) > 0);
	    }

	  /* compare this suffix_t and suffix_index_t to previous ones in cycle */
	  for(k = j; k < index; k += CYCLE)
	    {
	      assert(k < index); assert(i > 0); 
	      assert(((index - k)%CYCLE) == 0);
	      assert(suffix_is_valid(suffixes[k], corpus));
	      assert(suffix_includes(suffix_length(suffixes[index], corpus), 
				     suffixes[index],
				     suffix_length(suffixes[k], corpus), 
				     suffixes[k]));
	      assert(suffix_compare(suffix_length(suffixes[index], corpus), 
				    suffixes[index],
				    suffix_length(suffixes[k], corpus), 
				    suffixes[k]) > 0);
	      assert(suffix_corpus_compare(suffixes+k, suffixes+index, corpus) < 0);
	      assert(suffix_index_compare(indices+k, indices+index, corpus) < 0);
	    }
	}
    }
  assert(string_length == CYCLIC_STRING_LENGTH);
  assert(vector_length(corpus) == string_length);
  assert(vector_length(indices) == string_length);
  assert(vector_length(suffixes) == string_length);
  for(i = 0; i < string_length; i++)
    { 
      assert(indices[i] == i);
      assert(suffixes[i] == vector_element_p(corpus, i+1));
    }
#if BE_VERBOSE
  printf(".\n");
  printf("%u\t manually verified %d-byte suffix_t and suffix_index_t arrays.\n",
	 memory_total_bytes(), symbol_size);
#endif

  /* verify that comparison functions are good */
  assert(!vector_is_sorted(indices, &suffix_index_compare, corpus));
  assert(!suffix_indices_are_sorted(indices, corpus));
  suffix_indices_sort(indices, corpus);
  assert(vector_is_sorted(indices, &suffix_index_compare, corpus));
  assert(suffix_indices_are_sorted(indices, corpus));

  while(suffix_indices_are_sorted(indices, corpus)) { vector_shuffle(indices); }
  suffix_indices_byte_sort(indices, corpus, SPAN_LIMIT);
  assert(vector_is_sorted(indices, &suffix_index_compare, corpus));
  assert(suffix_indices_are_sorted(indices, corpus));

  assert(!vector_is_sorted(suffixes, &suffix_corpus_compare, corpus));
  vector_quicksort(suffixes, &suffix_corpus_compare, corpus);
  assert(vector_is_sorted(suffixes, &suffix_corpus_compare, corpus));
  assert(vector_length(indices) == string_length);
  for(i = 0; i < string_length; i++)
    {
      assert(suffix_is_valid(suffixes[i], corpus));
      assert(suffix_index_is_valid(indices[i], corpus));
      assert(suffix_length(suffixes[i],corpus) == suffix_index2length(indices[i], corpus));
      assert(suffix2index(suffixes[i], corpus) == indices[i]);
      assert(suffixes[i] == suffix_index2suffix(indices[i], corpus));
    }
#if BE_VERBOSE
  printf("%u\t verified _sort() and _byte_sort() on %d-byte suffix_t and suffix_index_t\n", 
	 memory_total_bytes(), symbol_size);
#endif

  for(depth = 0; depth < CYCLIC_DEPTH_LIMIT; depth++)
    {
      enclosure.bound = depth;
      enclosure.corpus = corpus;
      while(suffix_indices_are_bounded_sorted(indices, corpus, depth))
	{ vector_shuffle(indices); }
      assert(!vector_is_sorted(indices, &suffix_index_compare, corpus));
      assert(!vector_is_sorted(indices, &suffix_index_bounded_compare, &enclosure));
      assert(!suffix_indices_are_bounded_sorted(indices, corpus, depth));
      suffix_indices_bounded_sort(indices, corpus, depth);
      assert(vector_is_sorted(indices, &suffix_index_bounded_compare, &enclosure));
      assert(suffix_indices_are_bounded_sorted(indices, corpus, depth));

      while(suffix_indices_are_bounded_sorted(indices, corpus, depth))
	{ vector_shuffle(indices); }
      assert(!suffix_indices_are_bounded_sorted(indices, corpus, depth));
      suffix_indices_bounded_byte_sort(indices, corpus, depth, SPAN_LIMIT);
      assert(suffix_indices_are_bounded_sorted(indices, corpus, depth));
    }
#if BE_VERBOSE
  printf("%u\t verified _bounded_sort() [0,%d] on %d-byte suffix_index_t array\n", 
	 memory_total_bytes(), (int)CYCLIC_DEPTH_LIMIT, symbol_size);
#endif

  memory_destroy(symbol_p); symbol_p = NULL;
  vector_destroy(corpus); corpus = NULL;
  vector_destroy(indices); indices = NULL;
  vector_destroy(suffixes); suffixes = NULL;
#if BE_VERBOSE
  printf("%u\t destroyed %d-byte corpus and all cyclic arrays\n", 
	 memory_total_bytes(), symbol_size);
#endif
#if !BE_VERBOSE
  printf("%u\tverified all %d-byte _compare() and _sort() on %dx%d cyclic corpus\n", 
	 memory_total_bytes(), symbol_size, (int)CYCLE_COUNT, (int)CYCLE);
#endif

  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  return(TRUE);
}


int main(int argc, char **argv)
{
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  fprintf(stdout, "\n\tTesting suffix_t and suffix_index_t routines...\n");

#if TEST_MONOTONOUS
  assert(montonous_test(1));
  assert(montonous_test(2));
  assert(montonous_test(4));
#endif
#if TEST_CYCLIC
  assert(cyclic_test(1));
  assert(cyclic_test(2));
  assert(cyclic_test(4));
#endif

  fprintf(stdout, "\tmemory leak == %u bytes in %u blocks\n", 
	  memory_total_bytes(), memory_total_blocks());
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  fprintf(stdout, "\tsuffix_t library passes all tests.\n");
  return(0);
}

