/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include "memory.spartan.h"
#include "table.h"

static char table_rcsid[] = "$Id: table.c,v 1.10 1998/02/07 03:45:48 libpa Exp $";

#define ANNOUNCE_RCSID  FALSE
#define keycmp(x,y,s)   (memcmp(x,y,s))


table_t *table_create(unsigned key_size, unsigned datum_size)
{
  table_t *table;
#if ANNOUNCE_RCSID
  static boolean_t is_initialized = FALSE;
  if (!is_initialized)
    {
#ifndef NDEBUG
      fprintf(stderr, "%s\n", table_rcsid);
#else
      fprintf(stderr, "%s\t>unsafe<\n", table_rcsid);
#endif
      is_initialized = TRUE;
    }
  assert(is_initialized);
#endif

  assert(key_size > 0); assert(key_size <= TABLE_MAX_KEYSIZE);
  assert(datum_size > 0); assert(datum_size <= TABLE_MAX_DATUMSIZE);

  table = memory_spartan_create(sizeof(table_t));
  assert(table != NULL);
  table_initialize(table, key_size, datum_size);
  return(table);
}

/* Use this to initialize a vector_t of table_t */
void table_initialize(table_t *table, unsigned key_size, unsigned datum_size)
{
  assert(table != NULL);
  assert(key_size > 0);
  assert(datum_size > 0);
  assert(key_size <= TABLE_MAX_KEYSIZE);
  assert(datum_size <= TABLE_MAX_DATUMSIZE);

  table->key_size = key_size;
  table->datum_size = datum_size;
  table->size = 0;
  table->workspace = UINT_MAX;
  table->keys = NULL;
  table->datums = NULL;

  assert(table_is_valid(table));
}

table_t *table_copy(const table_t *original)
{
  table_t *copy;
  unsigned key_size, datum_size, size;

  assert(original != NULL);
  assert(table_is_valid(original));

  size = table_size(original);
  key_size = table_key_size(original);
  datum_size = table_datum_size(original);
  
  copy = memory_spartan_create(sizeof(table_t));
  copy->key_size = key_size;
  copy->datum_size = datum_size;
  copy->size = size;
  copy->workspace = original->workspace;
  copy->keys = memory_spartan_copy(original->keys, size * key_size);
  copy->datums = memory_spartan_copy(original->datums, size  * datum_size);

  assert(copy != NULL);
  assert(table_is_valid(copy));
  assert(table_are_equal(copy, original));
  return(copy);
}

void table_destroy(table_t *table)
{
  assert(table_is_valid(table));
  table_finalize(table);
  assert(table->keys == NULL);
  assert(table->datums == NULL);
  memory_spartan_destroy(table, sizeof(table_t));
}

void table_finalize(table_t *table)
{
  assert(table_is_valid(table));
  if (table->size != 0) 
    {
      unsigned key_size, datum_size, size;
      size = table_size(table);
      key_size = table_key_size(table);
      datum_size = table_datum_size(table);
      memory_spartan_destroy(table->keys, size * key_size);
      memory_spartan_destroy(table->datums, size * datum_size);
      table->keys = NULL;
      table->datums = NULL;
    }
  assert(table->keys == NULL);
  assert(table->datums == NULL);
  table->size = 0;
  table->key_size = 0;
  table->datum_size = 0;
  table->workspace = UINT_MAX;
}

boolean_t table_is_valid(const table_t *table)
{
  unsigned size, key_size, datum_size, i;
  unsigned char *key_p;

  if (table == NULL) { return(FALSE); }

  size = table_size(table);
  key_size = table_key_size(table);
  datum_size = table_datum_size(table);

  assert(!memory_spartan_address_is_valid(table->keys, size * key_size, 
					  (table->keys + (size * key_size))));
  assert(!memory_spartan_address_is_valid(table->datums, size * datum_size, 
					  (table->datums + (size * datum_size))));
  
  if ((size > 0) ? 
      ((table->keys == NULL) || (table->datums == NULL)) :
      ((table->keys != NULL) || (table->datums != NULL)))
    { return(FALSE); }

  key_p = table->keys;
  for(i = 1; i < size; i++)
    { 
      assert(memory_spartan_address_is_valid(table->keys, size*key_size, key_p));
      assert(memory_spartan_address_is_valid(table->keys, size*key_size, key_p+key_size));
      if (keycmp(key_p, key_p+key_size, key_size) >= 0) 
	{ return(FALSE); } 
    }
  return(TRUE);
}

boolean_t table_are_equal(const table_t *left, const table_t *right)
{
  unsigned size, key_size, datum_size;

  assert(left != NULL); assert(right != NULL);

  size = table_size(left);
  key_size = table_key_size(left);
  datum_size = table_datum_size(left);

  if ((size != table_size(right))
      || (key_size != table_key_size(right))
      || (datum_size != table_datum_size(right)))
    { return(FALSE); }

  return((memcmp(left->keys, right->keys, size*key_size) == 0)
	 && (memcmp(left->datums, right->datums, size*datum_size) == 0));
}

boolean_t table_is_empty(const table_t *table)
{
  assert(table != NULL);
  assert((table->size == 0) || ((table->keys != NULL) && (table->datums != NULL)));
  assert((table->size > 0) || ((table->keys == NULL) && (table->datums == NULL)));
  return(table->size == 0);
}

boolean_t table_includes(const table_t *table, const void *key_p)
{
  return(table_lookup(table, key_p) != NULL);
}

void *table_lookup(const table_t *table, const void *key_p)
{
  unsigned char *key_loc, *datum_p;
  unsigned  size, key_size, datum_size;
  unsigned  lower, middle, upper;
  int       compare;

  assert(table != NULL); assert(key_p != NULL);
  size = table_size(table);
  key_size = table_key_size(table);

  lower = 0;
  upper = size;
  while(lower < upper)
    {
      middle = lower + ((upper - lower) / 2);
      assert(lower <= middle);
      assert(middle < upper);

      key_loc = table->keys + (middle * key_size);
      assert(memory_spartan_address_is_valid(table->keys, size*key_size, key_loc));
      assert(memory_spartan_address_is_valid(table->keys, size*key_size, key_loc+key_size-1));

      compare = keycmp(key_p, key_loc, key_size);
      if (compare < 0) { upper = middle; }
      else if (compare > 0) { lower = middle + 1; }
      else    /* found key location */
	{
	  datum_size = table_datum_size(table);
	  datum_p = table->datums + (middle * datum_size);
	  return(datum_p);
	}
    }
  return(NULL); /* key not found */
}

void *table_insert(table_t *table, const void *key_p, const void *datum_p)
{
  unsigned char *datum_loc, *key_loc;
  unsigned  key_size, datum_size, size;
  unsigned  lower, middle, upper;
  int       compare;

  assert(table != NULL);
  assert(key_p != NULL); assert(datum_p != NULL);

  size = table_size(table);
  key_size = table_key_size(table);
  datum_size = table_datum_size(table);

  if (table->size == 0)
    {
      assert(table->keys == NULL);
      assert(table->datums == NULL);

      table->size = 1;
      table->keys = memory_spartan_create(key_size);
      table->datums = memory_spartan_create(datum_size);
      assert(table->keys != NULL);
      assert(table->datums != NULL);

      key_loc = table->keys;
      datum_loc = table->datums;
      memcpy(key_loc, key_p, key_size);
      memcpy(datum_loc, datum_p, datum_size);
      return(table->datums);
    }
  assert(table->keys != NULL);
  assert(table->datums != NULL);
  assert(size > 0);

  lower = 0;
  upper = size;
  compare = 0;
  assert(lower < upper);
  do
    {
      middle = lower + ((upper - lower) / 2);
      assert(lower <= middle);
      assert(middle < upper);

      key_loc = table->keys + (middle * key_size);
      assert(memory_spartan_address_is_valid(table->keys, size*key_size, key_loc));
      assert(memory_spartan_address_is_valid(table->keys, size*key_size, key_loc+key_size-1));

      compare = keycmp(key_p, key_loc, key_size);
      if (compare < 0) { upper = middle; }
      else if (compare > 0) { lower = middle + 1; }
      else        /* found key location */
	{
	  datum_loc = table->datums + (middle * datum_size);
	  assert(memory_spartan_address_is_valid(table->datums, size*datum_size, datum_loc));
	  assert(memory_spartan_address_is_valid(table->datums, size*datum_size, 
						 datum_loc+datum_size-1));
	  memcpy(datum_loc, datum_p, datum_size);
	  return(datum_loc);
	}
    } while(lower < upper);
  assert(lower >= upper);
  assert(compare != 0);

  /* can't insert new key into full table */
  if (size == TABLE_MAX_SIZE) { return(NULL); }
  assert(size > 0);

  /* must insert new key into table */
  table->keys = memory_spartan_recreate(table->keys, size*key_size, (size+1)*key_size);
  table->datums = memory_spartan_recreate(table->datums, size*datum_size, 
					  (size+1)*datum_size);
  assert(table->keys != NULL);
  assert(table->datums != NULL);

  if (compare > 0) { middle++; }

  key_loc = table->keys + (middle * key_size);
  datum_loc = table->datums + (middle * datum_size);
  assert(memory_spartan_address_is_valid(table->keys, (size+1)*key_size, key_loc));
  assert(memory_spartan_address_is_valid(table->datums, (size+1)*datum_size, datum_loc));
  assert(memory_spartan_address_is_valid(table->keys, (size+1)*key_size, key_loc+key_size-1));
  assert(memory_spartan_address_is_valid(table->datums, (size+1)*datum_size, 
					 datum_loc+datum_size-1));

  /* overlapping shift up */
  memory_move(key_loc+key_size, key_loc, (size-middle)*key_size);
  memory_move(datum_loc+datum_size, datum_loc, (size-middle)*datum_size);
  memcpy(key_loc, key_p, key_size); 
  memcpy(datum_loc, datum_p, datum_size); 
  table->size++;

  assert(table_size(table) == size+1);
  assert(memcmp(key_p, key_loc, key_size) == 0);
  assert((key_loc == table->keys + (size * key_size))
	 || (memcmp(key_p, key_loc+key_size, key_size) < 0));
  assert((key_loc == table->keys) 
	 || (memcmp(key_p, key_loc-key_size, key_size) > 0));
  assert(table_lookup(table, key_p) != NULL);
  assert(memcmp(table_lookup(table, key_p), datum_p, datum_size) == 0);

  return(datum_loc);
}

void table_remove(table_t *table, const void *key_p)
{
  unsigned  key_size, datum_size, size, location;
  unsigned char *datum_loc, *key_loc, *last_key_loc, *last_datum_loc;

  assert(table != NULL); assert(key_p != NULL);
  datum_loc = table_lookup(table, key_p);
  if (datum_loc == NULL) { return; }

  size = table_size(table);
  if (size == 1)
    {
      assert(keycmp(table->keys, key_p, table_key_size(table)) == 0); 
      table_clear(table);
      assert(table_size(table) == 0);
      assert(table->keys == NULL);
      assert(table->datums == NULL);
      return;
    }
  assert(size > 1);

  key_size = table_key_size(table);
  datum_size = table_datum_size(table);
  location = (datum_loc - table->datums) / datum_size;
  key_loc = table->keys + (location * key_size);
  last_key_loc = table->keys + (key_size * (size - 1));
  last_datum_loc = table->datums + (datum_size * (size - 1));

  assert(last_key_loc >= key_loc);
  assert(last_datum_loc >= datum_loc);
  assert(((last_key_loc - key_loc) % key_size) == 0);
  assert(((last_datum_loc - datum_loc) % datum_size) == 0);
  assert(((key_loc - table->keys) % key_size) == 0);
  assert(((datum_loc - table->datums) % datum_size) == 0);
  assert(memcmp(key_p, key_loc, key_size) == 0);
  assert(datum_loc == table->datums + (location * datum_size));
  assert(memory_spartan_address_is_valid(table->keys, size*key_size, key_loc));
  assert(memory_spartan_address_is_valid(table->datums, size*datum_size, datum_loc));
  assert(memory_spartan_address_is_valid(table->keys, size*key_size, 
					 last_key_loc+key_size-1));
  assert(memory_spartan_address_is_valid(table->datums, size*datum_size, 
					 last_datum_loc+datum_size-1));

  /* massive overlapping shift down */
  memory_move(key_loc, key_loc+key_size, last_key_loc - key_loc);
  memory_move(datum_loc, datum_loc+datum_size, last_datum_loc - datum_loc);
  table->keys = memory_spartan_recreate(table->keys, size * key_size, 
					(size-1) * key_size);
  table->datums = memory_spartan_recreate(table->datums, size * datum_size, 
					  (size-1) * datum_size);
  table->size--;

  assert(table->keys != NULL);
  assert(table->datums != NULL);
  assert(table_size(table) > 0);
  assert(table_size(table) == size - 1);
  assert(table_lookup(table, key_p) == NULL);
}


void table_clear(table_t *table)
{
  unsigned key_size, datum_size, size;

  assert(table != NULL);
  assert(table_key_size(table) > 0);  
  assert(table_key_size(table) <= TABLE_MAX_KEYSIZE);
  assert(table_datum_size(table) > 0);
  assert(table_datum_size(table) <= TABLE_MAX_DATUMSIZE);
  assert(table_size(table) >= 0);

  size = table_size(table);
  if (size == 0) 
    {
      assert(table->keys == NULL);
      assert(table->datums == NULL);
      return;
    }
  key_size = table_key_size(table);
  datum_size = table_datum_size(table);

  memory_spartan_destroy(table->keys, size * key_size);
  memory_spartan_destroy(table->datums, size * datum_size);
  table->keys = NULL;
  table->datums = NULL;
  table->size = 0;
}

void table_resize_datum(table_t *table, unsigned new_datum_size,
			void (*coerce)(const void *old_datum_p, void *new_datum_p, 
				       void *parameters),
			void *parameters)
{
  unsigned  size, old_datum_size;
  unsigned  new, old, i;
  unsigned char *new_datums, *old_datums;

  assert(table != NULL);
  assert(new_datum_size > 0);
  assert(new_datum_size <= TABLE_MAX_DATUMSIZE);

  old_datum_size = table_datum_size(table);
  table->datum_size = new_datum_size;

  size = table_size(table);
  if (size == 0) 
    { 
      assert(table->keys == NULL);
      assert(table->datums == NULL);
      return; 
    }

  /* desparate attempt to preserve locality betwen keys,datums */
  old_datums = memory_spartan_copy(table->datums, size * old_datum_size);
  new_datums = memory_spartan_recreate(table->datums, 
				       size * old_datum_size, 
				       size * new_datum_size);
  table->datums = new_datums;
  assert(old_datums != NULL);
  assert(new_datums != NULL);

  old = 0;
  new = 0;
  for(i = 0; i < size; i++)
    {
      coerce(old_datums+old, new_datums+new, parameters);
      old += old_datum_size;
      new += new_datum_size;
    }
  memory_spartan_destroy(old_datums, size * old_datum_size);  /* this was just a copy */
}


void table_map(table_t *table, 
	       boolean_t (*proc)(const void *key_p, void *datum_p, void *parameters),
	       void *parameters)
{
  unsigned  i, size;
  unsigned  key_size, datum_size;
  unsigned char *key_p, *datum_p;
  boolean_t is_alive;

  assert(table != NULL); assert(proc != NULL);
  size = table_size(table);
  if (size == 0) 
    {
      assert(table->keys == NULL);
      assert(table->datums == NULL);
      return;
    }

  key_size = table_key_size(table);
  datum_size = table_datum_size(table);

  key_p = table->keys;
  datum_p = table->datums;
  for(i = 0; i < size; i++)
    {
      assert(memory_spartan_address_is_valid(table->keys, size*key_size, key_p+key_size-1));
      assert(memory_spartan_address_is_valid(table->datums, size*datum_size, 
					     datum_p+datum_size-1));
      is_alive = proc(key_p, datum_p, parameters);
      key_p += key_size;
      datum_p += datum_size;
      if (!is_alive) { return; }
    }
}

