/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 *
 * $Id: table.h,v 1.7 1998/02/07 03:45:48 libpa Exp $
 *
 */

#ifndef TABLE_INCLUDED
#define TABLE_INCLUDED

#include "common.h"

/* Description.
 *  Compact table for storing symbolic <key,datum> pairs.
 *  Lookup is O(log n) worst case. Insertion is O(n) worst case.
 *  Use this table_t if you are short on memory and lookups dominate inserts.
 *  Warning: keys must be ordered by memcmp(), ie., not contain pointers.
 *
 */

#define TABLE_MAX_KEYSIZE    UCHAR_MAX
#define TABLE_MAX_DATUMSIZE  UCHAR_MAX
#define TABLE_MAX_SIZE       USHRT_MAX

typedef struct table_tag
{
  unsigned char   key_size;
  unsigned char   datum_size;
  unsigned short  size;
  unsigned int    workspace;     /* four bytes for caller */
  unsigned char  *keys;
  unsigned char  *datums;
} table_t;

/************************************************************
 *  CONSTRUCTORS/DESTRUCTOR
 ************************************************************/

table_t   *table_create(unsigned key_size, unsigned datum_size);
           /* Requires: key_size >= 1, datum_size >= 1                     */
           /* Returns: table for <key,datum> pairs of specified sizes      */
           /* Warning: keys must contain comparable data, ie., no pointers */
table_t   *table_copy(const table_t *table);
           /* Returns: exact copy of given table                           */
void       table_destroy(table_t *table);
           /* Effect: deallocates all memory used by given table           */

void       table_initialize(table_t *table, unsigned key_size, unsigned datum_size);
           /* Arguments: pointer to caller-allocated table_t               */
           /* Requires: key_size >= 1, datum_size >= 1                     */
           /* Effect: makes table into a valid table_t of specified size   */
           /* Note: use this to create a space-efficient array of table_t  */
void       table_finalize(table_t *table);
           /* Note: pre-destructor for caller-allocated table_t            */

/************************************************************
 *  SELECTORS
 ************************************************************/

#define   table_size(table_p)           ((unsigned)((table_p)->size))
          /* Returns: number of entries in table                        */
#define   table_key_size(table_p)       ((unsigned)((table_p)->key_size))
          /* Returns: number of bytes necessary to store each key       */
#define   table_datum_size(table_p)     ((unsigned)((table_p)->datum_size))
          /* Returns: number of bytes necessary to store each datum     */
#define   table_workspace_p(table_p)    (&((table_p)->workspace))
          /* Note: these four bytes, which would be otherwise wasted    */
          /*       by aligment on the two pointers, are for the caller. */

/************************************************************
 *  PREDICATES
 ************************************************************/

boolean_t  table_is_valid(const table_t *table);
           /* Returns: TRUE iff table is a valid table          */
boolean_t  table_are_equal(const table_t *left, const table_t *right);
           /* Returns: TRUE iff both tables are equivalent      */
boolean_t  table_is_empty(const table_t *table);
           /* Returns: TRUE iff table contains no entries       */
boolean_t  table_includes(const table_t *table, const void *key_p);
           /* Returns: TRUE iff table contains pair with *key_p */

/************************************************************
 *  BASIC OPERATIONS
 ************************************************************/

void      *table_lookup(const table_t *table, const void *key_p);
           /* Returns: datum_p associated with *key_p or NULL   */
void      *table_insert(table_t *table, const void *key_p, const void *datum_p);
           /* Returns: datum_p inside table after insertion     */
void       table_remove(table_t *table, const void *key_p);
           /* Effect: removes any <*key_p,datum> pairs in table */

void       table_clear(table_t *table);
           /* Effect: removes all entries from table                   */
void       table_resize_datum(table_t *table, unsigned new_datum_size,
			      void (*coerce)(const void *old_datum_p, 
					     void *new_datum_p,
					     void *parameters),
			      void *parameters);
           /* Arguments: coerce() mutates old *datum_p to new *datum_p */
           /* Effect: mutates table so that datum have new size        */
           /* Note: coercion occurs even if old and new have same size */

/************************************************************
 *  HIGHER ORDER FUNCTION
 ************************************************************/

void       table_map(table_t *table, 
		     boolean_t (*proc)(const void *key_p, void *datum_p, void *parameters),
		     void *parameters);
           /* Effect: calls proc() on every entry in table while TRUE  */


#endif




