/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include "memory.h"
#include "table.h"

static char table_test_rcsid[] = "$Id: table.test.c,v 1.9 1998/02/07 03:45:48 libpa Exp $";

#define DEBUG                 FALSE
#define BE_VERBOSE            FALSE

typedef unsigned char    tkey_t;
typedef unsigned short   tdatum_t;
typedef int              tbig_datum_t;
#define MAX_KEY               UCHAR_MAX
#define KEY_SIZE              (sizeof(tkey_t))
#define DATUM_SIZE            (sizeof(tdatum_t))
#define BIG_DATUM_SIZE        (sizeof(tbig_datum_t))

#define INSERT_TEST_SIZE      USHRT_MAX
#define REMOVE_TEST_SIZE      (MAX_KEY/2)
#define SEED                  8907489

extern double drand48(void);
extern void   srand48(long);

/* Note:
     remove things that are not in the table for various table sizes
 */


void little2big(const void *source_p, void *sink_p, void *parameters)
{
  tdatum_t     *old_p;
  tbig_datum_t *new_p;

  old_p = (tdatum_t *)source_p;
  new_p = (tbig_datum_t *)sink_p;

  *new_p = (tbig_datum_t)*old_p;
}
void big2little(const void *source_p, void *sink_p, void *parameters)
{
  tbig_datum_t *old_p;
  tdatum_t     *new_p;

  old_p = (tbig_datum_t *)source_p;
  new_p = (tdatum_t *)sink_p;

  *new_p = (tdatum_t)*old_p;
}

boolean_t negate_vector(const void *key_p, void *datum_p, void *parameters)
{
  int      *vector;
  tkey_t    key;
  tdatum_t  datum;

  key = *(tkey_t *)key_p;
  datum = *(tdatum_t *)datum_p;
  vector = (int *)parameters;

  assert(datum > 0);
  assert(datum == (tdatum_t)vector[key]);

  vector[key] *= -1;
  return(TRUE);
}

int main(void)
{
  int          vector[MAX_KEY];
  table_t     *table, *copy;
  tkey_t       key, *key_p;
  tdatum_t     datum, *datum_p;
  tbig_datum_t big_datum, *big_datum_p;
  unsigned     i, size;

  printf("\n\tTesting table_t library.\n");

  /* initialize test state variables */
  srand48(SEED);
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);
  for(i = 0; i < MAX_KEY; i++) { vector[i] = 0; }

  /* create table */
  table = table_create(KEY_SIZE, DATUM_SIZE);
  assert(table != NULL);
  assert(table_size(table) == 0);
  assert(table_key_size(table) == KEY_SIZE);
  assert(table_datum_size(table) == DATUM_SIZE);
  assert(!table_is_valid(NULL));
  assert(table_is_valid(table));
  printf("%u\tcreated empty table(%u,%u)\n", 
	  memory_total_bytes(), table_key_size(table), table_datum_size(table));

  /* resize empty table */
  assert(BIG_DATUM_SIZE > DATUM_SIZE);
  table_resize_datum(table, BIG_DATUM_SIZE, &little2big, NULL);
  assert(table != NULL);
  assert(table_size(table) == 0);
  assert(table_key_size(table) == KEY_SIZE);
  assert(table_datum_size(table) == BIG_DATUM_SIZE);
  table_resize_datum(table, DATUM_SIZE, &big2little, NULL);
  assert(table != NULL);
  assert(table_size(table) == 0);
  assert(table_key_size(table) == KEY_SIZE);
  assert(table_datum_size(table) == DATUM_SIZE);
  printf("%u\tcorrectly resized empty table up and down again\n", memory_total_bytes());

  /* remove some random keys from the empty table */
  for(i = 0 ; i < REMOVE_TEST_SIZE; i++)
    {
      key = (tkey_t)(drand48() * (double)MAX_KEY);
      assert(key >= 0);
      assert(key < MAX_KEY); 

      key_p = &key;
      assert(!table_includes(table, key_p));
      table_remove(table, key_p);
    }
  assert(table_is_valid(table));
  printf("%u\tcorrectly failed to remove any keys from empty table\n", memory_total_bytes());

  /* insert keys into empty table */
  size = 0;
#if BE_VERBOSE
  printf("\t");
#endif
  for(i = 0 ; i < INSERT_TEST_SIZE; i++)
    {
      key = (tkey_t)(drand48() * (double)MAX_KEY);
      assert(key >= 0);
      assert(key < MAX_KEY); 
      
      key_p = &key;
      datum_p = table_lookup(table, key_p);
      if (datum_p == NULL)
	{
#if BE_VERBOSE
	  printf("%u ", i);
#endif
	  assert(vector[key] == 0);
	  datum = 0;
	  size++;
	}
      else
	{
	  datum = *datum_p;
	  assert(datum > 0);
	  assert(datum == (tdatum_t)vector[key]);
	}

      vector[key]++;

      datum++;
      datum_p = table_insert(table, key_p, &datum);

      assert(datum == (tdatum_t)*datum_p);
      assert(datum_p == table_lookup(table, key_p));
      assert(table_size(table) == size);
    }
#if BE_VERBOSE
  printf("\n");
#endif
  assert(table_is_valid(table));
  assert(!table_is_empty(table));
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      datum_p = table_lookup(table, key_p);
      datum = ((datum_p == NULL) ? 0 : *datum_p);
      assert(datum == (tdatum_t)vector[key]);
    }
  printf("%u\tfilled table with %u entries\n", memory_total_bytes(), table_size(table));

  /* remove a few keys from the table (if they're in it) */
  for(i = 0 ; i < REMOVE_TEST_SIZE; i++)
    {
      key = (tkey_t)(drand48() * (double)MAX_KEY);
      assert(key >= 0);
      assert(key < MAX_KEY); 
      
      key_p = &key;
      if (table_includes(table, key_p)) { size--; }
      else { assert(vector[key] == 0); }
      table_remove(table, key_p);
      assert(table_size(table) == size);
      assert(table_lookup(table, key_p) == NULL);

      vector[key] = 0;
    }
  assert(table_is_valid(table));
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      datum_p = table_lookup(table, key_p);
      datum = ((datum_p == NULL) ? 0 : *datum_p);
      assert(datum == (tdatum_t)vector[key]);
    }
  printf("%u\tshrunk table to %u entries\n", 
	  memory_total_bytes(), table_size(table));

  /* verify table_copy() */
  copy = table_copy(table);
  assert(copy);
  assert(table_are_equal(table, copy));
  printf("%u\tverified table_copy()\n", memory_total_bytes());

  /* verify table_resize_datum() */
  assert(2*DATUM_SIZE == BIG_DATUM_SIZE);
  table_resize_datum(table, BIG_DATUM_SIZE, &little2big, NULL);
  assert(table_is_valid(table));
  assert(!table_are_equal(table, copy));
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      big_datum_p = table_lookup(table, key_p);
      big_datum = ((big_datum_p == NULL) ? 0 : *big_datum_p);
      assert(big_datum == (tbig_datum_t)vector[key]);
    }
  printf("%u\tresized table from %d bytes up to %d bytes\n", 
	  memory_total_bytes(), (int)DATUM_SIZE, (int)BIG_DATUM_SIZE);

  table_resize_datum(table, DATUM_SIZE, &big2little, NULL);
  assert(table_is_valid(table));
  assert(table_are_equal(table, copy));
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      datum_p = table_lookup(table, key_p);
      datum = ((datum_p == NULL) ? 0 : *datum_p);
      assert(datum == (tdatum_t)vector[key]);
    }
  printf("%u\tresized table from %d bytes back down to %d bytes\n", 
	  memory_total_bytes(), (int)BIG_DATUM_SIZE, (int)DATUM_SIZE);
  
  assert(table_are_equal(copy, table));
  table_destroy(copy);
  copy = NULL;
  printf("%u\tdestroyed table copy\n", memory_total_bytes());

  /* verify table_map() */
  table_map(table, &negate_vector, (void *)vector);
  for(key = 0; key < MAX_KEY; key++)
    {
      key_p = &key;
      datum_p = table_lookup(table, key_p);
      datum = ((datum_p == NULL) ? 0 : *datum_p);
      assert(vector[key] <= 0);
      assert(datum == (tdatum_t)(-vector[key]));
    }
  printf("%u\tverified table_map()\n", memory_total_bytes());

  /* verify table_clear() */
  table_clear(table);
  assert(table_size(table) == 0);
  assert(table_is_empty(table));
  for(key = 0; key < MAX_KEY; key++) 
    { assert(table_lookup(table, &key) == NULL); }
  printf("%u\tverified table_clear()\n", memory_total_bytes());

  table_destroy(table);
  table = NULL;

  printf("\tsizeof(table_t) == %d bytes (no overhead)\n", (int)sizeof(table_t));
  printf("\tmemory leak == %u bytes in %u blocks\n",
	  memory_total_bytes(), memory_total_blocks());
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  printf("\ttable_t library passes all tests.\n");
  return(0);
}



