/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "memory.h"
#include "vector.h"
#include "trie.h"

static char trie_rcsid[] = "$Id: trie.c,v 1.11 1998/02/04 06:38:39 libpa Exp $";

#define DEBUG                   FALSE
#define ANNOUNCE_RCSID          FALSE

/************************************************************
 NOTES
 . convert to patricia trie: compare depth to index length
 ************************************************************/

typedef unsigned  index_t;

/************************************************************
 * TRIE_NODE_T OPERATIONS
 ************************************************************/

#define trie_node_is_leaf(node_p)              (table_datum_size(node_p) == sizeof(index_t))
#define trie_node_is_interior(node_p)          (table_datum_size(node_p) == sizeof(table_t))
#define trie_node_stub(node_p)                 ((node_p)->workspace)
#define trie_node_set_stub(node_p,index)       ((node_p)->workspace = (index))
#define trie_node_degree(node_p)               (table_size(node_p))
#define trie_node_lookup(node_p,symbol_p)      (table_lookup((node_p),(symbol_p)))
#define trie_node_insert(node_p,symbol_p,child_p)  (table_insert(node_p,symbol_p,child_p))
#define trie_leaf_insert(leaf_p,symbol_p,index_p)  (table_insert(leaf_p,symbol_p,index_p))
#define trie_node_map(node_p,proc,parameters)  (table_map((node_p),(proc),(parameters)))
#define trie_leaf_map(leaf_p,proc,parameters)  (table_map((leaf_p),(proc),(parameters)))

/* all nodes begin as leaf nodes */
static void trie_initialize_leaf(const trie_t *trie, trie_node_t *leaf)
{
  table_initialize(leaf, trie_symbol_size(trie), sizeof(index_t));
  trie_node_set_stub(leaf, TRIE_NOT_FOUND);
  assert(trie_node_is_leaf(leaf));
}

static void index2node(const void *index_p, void *node_p, void *parameters)
{
  index_t      index;
  trie_node_t *node;
  trie_t      *trie;

  assert(index_p); assert(node_p); assert(parameters);

  index = *(index_t *)index_p;
  node = node_p;
  trie = parameters;

  assert(index < trie_size(trie));
  trie_initialize_leaf(trie, node);
  assert(trie_node_is_leaf(node));
  trie_node_set_stub(node, index);
}

static void trie_leaf2interior(trie_t *trie, trie_node_t *node)
{
  assert(trie_node_is_leaf(node));
  table_resize_datum(node, sizeof(table_t), &index2node, trie);
  assert(trie_node_is_interior(node));
}

static boolean_t trie_node_is_valid(const trie_t *trie, const trie_node_t *node)
{
  assert(trie);
  return((node != NULL) &&
	 (table_key_size(node) == trie_symbol_size(trie)) &&
	 (trie_node_is_leaf(node) || trie_node_is_interior(node)) &&
	 ((trie_node_stub(node) == TRIE_NOT_FOUND) ||
	  (trie_node_stub(node) < trie_size(trie))) &&
	 (table_is_valid(node))
	 );
}


/************************************************************
 *  CONSTRUCTOR/DESTRUCTOR
 ************************************************************/

trie_t *trie_create(unsigned symbol_size)
{
  trie_t *trie;
#if ANNOUNCE_RCSID
  static boolean_t is_initialized = FALSE;
  if (!is_initialized)
    {
#ifndef NDEBUG
      fprintf(stderr, "%s\n", trie_rcsid);
#else
      fprintf(stderr, "%s\t>unsafe<\n", trie_rcsid);
#endif
      is_initialized = TRUE;
    }
  assert(is_initialized);
#endif
  assert(sizeof(trie_node_t) != sizeof(index_t));

  trie = memory_create(sizeof(trie_t));
  assert(trie);
  trie->symbol_size = symbol_size;
  trie->depth = 0;
  trie->entry_count = 0;
  trie_initialize_leaf(trie, &(trie->root));
  assert(trie_node_is_leaf(&(trie->root)));
  assert(trie_node_stub(&(trie->root)) == TRIE_NOT_FOUND);
  assert(trie_node_degree(&(trie->root)) == 0);

  trie->strings = vector_create(symbol_size, 0);
  assert(vector_is_valid(trie->strings));
  trie->index2offset = vector_create(sizeof(unsigned), 1);
  assert(vector_is_valid(trie->index2offset));
  trie->index2offset[0] = 0;

  assert(trie_is_valid(trie));
  return(trie);
}


static boolean_t trie_entry_copier(unsigned string_length, const void *string, 
				   unsigned index, void *parameters)
{
  trie_t *trie = parameters;
  unsigned copy_index;
  assert(index == trie_size(trie));
  copy_index = trie_insert(trie, string_length, string);
  assert(copy_index == index);
  assert(index+1 == trie_size(trie));
  return(TRUE);
}

trie_t *trie_copy(const trie_t *original)
{
  trie_t *copy;
  copy = trie_create(trie_symbol_size(original));
  trie_map(original, &trie_entry_copier, copy);
  assert(trie_is_valid(copy));
  assert(trie_are_equal(copy, original));
  return(copy);
}


static boolean_t trie_node_destroyer(const void *key_p, void *datum_p, void *parameters)
{
  trie_node_t *node;

  node = datum_p;
  if (trie_node_is_interior(node))
    { trie_node_map(node, &trie_node_destroyer, parameters); }
  table_finalize(node);
  return(TRUE);
}

void trie_destroy(trie_t *trie)
{
  trie_node_t *root_p;

  assert(trie_is_valid(trie));
  root_p = &(trie->root);
  trie_node_destroyer(NULL, root_p, NULL);
  vector_destroy(trie->index2offset); 
  vector_destroy(trie->strings); 
#ifndef NDEBUG
  trie->symbol_size = UINT_MAX;
  trie->depth = UINT_MAX;
  trie->entry_count = UINT_MAX;
  trie->index2offset = NULL;
  trie->strings = NULL;
#endif
  memory_destroy(trie);
}

void trie_clear(trie_t *trie)
{
  trie_node_t *root_p;

  assert(trie_is_valid(trie));
  root_p = &(trie->root);
  trie_node_destroyer(NULL, root_p, NULL);
  trie->depth = 0;
  trie->entry_count = 0;
  trie_initialize_leaf(trie, root_p);
  trie->strings = vector_resize(trie->strings, 0);
  trie->index2offset = vector_resize(trie->index2offset, 1);
  trie->index2offset[0] = 0;
  assert(trie_is_valid(trie));
}


/************************************************************
 *  PREDICATE
 ************************************************************/

typedef struct validity_tag
{
  trie_t    *trie;
  unsigned   entry_total;
  unsigned   current_depth;
  unsigned   maximum_depth;
  char      *is_visited;
  boolean_t  is_valid;
} validity_t;


static boolean_t trie_leaf_validator(const void *key_p, void *datum_p, void *parameters)
{
  validity_t  *validity = parameters;
  index_t      index = *(index_t *)datum_p;

  assert(validity->is_valid);
  if (index >= trie_size(validity->trie))
    { 
      fprintf(stderr, "trie_t error: invalid leaf index %u\n", index);
      validity->is_valid = FALSE; 
    }
  else if (validity->is_visited[index])
    { 
      fprintf(stderr, "trie_t error: non-unique leaf index %u\n", index);
      validity->is_valid = FALSE; 
    }
  validity->is_visited[index] = TRUE;
  validity->entry_total++;
  return(validity->is_valid);
}

static boolean_t trie_node_validator(const void *key_p, void *datum_p, void *parameters)
{
  validity_t  *validity = parameters;
  trie_node_t *node_p = datum_p;
  index_t      stub_index;

  assert(validity->is_valid);
  if (!trie_node_is_valid(validity->trie, node_p))
    { 
      fprintf(stderr, "trie_t error: invalid node\n");
      validity->is_valid = FALSE; 
      return(FALSE); 
    }

  stub_index = trie_node_stub(node_p);
  if (stub_index != TRIE_NOT_FOUND) 
    {
      if (stub_index >= trie_size(validity->trie)) 
	{
	  fprintf(stderr, "trie_t error: invalid stub index %u\n", stub_index);
	  validity->is_valid = FALSE; 
	  return(FALSE); 
	} 
      else if (validity->is_visited[stub_index])
	{
	  fprintf(stderr, "trie_t error: non-unique stub index %u\n", stub_index);
	  validity->is_valid = FALSE; 
	  return(FALSE); 
	}
      if (validity->current_depth > validity->maximum_depth)
	{ validity->maximum_depth = validity->current_depth; }
      validity->is_visited[stub_index] = TRUE;
      validity->entry_total++; 
    }

  if (trie_node_is_interior(node_p))
    { 
      validity->current_depth++;
      trie_node_map(node_p, &trie_node_validator, parameters); 
      validity->current_depth--;
    }
  else 
    { 
      assert(trie_node_is_leaf(node_p));
      trie_leaf_map(node_p, &trie_leaf_validator, parameters); 
      if ((trie_node_degree(node_p) > 0) &&
	  (validity->current_depth >= validity->maximum_depth))
	{ validity->maximum_depth = validity->current_depth+1; }
    }
  return(validity->is_valid);
}

boolean_t trie_is_valid(const trie_t *trie)
{
  validity_t validity;
  trie_node_t *root;
  unsigned i, entry_count;

  if ((trie == NULL) ||
      (trie_symbol_size(trie) < 1) ||
      !vector_is_valid(trie->index2offset) ||
      !vector_is_valid(trie->strings) ||
      (trie->index2offset[0] != 0) ||
      (vector_length(trie->index2offset) != trie_size(trie)+1))
    { fprintf(stderr, "trie_t error: invalid structure\n"); return(FALSE); }

  root = (trie_node_t *)(&(trie->root));
  validity.trie = (trie_t *)trie;
  validity.is_visited = vector_create(sizeof(char), trie_size(trie));
  validity.is_valid = TRUE;
  validity.current_depth = 1;
  validity.maximum_depth = 0;
  validity.entry_total = 0;

  entry_count = trie_size(trie);
  for(i = 0; i < entry_count; i++) 
    { validity.is_visited[i] = FALSE; }

  if (trie_node_stub(root) != TRIE_NOT_FOUND) 
    { validity.entry_total++; validity.is_visited[trie_node_stub(root)] = TRUE; }
  if (trie_node_degree(root) > 0) { validity.maximum_depth = 1; }
  if (trie_node_is_leaf(root)) { trie_leaf_map(root, &trie_leaf_validator, &validity); }
  else { trie_node_map(root, &trie_node_validator, &validity); }
  assert(validity.current_depth == 1);

  if (validity.is_valid) {
    for(i = 0; i < entry_count; i++) { 
      if (!validity.is_visited[i]) {
	fprintf(stderr, "trie_t error: missing entry %u\n", i);
	validity.is_valid = FALSE;
      }
    }
  }
  vector_destroy(validity.is_visited); validity.is_visited = NULL;

  if (!validity.is_valid)
    { fprintf(stderr, "trie_t error: invalid node exists\n"); return(FALSE); }
  if (validity.maximum_depth != trie->depth)
    { fprintf(stderr, "trie_t error: incorrect depth\n"); return(FALSE); }
  if (validity.entry_total != trie->entry_count)
    { fprintf(stderr, "trie_t error: incorrect size\n"); return(FALSE); }
  return(TRUE);
	 
}


typedef struct comparator_tag {
  const trie_t *right;
  boolean_t     is_equal;
} comparator_t;

static boolean_t trie_entry_comparator(unsigned left_length, const void *left_string, 
				       unsigned index, void *parameters)
{
  comparator_t *enclosure = parameters;
  const trie_t *right = enclosure->right;
  unsigned right_length, bytes;
  const void *right_string;
  assert(enclosure->is_equal);
  assert(index < trie_size(right));
  right_length = trie_string_length(right, index);
  right_string = trie_string(right, index);
  bytes = right_length * trie_symbol_size(right);
  enclosure->is_equal = ((left_length == right_length) &&
			 (memcmp(left_string, right_string, bytes) == 0));
  return(enclosure->is_equal);
}

boolean_t trie_are_equal(const trie_t *left, const trie_t *right)
{
  comparator_t enclosure;
  assert(left != NULL); assert(right != NULL);
#ifndef NDEBUG
  if (left == right) { return(TRUE); }
#endif
  if ((trie_symbol_size(left) != trie_symbol_size(right)) ||
      (trie_depth(left) != trie_depth(right)) ||
      (trie_size(left) != trie_size(right)))
    { return(FALSE); }

  enclosure.right    = right;
  enclosure.is_equal = TRUE;
  trie_map(left, &trie_entry_comparator, &enclosure);
  return(enclosure.is_equal);
}


/************************************************************
 *  SELECTORS
 ************************************************************/

unsigned trie_lookup(const trie_t *trie, unsigned string_length, const void *string)
{
  const unsigned char *symbol_p;
  const trie_node_t   *node;
  index_t *index_p;
  unsigned symbol_size;

  if ((string_length > trie->depth) || (trie_size(trie) == 0)) 
    { return(TRIE_NOT_FOUND); }

  node = &(trie->root);
  symbol_p = string;
  symbol_size = trie_symbol_size(trie);
  while(string_length > 0)
    {
      assert(node);
      assert(trie_node_is_valid(trie, node));
      if (trie_node_is_interior(node))
	{
	  node = trie_node_lookup(node, symbol_p);
	  if (node == NULL) { return(TRIE_NOT_FOUND); }
	  string_length--;
	  symbol_p += symbol_size;
	}
      else
	{
	  assert(trie_node_is_leaf(node));
	  if (string_length > 1) { return(TRIE_NOT_FOUND); }
	  assert(string_length == 1);
	  index_p = trie_node_lookup(node, symbol_p);
	  if (index_p == NULL) { return(TRIE_NOT_FOUND); }
	  assert(*index_p < trie_size(trie));
	  return(*index_p);
	}
    }
  assert(string_length == 0);
  return(trie_node_stub(node));
}

unsigned trie_diversity(const trie_t *trie, unsigned string_length, const void *string)
{
  const unsigned char *symbol_p;
  const trie_node_t   *node;
  unsigned             symbol_size;

  if ((string_length >= trie->depth) || (trie_size(trie) == 0)) { return(0); }

  node = &(trie->root);
  symbol_p = string;
  symbol_size = trie_symbol_size(trie);
  while(string_length > 0)
    {
      assert(node);
      assert(trie_node_is_valid(trie, node));
      if (trie_node_is_interior(node))
	{
	  node = trie_node_lookup(node, symbol_p);
	  if (node == NULL) { return(0); }
	  string_length--;
	  symbol_p += symbol_size;
	}
      else
	{
	  assert(trie_node_is_leaf(node));
	  return(0);
	}
    }
  assert(string_length == 0);
  return(trie_node_degree(node));
}

unsigned trie_string_length(const trie_t *trie, unsigned index)
{
  unsigned offset, length;
  
  assert(trie != NULL);
  assert(trie_index_is_valid(trie, index));
  
  assert(vector_is_valid(trie->index2offset));
  assert(vector_index_is_valid(trie->index2offset, index));
  assert(vector_index_is_valid(trie->index2offset, index+1));
  assert(trie->index2offset[index] <= trie->index2offset[index+1]);

  offset = trie->index2offset[index];
  length = trie->index2offset[index+1] - offset;
  return(length);
}

const void *trie_string(const trie_t *trie, unsigned index)
{
  unsigned offset;
  const void *string;

  assert(trie != NULL);
  assert(trie_index_is_valid(trie, index));

  assert(vector_is_valid(trie->index2offset));
  assert(vector_index_is_valid(trie->index2offset, index));
  assert(vector_index_is_valid(trie->index2offset, index+1));
  assert(vector_object_size(trie->strings) == trie_symbol_size(trie));
  assert(trie->index2offset[index] <= trie->index2offset[index+1]);

  offset = trie->index2offset[index];
  if (offset == trie->index2offset[index+1]) { return(NULL); }
  assert(vector_index_is_valid(trie->strings, offset));
  string = vector_get(trie->strings, offset);
  return(string);
}


/************************************************************
 *  MUTATOR
 ************************************************************/

unsigned trie_insert(trie_t *trie, unsigned string_length, const void *string)
{
  const unsigned char *symbol_p;
  trie_node_t   *node, *child;
  index_t       *index_p, ultimate_index;
  unsigned       symbol_size, symbol_count, this_offset, next_offset;
  void          *destination;

  if (string_length > trie->depth) { trie->depth = string_length; }

  node = (trie_node_t *)(&(trie->root));
  symbol_p = string;
  symbol_size = trie_symbol_size(trie);
  symbol_count = string_length;
  while(symbol_count > 0)
    {
      assert(symbol_count > 0);
      assert(trie_node_is_valid(trie, node));
      if (trie_node_is_interior(node))
	{
	  child = trie_node_lookup(node, symbol_p);
	  if (child == NULL) 
	    { 
	      trie_node_t blank_child;
	      assert(trie_node_is_interior(node));
	      assert(trie_node_is_valid(trie, node));
	      trie_initialize_leaf(trie, &blank_child);
	      child = trie_node_insert(node, symbol_p, &blank_child);
	      assert(child); assert(child != &blank_child);
	      assert(trie_node_is_leaf(child));
	    }
	  assert(trie_node_is_valid(trie, child));
	  assert(child == trie_node_lookup(node, symbol_p));
	  node = child;
	  symbol_count--;
	  symbol_p += symbol_size;
	}
      else
	{
	  assert(trie_node_is_leaf(node));
	  if (symbol_count == 1)  /* just insert the last symbol */
	    {
	      index_p = trie_node_lookup(node, symbol_p);
	      if (index_p != NULL)
		{ 
		  assert(*index_p < trie_size(trie));
		  return(*index_p);   /* string already included */
		}
	      ultimate_index = trie->entry_count;
	      index_p = trie_leaf_insert(node, symbol_p, &ultimate_index);
	      assert(index_p);
	      assert(*index_p == ultimate_index);
	      trie->entry_count++;

	      /* actually append the string */
	      assert(vector_index_is_valid(trie->index2offset, ultimate_index));
	      assert(vector_length(trie->strings) == trie->index2offset[ultimate_index]);
	      this_offset = trie->index2offset[ultimate_index];
	      next_offset = this_offset + string_length;
	      trie->index2offset = vector_append(trie->index2offset, &next_offset);
	      trie->strings = vector_expand(trie->strings, string_length);
	      destination = vector_get(trie->strings, this_offset); 
	      (void) memcpy(destination, string, string_length * symbol_size);
	      assert(trie_string_length(trie, ultimate_index) == string_length);
	      assert(memcmp(trie_string(trie, ultimate_index), string,
			    string_length * symbol_size) == 0);

	      return(ultimate_index);
	    }
	  assert(symbol_count > 1);
	  trie_leaf2interior(trie, node);
	  assert(trie_node_is_interior(node));
	}
    }
  assert(symbol_count == 0);
  assert(trie_node_is_valid(trie, node));
  if (trie_node_stub(node) == TRIE_NOT_FOUND)
    {
      ultimate_index = trie->entry_count;
      trie_node_set_stub(node, ultimate_index);
      trie->entry_count++;

      /* actually append the string */
      /* note: can save space here by using a successor */
      assert(vector_index_is_valid(trie->index2offset, ultimate_index));
      assert(vector_length(trie->strings) == trie->index2offset[ultimate_index]);
      this_offset = trie->index2offset[ultimate_index];
      next_offset = this_offset + string_length;
      trie->index2offset = vector_append(trie->index2offset, &next_offset);
      trie->strings = vector_expand(trie->strings, string_length);
      destination = vector_get(trie->strings, this_offset); 
      (void) memcpy(destination, string, string_length * symbol_size);
      assert(trie_string_length(trie, ultimate_index) == string_length);
      assert(memcmp(trie_string(trie, ultimate_index), string,
		    string_length * symbol_size) == 0);
      assert(trie_node_stub(node) == ultimate_index);
    }
  assert(trie_node_stub(node) < trie_size(trie));
  return(trie_node_stub(node));
}

/************************************************************
 *  HIGHER ORDER PROCEDURE
 ************************************************************/

typedef struct enclosure_tag
{
  vector_t   *string;
  boolean_t (*proc)(unsigned, const void *, index_t, void *);
  void       *parameters;
  boolean_t   is_alive;  
} enclosure_t;

static boolean_t trie_leaf_mapper(const void *symbol_p, void *index_p, void *parameters)
{
  enclosure_t *enclosure = parameters;
  index_t      index = *(index_t *)index_p;
  unsigned     string_length;

  assert(enclosure->is_alive);
  assert(vector_is_valid(enclosure->string));
  enclosure->string = vector_append(enclosure->string, symbol_p);
  assert(vector_is_valid(enclosure->string));
  string_length = vector_length(enclosure->string);
  enclosure->is_alive = enclosure->proc(string_length, enclosure->string, index, 
					enclosure->parameters);
  enclosure->string = vector_contract(enclosure->string, 1);
  return(enclosure->is_alive);
}

static boolean_t trie_node_mapper(const void *symbol_p, void *node_p, void *parameters)
{
  enclosure_t *enclosure = parameters;
  trie_node_t *node = node_p;

  assert(enclosure->is_alive);
  assert(vector_is_valid(enclosure->string));
  enclosure->string = vector_append(enclosure->string, symbol_p);
  assert(vector_is_valid(enclosure->string));

  if (trie_node_stub(node) != TRIE_NOT_FOUND)
    {
      unsigned string_length;
      index_t  index;

      index = trie_node_stub(node);
      string_length = vector_length(enclosure->string);
      enclosure->is_alive = enclosure->proc(string_length, enclosure->string, index, 
					    enclosure->parameters);
    }
  if (enclosure->is_alive) 
    { 
      if (trie_node_is_interior(node)) 
	{ trie_node_map(node, &trie_node_mapper, parameters); }
      else
	{ trie_leaf_map(node, &trie_leaf_mapper, parameters); }
    }
  enclosure->string = vector_contract(enclosure->string, 1);
  assert(vector_is_valid(enclosure->string));
  return(enclosure->is_alive);
}

void trie_depth_map(const trie_t *trie, 
		    boolean_t (*proc)(unsigned string_length, const void *string, 
				      index_t string_index, void *parameters),
		    void *parameters)
{
  enclosure_t  enclosure;
  trie_node_t *root;

  enclosure.proc = proc;
  enclosure.parameters = parameters;
  enclosure.string = vector_create_surplus(trie_symbol_size(trie), 0, trie_depth(trie));
  enclosure.is_alive = TRUE;
  assert(vector_is_valid(enclosure.string));
  assert(vector_length(enclosure.string) == 0);

  root = (trie_node_t *)(&(trie->root));
  if (trie_node_stub(root) != TRIE_NOT_FOUND)
    {
      index_t  index;
      index = trie_node_stub(root);
      enclosure.is_alive = proc(0, NULL, index, parameters);
    }

  if (enclosure.is_alive) 
    { 
      if (trie_node_is_interior(root))
	{ trie_node_map(root, &trie_node_mapper, &enclosure); }
      else
	{ trie_leaf_map(root, &trie_leaf_mapper, &enclosure); }
    }

  assert(vector_is_valid(enclosure.string));
  assert(vector_length(enclosure.string) == 0);
  vector_destroy(enclosure.string);
}

void trie_map(const trie_t *trie, 
	      boolean_t (*proc)(unsigned, const void *, index_t, void *),
	      void *parameters)
{
  unsigned entry_count, string_length;
  const void *string;
  index_t index;
  boolean_t is_alive;

  entry_count = trie_size(trie);
  for(index = 0; index < entry_count; index++)
    {
      string_length = trie_string_length(trie, index);
      string = trie_string(trie, index);
      assert(trie_lookup(trie, string_length, string) == index);
      is_alive = proc(string_length, string, index, parameters);
      if (!is_alive) { return; }
    }
}



