/*
 * (c) 1996 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include "memory.h"
#include "vector.h"
#include "stopwatch.h"
#include "trie.h"

static char trie_corpus_rcsid[] = "$Id: trie.corpus.c,v 1.4 1997/06/05 22:08:53 ristad Exp $";


#define DEBUG                 FALSE

typedef struct diversity_enclosure_tag
{
  trie_t *trie;
  double  children;
  double  parents;
} diversity_enclosure_t;

void diversity_counter(unsigned string_length, const void *string,
		       unsigned string_index, void *parameters)
{
  diversity_enclosure_t *enclosure;
  double diversity;

  enclosure = parameters;
  diversity = trie_diversity(enclosure->trie, string_length-1, string);
  assert(diversity > 0);

  enclosure->children += 1;
  enclosure->parents  += 1/diversity;
}


extern double drand48(void);

int main(int argc, char **argv)
{
  diversity_enclosure_t enclosure;
  stopwatch_t *stopwatch;
  void *corpus;
  FILE *corpus_stream;
  char *corpus_file;
  unsigned symbol_size, depth_limit;
  unsigned char  uchar_symbol;
  unsigned short ushrt_symbol;
  unsigned int   uint_symbol;
  trie_t *trie;
  double corpus_time, trie_load_time;
  unsigned corpus_memory_usage, trie_memory_usage;
  unsigned index, corpus_length;
  int i, byte;

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  if (argc >= 2) { symbol_size = atoi(argv[1]); }
  if (argc >= 3) { depth_limit = atoi(argv[2]); }
  if (argc >= 4) { corpus_file = argv[3]; }

  if ((argc != 4) || (depth_limit < 1) ||
      !((symbol_size == 1) || (symbol_size == 2) || (symbol_size == 4)))
    {
      printf("Usage:\t%s symbol-size depth-limit corpus-file\n", argv[0]);
      return(-1);
    }

  printf("\n\tReading corpus %s into memory (%u bytes/symbol)...", 
	 corpus_file, symbol_size);

  /* read in the corpus using fscanf() */
  stopwatch = stopwatch_create();    /* allocate stopwatch */
  stopwatch_start(stopwatch);
  corpus_stream = fopen(corpus_file, "r");
  if (corpus_stream == NULL)
    {
      printf("%s:\tcan't open corpus file %s\n", argv[0], corpus_file);
      return(-1);
    }
  corpus = vector_create(symbol_size, 0);
  assert(vector_is_valid(corpus));
  assert(vector_length(corpus) == 0);

  switch(symbol_size)
    {
    case sizeof(unsigned char):
      byte = fgetc(corpus_stream);
      while(byte != EOF)
	{
	  assert(byte >= 0);
	  assert(byte <= UCHAR_MAX);
	  uchar_symbol = (unsigned char)byte;
	  corpus = vector_append(corpus, &uchar_symbol);
	  assert(vector_is_valid(corpus));
	  byte = fgetc(corpus_stream);
	}
      break;
    case sizeof(unsigned short):
      while(fscanf(corpus_stream, "%u", &uint_symbol) == 1)
	{
	  assert(uint_symbol <= USHRT_MAX);
	  ushrt_symbol = (unsigned short)uint_symbol;
	  corpus = vector_append(corpus, &ushrt_symbol);
	  assert(vector_is_valid(corpus));
	}
      break;
    case sizeof(unsigned int):
      while(fscanf(corpus_stream, "%u", &uint_symbol) == 1)
	{
	  corpus = vector_append(corpus, &uint_symbol);
	  assert(vector_is_valid(corpus));
	}
      break;
    default:
      abort();
      break;
    }
  assert(vector_is_valid(corpus));
  assert(feof(corpus_stream));
  fclose(corpus_stream); corpus_stream = NULL;
  corpus_length = vector_length(corpus);

  corpus_time = stopwatch_split(stopwatch);
  corpus_memory_usage = memory_total_bytes();
  printf("done\n\t (%u symbols, %.2f cpu).\n", vector_length(corpus), corpus_time);
  stopwatch_restart(stopwatch);

  printf("\tLoading corpus into %u,%u trie_t...", symbol_size, depth_limit);
  /* create the trie_t and insert all corpus substrings */
  trie = trie_create(symbol_size);
  assert(trie);
  assert(trie_size(trie) == 0);
  assert(trie_symbol_size(trie) == symbol_size);
  assert(vector_object_size(corpus) == symbol_size);

  for(i = 0; i <= corpus_length-depth_limit; i++)
    {
      assert(vector_index_is_valid(corpus, i));
      index = trie_insert(trie, depth_limit, vector_element_p(corpus, i));
      assert(index != TRIE_NOT_FOUND);
    }
  assert(trie_depth(trie) == depth_limit);
  assert(trie_symbol_size(trie) == symbol_size);

  trie_load_time = stopwatch_split(stopwatch);
  stopwatch_stop(stopwatch);
  stopwatch_destroy(stopwatch); stopwatch = NULL;
  vector_destroy(corpus); corpus = NULL;

  trie_memory_usage = memory_total_bytes();

  enclosure.trie = trie;
  enclosure.children = 0;
  enclosure.parents = 0;
  trie_map(trie, &diversity_counter, &enclosure);
  assert(enclosure.children == trie_size(trie));

  printf("done\n\t (%d strings, %d bytes, %.2fs).\n", 
	 (int)trie_size(trie), (int)trie_memory_usage, trie_load_time);
  printf("\tmean branching factor %.2f at level %u\n",
	 enclosure.children/enclosure.parents, depth_limit-1);

  printf("\n%u,%u\t%d\t%d\t%.2f\n",
	 symbol_size, depth_limit,
	 (int)trie_size(trie), (int)trie_memory_usage, trie_load_time);

  /* cleanup */
  trie_destroy(trie); trie = NULL;
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);

  return(0);
}



