/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 *
 * $Id: trie.h,v 1.6 1998/01/28 17:45:36 libpa Exp $
 *
 */

#ifndef TRIE_INCLUDED
#define TRIE_INCLUDED

#include <stdio.h>
#include <limits.h>
#include "common.h"
#include "table.h"

/* Description.
 *  Simple implementation of traditional trie, which maps strings to indices.
 *  Stores arbitrary strings over arbitrary alphabets.
 */
 
#define TRIE_NOT_FOUND           (UINT_MAX)
#define TRIE_INDEX_MAX           (UINT_MAX-1)


typedef table_t trie_node_t;
typedef struct trie_tag
{
  unsigned     symbol_size;            /* bytes per symbol        */
  unsigned     depth;                  /* length of longest entry */
  unsigned     entry_count;            /* number of entries       */
  trie_node_t  root;
  unsigned    *index2offset;
  void        *strings;
} trie_t;

/************************************************************
 *  CONSTRUCTOR/DESTRUCTOR
 ************************************************************/

trie_t     *trie_create(unsigned symbol_size);
            /* Returns: empty trie for strings over symbol_size-byte symbols */
trie_t     *trie_copy(const trie_t *trie);
            /* Returns: exact copy of given trie             */
void        trie_destroy(trie_t *trie);
            /* Effect: deallocate all memory used by trie_t  */
void        trie_clear(trie_t *trie);
            /* Effect: removes all strings from given trie_t */

/************************************************************
 *  SELECTORS/PREDICATE
 ************************************************************/

#define     trie_size(trie_p)                ((trie_p)->entry_count)
            /* Returns: number of distinct strings in trie_t */
#define     trie_depth(trie_p)               ((trie_p)->depth)
            /* Returns: length of longest string in trie_t   */
#define     trie_symbol_size(trie_p)         ((trie_p)->symbol_size)
            /* Returns: bytes per symbol for trie_t strings  */
#define     trie_index_is_valid(trie_p,i)    ((i) < trie_size(trie_p))
boolean_t   trie_is_valid(const trie_t *trie);
            /* Returns: TRUE iff given trie is valid         */
boolean_t   trie_are_equal(const trie_t *left, const trie_t *right);
            /* Returns: TRUE iff both tables are equivalent  */

unsigned    trie_lookup(const trie_t *trie, unsigned string_length, const void *string);
            /* Returns: index for string if found else TRIE_NOT_FOUND  */
unsigned    trie_diversity(const trie_t *trie, unsigned string_length, const void *string);
            /* Returns: number of symbols that can follow given string */

unsigned    trie_string_length(const trie_t *trie, unsigned index);
            /* Requires: trie_index_is_valid(trie, index)            */
            /* Returns: length of string associated with given index */
const void *trie_string(const trie_t *trie, unsigned index);
            /* Requires: trie_index_is_valid(trie, index)            */
            /* Returns: actual string associated with given index    */

/************************************************************
 *  MUTATOR
 ************************************************************/

unsigned    trie_insert(trie_t *trie, unsigned string_length, const void *string);
            /* Effect: adds given string to trie if not already included */
            /* Note: strings are indexed consecutively starting from 0   */
            /* Returns: proper index of given string                     */

/************************************************************
 *  HIGHER-ORDER FUNCTION
 ************************************************************/

void        trie_map(const trie_t *trie, 
		     boolean_t (*proc)(unsigned string_length, const void *string, 
				       unsigned string_index, void *parameters),
		     void *parameters);
            /* Effect: calls proc() on all strings in order of their index  */
            /*         while proc() returns TRUE                            */
            /* Note: proc() may not mutate string without invalidating trie */

void        trie_depth_map(const trie_t *trie, 
			   boolean_t (*proc)(unsigned string_length, const void *string, 
					     unsigned string_index, void *parameters),
			   void *parameters);
            /* Effect: calls proc() on all strings in df lexographic order  */
            /*         while proc() returns TRUE                            */
            /* Note: proc() may not mutate string without invalidating trie */

#endif

