/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>  /* for memset()  */
#include <math.h>    /* for drand48() */
#include "common.h"
#include "memory.h"
#include "vector.h"
#include "trie.h"

static char trie_test_rcsid[] = "$Id: trie.test.c,v 1.6 1998/01/28 17:45:36 libpa Exp $";

#define DEBUG                 FALSE
#define BE_VERBOSE            FALSE
#define SUBSTRING_LIMIT       8
#define TERMINATE_THRESHOLD   0.9
#define TERMINATE_TRIALS      100

static int Corpus_Size[] = { 0, UCHAR_MAX, UCHAR_MAX, 0, UCHAR_MAX };


extern double drand48(void);
extern void   srand48(long);


void set_symbol(void *symbol_p, int symbol_size, unsigned value)
{
  switch(symbol_size)
    {
    case 1: *((unsigned char  *)symbol_p) = value; break;
    case 2: *((unsigned short *)symbol_p) = value; break;
    case 4: *((unsigned int   *)symbol_p) = value; break;
    default: abort();
    }
}

unsigned get_symbol(const void *symbol_p, int symbol_size)
{
  switch(symbol_size)
    {
    case 1: return(*((unsigned char  *)symbol_p));
    case 2: return(*((unsigned short *)symbol_p));
    case 4: return(*((unsigned int   *)symbol_p));
    default: abort();
    }
}

typedef struct map_enclosure_tag
{
  trie_t     *trie;
  boolean_t  *is_visited;
} map_enclosure_t;

boolean_t map_validator(unsigned length, const void *string,
			unsigned index, void *parameters)
{
  map_enclosure_t *enclosure = parameters;
  boolean_t *is_visited = enclosure->is_visited;
  trie_t *trie = enclosure->trie;
  unsigned symbol_size;

  symbol_size = trie_symbol_size(trie);

  assert(vector_is_valid(is_visited));
  assert(vector_object_size(is_visited) == sizeof(boolean_t));
  assert(vector_index_is_valid(is_visited, index));
  assert(!is_visited[index]);
  is_visited[index] = TRUE;

  assert(trie_index_is_valid(trie, index));
  assert(trie_lookup(trie, length, string) == index);
  assert(trie_string_length(trie, index) == length);
  assert(memcmp(trie_string(trie, index), string, length * symbol_size) == 0);
  return(TRUE);
}

boolean_t map_terminator(unsigned length, const void *string,
			 unsigned index, void *parameters)
{
  boolean_t *is_terminated_p = parameters;
  double throw;

  if (*is_terminated_p) { fprintf(stderr, "failure to terminate\n"); abort(); }
  throw = drand48();
  if (throw > TERMINATE_THRESHOLD) { *is_terminated_p = TRUE; return(FALSE); }
  return(TRUE);
}


/* Note: generate unique string and store all its substrings of all lengths */
boolean_t unique_test(unsigned symbol_size)
{
  trie_t *trie, *copy;
  unsigned corpus_length, string_count, index, diversity;
  void *symbol_p, *corpus;
  map_enclosure_t enclosure;
  boolean_t is_terminated;
  int t, i, length;

  symbol_p = memory_create(symbol_size);
  assert(symbol_p);
  assert(memory_bytes(symbol_p) == symbol_size);

  corpus = vector_create_surplus(symbol_size, 0, Corpus_Size[symbol_size]);
  assert(vector_is_valid(corpus));
  assert(vector_length(corpus) == 0);

  trie = trie_create(symbol_size);
  assert(trie);
  assert(trie_size(trie) == 0);
  assert(trie_depth(trie) == 0);
  assert(trie_symbol_size(trie) == symbol_size);
  assert(trie_is_valid(trie));
  assert(trie_are_equal(trie, trie));
#if BE_VERBOSE
  printf("%u\tcreated %u-byte symbol, corpus, and trie_t\n", 
	 memory_total_bytes(), symbol_size);
#endif

  copy = trie_copy(trie);
  assert(trie_are_equal(copy, trie));
#if BE_VERBOSE
  printf("%u\tverified %u-byte _copy and _are_equal() on empty trie_t\n",
	 memory_total_bytes(), symbol_size);
#endif

  /* generate unique string */
#if BE_VERBOSE
  printf("\tgenerating %u-byte unique string of length %d", 
	 symbol_size, (int)Corpus_Size[symbol_size]);
#endif
  corpus_length = 0;
  for(i = 0; i < Corpus_Size[symbol_size]; i++)
    {
      set_symbol(symbol_p, symbol_size, i);
      assert(get_symbol(symbol_p, symbol_size) == i);
      corpus = vector_append(corpus, symbol_p);
      corpus_length++;
    }
  assert(vector_is_valid(corpus));
  assert(vector_length(corpus) == corpus_length);
  assert(corpus_length == Corpus_Size[symbol_size]);
#if BE_VERBOSE
  printf(" done.\n");
#endif

  /* insert shorter strings before longer strings */
  printf("%u\tinserting all %u-byte substrings of length 0", 
	 memory_total_bytes(), symbol_size);
  assert(trie_lookup(trie, 0, NULL) == TRIE_NOT_FOUND);
  index = trie_insert(trie, 0, NULL);
  string_count = 1;
  assert(index == string_count-1);
  assert(trie_size(trie) == string_count);
  assert(trie_lookup(trie, 0, NULL) == string_count-1);
  assert(trie_insert(trie, 0, NULL) == string_count-1);
  assert(trie_string_length(trie, string_count-1) == 0);
  assert(trie_string(trie, string_count-1) == NULL);
  assert(trie_is_valid(trie));
  for(length = 1; length <= SUBSTRING_LIMIT; length++)
    {
      printf(" %d", length);
      assert(trie_depth(trie) == length-1);
      for(i = 0; i <= Corpus_Size[symbol_size]-length; i++)
	{
	  assert(vector_index_is_valid(corpus, i));

	  /* make sure its not there */
	  index = trie_lookup(trie, length, vector_element_p(corpus, i));
	  assert(index == TRIE_NOT_FOUND);
	  assert(trie_size(trie) == string_count);

	  /* actually insert it */
	  index = trie_insert(trie, length, vector_element_p(corpus, i));
	  string_count++;
	  assert(index == string_count-1);
	  assert(trie_size(trie) == string_count);
	  assert(trie_depth(trie) == length);

	  /* make sure its really there */
	  index = trie_lookup(trie, length, vector_element_p(corpus, i));
	  assert(index == string_count-1);
	  assert(trie_size(trie) == string_count);
	  assert(trie_index_is_valid(trie, index));
	  assert(trie_string_length(trie, index) == length);
	  assert(memcmp(trie_string(trie, index), vector_element_p(corpus, i),
			length * symbol_size) == 0);

	  /* make sure that repeated insert doesn't change anything */
	  index = trie_insert(trie, length, vector_element_p(corpus, i));
	  assert(index == string_count-1);
	  assert(trie_size(trie) == string_count);
	  assert(trie_depth(trie) == length);
	  assert(trie_is_valid(trie));
	}
      assert(trie_depth(trie) == length);
    }
  assert(trie_size(trie) == string_count);
  printf(" done.\n");

  /* test equality and copying */
  assert(trie_are_equal(trie, trie));
  assert(!trie_are_equal(trie, copy));
  trie_destroy(copy); copy = NULL;
  copy = trie_copy(trie);
  assert(trie_are_equal(trie, copy));
#if BE_VERBOSE
  printf("%u\tverified %u-byte _copy and _are_equal() on full trie_t\n",
	 memory_total_bytes(), symbol_size);
#endif

  /* verify that all strings are still there with same indices */
#if BE_VERBOSE
  printf("%u\tverifying indexing of all %u-byte substrings of length 0", 
	 memory_total_bytes(), symbol_size);
#endif
  index = trie_lookup(trie, 0, NULL);
  string_count = 1;
  assert(index == string_count-1);
  assert(trie_string_length(trie, index) == 0);
  assert(trie_string(trie, index) == NULL);
  diversity = trie_diversity(trie, 0, NULL);
  assert(diversity == Corpus_Size[symbol_size]);
  for(length = 1; length <= SUBSTRING_LIMIT; length++)
    {
#if BE_VERBOSE
      printf(" %d", length);
#endif
      for(i = 0; i <= Corpus_Size[symbol_size]-length; i++)
	{
	  index = trie_lookup(trie, length, vector_element_p(corpus, i));
	  assert(trie_index_is_valid(trie, index));
	  assert(trie_string_length(trie, index) == length);
	  assert(memcmp(trie_string(trie, index), vector_element_p(corpus, i),
			length * symbol_size) == 0);

	  string_count++;
	  assert(index == string_count-1);
	  if (length > 1)
	    {
	      diversity = trie_diversity(trie, length-1, vector_element_p(corpus, i));
	      assert(diversity == 1);
	    }
	}
    }
  assert(trie_size(trie) == string_count);
#if BE_VERBOSE
  printf(" done.\n");
#endif

  enclosure.trie = trie;
  enclosure.is_visited = vector_create(sizeof(boolean_t), string_count);
  assert(vector_is_valid(enclosure.is_visited));
  for(i = 0; i < string_count; i++) 
    { enclosure.is_visited[i] = FALSE; }
  trie_map(trie, &map_validator, &enclosure);
  for(i = 0; i < string_count; i++) 
    { assert(enclosure.is_visited[i]); enclosure.is_visited[i] = FALSE; }
  trie_depth_map(trie, &map_validator, &enclosure);
  for(i = 0; i < string_count; i++) 
    { assert(enclosure.is_visited[i]); enclosure.is_visited[i] = FALSE; }
  for(t = 0; t < TERMINATE_TRIALS; t++)
    {
      is_terminated = FALSE; 
      trie_map(trie, &map_terminator, &is_terminated);
      is_terminated = FALSE; 
      trie_depth_map(trie, &map_terminator, &is_terminated);
    }
#if BE_VERBOSE
  printf("%u\tverified trie_map() and trie_depth_map().\n", memory_total_bytes());
#endif

  trie_clear(trie);
  assert(trie_size(trie) == 0);
  assert(trie_depth(trie) == 0);
  assert(trie_is_valid(trie));
  assert(trie_are_equal(trie, trie));
  assert(!trie_are_equal(trie, copy));
#if BE_VERBOSE
  printf("%u\tcleared trie_t succesfully.\n", memory_total_bytes());
#endif

  /* now insert longer strings before shorter strings */
  printf("%u\tinserting all %u-byte substrings of length", 
	 memory_total_bytes(), symbol_size);
  string_count = 0;
  for(length = SUBSTRING_LIMIT; length > 0; length--)
    {
      printf(" %d", length);
      for(i = 0; i <= Corpus_Size[symbol_size]-length; i++)
	{
	  assert(vector_index_is_valid(corpus, i));

	  /* make sure its not there */
	  index = trie_lookup(trie, length, vector_element_p(corpus, i));
	  assert(index == TRIE_NOT_FOUND);
	  assert(trie_size(trie) == string_count);
	  assert(trie_is_valid(trie));

	  /* actually insert it */
	  index = trie_insert(trie, length, vector_element_p(corpus, i));
	  string_count++;
	  assert(index == string_count-1);
	  assert(trie_size(trie) == string_count);
	  assert(trie_is_valid(trie));

	  /* make sure its really there */
	  index = trie_lookup(trie, length, vector_element_p(corpus, i));
	  assert(index == string_count-1);
	  assert(trie_size(trie) == string_count);
	  assert(trie_index_is_valid(trie, index));
	  assert(trie_string_length(trie, index) == length);
	  assert(memcmp(trie_string(trie, index), vector_element_p(corpus, i),
			length * symbol_size) == 0);

	  /* make sure that repeated insert doesn't change anything */
	  index = trie_insert(trie, length, vector_element_p(corpus, i));
	  assert(index == string_count-1);
	  assert(trie_size(trie) == string_count);
	  assert(trie_is_valid(trie));
	}
      assert(trie_depth(trie) == SUBSTRING_LIMIT);
      assert(trie_size(trie) == string_count);
    }
  assert(trie_lookup(trie, 0, NULL) == TRIE_NOT_FOUND);
  index = trie_insert(trie, 0, NULL);
  assert(trie_index_is_valid(trie, index));
  assert(trie_string_length(trie, index) == 0);
  assert(trie_string(trie, index) == NULL);
  string_count++;
  assert(index == string_count-1);
  assert(trie_size(trie) == string_count);
  assert(trie_lookup(trie, 0, NULL) == string_count-1);
  assert(trie_insert(trie, 0, NULL) == string_count-1);
  printf(" 0 done.\n");

  /* verify that all strings are still there with same indices */
#if BE_VERBOSE
  printf("%u\tverifying indexing of all %u-byte substrings of length", 
	 memory_total_bytes(), symbol_size);
#endif
  string_count = 0;
  for(length = SUBSTRING_LIMIT; length > 0; length--)
    {
#if BE_VERBOSE
      printf(" %d", length);
#endif
      for(i = 0; i <= Corpus_Size[symbol_size]-length; i++)
	{
	  index = trie_lookup(trie, length, vector_element_p(corpus, i));
	  assert(trie_index_is_valid(trie, index));
	  assert(trie_string_length(trie, index) == length);
	  assert(memcmp(trie_string(trie, index), vector_element_p(corpus, i),
			length * symbol_size) == 0);

	  string_count++;
	  assert(index == string_count-1);
	  if (length > 1)
	    {
	      diversity = trie_diversity(trie, length-1, vector_element_p(corpus, i));
	      assert(diversity == 1);
	    }
	}
    }
  index = trie_lookup(trie, 0, NULL);
  assert(trie_index_is_valid(trie, index));
  assert(trie_string_length(trie, index) == 0);
  assert(trie_string(trie, index) == NULL);
  string_count++;
  assert(index == string_count-1);
  diversity = trie_diversity(trie, 0, NULL);
  assert(diversity == Corpus_Size[symbol_size]);
  assert(string_count == trie_size(trie));
#if BE_VERBOSE
  printf(" 0 done.\n");
#endif

  assert(trie_symbol_size(copy) == trie_symbol_size(trie));
  assert(trie_depth(copy) == trie_depth(trie));
  assert(trie_size(copy) == trie_size(trie));
  assert(!trie_are_equal(trie, copy));
  trie_destroy(copy); copy = NULL;
#if BE_VERBOSE
  printf("%u\tverified %u-byte !trie_are_equal() on reconstituted trie_t\n",
	 memory_total_bytes(), symbol_size);
#endif

  assert(enclosure.trie == trie);
  assert(vector_is_valid(enclosure.is_visited));
  assert(vector_length(enclosure.is_visited) == string_count);
  for(i = 0; i < string_count; i++) 
    { assert(!enclosure.is_visited[i]); }
  trie_map(trie, &map_validator, &enclosure);
  for(i = 0; i < string_count; i++) 
    { assert(enclosure.is_visited[i]); enclosure.is_visited[i] = FALSE; }
  trie_depth_map(trie, &map_validator, &enclosure);
  for(i = 0; i < string_count; i++) 
    { assert(enclosure.is_visited[i]); }
#if BE_VERBOSE
  printf("%u\tverified trie_map() and trie_depth_map().\n", memory_total_bytes());
#endif

  /* clean up */
  vector_destroy(enclosure.is_visited); enclosure.is_visited = NULL;
  memory_destroy(symbol_p); symbol_p = NULL;
  vector_destroy(corpus); corpus = NULL;
  trie_destroy(trie); trie = NULL;
#if BE_VERBOSE
  printf("%u\tdestroyed valid %u-byte trie_t, corpus and symbol_p\n", 
	 memory_total_bytes(), symbol_size); 
#endif
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);

  return(TRUE);
}


int main(int argc, char **argv)
{
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("\n\tTesting trie_t library...\n");

  assert(unique_test(1));
  assert(unique_test(2));
  assert(unique_test(4));

  printf("\tmemory leak == %u bytes in %u blocks\n", 
	 memory_total_bytes(), memory_total_blocks());
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  printf("\ttrie_t library passes all tests.\n");
  return(0);
}


