/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "memory.spartan.h"
#include "unigram.h"

static char  unigram_rcsid[] = "$Id: unigram.c,v 1.8 1998/02/07 03:46:35 libpa Exp $";

#define ANNOUNCE_UPSIZE     FALSE


/**********************************************************************/
#define uni2tab(uni_p)     ((table_t *)(uni_p))
void    unigram_upsize(unigram_t *unigram);
        /* Effect: double counter size */

/**********************************************************************/

typedef unsigned char min_counter_t;
typedef unsigned int  max_counter_t;


/* Fast Coercion for Resizing */
static void uchar_up(const void *old_p, void *new_p, void *parameters)
{ *(USHORT *)new_p = *(UCHAR *)old_p; }
static void ushrt_up(const void *old_p, void *new_p, void *parameters)
{ *(UINT *)new_p = *(USHORT *)old_p;  }
static void (*Upsizers[])(const void *, void *, void *) =  { NULL, &uchar_up, 
							       &ushrt_up, NULL, NULL };

static void inc_uchar(void *counter_p) { (*(UCHAR *)counter_p)++;  }
static void inc_ushrt(void *counter_p) { (*(USHORT *)counter_p)++; }
static void inc_uint(void *counter_p)  { (*(UINT *)counter_p)++;   }
static void (*Incrementers[])(void *) = { NULL, &inc_uchar, &inc_ushrt, NULL, &inc_uint };

static unsigned get_uchar(const void *counter_p) { return((unsigned)*(UCHAR *)counter_p);  }
static unsigned get_ushrt(const void *counter_p) { return((unsigned)*(USHORT *)counter_p); }
static unsigned get_uint(const void *counter_p)  { return((unsigned)*(UINT *)counter_p);   }
static unsigned (*Getters[])(const void *) = { NULL, &get_uchar, &get_ushrt, 
						 NULL, &get_uint };

static void set_uchar(void *counter_p, unsigned v) { (*(UCHAR *)counter_p)  = v; }
static void set_ushrt(void *counter_p, unsigned v) { (*(USHORT *)counter_p) = v; }
static void set_uint(void *counter_p, unsigned v)  { (*(UINT *)counter_p)   = v; }
static void (*Setters[])(void *, unsigned) = { NULL, &set_uchar, &set_ushrt, 
						 NULL, &set_uint };

static unsigned Counter_Limits[] = { 0, UCHAR_MAX, USHRT_MAX, 0, UINT_MAX };


#define counter_limit(uni_p)         (Counter_Limits[unigram_counter_size(uni_p)])
#define counter_upsizer(uni_p)       (Upsizers[unigram_counter_size(uni_p)])
#define counter_increment(uni_p,c_p) (Incrementers[unigram_counter_size(uni_p)](c_p))
#define counter_get(uni_p,c_p)       (Getters[unigram_counter_size(uni_p)](c_p))
#define counter_set(uni_p,c_p,v)     (Setters[unigram_counter_size(uni_p)](c_p, v))

/**********************************************************************/


unigram_t *unigram_create(unsigned symbol_size)
{
  unigram_t *unigram;

  unigram = memory_spartan_create(sizeof(unigram_t));
  assert(unigram);
  unigram_initialize(unigram, symbol_size);
  assert(unigram_diversity(unigram) == 0);
  assert(unigram_symbol_size(unigram) == symbol_size);
  return(unigram);
}

void unigram_initialize(unigram_t *unigram, unsigned symbol_size)
{
  unsigned key_size, datum_size;
  assert(symbol_size > 0);
  key_size = symbol_size;
  datum_size = sizeof(min_counter_t);
  table_initialize(uni2tab(unigram), key_size, datum_size);
  unigram_trials(unigram) = 0;
  assert(unigram_trials(unigram) == 0);
  assert(unigram_is_valid(unigram));
}

void unigram_destroy(unigram_t *unigram)
{
  assert(unigram);
  unigram_finalize(unigram);
  memory_spartan_destroy(unigram, sizeof(unigram_t));
}

void unigram_finalize(unigram_t *unigram)
{
  assert(unigram);
  table_finalize(uni2tab(unigram));
}


void unigram_upsize(unigram_t *unigram)
{
  unsigned counter_size;

#if ANNOUNCE_UPSIZE
  fprintf(stderr, "unigram_upsize(): resizing counters from %u to %u bytes.\n",
	  unigram_counter_size(unigram), 2*unigram_counter_size(unigram));
#endif

  counter_size = unigram_counter_size(unigram);
  assert(counter_size <= sizeof(max_counter_t));
  if (counter_size == sizeof(max_counter_t))
    {
      fprintf(stderr, "unigram_t error: counter exceeds %d-byte range\n",
	      (int)sizeof(max_counter_t));
      abort();
    }
  table_resize_datum(uni2tab(unigram), 2*counter_size, counter_upsizer(unigram), NULL);
}

boolean_t unigram_verify_counters(const void *symbol_p, unsigned frequency, void *parameters)
{
  unsigned *total_p = (unsigned *)parameters;
  *total_p += frequency;
  return(TRUE);
}

boolean_t unigram_is_valid(const unigram_t *unigram)
{
  unsigned total;
  if (!table_is_valid(uni2tab(unigram))) 
    {
      fprintf(stderr, "unigram_is_valid(): invalid table\n");
      return(FALSE);
    }
  total = 0;
  unigram_map(unigram, &unigram_verify_counters, &total);
  assert(total >= unigram_diversity(unigram));
  if (total != unigram_trials(unigram))
    {
      fprintf(stderr, "unigram_is_valid(): trials %d != sum %d\n",
	      unigram_trials(unigram), total);
      return(FALSE);
    }
  return(TRUE);
}

unsigned unigram_frequency(const unigram_t *unigram, const void *symbol_p)
{
  void *counter_p;
  counter_p = table_lookup(uni2tab(unigram), symbol_p);
  if (counter_p == NULL) { return(0); }
  else { return(counter_get(unigram, counter_p)); }
}


typedef struct maximizer_tag {
  unsigned  maximum;
  unsigned  residue;
} maximizer_t;

boolean_t maximum_mapper(const void *symbol_p, unsigned frequency, void *parameters)
{ 
  maximizer_t *enclosure = parameters; 
  assert(frequency > 0); assert(enclosure->maximum < enclosure->residue); 
  if (frequency > enclosure->maximum) { enclosure->maximum = frequency; } 
  assert(enclosure->residue >= frequency);
  enclosure->residue -= (frequency - 1);
  return(enclosure->maximum < enclosure->residue);
}

unsigned unigram_maximum(const unigram_t *unigram)
{
  maximizer_t enclosure;
  unsigned    trials, diversity;

  trials = unigram_trials(unigram);
  diversity = unigram_diversity(unigram);
  if (trials == 0) { return(0); }
  enclosure.maximum = 1;
  enclosure.residue = trials + 1 - diversity;
  if (enclosure.residue <= 2) { return(enclosure.residue); }
  unigram_map(unigram, &maximum_mapper, &enclosure);
  assert((diversity == 0) ? (enclosure.maximum == 0) : (enclosure.maximum > 0));
  assert(enclosure.maximum <= trials + 1 - diversity);
  return(enclosure.maximum);
}

void unigram_set_frequency(unigram_t *unigram, const void *symbol_p, unsigned frequency)
{
  void *counter_p;

  counter_p = table_lookup(uni2tab(unigram), symbol_p);
  if (counter_p) 
    { 
      assert(unigram_trials(unigram) >= counter_get(unigram, counter_p));
      unigram_trials(unigram) -= counter_get(unigram, counter_p); 
    }
  if (frequency == 0)
    {
      table_remove(uni2tab(unigram), symbol_p);
      /* Note: opportunity to downsize counters here! */
    }
  else
    {
      max_counter_t  zero;

      while(frequency > counter_limit(unigram))
	{ unigram_upsize(unigram); }
      assert(frequency <= counter_limit(unigram));
      assert(unigram_trials(unigram) <= (unsigned)UINT_MAX - frequency);

      zero = 0;
      counter_p = table_insert(uni2tab(unigram), symbol_p, &zero);
      counter_set(unigram, counter_p, frequency);
      unigram_trials(unigram) += frequency;
    }
  assert(unigram_frequency(unigram, symbol_p) == frequency);
  assert(unigram_trials(unigram) >= unigram_diversity(unigram));
}

void unigram_increment_frequency(unigram_t *unigram, const void *symbol_p)
{
  void *counter_p;

  counter_p = table_lookup(uni2tab(unigram), symbol_p);
  if (counter_p == NULL)
    {
      max_counter_t zero;
      zero = 0;
      counter_p = table_insert(uni2tab(unigram), symbol_p, &zero);
    }
  else
    {
      unsigned frequency;
      frequency = counter_get(unigram,counter_p);
      if (frequency == counter_limit(unigram))
	{
	  unigram_upsize(unigram);
	  counter_p = table_lookup(uni2tab(unigram), symbol_p);
	  assert(counter_p != NULL);
	  assert(counter_get(unigram,counter_p) == frequency);
	}
      assert(frequency < counter_limit(unigram));
    }
  assert(unigram_trials(unigram) < UINT_MAX);  /* else trial counter overflow */
  counter_increment(unigram,counter_p);
  unigram_trials(unigram)++;
}

void unigram_clear(unigram_t *unigram)
{
  unigram_trials(unigram) = 0;
  table_clear(uni2tab(unigram));
  table_resize_datum(uni2tab(unigram), sizeof(min_counter_t), NULL, NULL);
  assert(unigram_trials(unigram) == 0);
  assert(unigram_diversity(unigram) == 0);
  assert(unigram_counter_size(unigram) == sizeof(min_counter_t));
}


typedef struct enclosure_tag
{
  boolean_t      (*proc)(const void *, unsigned, void *);
  void            *parameters;
  const unigram_t *unigram;
} enclosure_t;

boolean_t unigram_mapper(const void *key_p, void *datum_p, void *parameters)
{
  enclosure_t *enclosure_p = (enclosure_t *)parameters;
  unsigned     frequency;
  boolean_t    is_alive;

  frequency = counter_get(enclosure_p->unigram, datum_p);
  assert(frequency > 0);

  is_alive = enclosure_p->proc(key_p, frequency, enclosure_p->parameters);
  return(is_alive);
}

void unigram_map(const unigram_t *unigram, 
		 boolean_t (*proc)(const void *symbol_p, unsigned frequency, 
				   void *parameters),
		 void *parameters)
{
  enclosure_t enclosure;

  enclosure.proc = proc;
  enclosure.unigram = unigram;
  enclosure.parameters = parameters;
  table_map(uni2tab(unigram), &unigram_mapper, &enclosure);
}


