/*
 * (c) 1998 Eric Sven Ristad.  All rights reserved.
 * 
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "memory.h"
#include "unigram.h"

static char unigram_test_rcsid[] = "$Id: unigram.test.c,v 1.6 1998/02/07 03:46:35 libpa Exp $";

#define DEBUG                 FALSE
#define BE_VERBOSE            FALSE
#define MAX_SYMBOL            UCHAR_MAX
#define OBJECT_FILE           "unigram.test"
#define TEXT_FILE             "unigram.test.c"
#define WIERD_FILE            "libunigram.a"
#define BYTE_LIMIT            100000

boolean_t verify_counts(const void *symbol_p, unsigned frequency, void *parameters)
{
  unsigned *vector;
  unsigned char symbol;

  vector = (unsigned *)parameters;
  symbol = *((unsigned char *)symbol_p);
  assert(frequency > 0);
  assert(symbol <= MAX_SYMBOL);
  assert(vector[symbol] == frequency);
  return(TRUE);
}

/* Effect: verifies unigram_t module on given file */
void verify_file_counts(char *filename)
{
  unsigned initial_allocation;
  unsigned vector[MAX_SYMBOL+1], n, count, maximum;
  unigram_t *unigram;
  FILE *stream;
  unsigned char symbol;
  int        i, q, byte;

  initial_allocation = memory_total_bytes();

  /* create empty unigram_t */
  unigram = unigram_create(sizeof(char));
  assert(unigram);
  assert(unigram_trials(unigram) == 0);
  assert(unigram_diversity(unigram) == 0);
  assert(unigram_counter_size(unigram) == 1);
  assert(unigram_symbol_size(unigram) == sizeof(char));

  /* initialize vector[] of frequencies to zero */
  for(i = 0; i <= MAX_SYMBOL; i++) { vector[i] = 0; }

  /* open stream to specified file */
  stream = fopen(filename, "r");
  assert(stream);

  q = 0;
  n = 0;
  maximum = 0;
  while((byte = fgetc(stream)) != EOF)
    {
      symbol = byte;
      assert(symbol <= MAX_SYMBOL); 
      
      assert(unigram_frequency(unigram, &symbol) == vector[byte]);

      if (vector[byte] == 0) { q++; }
      vector[byte]++;
      count = vector[byte];
      n++;

      if (count > maximum) { maximum = count; }

      unigram_increment_frequency(unigram, &symbol);
      assert(unigram_frequency(unigram, &symbol) == vector[byte]);
      assert(unigram_diversity(unigram) == q);
      assert(unigram_trials(unigram) == n);
      assert(unigram_maximum(unigram) == maximum);

      unigram_set_frequency(unigram, &symbol, 0);
      assert(unigram_frequency(unigram, &symbol) == 0);
      assert(unigram_diversity(unigram) == q-1);
      assert(unigram_trials(unigram) == n - count);

      unigram_set_frequency(unigram, &symbol, count);
      assert(unigram_frequency(unigram, &symbol) == vector[byte]);
      assert(unigram_diversity(unigram) == q);
      assert(unigram_trials(unigram) == n);
      assert(count == vector[byte]);

      if (n >= BYTE_LIMIT) { break; }
    }
  fclose(stream);
  stream = NULL;
  assert(unigram_is_valid(unigram));
  assert(unigram_maximum(unigram) == maximum);

  /* verify file counts both ways */
  unigram_map(unigram, &verify_counts, vector);
  for(i = 0; i <= MAX_SYMBOL; i++) 
    {
      symbol = i;
      assert(vector[i] == unigram_frequency(unigram, &symbol));
    }
  printf("%u\tunigram for %s has %u symbols and %u-byte counters\n", 
	  memory_total_bytes(), filename, unigram_diversity(unigram),
	  unigram_counter_size(unigram));

  /* clear file unigram */
  unigram_clear(unigram);
  assert(unigram_trials(unigram) == 0);
  assert(unigram_diversity(unigram) == 0);
  assert(unigram_is_valid(unigram));
#if BE_VERBOSE
  printf("%u\tcleared %s unigram\n", memory_total_bytes(), filename);
#endif

  /* destroy file unigram */
  unigram_destroy(unigram);
  unigram = NULL;
#if BE_VERBOSE
  printf("%u\tdestroyed %s unigram\n", memory_total_bytes(), filename);
#endif
  assert(memory_total_bytes() == initial_allocation);
}

int main(void)
{
  unigram_t *unigram;
  unsigned char symbol;

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("\n\tTesting unigram_t library...\n");

  /* allocate and initilize our first unigram_t */
  unigram = unigram_create(sizeof(char));
  assert(unigram);
  assert(unigram_symbol_size(unigram) == sizeof(char));
  assert(unigram_diversity(unigram) == 0);
  assert(unigram_trials(unigram) == 0);
  printf("%u\tcreated empty unigram with %u-byte counters\n", 
	  memory_total_bytes(), unigram_counter_size(unigram));

  /* test minimum frequency increment */
  symbol = 0;
  unigram_set_frequency(unigram, &symbol, 0);
  assert(unigram_trials(unigram) == 0);
  assert(unigram_frequency(unigram, &symbol) == 0);
  assert(unigram_diversity(unigram) == 0);
  assert(unigram_maximum(unigram) == 0);
  printf("%u\tset symbol frequency to minimum (%u-byte counters)\n",
	  memory_total_bytes(), unigram_counter_size(unigram));

  /* test maximum frequency increment (1 bytes to 4 bytes) */
  symbol = 0;
  unigram_set_frequency(unigram, &symbol, UINT_MAX);
  assert(unigram_trials(unigram) == UINT_MAX);
  assert(unigram_frequency(unigram, &symbol) == UINT_MAX);
  assert(unigram_diversity(unigram) == 1);
  assert(unigram_maximum(unigram) == UINT_MAX);
  printf("%u\tset symbol frequency to maximum (%u-byte counters)\n",
	  memory_total_bytes(), unigram_counter_size(unigram));

  unigram_set_frequency(unigram, &symbol, 0);
  assert(unigram_trials(unigram) == 0);
  assert(unigram_frequency(unigram, &symbol) == 0);
  assert(unigram_diversity(unigram) == 0);
  assert(unigram_maximum(unigram) == 0);
  assert(unigram_counter_size(unigram) == sizeof(unsigned int));
  unigram_clear(unigram);
  assert(unigram_trials(unigram) == 0);
  assert(unigram_diversity(unigram) == 0);
  assert(unigram_maximum(unigram) == 0);
  assert(unigram_counter_size(unigram) == sizeof(unsigned char));
  printf("%u\tcleared jumping unigram (%u-byte counters)\n", 
	  memory_total_bytes(), unigram_counter_size(unigram));

  unigram_destroy(unigram);
  unigram = NULL;
  printf("%u\tdestroyed jumping unigram\n", memory_total_bytes());
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);

  verify_file_counts(OBJECT_FILE);   /* object code file (q=256) */
  verify_file_counts(TEXT_FILE);     /* text file (q=84)         */
  verify_file_counts(WIERD_FILE);    /* archive (q=249)          */

  printf("\tmemory leak == %u bytes in %u blocks\n", 
	  memory_total_bytes(), memory_total_blocks());
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  printf("\tsizeof(unigram_t) == %d bytes (with no overhead!)\n", (int)sizeof(unigram_t));
  printf("\tunigram_t library passes all tests.\n");
  return(0);
}



