#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <limits.h>
#include <netinet/in.h>  /* htonl() and ntohl() */
#include "common.h"
#include "memory.spartan.h"
#include "vector.h"
#include "vector.private.h"

extern double drand48(void);

#define DEBUG              FALSE
#define ANNOUNCE_ERRORS    TRUE
#define ANNOUNCE_WARNINGS  TRUE

/************************************************************
 * INTERNAL PROTOTYPES
 ************************************************************/
/* WARNING: public use of the following resize() operations is discouraged  */
/*          because it may result in uninitialized vector elements.         */
static vector_t *vector_strict_resize(vector_t *vector, unsigned new_length);
                 /* Effect: reallocates vector to exactly new_length entries        */
                 /*         without any surplus entries.                            */
static vector_t *vector_contracting_resize(vector_t *vector, unsigned new_length);
                 /* Effect: reallocates vector to contain new_length entries        */
                 /*         and at most VECTOR_MAXSURPLUS surplus entries           */
static vector_t *vector_expanding_resize(vector_t *vector, unsigned new_length);
                 /* Effect: reallocates vector to contain new_length entries        */
                 /*         and at most VECTOR_MAXSURPLUS surplus entries           */


/************************************************************
 * CONSTRUCTORS/DESTRUCTOR
 ************************************************************/

vector_t *vector_create(unsigned object_size, unsigned length)
{
  vector_header_t *header;
  vector_t        *vector;
  unsigned         real_length, maximum_real_length, bytes;

  if (object_size > VECTOR_OBJECT_SIZE_MAX) 
    { 
#if ANNOUNCE_ERRORS
      fprintf(stderr, "vector_t error: object size %u exceeds limit %u\n",
	      object_size, (unsigned)VECTOR_OBJECT_SIZE_MAX);
#endif
      return(NULL);
    }

  real_length = length_initial(length);
  maximum_real_length = length+VECTOR_SURPLUS_LENGTH_MAX;
  real_length = MIN(real_length, maximum_real_length);
  assert(real_length >= length);
  assert(real_length - length <= VECTOR_SURPLUS_LENGTH_MAX);
  bytes = sizeof(vector_header_t) + real_length*object_size;
  assert(bytes >= sizeof(vector_header_t));
  header = memory_spartan_create(bytes);
  if (header == NULL) { return NULL; }  /* invalid vector_t */

  header->length = length;
  header->surplus_length = real_length - length;
  header->object_size = object_size;

#if DEBUG
  printf("%u bytes for vector_create(%u,%u) with %u surplus slots\n", 
	 bytes, object_size, length, real_length - length);
#endif

  vector = header2vector(header);
  assert(vector_header(vector) == header);
  assert(vector_is_valid(vector));
  assert(vector_length(vector) == length);
  assert(vector_object_size(vector) == object_size);
  assert(vector_surplus_length(vector) == real_length - length);
  assert(vector_allocated_length(vector) == real_length);
  assert(vector_bytes(vector) == bytes);

  return(vector);
}

vector_t *vector_create_surplus(unsigned object_size, unsigned length, unsigned surplus)
{
  vector_t *vector;

  if (surplus > VECTOR_SURPLUS_LENGTH_MAX)
    { surplus = VECTOR_SURPLUS_LENGTH_MAX; }
  if (length > VECTOR_LENGTH_MAX - surplus)
    { surplus = VECTOR_LENGTH_MAX - length; }

  vector = vector_create(object_size, length+surplus);
  assert(vector_length(vector) == length+surplus);
  vector = vector_contract(vector, surplus);
  assert(vector_length(vector) == length);
  assert(vector_surplus_length(vector) == surplus);
  return(vector);
}

void vector_destroy(vector_t *vector)
{
  vector_header_t *header;
  unsigned         bytes;

  assert(vector_is_valid(vector));

  bytes = vector_bytes(vector);
  header = vector_header(vector);
  /* invalidate header for extra safety */
  header->length = 0;
  header->surplus_length = 0;
  header->object_size = 0;
  memory_spartan_destroy(header, bytes);
}

vector_t *vector_copy(const vector_t *original)
{
  vector_t *copy;
  unsigned  length, object_size;

  assert(vector_is_valid(original));

  length = vector_length(original);
  object_size = vector_object_size(original);
  copy = vector_create(object_size, length);
  if (copy == NULL) { return(NULL); }
  vector_mirror(original, copy);

  assert(copy);
  assert(vector_surplus_length(copy) == 0);
  assert(vector_are_equivalent(original, copy));

  return(copy);
}

void vector_mirror(const vector_t *source, vector_t *target)
{
  unsigned length, object_size;

  assert(source != target);
  assert(vector_is_valid(source));
  assert(vector_is_valid(target));
  assert(vector_object_size(source) == vector_object_size(target));
  assert(vector_length(source) == vector_length(target));

  length = vector_length(source);
  object_size = vector_object_size(source);
  (void) memcpy(target, source, length * object_size);

  assert(vector_are_equivalent(source, target));
  assert(vector_parity(source) == vector_parity(target));
}

/************************************************************
 * 
 ************************************************************/

unsigned vector_parity(const vector_t *vector)
{
  unsigned length, object_size, byte_count;
  const unsigned int  *contents_p;
  const unsigned char *remainder_p;
  unsigned parity_value;
  unsigned char *parity_bytes;

  assert(vector_is_valid(vector));
  length = vector_length(vector);
  object_size = vector_object_size(vector);
  byte_count = length * object_size;

  contents_p = vector;
  parity_value = 0;
  while(byte_count >= sizeof(unsigned))
    {
      parity_value = parity_value ^ *contents_p;
      contents_p++;
      byte_count -= sizeof(unsigned);
    }
  assert(byte_count < sizeof(unsigned));

  parity_bytes = (unsigned char *) &parity_value;
  remainder_p = (const unsigned char *) contents_p;
  while(byte_count > 0)
    {
      byte_count--;
      parity_bytes[byte_count] = parity_bytes[byte_count] ^ *remainder_p;
      remainder_p++;
    }

  return(parity_value);
}

int vector_compare(const vector_t *left, const vector_t *right)
{
  unsigned  left_length, right_length, left_object_size, right_object_size, bytes;
  long      delta;
  int       comparison;

  assert(vector_is_valid(left));
  assert(vector_is_valid(right));

  /* primary order is by object size */
  left_object_size = vector_object_size(left);
  right_object_size = vector_object_size(right);
  delta = (long)left_object_size - right_object_size;
  if (delta != 0) { return(delta); }
  assert(left_object_size == right_object_size);

  /* secondary order is by number of elements (length) */
  left_length = vector_length(left);
  right_length = vector_length(right);
  delta = (long)left_length - right_length;
  if (delta != 0) { return(delta); }
  assert(left_length == right_length);

  /* tertiary order is by contents of vector */
  bytes = left_object_size * left_length;
  assert(bytes == right_object_size * right_length);
  comparison = memcmp(left, right, bytes);
  return(comparison);
}

/************************************************************
 * 
 ************************************************************/

void vector_set(vector_t *vector, unsigned index, const void *object_p)
{
  unsigned length, object_size;
  void *element_p;

  assert(object_p != NULL);
  assert(vector_is_valid(vector));
  assert(vector_index_is_valid(vector, index));

  length = vector_length(vector);
  object_size = vector_object_size(vector);
  element_p = vector_element_p(vector, index);
  memory_move(element_p, object_p, object_size);
}


vector_t *vector_append(vector_t *vector, const void *object_p)
{
  unsigned length, object_size;

  assert(object_p != NULL);
  assert(vector_is_valid(vector));

  length = vector_length(vector);
  object_size = vector_object_size(vector);

  if (length == VECTOR_LENGTH_MAX) { vector_destroy(vector); return(NULL); }
  vector = vector_expanding_resize(vector, length+1);
  if (vector == NULL) { return(NULL); }
  assert(vector_is_valid(vector));
  assert(vector_length(vector) == length+1);
  assert(vector_index_is_valid(vector,length));

  (void) memcpy(vector_element_p(vector,length), object_p, object_size);
  return(vector);
}

vector_t *vector_insert(vector_t *vector, unsigned index, const void *object_p)
{
  unsigned object_size, length;

  assert(object_p != NULL);
  assert(vector_is_valid(vector));
  assert(index <= vector_length(vector));

  length = vector_length(vector);
  object_size = vector_object_size(vector);

  if (length == VECTOR_LENGTH_MAX) { vector_destroy(vector); return(NULL); }
  vector = vector_expanding_resize(vector, length+1);
  if (vector == NULL) { return(NULL); }

  if (index < length)
    {
      unsigned displaced_length;
      displaced_length = length - index;
      (void) memory_move(vector_element_p(vector,index+1), 
			 vector_element_p(vector,index),
			 displaced_length*object_size);
    }
  (void) memcpy(vector_element_p(vector,index), object_p, object_size);
  return(vector);
}

vector_t *vector_delete(vector_t *vector, unsigned index)
{
  unsigned length, suffix_length, object_size;

  assert(vector_is_valid(vector));
  assert(vector_length(vector) > 0);
  assert(vector_index_is_valid(vector, index));
	
  /* overwrite the deleted element using overlapping move */
  length = vector_length(vector);
  object_size  = vector_object_size(vector);
  suffix_length = length - (index + 1);
  (void) memory_move(vector_element_p(vector,index), 
		     vector_element_p(vector,index+1),
		     suffix_length*object_size);

  /* resize the vector to eliminate the last element */
  vector = vector_contracting_resize(vector, length-1);
  assert(vector_is_valid(vector));
  assert(vector_length(vector) == length-1);

  return(vector);
}

/************************************************************
 *
 ************************************************************/

vector_t *vector_resize(vector_t *vector, unsigned length)
{
  assert(vector_is_valid(vector));
  return(vector_strict_resize(vector, length));
}

vector_t *vector_contract(vector_t *vector, unsigned contraction)
{
  unsigned length;
  assert(vector_is_valid(vector));
  length = vector_length(vector);
  assert(length >= contraction);
  vector = vector_contracting_resize(vector, length - contraction);
  if (vector == NULL) { return(NULL); }
  assert((contraction == 0) || vector_surplus_length(vector) > 0);
  assert(vector_is_valid(vector));
  assert(vector_length(vector) == length - contraction);
  return(vector);
}

vector_t *vector_expand(vector_t *vector, unsigned expansion)
{
  unsigned length;
  assert(vector_is_valid(vector));
  length = vector_length(vector);
  assert((VECTOR_LENGTH_MAX - expansion) >= length);
  vector = vector_expanding_resize(vector, length + expansion);
  if (vector == NULL) { return(NULL); }
  assert(vector_is_valid(vector));
  assert(vector_length(vector) == length + expansion);
  return(vector);
}


vector_t *vector_insert_vector(vector_t *vector, unsigned index, const vector_t *subvector)
{
  unsigned length, sublength, displaced_length, object_size;

  assert(vector_is_valid(vector)); 
  assert(vector_is_valid(subvector));
  assert(vector_object_size(vector) == vector_object_size(vector));
  assert(index <= vector_length(vector));

  length = vector_length(vector);
  sublength = vector_length(subvector);
  object_size = vector_object_size(vector);

  if (sublength == 0) { return(vector); }
  if (length > VECTOR_LENGTH_MAX - sublength) { vector_destroy(vector); return(NULL); }
  vector = vector_expanding_resize(vector, length + sublength);
  if (vector == NULL) { return(NULL); }
  assert(vector_length(vector) == length + sublength);

  displaced_length = length - index;  /* number of displaced elements */
  (void) memory_move(vector_element_p(vector,index+sublength), 
		     vector_element_p(vector,index),
		     displaced_length*object_size);
  (void) memcpy(vector_element_p(vector,index), 
		subvector, 
		sublength*object_size);
  assert(vector_is_valid(vector));
  return(vector);
}

vector_t *vector_concatenate(vector_t *root, const vector_t *suffix)
{
  unsigned root_length, suffix_length, object_size;

  assert(root != suffix);
  assert(vector_is_valid(root));
  assert(vector_is_valid(suffix));
  assert(vector_object_size(root) == vector_object_size(suffix));
  
  root_length = vector_length(root);
  suffix_length = vector_length(suffix);
  if (suffix_length == 0) { return(root); }

  root = vector_expanding_resize(root, root_length + suffix_length);
  if (root == NULL) { return(NULL); }  /* failure! */
  object_size = vector_object_size(root);
  (void) memcpy(vector_element_p(root,root_length), suffix, suffix_length*object_size);
  assert(vector_is_valid(root));
  assert(vector_length(root) == root_length+suffix_length);
  assert(memcmp(vector_element_p(root,root_length), suffix, suffix_length*object_size) == 0);
  return(root);
}

vector_t *vector_delete_subvector(vector_t *vector, unsigned index, unsigned sublength)
{
  unsigned length, object_size, displaced_length;
	
  assert(vector_is_valid(vector));
  assert(vector_index_is_valid(vector, index));
  assert(index+sublength <= vector_length(vector));

  /* overwrite deleted subvector using overlapping move */
  length = vector_length(vector);
  object_size = vector_object_size(vector);
  displaced_length = length - (index + sublength);
  (void) memory_move(vector_element_p(vector,index), 
		     vector_element_p(vector,index+sublength),
		     displaced_length*object_size);

  /* resize the vector to eliminate the last sublength elements */
  vector = vector_contracting_resize(vector, length-sublength);
  assert(vector_is_valid(vector));
  assert(vector_length(vector) == length-sublength);

  return(vector);
}

vector_t *vector_prune(vector_t *vector, 
		       boolean_t (*prune_me)(const void *element_p, void *parameters),
		       void *parameters)
{
  unsigned object_size, length, deleted_count;
  unsigned char *left_p, *right_p;

  assert(vector_is_valid(vector));
  assert(prune_me != NULL);

  length = vector_length(vector);
  object_size = vector_object_size(vector);
  assert(object_size > 0);

  deleted_count = 0;
  left_p = vector_element_p(vector, 0);
  right_p = left_p;
  while(length > 0)
    {
      assert((unsigned char *)vector <= left_p); assert(left_p <= right_p);
      if (prune_me(right_p, parameters))
	{ 
	  right_p += object_size;
	  deleted_count++;
	}
      else
	{
	  if (left_p != right_p) { memcpy(left_p, right_p, object_size); }
	  left_p += object_size;
	  right_p += object_size;
	}
      length--;
    }
  assert(deleted_count <= vector_length(vector));
  vector = vector_contract(vector, deleted_count);
  return(vector);
}

/************************************************************
 *
 ************************************************************/

static unsigned char *vector_quickpartition(unsigned char *lower_p, 
					    unsigned char *upper_p, 
					    void *spare_element_p, 
					    unsigned object_size,
					    int (*compare)(const void *left_p, 
							   const void *right_p, 
							   const void *parameters),
					    const void *parameters)
{
  unsigned char *pivot_element_p;
  unsigned span, pivot_index;
#ifndef NDEBUG
  unsigned char *hard_lower_p, *hard_upper_p;
  hard_lower_p = lower_p;
  hard_upper_p = upper_p;
#endif

  assert(object_size > 0);
  assert(lower_p < upper_p);
  assert((upper_p - lower_p) >= object_size);
  assert(((upper_p - lower_p)%object_size) == 0);

  /* select random pivot element */
  span =  (upper_p - lower_p) / object_size;
  pivot_index = drand48() * (double)(span+1);
  assert(pivot_index >= 0); assert(pivot_index <= span);
  pivot_element_p = lower_p + (pivot_index * object_size);

  /* repeatedly swap around the pivot element */
  while(1)
    {
      assert(lower_p <= hard_upper_p);
      assert(upper_p >= hard_lower_p);
      assert(hard_lower_p <= pivot_element_p);
      assert(pivot_element_p <= hard_upper_p);

      while (compare(upper_p, pivot_element_p, parameters) > 0)
	{ upper_p -= object_size; assert(upper_p >= hard_lower_p); }
      while (compare(lower_p, pivot_element_p, parameters) < 0)
	{ lower_p += object_size; assert(lower_p <= hard_upper_p); }

      if (lower_p >= upper_p) { return(upper_p); }

      /* Danger! any swap can change location of the pivot element */
      assert(lower_p < upper_p);
      if (lower_p == pivot_element_p) { pivot_element_p = upper_p; }
      else if (upper_p == pivot_element_p) { pivot_element_p = lower_p; }

      /* actually perform the exchange using spare_element_p */
      memcpy(spare_element_p, lower_p, object_size);
      memcpy(lower_p, upper_p, object_size);
      memcpy(upper_p, spare_element_p, object_size);

      lower_p += object_size;
      upper_p -= object_size;
      assert(lower_p <= hard_upper_p);
      assert(upper_p >= hard_lower_p);
    }
}


static void vector_internal_quicksort(unsigned char *lower_p, 
				      unsigned char *upper_p,
				      void *spare_element_p, 
				      unsigned object_size,
				      int (*compare)(const void *left_p, 
						     const void *right_p, 
						     const void *parameters),
				      const void *parameters)
{
  unsigned char *middle_p;

  assert(object_size > 0);
  assert(lower_p <= upper_p+object_size);
  assert(((upper_p - lower_p)%object_size) == 0);

  if (lower_p >= upper_p) { return; }
  assert(lower_p < upper_p);

  if (lower_p + object_size == upper_p)
    {
      if (compare(lower_p, upper_p, parameters) > 0)
	{
	  memcpy(spare_element_p, lower_p, object_size);
	  memcpy(lower_p, upper_p, object_size);
	  memcpy(upper_p, spare_element_p, object_size);
	}
      assert(compare(lower_p, upper_p, parameters) <= 0);
      return;
    }

  middle_p = vector_quickpartition(lower_p, upper_p, spare_element_p, 
				   object_size, compare, parameters);
  assert(lower_p <= middle_p); assert(middle_p <= upper_p);
  vector_internal_quicksort(lower_p, middle_p, spare_element_p, 
			    object_size, compare, parameters);
  vector_internal_quicksort(middle_p+object_size, upper_p, spare_element_p, 
			    object_size, compare, parameters);
}

void vector_quicksort(vector_t *vector,
		      int (*compare)(const void *, const void *, const void *),
		      const void *parameters)
{
  unsigned length;
  assert(vector_is_valid(vector));
  assert(compare != NULL);
  length = vector_length(vector);
  if (length > 1) { vector_quicksubsort(vector, 0, length-1, compare, parameters); }
  assert(vector_is_sorted(vector, compare, parameters));
}

void vector_quicksubsort(vector_t *vector, unsigned start, unsigned stop,
			 int (*compare)(const void *, const void *, const void *),
			 const void *parameters)
{
  unsigned char *lower_p, *upper_p;
  unsigned object_size, length;
  void *spare_element_p;
  boolean_t spare_is_allocated;

  if (start >= stop) { return; }
  assert(vector_is_valid(vector));
  assert(start < stop); assert(stop < vector_length(vector));

  /* allocate the spare_element_p for swapping elements */
  spare_element_p = NULL;
  object_size = vector_object_size(vector);
  if (vector_surplus_length(vector) > 0)
    {
      length = vector_length(vector);
      assert(!vector_index_is_valid(vector, length));
      spare_element_p = vector_element_p(vector, length);
      spare_is_allocated = FALSE;
    }
  else
    {
      spare_element_p = memory_spartan_create(object_size);
      spare_is_allocated = TRUE;
      assert(spare_element_p);
    }
  assert(spare_element_p);
  assert(vector_index_is_valid(vector, start));
  assert(vector_index_is_valid(vector, stop));

  lower_p = vector_element_p(vector, start);
  upper_p = vector_element_p(vector, stop);
  vector_internal_quicksort(lower_p, upper_p, 
			    spare_element_p, object_size,
			    compare, parameters);
  if (spare_is_allocated) 
    { 
      memory_spartan_destroy(spare_element_p, object_size); 
      spare_element_p = NULL;
    }
}

boolean_t vector_is_sorted(const vector_t *vector,
			   int (*compare)(const void *left_p, const void *right_p, 
					  const void *parameters),
			   const void *parameters)
{
  unsigned object_size, length;
  unsigned char *left_p, *right_p;

  assert(vector_is_valid(vector));

  length = vector_length(vector);
  object_size = vector_object_size(vector);
  assert(object_size > 0);

  left_p = vector_element_p(vector, 0);
  right_p = left_p + object_size;
  assert(right_p == vector_element_p(vector, 1));
  while(length > 1)
    {
      assert(right_p - left_p == object_size);
      if (compare(left_p, right_p, parameters) > 0) { return(FALSE); }
      left_p = right_p;
      right_p += object_size;
      length--;
    }
  return(TRUE);
}

unsigned vector_sorted_lookup(const vector_t *vector, const void *object_p,
			      int (*compare)(const void *left_p, const void *right_p,
					     const void *parameters),
			      const void *parameters)
{
  unsigned object_size;
  unsigned lower, middle, upper;
  unsigned char *element_p;
  unsigned length;
  int comparison;

  assert(object_p != NULL);
  assert(vector_is_valid(vector));
  assert(vector_is_sorted(vector, compare, parameters));

  length = vector_length(vector);
  if (length == 0) { return(0); }

  object_size = vector_object_size(vector);
  lower = 0;
  upper = length;
  do
    {
      middle = lower + ((upper - lower) / 2);
      assert(lower <= middle);
      assert(middle < upper);
      assert(vector_index_is_valid(vector, middle));

      element_p = vector_element_p(vector, middle);
      comparison = compare(object_p, element_p, parameters);
      if (comparison < 0) { upper = middle; }
      else if (comparison > 0) { lower = middle + 1; }
      else { return(middle); }  /* found object_p exactly! */
    } while(lower < upper);
  assert(lower == upper);
  assert((lower == 0) || 
	 (compare(vector_element_p(vector,lower-1), object_p, parameters) <= 0));
  assert((lower == vector_length(vector)) || 
	 (compare(object_p, vector_element_p(vector,lower), parameters) <= 0));
  return(lower);
}

vector_t *vector_sorted_insert(vector_t *vector, const void *object_p,
			       int (*compare)(const void *left_p, const void *right_p, 
					      const void *parameters),
			       const void *parameters)
{
  unsigned index, length;

  assert(object_p != NULL);
  assert(vector_is_valid(vector));
  length = vector_length(vector);
  index = vector_sorted_lookup(vector, object_p, compare, parameters);
  assert(index <= length);
  vector = vector_insert(vector, index, object_p);
  if (vector == NULL) { return(NULL); }
  assert(vector_is_sorted(vector, compare, parameters));
  return(vector);
}

vector_t *vector_sorted_unique(vector_t *vector, 
			       int (*compare)(const void *left_p, const void *right_p, 
					      const void *parameters),
			       const void *parameters)
{
  unsigned object_size, length, deleted_count;
  unsigned char *left_p, *right_p;

  assert(vector_is_valid(vector));
  assert(vector_is_sorted(vector, compare, parameters));

  length = vector_length(vector);
  object_size = vector_object_size(vector);
  assert(object_size > 0);

  deleted_count = 0;
  left_p = vector_element_p(vector, 0);
  right_p = left_p + object_size;
  assert(right_p == vector_element_p(vector, 1));
  while(length > 1)
    {
      assert((unsigned char *)vector <= left_p); assert(left_p < right_p);
      if (compare(left_p, right_p, parameters) == 0) 
	{ 
	  assert(compare(right_p - object_size, right_p, parameters) == 0);
	  right_p += object_size;
	  deleted_count++;
	}
      else
	{
	  assert(compare(right_p - object_size, right_p, parameters) != 0);
	  left_p += object_size;
	  if (left_p != right_p) { memcpy(left_p, right_p, object_size); }
	  right_p += object_size;
	}
      length--;
    }
  assert(((vector_length(vector) == 0) && (deleted_count == 0)) || 
	 (deleted_count < vector_length(vector)));
  vector = vector_contract(vector, deleted_count);
  return(vector);
}


#if FALSE
vector_t *vector_sorted_merge(const vector_t *left, const vector_t *right,
			      int (*compare)(const void *left_p, const void *right_p,
					     const void *parameters),
			      const void *parameters);
          /* Returns: sorted concatentation of two sorted vector_t */

vector_t *vector_sorted_merge(const vector_t *left, const vector_t *right,
			      int (*compare)(const void *left_p, const void *right_p,
					     const void *parameters),
			      const void *parameters)
{
  unsigned object_size, left_length, right_length, length, index;
  unsigned char *left_p, *right_p;
  unsigned char *left_limit_p, *right_limit_p;
  vector_t *merged;

  assert(vector_is_valid(left));
  assert(vector_is_valid(right));

  object_size = vector_object_size(left);
  assert(object_size == vector_object_size(right));
  left_length = vector_length(left);
  right_length = vector_length(right);
  if (right_length > VECTOR_LENGTH_MAX - left_length)
    { 
#if ANNOUNCE_ERRORS
      fprintf(stderr, "vector_t error: two vectors are too long to be merged\n");
#endif
      return(NULL);
    }
  length = left_length + right_length;
  
  merged = vector_create(object_size, length);
  assert(vector_is_valid(merged));

  index = 0;
  left_p = vector_element_p(left, left_index);
  right_p = vector_element_p(right, right_index);
  left_limit_p = left_p + (object_size * left_length);
  right_limit_p = right_p + (object_size * right_length);
  while((left_p < left_limit_p) && (right_p < right_limit_p))
    {
      comparison = compare(left_p, right_p, parameters);
      if (comparison <= 0)
	{
	  vector_set(merged, index, left_p);
	  index++;
	  left_p += object_size;
	}
      if (comparison >= 0)
	{
	  vector_set(merged, index, right_p);
	  index++;
	  right_p += object_size;
	}
    }
  assert((left_p == left_limit_p) || (right_p == right_limit_p));
  while(left_p < left_limit_p) 
    { 
      vector_set(merged, index, left_p);
      index++;
      left_p += object_size;
    }
  while(right_p < right_limit_p) 
    { 
      vector_set(merged, index, right_p);
      index++;
      right_p += object_size;
    }
  assert(vector_is_valid(merged));
  assert(index == vector_length(merged));
  assert(vector_is_sorted(merged, compare, parameters));
#if EXTRA_SAFE
  {
    void *union;
    union = vector_concatenate(left, right);
    assert(vector_length(union) == vector_length(merged));
    assert(vector_are_permuted(merged, union));
    vector_destroy(union);
  }
#endif
}

#endif

/************************************************************
 *
 ************************************************************/

void vector_reverse(vector_t *vector)
{
  boolean_t buffer_is_allocated;
  unsigned length, object_size;
  void *buffer_p;
  unsigned i, j;

  assert(vector_is_valid(vector));

  length = vector_length(vector);
  if (length < 2) { return; } /* identity operation */

  object_size = vector_object_size(vector);
  if (vector_surplus_length(vector) == 0)
    {
      buffer_p = memory_spartan_create(object_size);
      buffer_is_allocated = TRUE;
    }
  else
    {
      assert(!vector_index_is_valid(vector, length));
      buffer_p = vector_element_p(vector, length);
      buffer_is_allocated = FALSE;
    }
  assert(buffer_p != NULL);

  for (i=0, j=length-1; i<j; i++, j--)
    {
      memcpy(buffer_p, vector_element_p(vector,i), object_size);   
      memcpy(vector_element_p(vector,i), vector_element_p(vector,j), object_size); 
      memcpy(vector_element_p(vector,j), buffer_p, object_size);
    }
  if (buffer_is_allocated) 
    { memory_spartan_destroy(buffer_p, object_size); }
}

boolean_t vector_are_reversed(const vector_t *left, const vector_t *right)
{
  unsigned length, object_size;
  unsigned char *left_p, *right_p;

  assert(vector_is_valid(left));
  assert(vector_is_valid(right));

  length = vector_length(left);
  object_size = vector_object_size(right);

  if ((vector_length(right) != length) ||
      (vector_object_size(right) != object_size))
    { return(FALSE); }
  else if (length == 0) { return(TRUE); }

  assert(length > 0);
  left_p = vector_element_p(left, 0);
  right_p = vector_element_p(left, length-1);
  while(right_p > (unsigned char *)right)
    {
      if (memcmp(left_p, right_p, object_size) != 0) { return(FALSE); }
      left_p += object_size;
      right_p -= object_size;
    }
  return(TRUE);
}


void vector_shuffle(vector_t *vector)
{
  unsigned length, object_size;
  unsigned char *buffer_p, *target_p, *source_p;
  boolean_t buffer_is_allocated;
  unsigned i, j, unassigned_count;
  
  assert(vector_is_valid(vector));

  length = vector_length(vector);
  object_size = vector_object_size(vector);

  if (length < 2) { return; }  /* nothing to shuffle */

  /* obtain temporary storage needed to permute in place */
  /* first check if vector has extra space, to avoid malloc() */
  if (vector_surplus_length(vector) > 0)
    {
      /* good news - vector has extra space, no malloc() needed */
      assert(!vector_index_is_valid(vector, length));
      buffer_p = vector_element_p(vector, length);
      buffer_is_allocated = FALSE;
    }
  else
    {
      /* bad news - no extra space in vector, malloc() required */
      buffer_p = memory_spartan_create(object_size);
      buffer_is_allocated = TRUE;
    }
  assert(buffer_p);
  assert(length > 1);

  /* repeatedly move randomly chosen element to front of vector */
  target_p = (unsigned char *)vector;
  unassigned_count = length;
  for(i = 0; i < length-1; i++)
    {
      assert(vector_index_is_valid(vector, i));
      assert((unassigned_count > 0) && (unassigned_count <= length));
#ifndef NDEBUG
      if (i+unassigned_count != length)
	{ fprintf(stdout, "%u+%u != %u\n", i, unassigned_count, length); }
#endif
      assert(i+unassigned_count == length);
      assert((unsigned char *)vector + object_size * i == target_p);

      j = drand48() * (double)unassigned_count;
      assert((j >= 0) && (j < unassigned_count));
      if (j > 0)
	{
	  source_p = target_p + (object_size * j);
	  assert(source_p > target_p);
	  (void) memcpy(buffer_p, target_p, object_size);
	  (void) memcpy(target_p, source_p, object_size);
	  (void) memcpy(source_p, buffer_p, object_size);
	}
      target_p += object_size;
      unassigned_count--;
    }
  assert(unassigned_count == 1);  /* the untouched element */

  /* free temporary storage if we allocated it */
  if (buffer_is_allocated) 
    { memory_spartan_destroy(buffer_p, object_size); }
}


boolean_t vector_are_permuted(const vector_t *left, const vector_t *right)
{
  unsigned length, object_size;
  unsigned const char *source_p, *target_p;
  char *is_matched;
  boolean_t target_is_found;
  unsigned i, j;

  assert(vector_is_valid(left));  
  assert(vector_is_valid(right));

  length = vector_length(left);
  object_size = vector_object_size(left);
  if ((length != vector_length(right)) ||
      (object_size != vector_object_size(right)))
    { return(FALSE); }

  is_matched = vector_create(sizeof(char), length);
  assert(vector_is_valid(is_matched));
  assert(vector_length(is_matched) == length);
  memset(is_matched, 0, sizeof(char) * length);

  source_p = left;
  for(i = 0; i < length; i++)
    {
      assert(source_p == vector_element_p(left, i));
      target_p = right;
      target_is_found = FALSE;
      for(j = 0; j < length; j++)
	{
	  assert(!target_is_found);
	  assert(target_p == vector_element_p(right, j));
	  if (!is_matched[j] && (memcmp(source_p, target_p, object_size) == 0))
	    {
	      /* we match left[i] with right[j] */
	      is_matched[j] = TRUE;
	      target_is_found = TRUE;
	      break;
	    }
	  target_p += object_size;
	}
      if (!target_is_found)
	{
	  assert(j == length);
	  vector_destroy(is_matched);
	  return(FALSE);
	}
      source_p += object_size;
    }
  vector_destroy(is_matched);  
  return(TRUE);
}

/************************************************************
 *
 ************************************************************/

unsigned vector_fprintf(const vector_t *vector, FILE *stream, 
			void (*element_fprintf)(const void *element_p, FILE *stream))
{
  unsigned length, object_size, i;
  const unsigned char *element_p;
  int chars_printed;

  assert(stream);
  assert(!ferror(stream));
  assert(vector_is_valid(vector));

  length = vector_length(vector);
  object_size = vector_object_size(vector);

  /* fprintf the header */
  chars_printed = fprintf(stream, "[%u %u:", object_size, length);
  if ((chars_printed < 3) || ferror(stream)) { return(0); }

  element_p = vector;
  for(i = 0; i < length; i++)
    {
      assert(element_p == vector_element_p(vector, i));
      chars_printed = fprintf(stream, " ");
      element_fprintf(element_p, stream);
      if (ferror(stream)) { return(i); }
      element_p += object_size;
    }
  chars_printed = fprintf(stream, " ]");
  return(length);
}

vector_t *vector_fscanf(FILE *stream, 
			void (*element_fscanf)(void *element_p, FILE *stream))
{
  unsigned length, object_size, i;
  unsigned char *element_p;
  vector_t *vector;
  int tokens_scanned;
  
  tokens_scanned = fscanf(stream, "[%u %u:", &object_size, &length);
  if ((tokens_scanned != 2) || ferror(stream)) { return(NULL); }
  vector = vector_create(object_size, length);
  assert(vector_is_valid(vector));

  element_p = vector;
  for(i = 0; i < length; i++)
    {
      assert(element_p == vector_element_p(vector, i));
      element_fscanf(element_p, stream);
      if (ferror(stream)) { vector_destroy(vector); return(NULL); }
      element_p += object_size;
    }
  fscanf(stream, " ]");
  if (ferror(stream)) { vector_destroy(vector); return(NULL); }
  assert(vector_is_valid(vector));
  assert(vector_length(vector) == length);
  assert(vector_object_size(vector) == object_size);
  return(vector);
}


unsigned vector_write(const vector_t *vector, FILE *stream,
		      void (*object2nbo)(const void *object_p, void *nbo_p))
{
  unsigned object_size, length, nbo, objects_written;

  assert(vector_is_valid(vector));
  assert(stream != NULL); assert(!ferror(stream));
  assert(UINT_MAX >= VECTOR_OBJECT_SIZE_MAX);

  length = vector_length(vector);
  object_size = vector_object_size(vector);

  nbo = htonl(length);
  objects_written = fwrite(&nbo, sizeof(unsigned), 1, stream);
  nbo = htonl(object_size);
  objects_written += fwrite(&nbo, sizeof(unsigned), 1, stream);
  if (objects_written != 2) { return(0); }

  objects_written = 0;
  if (object2nbo)
    {
      const unsigned char *element_p;
      void                *nbo_p;
      unsigned             i;

      nbo_p = memory_create(object_size);
      assert(memory_is_valid(nbo_p));
      element_p = vector;
      for(i = 0; i < length; i++)
	{
	  object2nbo(element_p, nbo_p);
	  objects_written += fwrite(nbo_p, object_size, 1, stream);	  
	  if (ferror(stream)) { break; }
	  element_p += object_size;
	}  
      memory_destroy(nbo_p); nbo_p = NULL;
      return(objects_written);
    }
  else
    {
#if (ANNOUNCE_WARNINGS && (BYTE_ORDER != BIG_ENDIAN))
      static boolean_t is_warned = FALSE;
      if (!is_warned)
	{ 
	  fprintf(stderr, 
		  "vector_t warning: vector_write() is not portable w/o object2nbo()\n"); 
	  is_warned = TRUE; 
	}
#endif
      objects_written = fwrite(vector, object_size, length, stream);      
      return(objects_written);
    }
}

vector_t *vector_read(FILE *stream,
		      void (*nbo2object)(const void *nbo_p, void *object_p))
{
  unsigned  length, object_size, nbo, objects_read;
  vector_t *vector;

  assert(stream != NULL); assert(!ferror(stream));

  /* read the header */
  objects_read = fread(&nbo, sizeof(unsigned), 1, stream);
  length = ntohl(nbo);
  objects_read += fread(&nbo, sizeof(unsigned), 1, stream);
  object_size = ntohl(nbo);
  if (objects_read != 2) { return(NULL); }
  assert(object_size > 0);

  /* read the elements */
  vector = vector_create(object_size, length);
  assert(vector_is_valid(vector));
  if (nbo2object)
    {
      unsigned char *element_p;
      void          *nbo_p;
      unsigned       i;

      nbo_p = memory_create(object_size);
      assert(memory_is_valid(nbo_p));
      element_p = vector;
      objects_read = 0;
      for(i = 0; i < length; i++)
	{
	  objects_read += fread(nbo_p, object_size, 1, stream);
	  if (ferror(stream)) { break; }
	  nbo2object(nbo_p, element_p);
	  element_p += object_size;
	}
      memory_destroy(nbo_p); nbo_p = NULL;
    }
  else
    {
#if (ANNOUNCE_WARNINGS && (BYTE_ORDER != BIG_ENDIAN))
      static boolean_t is_warned = FALSE;
      if (!is_warned)
	{ 
	  fprintf(stderr, 
		  "vector_t warning: vector_read() is not portable w/o nbo2object()\n"); 
	  is_warned = TRUE; 
	}
#endif
      objects_read = fread(vector, object_size, length, stream);
    }
  if ((objects_read != length) || ferror(stream))
    { vector_destroy(vector); return(NULL); }
  return(vector);
}


/************************************************************
 *   INTERNAL ROUTINES
 ************************************************************/

/* Effect: resize vector so allocated_length(vector) == length(vector) */
static vector_t *vector_strict_resize(vector_t *vector, unsigned new_length)
{
  unsigned allocated_length, object_size;
  unsigned old_bytes, new_bytes;
  vector_header_t *header;

  assert(vector_is_valid(vector));

  object_size = vector_object_size(vector);
  allocated_length = vector_allocated_length(vector);
  old_bytes = sizeof(vector_header_t) + object_size * allocated_length;
  new_bytes = sizeof(vector_header_t) + object_size * new_length;
  
  header = vector_header(vector);
  header = memory_spartan_recreate(header, old_bytes, new_bytes);
  if (header == NULL) { return(NULL); }
  header->length = new_length;
  header->surplus_length = 0;
  vector = header2vector(header);

  assert(vector_is_valid(vector));
  assert(vector_length(vector) == new_length);
  assert(vector_surplus_length(vector) == 0);
  assert(vector_object_size(vector) == object_size);

  return(vector);
}


static vector_t *vector_contracting_resize(vector_t *vector, unsigned new_length)
{
  unsigned allocated_length, new_contracted_surplus;
  vector_header_t *header;
  
  assert(vector_is_valid(vector));

  /* verify that caller chose proper resize function */
  assert(new_length <= vector_length(vector));
  assert(vector_allocated_length(vector) >= new_length);

  allocated_length = vector_allocated_length(vector);
  new_contracted_surplus = allocated_length - new_length;
  if (new_contracted_surplus > VECTOR_SURPLUS_LENGTH_MAX)
    { 
      unsigned new_contracted_length;
      new_contracted_length = new_length + VECTOR_SURPLUS_TARGET;
      vector = vector_strict_resize(vector, new_contracted_length);
      if (vector == NULL) { return(NULL); }
      header = vector_header(vector);
      header->length = new_length;
      header->surplus_length = new_contracted_length - new_length;
      assert(vector_is_valid(vector));
      assert(vector_length(vector) == new_length);
      assert(vector_surplus_length(vector) == VECTOR_SURPLUS_TARGET);
      assert(vector_allocated_length(vector) == new_contracted_length);
      return(vector);
    }
  else
    {
      assert(allocated_length >= new_length);
      assert((allocated_length - new_length) <= VECTOR_SURPLUS_LENGTH_MAX);
      assert((allocated_length - new_length) >= vector_surplus_length(vector));
      assert(new_contracted_surplus == allocated_length - new_length);
      header = vector_header(vector);
      header->length = new_length;
      header->surplus_length = new_contracted_surplus;
      assert(vector_is_valid(vector));
      assert(vector_length(vector) == new_length);
      return(vector);
    }
}

static vector_t *vector_expanding_resize(vector_t *vector, unsigned new_length)
{
  unsigned allocated_length;
  vector_header_t *header;
  
  assert(vector_is_valid(vector));

  /* verify that new_length did not overflow before call */
  assert(new_length >= vector_length(vector));

  allocated_length = vector_allocated_length(vector);
  if (new_length > allocated_length)  /* insufficient surplus */
    {
      unsigned long new_expanded_length;
      new_expanded_length = length_successor(new_length);
      if ((new_expanded_length > VECTOR_SURPLUS_LENGTH_MAX) && /* prevent underflow */
	  ((new_expanded_length - VECTOR_SURPLUS_LENGTH_MAX) > new_length))
	{ new_expanded_length = new_length + VECTOR_SURPLUS_LENGTH_MAX; }
      assert(new_expanded_length >= new_length);
      assert(new_expanded_length - new_length <= VECTOR_SURPLUS_LENGTH_MAX);
      vector = vector_strict_resize(vector, new_expanded_length);
      if (vector == NULL) { return(NULL); }
      header = vector_header(vector);
      header->length = new_length;
      header->surplus_length = new_expanded_length - new_length;
      assert(vector_is_valid(vector));
      assert(vector_length(vector) == new_length);
      assert(vector_allocated_length(vector) == new_expanded_length);
      return(vector);
    }
  else /* sufficient surplus exists */
    {
      assert(allocated_length >= new_length);
      assert((allocated_length - new_length) <= VECTOR_SURPLUS_LENGTH_MAX);
      assert((allocated_length - new_length) <= vector_surplus_length(vector));
      header = vector_header(vector);
      header->length = new_length;
      header->surplus_length = allocated_length - new_length;
      assert(vector_is_valid(vector));
      assert(vector_length(vector) == new_length);
      return(vector);
    }
}

