/*
 * (c) 1997 Eric Sven Ristad and Peter N. Yianilos.  All rights reserved.
 *
 * $Id: vector.h,v 1.22 1998/02/04 16:55:28 libpa Exp $
 *
 */

#ifndef VECTOR_INCLUDED
#define VECTOR_INCLUDED

#include <limits.h>
#include <stdio.h>
#include "common.h"

/* Description.
 *   Self-contained one dimensional array of arbitary object types --
 *   the fundamental data structure of efficient C programs.
 *   Object types are parameterized by byte size.
 *   All valid vector_t are non-NULL, even empty vector_t.
 *
 */

#define VECTOR_LENGTH_MAX          ((unsigned)UINT_MAX)
#define VECTOR_OBJECT_SIZE_MAX     ((unsigned)USHRT_MAX)
#define VECTOR_SURPLUS_LENGTH_MAX  ((unsigned)USHRT_MAX)

typedef void vector_t;
typedef struct vector_header_tag 
{
  unsigned short object_size;     /* bytes/entry */
  unsigned short surplus_length;  /* additional entries in reserve */
  unsigned int   length;          /* number of current entries */
} vector_header_t;

/************************************************************
 *  CONSTRUCTORS/DESTRUCTORS
 ************************************************************/

vector_t   *vector_create(unsigned object_size, unsigned length);
            /* Returns: vector_t with length objects of size object_size   */
            /* Note: vector_create(*,0) != NULL unless error occurs        */
vector_t   *vector_create_surplus(unsigned object_size, unsigned length, unsigned surplus);
            /* Returns: vector_t with length objects of size object_size   */
            /*          with surplus additional "hidden" slots             */
            /* Note: vector_create_surplus(*,0,0) != NULL unless error     */
vector_t   *vector_copy(const vector_t *vector);
            /* Returns: exact copy of given vector or NULL if not possible */
void        vector_mirror(const vector_t *source, vector_t *target);
            /* Requires: source and target have same length, object_size   */
            /* Effect: target contains an exact copy of source vector      */
void        vector_destroy(vector_t *vector);
            /* Effect: frees all memory used by given vector               */

/************************************************************
 *  SELECTORS
 ************************************************************/

#define     vector_length(vector)          ((unsigned)(vector_header(vector)->length))
            /* Returns: number of elements currently in vector */
#define     vector_object_size(vector)     ((unsigned)(vector_header(vector)->object_size))
            /* Returns: size of individual entries, in bytes   */
#define     vector_get(vector,i)           \
            ((void *)(((char *)(vector))+((i)*vector_object_size(vector))))
            /* Returns: pointer to i^th element in vector      */
#define     vector_element_p(vector,i)     (vector_get(vector,i))
            /* Returns: pointer to i^th element in vector      */

unsigned    vector_parity(const vector_t *vector);
            /* Returns: parity value for vector contents only  */
            /* Warning: value is not portable (byte ordering)  */

/************************************************************
 *  SINGLE ELEMENT MUTATORS
 ************************************************************/

void        vector_set(vector_t *vector, unsigned index, const void *object_p);
            /* Effect: sets vector[index] to *object_p for valid index      */

vector_t   *vector_append(vector_t *vector, const void *object_p);
            /* Effect: appends *object_p to end of vector                   */
            /* Returns: address of modified vector or NULL if not possible  */
            /* Note: repeated vector_append() is time-efficient             */
vector_t   *vector_insert(vector_t *vector, unsigned index, const void *object_p);
            /* Requires: index >= 0 and index <= vector_length(vector)      */
            /* Effect: inserts *object_p into vector at position index      */
            /* Returns: address of modified vector                          */
vector_t   *vector_delete(vector_t *vector, unsigned index);
            /* Effect: deletes object at index and reindexes successors     */
            /* Returns: address of modified vector                          */

/************************************************************
 *  MULTIPLE ELEMENT MUTATORS
 ************************************************************/

vector_t   *vector_resize(vector_t *vector, unsigned length);
            /* Effect: modifies vector to be of specified length            */
            /* Returns: address of modified vector                          */
vector_t   *vector_contract(vector_t *vector, unsigned contraction);
            /* Effect: removes last contraction entries of vector           */
            /* Returns: address of modified vector                          */
vector_t   *vector_expand(vector_t *vector, unsigned expansion);
            /* Effect: adds expansion new (uninitialized) entries to vector */
            /* Returns: address of modified vector                          */

vector_t   *vector_insert_vector(vector_t *vector, unsigned index, const vector_t *subvector);
            /* Requires: vector and subvector have same object size         */
            /* Effect: inserts copy of subvector into vector at index       */
            /* Returns: address of modified vector                          */
vector_t   *vector_concatenate(vector_t *root, const vector_t *suffix);
            /* Requires: root and suffix are distinct w/ same object size   */
            /* Effect: concatenates copy of suffix onto end of root vector  */
            /* Note: caller is responsible for destroying suffix vector     */
            /* Returns: address of modified vector                          */
vector_t   *vector_delete_subvector(vector_t *vector, unsigned index, unsigned sublength);
            /* Effect: deletes sublength objects starting at position index */
            /* Returns: address of modified vector                          */

vector_t   *vector_prune(vector_t *vector, 
			 boolean_t (*prune_me)(const void *element_p, void *parameters),
			 void *parameters);
            /* Effect: removes all entries for which prune_me() is true     */
            /* Returns: address of modified vector                          */

/************************************************************
 *  PREDICATES/COMPARISON
 ************************************************************/

#define     vector_are_equivalent(left,right)  (vector_compare((left),(right)) == 0)
            /* Returns: TRUE iff vectors are indistinguishable by selectors */
#define     vector_is_valid(vector)                                     \
            (((vector) != NULL) && (vector_object_size(vector) > 0) &&  \
	     (vector_length(vector) <= VECTOR_LENGTH_MAX - vector_surplus_length(vector)))
            /* Returns: TRUE iff vector is valid, ie., not an error         */
#define     vector_index_is_valid(vector,i)    ((0<=(i)) && ((i)<vector_length(vector)))
            /* Returns: TRUE iff index i is valid for vector                */

int         vector_compare(const vector_t *left, const vector_t *right);
            /* Returns: total order on vectors, 0 iff they are equivalent   */

/************************************************************
 *  PERMUTATION OPERATORS
 ************************************************************/

void        vector_reverse(vector_t *vector);
            /* Effect: reverses elements of given vector in place           */
void        vector_shuffle(vector_t *vector);
            /* Effect: randomly permutes elements of vector in place        */
            /* Requires: double drand48(void)                               */
boolean_t   vector_are_reversed(const vector_t *left, const vector_t *right);
            /* Returns: TRUE iff vectors are reversals of each other        */
boolean_t   vector_are_permuted(const vector_t *left, const vector_t *right);
            /* Returns: TRUE iff vectors are permutations of each other     */

/************************************************************
 *  SORTED VECTORS
 *  if left < right then compare(left, right) < 0 
 *  else if left = right then compare(left, right) == 0 
 *  else left > right then compare(left, right) > 0 
 ************************************************************/

void        vector_quicksort(vector_t *vector,
			     int (*compare)(const void *left_p, const void *right_p,
					    const void *parameters),
			     const void *parameters);
            /* Effect: random quicksorts vector according to compare()    */
            /* Note: caller parameters are passed to compare() function.  */
            /*       No need to vector_shuffle(), callee randomizes pivot */
void        vector_quicksubsort(vector_t *vector,
				unsigned start, unsigned stop,
				int (*compare)(const void *left_p, const void *right_p,
					       const void *parameters),
				const void *parameters);
            /* Effect: sorts vector[start...stop] using quicksort         */

boolean_t   vector_is_sorted(const vector_t *vector,
			     int (*compare)(const void *left_p, const void *right_p,
					    const void *parameters),
			     const void *parameters);
            /* Returns: TRUE iff vector is ordered according compare()    */

unsigned    vector_sorted_lookup(const vector_t *vector, const void *object_p,
				 int (*compare)(const void *left_p, const void *right_p,
						const void *parameters),
				 const void *parameters);
            /* Arguments: sorted vector and implicit partial ordering compare() */
            /* Requires: vector_is_sorted(vector, compare, parameters)     */
            /* Returns: index where object_p is or should be                    */

vector_t   *vector_sorted_insert(vector_t *vector, const void *object_p,
				 int (*compare)(const void *left_p, const void *right_p, 
						const void *parameters),
				 const void *parameters);
            /* Effect: inserts *object_p into vector sorted by compare()   */
            /* Requires: vector_is_sorted(vector, compare, parameters)     */
            /* Returns: address of modified vector                         */
vector_t   *vector_sorted_unique(vector_t *vector, 
				 int (*compare)(const void *left_p, const void *right_p, 
						const void *parameters),
				 const void *parameters);
            /* Effect: deletes adjacent elements for which compare() == 0  */
            /* Requires: vector_is_sorted(vector, compare, parameters)     */
            /* Returns: address of modified vector                         */

/************************************************************
 *  INPUT/OUTPUT 
 *  Note: all input/output is self-deliminating
 ************************************************************/

unsigned   vector_fprintf(const vector_t *vector, FILE *stream, 
			  void (*element_fprintf)(const void *element_p, FILE *stream));
           /* Arguments: element_fprintf() formats element w/o space   */
           /* Effect: formats vector on ascii file stream              */
           /* Returns: number of entries successfully written          */
vector_t  *vector_fscanf(FILE *stream, 
			 void (*element_fscanf)(void *element_p, FILE *stream));
           /* Effect: reads vector from ascii file stream              */
           /* Requires: stream was created by vector_fprintf()         */

unsigned   vector_write(const vector_t *vector, FILE *stream,
			void (*object2nbo)(const void *object_p, void *nbo_p));
           /* Arguments: object2nbo() converts to network byte order   */
           /*            If NULL then no conversion is performed.      */
           /* Effect: writes given vector to binary file stream        */
           /* Returns: number of entries successfully written          */
           /* Note: file is portable iff object2nbo() is correct       */
vector_t  *vector_read(FILE *stream,
		       void (*nbo2object)(const void *nbo_p, void *object_p));
           /* Arguments: nbo2object() converts from network byte order */
           /*            If NULL then no conversion is performed       */
           /* Effect: reads a vector_t from binary file stream         */
           /* Returns: vector that was written to binary file stream   */
           /* Note: file is portable iff nbo2object() is correct       */

/************************************************************
 *  MISCELLANY
 ************************************************************/

#define vector_header(vector) (((vector_header_t *)(vector))-1)
#define vector_surplus_length(vector) ((unsigned)(vector_header(vector)->surplus_length))
        /* Returns: total number of additional (ie., empty) slots */
#define vector_allocated_length(vector) (vector_length(vector)+vector_surplus_length(vector))
        /* Returns: total number of allocated slots               */
#define vector_bytes(vector) (sizeof(vector_header_t) + \
			      (vector_allocated_length(vector)*vector_object_size(vector)))
        /* Returns: total bytes allocated for given vector, including overhead */

#endif	/* VECTOR_INCLUDED */



