/*
 * (c) 1997 Eric Sven Ristad and Peter N. Yianilos.  All rights reserved.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <limits.h>
#include <string.h>   /* for memcmp() */
#include <math.h>     /* for fabs() */
#include <unistd.h>   /* for getpid() */
#include "common.h"
#include "memory.spartan.h"
#include "vector.h"
#include "vector.private.h"

static char vector_test_rcsid[] = "$Id: vector.test.c,v 1.20 1997/05/27 17:02:55 ristad Exp $";

#define VERBOSE            FALSE
#define SHUFFLE_COUNT      8
#define VECTOR_COUNT       128
#define test_length(i)     (i)
#define test_entry(i,j)    (((i)*VECTOR_COUNT)+j)

unsigned Vector_Global_Parameter = UINT_MAX;

/* Note annoying assymetry between fprintf() and fscanf() for doubles */
void dbl_fprintf(const void *d_p, FILE *stream) { fprintf(stream, "%.16e", *(double *)d_p); }
void dbl_fscanf(void *d_p, FILE *stream) { fscanf(stream, "%le", (double *)d_p); }
int dbl_compare(const void *left_p, const void *right_p, const void *parameters)
{ 
  double left, right;
  assert(parameters == &Vector_Global_Parameter);
  left = *(double *)left_p;
  right = *(double *)right_p;
  if (left > right) { return(1); }
  else if (left < right) { return(-1); }
  else { return(0); }
}
/* interval [lower, upper) */
typedef struct dbl_interval_tag { double lower, upper; } dbl_interval_t;
boolean_t dbl_prune(const void *element_p, void *parameters)
{
  double element;
  dbl_interval_t interval;
  element = *(double *)element_p;
  interval = *(dbl_interval_t *)parameters;
  return((element < interval.lower) || (element >= interval.upper));
}
boolean_t prune_all(const void *element_p, void *parameters) { return(TRUE); }

void dbl2nbo(const void *dbl_p, void *nbo_p) 
{ 
  const unsigned *hbo32_p; unsigned *nbo32_p;
  hbo32_p = dbl_p; nbo32_p = nbo_p;
#if BYTE_ORDER == BIG_ENDIAN
  nbo32_p[0] = hbo32_p[0];
  nbo32_p[1] = hbo32_p[1];
#elif BYTE_ORDER == LITTLE_ENDIAN
  nbo32_p[0] = htonl(hbo32_p[1]);
  nbo32_p[1] = htonl(hbo32_p[0]);
#else
  abort();
#endif
}

void nbo2dbl(const void *nbo_p, void *dbl_p) 
{ 
  unsigned *hbo32_p; const unsigned *nbo32_p;
  hbo32_p = dbl_p; nbo32_p = nbo_p;
#if BYTE_ORDER == BIG_ENDIAN
  hbo32_p[0] = nbo32_p[0];
  hbo32_p[1] = nbo32_p[1];
#elif BYTE_ORDER == LITTLE_ENDIAN
  hbo32_p[0] = ntohl(nbo32_p[1]);
  hbo32_p[1] = ntohl(nbo32_p[0]);
#else
  abort();
#endif
}

int main(int argc, char **argv)
{
  double *vector[VECTOR_COUNT], *copy_vector, *prune_vector, element;
  dbl_interval_t interval;
  char   *byte_vector;
  unsigned middle, index, length, element_count;
  unsigned total_bytes, total_blocks, parity_value;
  FILE *stream;
  char  testfile[512];
  int i, j;

  printf("\tTesting vector_t library...\n");
  assert(sizeof(vector_header_t) == sizeof(double));

  /* test excessive resizing */
  byte_vector = vector_create_surplus(sizeof(char), 1, VECTOR_SURPLUS_LENGTH_MAX);
  assert(vector_is_valid(byte_vector));
  assert(vector_length(byte_vector) == 1);
  assert(vector_surplus_length(byte_vector) == VECTOR_SURPLUS_LENGTH_MAX);
  assert(vector_allocated_length(byte_vector) == VECTOR_SURPLUS_LENGTH_MAX+1);
  assert(vector_bytes(byte_vector) == VECTOR_SURPLUS_LENGTH_MAX+1 + sizeof(vector_header_t));
  assert(memory_total_bytes() == VECTOR_SURPLUS_LENGTH_MAX+1 + sizeof(vector_header_t));
  assert(memory_total_blocks() == 1);
  printf("%u\tallocated vector_t with maximal surplus\n", memory_total_bytes());
  byte_vector = vector_contract(byte_vector, 1);
  assert(vector_is_valid(byte_vector));
  assert(vector_length(byte_vector) == 0);
  assert(vector_surplus_length(byte_vector) == VECTOR_SURPLUS_TARGET);
  assert(vector_allocated_length(byte_vector) == VECTOR_SURPLUS_TARGET);
  assert(vector_bytes(byte_vector) == VECTOR_SURPLUS_TARGET + sizeof(vector_header_t));
  assert(memory_total_bytes() == VECTOR_SURPLUS_TARGET + sizeof(vector_header_t));
  assert(memory_total_blocks() == 1);
  vector_destroy(byte_vector); byte_vector = NULL;
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);
  printf("%u\tperformed minimal contraction to force reallocation\n", memory_total_bytes());

  total_bytes = 0;
  total_blocks = 0;
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);

  /* allocate and initialize our vector[][] */
  for(i = 0; i < VECTOR_COUNT; i++)
    {
      length = test_length(i);
      vector[i] = vector_create(sizeof(double), length);
      assert(vector_is_valid(vector[i]));
      assert(vector_length(vector[i]) == length);
      assert(vector_surplus_length(vector[i]) == 0);
      assert(vector_allocated_length(vector[i]) == length);
      assert(vector_bytes(vector[i]) == sizeof(vector_header_t) + (sizeof(double) * length));

      total_blocks++;
      total_bytes += vector_bytes(vector[i]);
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);

      assert(!vector_index_is_valid(vector[i], -1));
      assert(!vector_index_is_valid(vector[i], length));

      /* initialize in ascending order */
      for(j = 0; j < length; j++)
	{
	  assert(vector_index_is_valid(vector[i], j));
	  element = test_entry(i,j);
	  vector_set(vector[i], j, &element);
	  assert(vector[i][j] == test_entry(i,j));
	  assert(vector_get(vector[i],j) == vector[i]+j);
	  assert(vector_element_p(vector[i],j) == vector[i]+j);
	  assert((j == 0) || (test_entry(i,j-1) < test_entry(i,j)));
	}
      assert(vector_is_sorted(vector[i], &dbl_compare, &Vector_Global_Parameter));
      assert(vector_compare(vector[i], vector[i]) == 0);
      assert((i == 0) || !vector_are_equivalent(vector[i-1], vector[i]));
      assert((i == 0) || (vector_compare(vector[i-1], vector[i]) < 0));

#if VERBOSE
      printf("\t%d:\t", i);
      vector_fprintf(vector[i], stdout, &dbl_fprintf);
      printf("\n");
#endif
    }
  assert(memory_total_blocks() == VECTOR_COUNT);
  printf("%u\tcreated %d valid ascending vector_t\n", 
	 memory_total_bytes(), VECTOR_COUNT);

  /* verify copy and input/output procedures */
  sprintf(testfile, "/tmp/vector.tmp.%d", (int)getpid());
  for(i = 0; i < VECTOR_COUNT; i++)
    {
      /* copy operation */
      copy_vector = vector_copy(vector[i]);
      assert(vector_is_valid(copy_vector));
      assert(vector_length(vector[i]) == test_length(i));
      assert(vector_are_equivalent(copy_vector, vector[i]));
      assert(vector_parity(copy_vector) == vector_parity(vector[i]));
      assert(memcmp(copy_vector, vector[i], sizeof(double)*test_length(i)) == 0);
      assert(memory_total_bytes() == total_bytes + vector_bytes(copy_vector));
      assert(memory_total_blocks() == total_blocks + 1);
      vector_destroy(copy_vector);
      copy_vector = NULL;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);

      /* nonportable binary input/output */
      stream = tmpfile();
      assert(stream);
      element_count = vector_write(vector[i], stream, NULL);
      assert(!ferror(stream));
      assert(element_count == test_length(i));
      rewind(stream);
      copy_vector = vector_read(stream, NULL);
      assert(!ferror(stream));
      assert(fgetc(stream) == EOF);
      assert(vector_are_equivalent(copy_vector, vector[i]));
      assert(vector_parity(copy_vector) == vector_parity(vector[i]));
      vector_destroy(copy_vector); copy_vector = NULL;
      fclose(stream); stream = NULL;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);

      /* portable binary input/output */
      stream = tmpfile();
      assert(stream);
      element_count = vector_write(vector[i], stream, &dbl2nbo);
      assert(!ferror(stream));
      assert(element_count == test_length(i));
      rewind(stream);
      copy_vector = vector_read(stream, &nbo2dbl);
      assert(!ferror(stream));
      assert(fgetc(stream) == EOF);
      assert(vector_are_equivalent(copy_vector, vector[i]));
      assert(vector_parity(copy_vector) == vector_parity(vector[i]));
      vector_destroy(copy_vector); copy_vector = NULL;
      fclose(stream); stream = NULL;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);

      /* ascii input/output */
      stream = fopen(testfile, "w+");
      assert(stream);
      element_count = vector_fprintf(vector[i], stream, &dbl_fprintf);
      assert(!ferror(stream));
      assert(element_count == test_length(i));
      rewind(stream);
      copy_vector = vector_fscanf(stream, &dbl_fscanf);
      assert(!ferror(stream));
      assert(fgetc(stream) == EOF);
      assert(vector_length(copy_vector) == vector_length(vector[i]));
      assert(vector_object_size(copy_vector) == vector_object_size(vector[i]));
#if DEBUG
      printf(" ");
      (void) vector_fprintf(vector[i], stdout, &dbl_fprintf);
      (void) vector_fprintf(copy_vector, stdout, &dbl_fprintf);
      printf("\n");
#endif
      for(j = 0; j < vector_length(copy_vector); j++)
	{ assert(copy_vector[j] == vector[i][j]); }
      vector_destroy(copy_vector);
      copy_vector = NULL;
      fclose(stream);
      stream = NULL;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  remove(testfile);
  printf("%u\tcopied, write/read, fprintf/fscanf %d vectors correctly\n", 
	 memory_total_bytes(), VECTOR_COUNT);

  /* verify single element mutators */
  for(i = 1; i < VECTOR_COUNT; i++)
    {
      parity_value = vector_parity(vector[i]);
      length = vector_length(vector[i]);
      assert(length == test_length(i));
      assert(length >= 1);

      /* delete middle element */
      middle  = (length - 1)/2;
      element = vector[i][middle];
      assert(element == (double)test_entry(i,middle));
      vector[i] = vector_delete(vector[i], middle);
      assert(vector_is_valid(vector[i]));
      assert(vector_length(vector[i]) == length-1);
      assert(vector_surplus_length(vector[i]) == 1);
      assert(vector_is_sorted(vector[i], &dbl_compare, &Vector_Global_Parameter));
      assert(parity_value != vector_parity(vector[i]));

      /* insert it back */
      vector[i] = vector_insert(vector[i], middle, &element);
      assert(vector[i][middle] == (double)test_entry(i,middle));
      assert(vector_is_valid(vector[i]));
      assert(vector_length(vector[i]) == length);
      assert(vector_surplus_length(vector[i]) == 0);
      assert(vector_is_sorted(vector[i], &dbl_compare, &Vector_Global_Parameter));
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
      assert(parity_value == vector_parity(vector[i]));

      /* append irregular element to end of vector */
      assert(length == vector_length(vector[i]));
      element = test_entry(i,0);
      vector[i] = vector_append(vector[i], &element);
      total_bytes += sizeof(double) * (1+vector_surplus_length(vector[i]));
      assert(vector_is_valid(vector[i]));
      assert(*(double *)vector_get(vector[i], length) == element);
      assert(*(double *)vector_element_p(vector[i], length) == element);
      assert(vector_length(vector[i]) == length+1);
      assert(vector_allocated_length(vector[i]) == length_successor(length+1));
      assert(vector_allocated_length(vector[i]) == 
	     vector_length(vector[i]) + vector_surplus_length(vector[i]));
      assert(memory_total_blocks() == total_blocks);
      assert(memory_total_bytes() == total_bytes);
      assert((i < 2) || !vector_is_sorted(vector[i], &dbl_compare, &Vector_Global_Parameter));

      /* contract it back */
      vector[i] = vector_contract(vector[i], 1);
      assert(vector_length(vector[i]) == length);
      assert(vector_allocated_length(vector[i]) == 
	     length_predecessor(length_successor(length+1)));
      assert(vector_surplus_length(vector[i]) == length_successor(length+1) - length);
      assert(memory_total_blocks() == total_blocks);
      assert(memory_total_bytes() == total_bytes);
      assert(parity_value == vector_parity(vector[i]));

      /* verify null contraction */
      assert(vector[i] == vector_contract(vector[i], 0));
    }
  printf("%u\tdeleted, inserted, and contracted single element from %d valid vector_t\n", 
	 memory_total_bytes(), VECTOR_COUNT);

  /* verify multiple element mutators */
  for(i = 0; i < VECTOR_COUNT-1; i++)
    {
      /* prepare to adjust total bytes at end of loop */
      assert(vector_length(vector[i]) == test_length(i));
      assert((i == 0) || (vector_surplus_length(vector[i]) > 0));
      total_bytes -= sizeof(double) * vector_surplus_length(vector[i]);

      /* verify null _prune() */
      interval.lower = test_entry(i, 0);
      interval.upper = test_entry(i+1, 0);
      prune_vector = vector_prune(vector[i], &dbl_prune, &interval);
      assert(vector_is_valid(prune_vector));
      assert(prune_vector == vector[i]);
      prune_vector = NULL;

      /* verify total _prune() */
      prune_vector = vector_copy(vector[i+1]);
      assert(vector_is_valid(prune_vector));
      assert(prune_vector != vector[i+1]);
      prune_vector = vector_prune(prune_vector, &dbl_prune, &interval);
      assert(vector_is_valid(prune_vector));
      assert(vector_length(prune_vector) == 0);
      vector_destroy(prune_vector); prune_vector = NULL;

      /* concatentate suffix vector */
      assert(i+1 < VECTOR_COUNT);
      vector[i] = vector_concatenate(vector[i], vector[i+1]);
      assert(vector_is_valid(vector[i]));
      assert(vector_length(vector[i]) == test_length(i) + test_length(i+1));
      for(j = 0; j < test_length(i); j++)
	{ assert(vector[i][j] == test_entry(i,j)); }
      for(j = 0; j < test_length(i+1); j++)
	{ assert(vector[i][test_length(i)+j] == test_entry(i+1, j)); }
      assert(vector_is_sorted(vector[i], &dbl_compare, &Vector_Global_Parameter));

      /* concatentate prefix vector */
      vector[i] = vector_insert_vector(vector[i], 0, vector[i+1]);
      assert(vector_is_valid(vector[i]));
      assert(vector_length(vector[i]) == test_length(i) + 2*test_length(i+1));
      for(j = 0; j < test_length(i+1); j++)
	{ assert(vector[i][j] == test_entry(i+1,j)); }
      for(j = 0; j < test_length(i); j++)
	{ assert(vector[i][test_length(i+1)+j] == test_entry(i,j)); }
      for(j = 0; j < test_length(i+1); j++)
	{ assert(vector[i][test_length(i)+test_length(i+1)+j] == test_entry(i+1,j)); }
      assert((i < 2) || !vector_is_sorted(vector[i], &dbl_compare, &Vector_Global_Parameter));

      /* copy the engorged vector and then prune it */
      prune_vector = vector_copy(vector[i]);
      assert(vector_is_valid(prune_vector));
      assert(prune_vector != vector[i]);
      assert(vector_are_equivalent(prune_vector, vector[i]));
      interval.lower = test_entry(i, 0);
      interval.upper = test_entry(i+1, 0);
      prune_vector = vector_prune(prune_vector, &dbl_prune, &interval);
      assert(vector_is_valid(prune_vector));
      assert(!vector_are_equivalent(prune_vector, vector[i]));
      assert(vector_length(prune_vector) == test_length(i));

      /* remove the suffix vector */
      vector[i] = vector_contract(vector[i], test_length(i+1));
      assert(vector_is_valid(vector[i]));
      assert(vector_length(vector[i]) == test_length(i) + test_length(i+1));
      assert(vector_surplus_length(vector[i]) >= test_length(i+1));
      
      /* remove the prefix vector */
      vector[i] = vector_delete_subvector(vector[i], 0, test_length(i+1));
      assert(vector_is_valid(vector[i]));
      assert(vector_length(vector[i]) == test_length(i));
      assert(vector_surplus_length(vector[i]) >= 2*test_length(i+1));
      for(j = 0; j < test_length(i); j++)
	{ assert(vector[i][j] == test_entry(i,j)); }

      /* verify that contract/delete and prune both gave identical result */
      assert(vector[i] != prune_vector);
      assert(vector_are_equivalent(vector[i], prune_vector));
      prune_vector = vector_prune(prune_vector, &prune_all, NULL);
      assert(vector_is_valid(prune_vector));
      assert(vector_length(prune_vector) == 0);
      vector_destroy(prune_vector); prune_vector = NULL;
      assert(!vector_is_valid(prune_vector));

      /* combine vector_expand() and vector_resize() */
      vector[i] = vector_expand(vector[i], vector_length(vector[i+1]));
      assert(vector_is_valid(vector[i]));
      assert(vector_length(vector[i]) == test_length(i) + test_length(i+1));
      assert(vector_length(vector[i+1]) == test_length(i+1));
      vector[i] = vector_resize(vector[i], test_length(i));
      assert(vector_is_valid(vector[i]));
      assert(vector_length(vector[i]) == test_length(i));
      for(j = 0; j < test_length(i); j++)
	{ assert(vector[i][j] == test_entry(i,j)); }
      
      /* actually adjust total bytes */
      total_bytes += sizeof(double) * vector_surplus_length(vector[i]);
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  printf("%u\tconcatenated, deleted, and pruned subvectors to create %d valid vector_t\n", 
	 memory_total_bytes(), VECTOR_COUNT);

  /* verify permutation and sorting operations */
  for(i = 0; i < VECTOR_COUNT; i++)
    {
      copy_vector = vector_copy(vector[i]); 
      assert(vector_is_valid(copy_vector));
      assert(vector_is_sorted(copy_vector, &dbl_compare, &Vector_Global_Parameter));
      assert(vector_surplus_length(copy_vector) == 0);
      assert(vector_bytes(copy_vector) == 
	     sizeof(vector_header_t) + sizeof(double) * vector_length(vector[i]));
      
      total_bytes += sizeof(vector_header_t) + sizeof(double) * vector_length(vector[i]);
      total_blocks++;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);

      vector_reverse(copy_vector);
      assert((i < 2) || 
	     !vector_is_sorted(copy_vector, &dbl_compare, &Vector_Global_Parameter));
      assert((i < 2) || !vector_are_equivalent(copy_vector, vector[i]));
      assert(vector_are_permuted(copy_vector, vector[i]));

      vector_quicksort(copy_vector, &dbl_compare, &Vector_Global_Parameter);
      assert(vector_is_sorted(copy_vector, &dbl_compare, &Vector_Global_Parameter));
      assert(vector_are_equivalent(copy_vector, vector[i]));
      assert(vector_parity(copy_vector) == vector_parity(vector[i]));

      /* assert() with high probability but not with certainty */
      for(j = 0; j < SHUFFLE_COUNT; j++) { vector_shuffle(copy_vector); }
      assert((i < 3) || 
	     !vector_is_sorted(copy_vector, &dbl_compare, &Vector_Global_Parameter));
      assert((i < 3) || !vector_are_equivalent(copy_vector, vector[i]));
      assert(vector_are_permuted(copy_vector, vector[i]));

      if (i > 1)
	{
	  assert(vector_index_is_valid(copy_vector, 0));
	  copy_vector[0]++;
	  assert(!vector_are_permuted(copy_vector, vector[i]));
	  copy_vector[0]--;
	}

      /* sort twice to stress-test quicksort worst case */
      vector_quicksort(copy_vector, &dbl_compare, &Vector_Global_Parameter);
      vector_quicksort(copy_vector, &dbl_compare, &Vector_Global_Parameter);
      assert(vector_is_sorted(copy_vector, &dbl_compare, &Vector_Global_Parameter));
      assert(vector_are_equivalent(copy_vector, vector[i]));
      assert(vector_parity(copy_vector) == vector_parity(vector[i]));

      for(j = 0; j < vector_length(vector[i]); j++)
	{
	  element = test_entry(i, j);
	  assert(vector[i][j] == element);
	  assert(dbl_compare(&element, &element, &Vector_Global_Parameter) == 0);
	  index = vector_sorted_lookup(vector[i], &element, 
				       &dbl_compare, &Vector_Global_Parameter);
	  assert(index == j);
	}

      vector_destroy(copy_vector);
      copy_vector = NULL;
      total_bytes -= sizeof(vector_header_t) + sizeof(double) * vector_length(vector[i]);
      total_blocks--;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  printf("%u\tcorrectly reversed, shuffled and sorted %d vector_t\n", 
	 memory_total_bytes(), VECTOR_COUNT);

  /* verify vector_sorted_unique() */
  for(i = 0; i < VECTOR_COUNT; i++)
    {
      copy_vector = vector_copy(vector[i]);
      assert(vector_is_valid(copy_vector));
      assert(vector_length(copy_vector) == test_length(i));
      assert(vector_is_sorted(copy_vector, &dbl_compare, &Vector_Global_Parameter));

      prune_vector = vector_sorted_unique(copy_vector, &dbl_compare, 
					  &Vector_Global_Parameter);
      assert(prune_vector == copy_vector);
      assert(vector_are_equivalent(copy_vector, prune_vector));
      prune_vector = NULL;

      copy_vector = vector_concatenate(copy_vector, vector[i]);
      assert(vector_is_valid(copy_vector));
      assert(vector_length(copy_vector) == 2 * test_length(i));
      vector_quicksort(copy_vector, &dbl_compare, &Vector_Global_Parameter);
      assert(vector_length(copy_vector) == 2 * test_length(i));
      copy_vector = vector_sorted_unique(copy_vector, &dbl_compare, &Vector_Global_Parameter);
      assert(vector_is_valid(copy_vector));
      assert(vector_length(copy_vector) == test_length(i));
      assert(vector_is_sorted(copy_vector, &dbl_compare, &Vector_Global_Parameter));
      assert(vector_are_equivalent(copy_vector, vector[i]));
      
      vector_destroy(copy_vector);
      copy_vector = NULL;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  printf("%u\tcorrectly replicated, sorted and uniqued %d vectors\n", 
	 memory_total_bytes(), VECTOR_COUNT);

  /****************************** EPILOGUE *******************************/

  /* deallocate all vector_t */
  for(i = 0; i < VECTOR_COUNT; i++)
    {
      total_blocks--;
      total_bytes -= vector_bytes(vector[i]);
      vector_destroy(vector[i]);
      vector[i] = NULL;
      assert(memory_total_bytes() == total_bytes);
      assert(memory_total_blocks() == total_blocks);
    }
  printf("%u\tdestroyed all %d vector_t correctly\n", 
	 memory_total_bytes(), VECTOR_COUNT);

  /* verify absence of memory leak */
  printf("\tmemory leak == %u bytes in %u blocks\n", 
	 memory_total_bytes(), memory_total_blocks());
  assert(total_bytes == 0);
  assert(total_blocks == 0);
  assert(memory_total_bytes() == 0);
  assert(memory_total_blocks() == 0);

  printf("\tvector_t library passed all tests.\n\n");
  return(0);
}

#ifdef NDEBUG
#error test suite must not be compiled with NDEBUG
#endif


